/*
 *  Author: Li Luo 
 *  Email:  li.luo@siat.ac.cn
 */

#include <cusp/format.h>
#include <cusp/detail/utils_.h>
#include <assert.h>

namespace cusp
{
namespace detail
{
		
// row and column indices are copied from GPU to GPU
// assign MKL value from CPU to GPU
// only used in MKLILU0
template <typename MatrixType,
	      typename ValueType>
void create_fact_from_mkl(const MatrixType& A,
						  const ValueType* val,
						  MatrixType& A_fact,
						  cusp::csr_format)
{	
	// resize A_fact
	A_fact.resize(A.num_rows, A.num_cols, A.num_entries);
	
	cusp::blas::copy(A.row_offsets, A_fact.row_offsets);
	cusp::blas::copy(A.column_indices, A_fact.column_indices);
	A_fact.values.assign(val,val+A.num_entries);	
}
						  
// assign MKL data from CPU to GPU
// used in MKLILU0 or SPARSKIT2ILU
template <typename IndexType,
          typename ValueType,
		  typename MatrixType>
void create_fact_from_mkl(const IndexType num_rows, 
						  const IndexType num_cols,
						  const IndexType num_entries,						  
						  const IndexType* row_off, 
						  const IndexType* col_ind, 
						  const ValueType* val, 
						  MatrixType& A_fact,
						  cusp::csr_format)
{
	// MKL requests 
	assert(num_rows==num_cols);
	
	// resize A_fact
	// the real number of entries must less or equal to the maxmum number of entries
	IndexType num_entries_ = row_off[num_rows] - 1; // MKL is 1-based
	assert(num_entries_<=num_entries);	
	A_fact.resize(num_rows, num_cols, num_entries_);

	// MKL is 1-based while cusp is 0-based, transform row_offsets and column_indices to be minusing one
	A_fact.row_offsets.assign(thrust::make_transform_iterator(row_off, cusp::detail::ADD<IndexType>(-1)),
	thrust::make_transform_iterator(row_off+num_rows+1, cusp::detail::ADD<IndexType>(-1)));
	
	A_fact.column_indices.assign(thrust::make_transform_iterator(col_ind, cusp::detail::ADD<IndexType>(-1)),
	thrust::make_transform_iterator(col_ind+num_entries_, cusp::detail::ADD<IndexType>(-1)));
	
	A_fact.values.assign(val,val+num_entries_);	
}

} // end namespace detail

template <typename MatrixType,
	      typename ValueType>
void create_fact_from_mkl(const MatrixType& A,
						  const ValueType* val,
						  MatrixType& A_fact)
{
    CUSP_PROFILE_SCOPED();
	
	cusp::detail::create_fact_from_mkl(A, val, A_fact, typename MatrixType::format());
}
						  
template <typename IndexType,
          typename ValueType,
		  typename MatrixType>
void create_fact_from_mkl(const IndexType num_rows, 
						  const IndexType num_cols,
						  const IndexType num_entries,						  
						  const IndexType* row_off, 
						  const IndexType* col_ind, 
						  const ValueType* val, 
						  MatrixType& A_fact)
{
    CUSP_PROFILE_SCOPED();
	
	cusp::detail::create_fact_from_mkl(num_rows, num_cols, num_entries, row_off, col_ind, val, A_fact, typename MatrixType::format());
} 
} // end namespace cusp

