/*
 *  Author: Li Luo 
 *  Email:  li.luo@siat.ac.cn 
 */

/*! \file permutation.h
 *  \brief permute array or matrix
 */

#include <cusp/coo_matrix.h>
#include <cusp/csr_matrix.h>
#include <cusp/transpose.h>
#include <cusp/multiply.h>
#include <cusp/detail/format_utils.h>
#include <cusp/detail/utils_.h>

#include <thrust/scatter.h>
#include <thrust/fill.h>
#include <thrust/transform.h>
#include <thrust/scan.h>

#pragma once

namespace cusp {

// permutate by scatter
// pass pt as backward: from objective to original
// pass p as forward: from original to objective
template< typename Array1, typename Array2 >
void permute_array(const Array1 &p, const Array2 &x, Array2 &y)
{
    CUSP_PROFILE_SCOPED();
	
	thrust::scatter(x.begin(), x.end(), p.begin(), y.begin());		
}	

// get the inverse permutation
template< typename Array1, typename Array2 >
void get_inverse_permutation(const Array1 &p, Array2 &pt)
{
    CUSP_PROFILE_SCOPED();

	Array2 p_tmp(p);
	thrust::sequence(pt.begin(), pt.end());
	// pt[i] is the objective number of i in original number
	thrust::stable_sort_by_key(p_tmp.begin(), p_tmp.end(), pt.begin());
}

namespace detail {

// build the permutation matrix from permutation array
template< typename Array1d, typename Matrix >
void build_permutation_matrix(const Array1d &p, Matrix &A, cusp::coo_format)
{
    int n = p.size();
	A.resize(n,n,n);
	// permute the col to row
	// permute the original number p[i] to the objective number i
    thrust::sequence(A.row_indices.begin(), A.row_indices.end());	
	thrust::copy(p.begin(), p.end(), A.column_indices.begin());	
	thrust::fill(A.values.begin(), A.values.end(), 1);
}

template< typename Matrix, typename Array1d >
void build_column_indices_permutation(const Matrix &A, const Array1d &p, Array1d &p_row_indices, Array1d &indices_permutation, cusp::coo_format)
{
    typedef typename Matrix::index_type  IndexType;
	IndexType num_rows = A.num_rows;
	IndexType num_entries = A.num_entries;	
	
	// row_indices --> nnz
	Array1d nnz(num_rows);
	cusp::detail::indices_to_nnz(A.row_indices, nnz);
	
	// nnz --> permuted nnz
	Array1d p_nnz(num_rows);	
	cusp::permute_array(p,nnz,p_nnz);
	
	// permuted nnz --> permuted row_offsets
	Array1d p_row_offsets(num_rows+1,0);
	cusp::detail::nnz_to_offsets(p_nnz, p_row_offsets);	
	
	// permuted row_offsets --> permuted row_indices		
	p_row_indices.resize(num_entries);	
	cusp::detail::offsets_to_indices(p_row_offsets, p_row_indices);
	
	// row_indices --> row_offsets 
	Array1d row_offsets(num_rows+1,0);
	cusp::detail::indices_to_offsets(A.row_indices, row_offsets);
	
	// transform the original sort to the permutated global numbering		
	Array1d p_offsets(num_rows);
	
	// the head for the current line i will be interchange to the p_row_indices[p[i]]
	IndexType *p_offsets_ptr = thrust::raw_pointer_cast(&p_offsets[0]);
	IndexType *p_row_offsets_ptr = thrust::raw_pointer_cast(&p_row_offsets[0]);	
	const IndexType *p_ptr = thrust::raw_pointer_cast(&p[0]);
	cusp::detail::subscript<IndexType, IndexType>(num_rows,p_ptr,p_row_offsets_ptr,p_offsets_ptr, typename Array1d::memory_space( ));

	// extend the permutated global numbering with zero filling	
	indices_permutation.resize(num_entries);	
	thrust::fill(indices_permutation.begin(), indices_permutation.end(), 0);
    // convert compressed row offsets into uncompressed row indices_permutation
    thrust::scatter(p_offsets.begin(),
					   p_offsets.end(),
					   row_offsets.begin(),
                       indices_permutation.begin());
	// extend
	// A.row_indices serves as a key like: 0011122333
	// indices_permutation is 9030010600
	// output: 9933311666
    thrust::inclusive_scan_by_key(A.row_indices.begin(), A.row_indices.end(), indices_permutation.begin(), indices_permutation.begin());  
	
	// create counting numbers to be added to indices_permutation
	// A.row_indices serves as a key like: 001112233
	// output: 0101201012
	Array1d local_indices(num_entries);	
    thrust::exclusive_scan_by_key(A.row_indices.begin(), A.row_indices.end(), thrust::make_constant_iterator<IndexType>(1), local_indices.begin());  
	
	// add: 9933311666+0101201012=9 10 3 4 5 1 2 6 7 8
    thrust::transform(indices_permutation.begin(), indices_permutation.end(), local_indices.begin(), indices_permutation.begin(), thrust::plus<IndexType>());	
}

template< typename Matrix, typename Array1d >
void build_row_indices_permutation(const Matrix &A, const Array1d &p, Array1d &p_column_indices, Array1d &indices_permutation, cusp::coo_format)
{
    typedef typename Matrix::index_type  IndexType;
	IndexType num_cols = A.num_cols;
	IndexType num_entries = A.num_entries;	
	
	// column_indices --> nnz
	Array1d nnz(num_cols);
	cusp::detail::indices_to_nnz(A.column_indices, nnz);
	
	// nnz --> permuted nnz
	Array1d p_nnz(num_cols);	
	cusp::permute_array(p,nnz,p_nnz);
	
	// permuted nnz --> permuted column_offsets	
	Array1d p_column_offsets(num_cols+1,0);
	cusp::detail::nnz_to_offsets(p_nnz, p_column_offsets);	
	
	// permuted column_offsets --> permuted column_indices		
	p_column_indices.resize(num_entries);	
	cusp::detail::offsets_to_indices(p_column_offsets, p_column_indices);
	
	// column_indices --> column_offsets
	Array1d column_offsets(num_cols+1,0);
	cusp::detail::indices_to_offsets(A.column_indices, column_offsets);
	
	// transform the original sort to the permutated global numbering		
	Array1d p_offsets(num_cols);

	// the head for the current line i will be interchange to the p_column_offsets[p[i]]	
	const IndexType *p_column_offsets_ptr = thrust::raw_pointer_cast(&p_column_offsets[0]);	
	const IndexType *p_ptr = thrust::raw_pointer_cast(&p[0]);
	IndexType *p_offsets_ptr = thrust::raw_pointer_cast(&p_offsets[0]);
	cusp::detail::subscript<IndexType, IndexType>(num_cols, p_ptr, p_column_offsets_ptr, p_offsets_ptr, typename Array1d::memory_space( ));

	// extend the permutated global numbering with zero filling		
	indices_permutation.resize(num_entries);	
	thrust::fill(indices_permutation.begin(), indices_permutation.end(), 0);
    // convert compressed row offsets into uncompressed row indices_permutation
    thrust::scatter(p_offsets.begin(),
					   p_offsets.end(),
					   column_offsets.begin(),
                       indices_permutation.begin());
	// extend 			
	thrust::inclusive_scan_by_key(A.column_indices.begin(), A.column_indices.end(), indices_permutation.begin(), indices_permutation.begin());  

	// create counting numbers to be added to indices_permutation	
	Array1d local_indices(num_entries);	
    thrust::exclusive_scan_by_key(A.column_indices.begin(), A.column_indices.end(), thrust::make_constant_iterator<IndexType>(1), local_indices.begin());  

	// add	
    thrust::transform(indices_permutation.begin(), indices_permutation.end(), local_indices.begin(), indices_permutation.begin(), thrust::plus<IndexType>());	
}

template< typename Matrix, typename Array1d >
void permute_matrix_by_row(const Matrix &A, const Array1d &p_row_indices, const Array1d &indices_permutation, Matrix &pA, cusp::coo_format)
{
	pA.resize(A.num_rows, A.num_cols, A.num_entries);	
	pA.row_indices = p_row_indices;
	cusp::permute_array(indices_permutation,A.column_indices,pA.column_indices);	
	cusp::permute_array(indices_permutation,A.values,pA.values);
}

template< typename Matrix, typename Array1d >
void permute_matrix_by_column(const Matrix &A, const Array1d &p_column_indices, const Array1d &indices_permutation, Matrix &pA, cusp::coo_format)
{
	pA.resize(A.num_rows, A.num_cols, A.num_entries);	
	pA.column_indices = p_column_indices;
	cusp::permute_array(indices_permutation,A.row_indices,pA.row_indices);	
	cusp::permute_array(indices_permutation,A.values,pA.values);
}
} // namespace detail

template< typename Array1d, typename Matrix >
void build_permutation_matrix(const Array1d &p, Matrix &A)
{
    CUSP_PROFILE_SCOPED();
	
	detail::build_permutation_matrix(p, A, typename Matrix::format( ));
}

// directly permute column_indices and values for COO format
template< typename Matrix, typename Array1d >
void build_column_indices_permutation(const Matrix &A, const Array1d &p, Array1d &p_row_indices, Array1d &indices_permutation)
{
    CUSP_PROFILE_SCOPED();
	
	detail::build_column_indices_permutation(A, p, p_row_indices, indices_permutation, typename Matrix::format( ));
}

// directly permute row_indices and values for COO format
template< typename Matrix, typename Array1d >
void build_row_indices_permutation(const Matrix &A, const Array1d &p, Array1d &p_column_indices, Array1d &indices_permutation)
{
    CUSP_PROFILE_SCOPED();
	
	detail::build_row_indices_permutation(A, p, p_column_indices, indices_permutation, typename Matrix::format( ));
}

// permute matrix by row
template< typename Matrix, typename Array1d >
void permute_matrix_by_row(const Matrix &A, const Array1d &p_row_indices, const Array1d &indices_permutation, Matrix &pA)
{
    CUSP_PROFILE_SCOPED();

	detail::permute_matrix_by_row(A, p_row_indices, indices_permutation, pA, typename Matrix::format( ));
}

// permute matrix by row
template< typename Matrix, typename Array1d >
void permute_matrix_by_column(const Matrix &A, const Array1d &p_column_indices, const Array1d &indices_permutation, Matrix &pA)
{
    CUSP_PROFILE_SCOPED();

	detail::permute_matrix_by_column(A, p_column_indices, indices_permutation, pA, typename Matrix::format( ));
}

template< typename Matrix, typename Array1d >
void compute_ptAqt(const Matrix &A, const Array1d &p, const Array1d &q, Matrix &ptAqt)
{
    CUSP_PROFILE_SCOPED();
	
	Matrix temp;
	Array1d p_row_indices(A.num_entries);	
	Array1d p_column_indices(A.num_entries);	
	Array1d indices_permutation(A.num_entries);	
	
	// perform the ptA
	cusp::build_column_indices_permutation(A, p, p_row_indices, indices_permutation);
	cusp::permute_matrix_by_row(A, p_row_indices, indices_permutation, temp);	
	// sort by column first then by row
	cusp::detail::sort_by_row_and_column(temp.column_indices, temp.row_indices, temp.values);
	// perform the ptAp
	cusp::build_row_indices_permutation(temp, q, p_column_indices, indices_permutation);
	cusp::permute_matrix_by_column(temp, p_column_indices, indices_permutation, ptAqt);
	cusp::detail::sort_by_row_and_column(ptAqt.row_indices, ptAqt.column_indices, ptAqt.values);	
}

template< typename Matrix, typename Array1d >
void compute_ptAp(const Matrix &A, const Array1d &p, Matrix &ptAp)
{
    CUSP_PROFILE_SCOPED();
	
	Matrix temp;
	Array1d p_row_indices(A.num_entries);	
	Array1d p_column_indices(A.num_entries);	
	Array1d indices_permutation(A.num_entries);	
	
	// perform the ptA
	cusp::build_column_indices_permutation(A, p, p_row_indices, indices_permutation);
	cusp::permute_matrix_by_row(A, p_row_indices, indices_permutation, temp);	
	// sort by column first then by row
	cusp::detail::sort_by_row_and_column(temp.column_indices, temp.row_indices, temp.values);
	// perform the ptAp
	cusp::build_row_indices_permutation(temp, p, p_column_indices, indices_permutation);
	cusp::permute_matrix_by_column(temp, p_column_indices, indices_permutation, ptAp);
	cusp::detail::sort_by_row_and_column(ptAp.row_indices, ptAp.column_indices, ptAp.values);	
}

template< typename Matrix1, typename Matrix2 >
void compute_ptAp(const Matrix1 &A, const Matrix2 &pt_matrix, const Matrix2 &p_matrix, Matrix1 &ptAp)
{
    CUSP_PROFILE_SCOPED();
	
	Matrix1 temp;
	// perform the ptA, note: p_matrix*A is really ptA
	cusp::multiply(p_matrix,A,temp);
	// perform the ptAp
	cusp::multiply(temp,pt_matrix,ptAp);
}	

} // namespace cusp

