/*
 *  Author: Li Luo 
 *  Email:  li.luo@siat.ac.cn
 */


#pragma once

#include <thrust/detail/type_traits.h>
#include <cusp/detail/device/arch.h>
#include <cusp/detail/device/utils.h>
#include <thrust/device_ptr.h>

namespace cusp
{
namespace detail
{

template <typename IndexType>
__global__ void
triangular_counts_kernel(const IndexType num_rows,
                         const IndexType * Ap, 
                         const IndexType * Aj, 
						     IndexType * nnz_lo, 
                             IndexType * nnz_up)
{
    const IndexType thread_id = blockDim.x * blockIdx.x + threadIdx.x;
    const IndexType grid_size = gridDim.x * blockDim.x;

    for(IndexType row = thread_id; row < num_rows; row += grid_size)
    {
        const IndexType row_start = Ap[row];
        const IndexType row_end   = Ap[row+1];
        
		IndexType entries_per_row_lo = 0; 
		IndexType entries_per_row_up = 0; 		
        for (IndexType jj = row_start; jj < row_end; jj++) {
            if (Aj[jj] < row) // lower triangle without diagonal
			entries_per_row_lo++; 
			else	// upper triangle with diagonal
			entries_per_row_up++;      
		}
		entries_per_row_lo++; // plus the diagonal 1.0
		nnz_lo[row] += entries_per_row_lo;
		nnz_up[row] += entries_per_row_up;		
    }
}

template <typename IndexType, typename ValueType>
__global__ void
triangular_fill_kernel(const IndexType num_rows,
                       const IndexType * Ap, 
                       const IndexType * Aj, 
                       const ValueType * Ax, 
					   const IndexType * Ap_lo, 
							 IndexType * Aj_lo, 
							 ValueType * Ax_lo, 
					   const IndexType * Ap_up, 
							 IndexType * Aj_up, 
							 ValueType * Ax_up)
{
    const IndexType thread_id = blockDim.x * blockIdx.x + threadIdx.x;
    const IndexType grid_size = gridDim.x * blockDim.x;

    for(IndexType row = thread_id; row < num_rows; row += grid_size)
    {
        const IndexType row_start = Ap[row];
        const IndexType row_end   = Ap[row+1];
        
		IndexType row_offsets_lo = Ap_lo[row];	// start of this row in lower triangle
		IndexType row_offsets_up = Ap_up[row];	// start of this row in upper triangle
		IndexType entries_per_row_lo = 0; 
		IndexType entries_per_row_up = 0; 	
        for (IndexType jj = row_start; jj < row_end; jj++) {
            if (Aj[jj] < row) { // lower triangle without diagonal
			Aj_lo[row_offsets_lo + entries_per_row_lo] = Aj[jj];
			Ax_lo[row_offsets_lo + entries_per_row_lo] = Ax[jj];
			entries_per_row_lo++; 				
			} else {	// upper triangle with diagonal
			Aj_up[row_offsets_up + entries_per_row_up] = Aj[jj];
			Ax_up[row_offsets_up + entries_per_row_up] = Ax[jj];
			entries_per_row_up++;	
			}			
		}
		Aj_lo[row_offsets_lo + entries_per_row_lo] = row;  // col_indice of diagonal
		Ax_lo[row_offsets_lo + entries_per_row_lo] = 1.0;  // value of diagonal		
    }
}

template <typename Matrix, typename IndexType>
void triangular_counts(const Matrix&    A, 
						   IndexType* nnz_lo, 
                           IndexType* nnz_up,
						   cusp::csr_format)
{
    const size_t BLOCK_SIZE = 256;
    const size_t MAX_BLOCKS = cusp::detail::device::arch::max_active_blocks(triangular_counts_kernel<IndexType>, BLOCK_SIZE, (size_t) 0);
    const size_t NUM_BLOCKS = std::min<size_t>(MAX_BLOCKS, cusp::detail::device::DIVIDE_INTO(A.num_rows, BLOCK_SIZE));

    triangular_counts_kernel<IndexType> <<<NUM_BLOCKS, BLOCK_SIZE>>> 
        (A.num_rows,
         thrust::raw_pointer_cast(&A.row_offsets[0]),
         thrust::raw_pointer_cast(&A.column_indices[0]),
         nnz_lo,
		 nnz_up);
}

template <typename Matrix>
void triangular_fill(const Matrix&    A, 
						   Matrix&    A_lo,
						   Matrix&    A_up,
						   cusp::csr_format)
{
    typedef typename Matrix::index_type  IndexType;
    typedef typename Matrix::value_type  ValueType;
	
    const size_t BLOCK_SIZE = 256;
    const size_t MAX_BLOCKS = cusp::detail::device::arch::max_active_blocks(triangular_fill_kernel<IndexType,ValueType>, BLOCK_SIZE, (size_t) 0);
    const size_t NUM_BLOCKS = std::min<size_t>(MAX_BLOCKS, cusp::detail::device::DIVIDE_INTO(A.num_rows, BLOCK_SIZE));

    triangular_fill_kernel<IndexType,ValueType> <<<NUM_BLOCKS, BLOCK_SIZE>>> 
        (A.num_rows,
         thrust::raw_pointer_cast(&A.row_offsets[0]),
         thrust::raw_pointer_cast(&A.column_indices[0]),
         thrust::raw_pointer_cast(&A.values[0]),
         thrust::raw_pointer_cast(&A_lo.row_offsets[0]),		 
         thrust::raw_pointer_cast(&A_lo.column_indices[0]),
         thrust::raw_pointer_cast(&A_lo.values[0]),	
         thrust::raw_pointer_cast(&A_up.row_offsets[0]),				 
         thrust::raw_pointer_cast(&A_up.column_indices[0]),
         thrust::raw_pointer_cast(&A_up.values[0]));
}

template <typename Matrix, typename IndexType>
void triangular_counts(const Matrix&    A, 
						   IndexType* nnz_lo, 
                           IndexType* nnz_up)
{
    CUSP_PROFILE_SCOPED();
	
	triangular_counts(A, nnz_lo, nnz_up, typename Matrix::format( ));
}

template <typename Matrix>
void triangular_fill(const Matrix&    A, 
						   Matrix&    A_lo,
						   Matrix&    A_up)
{
    CUSP_PROFILE_SCOPED();
	
	triangular_fill(A, A_lo, A_up, typename Matrix::format( ));
}
						   
} // end namespace detail
} // end namespace cusp
