/*
 *  Author: Li Luo 
 *  Email:  li.luo@siat.ac.cn
 */
    

#include <cusp/format.h>
#include <cusp/detail/format_utils.h>
#include <cusp/array1d.h>
#include <cusp/detail/triangle.h>
#include <thrust/scan.h>
#include <thrust/iterator/constant_iterator.h>
#include <assert.h>

namespace cusp
{
namespace detail
{
    // ADD<T> computes the sum of a number f(x) -> x+alpha
	// the name "ADD" is chosen different from "PLUS" from blas
    template <typename T>
        struct ADD : public thrust::unary_function<T,T>
        {
            T alpha;

            ADD(T _alpha)
                : alpha(_alpha) {}
				
            __host__ __device__
                T operator()(T x)
                { 
                    return x + alpha;
                }
        };

    // INV<T> computes the inverse of a number f(x) -> a/x
    template <typename T>
        struct INV : public thrust::unary_function<T,T>
        {
            T alpha;

            INV(T _alpha)
                : alpha(_alpha) {}
				
            __host__ __device__
                T operator()(T x)
                { 
                    return alpha / x;
                }
        };
		
template <typename IndexType, typename ValueType>
__global__ void subscript_kernel(const IndexType n, const IndexType *indices, const ValueType *x, ValueType *y)
{
  const IndexType tx = threadIdx.x;
  const IndexType num_threads = blockDim.x * gridDim.x;
  const IndexType tid = blockIdx.x * blockDim.x + tx;
  for(IndexType i = tid; i < n ; i+= num_threads)
    {
      y[i] = x[indices[i]];
    }
}

template <typename IndexType, typename ValueType>
void subscript(const IndexType n, 
			   const IndexType *indices, 
			   const ValueType *x, 
					 ValueType *y,
					 cusp::device_memory)
{
    const size_t THREADS_PER_BLOCK  = 128;

    const size_t MAX_BLOCKS = cusp::detail::device::arch::max_active_blocks(subscript_kernel<IndexType, ValueType>, THREADS_PER_BLOCK, (size_t) 0);
    const size_t NUM_BLOCKS = std::min<size_t>(MAX_BLOCKS, cusp::detail::device::DIVIDE_INTO(n, THREADS_PER_BLOCK));
  
	subscript_kernel<IndexType, ValueType><<<NUM_BLOCKS, THREADS_PER_BLOCK>>>(n, indices, x, y);

}

template <typename IndexType, typename ValueType>
void subscript(const IndexType n, 
			   const IndexType *indices, 
			   const ValueType *x, 
					 ValueType *y,
					 cusp::host_memory)
{
	for (int i=0; i<n; i++)
	{
      y[i] = x[indices[i]];
	}
}
					 
template <typename Matrix>
void extract_triangle(const Matrix& A, Matrix& A_lo, Matrix& A_up, cusp::csr_format)
{
    typedef typename Matrix::index_type   IndexType;
    typedef typename Matrix::value_type   ValueType;
    typedef typename Matrix::memory_space MemorySpace;
	
	cusp::array1d<IndexType, MemorySpace> nnz_lo(A.num_rows, 0);
	cusp::array1d<IndexType, MemorySpace> nnz_up(A.num_rows, 0);	
									
	// count the number of non-zero entries in each row of lower and upper triangle								
	triangular_counts<Matrix, IndexType>(A, 
						   thrust::raw_pointer_cast(&*nnz_lo.begin()), 
                           thrust::raw_pointer_cast(&*nnz_up.begin()));		

	// create row_offsets of lower triangle from nnz_lo				
	A_lo.row_offsets.resize(A.num_rows+1);
	thrust::fill(A_lo.row_offsets.begin(), A_lo.row_offsets.end(), 0);
    thrust::inclusive_scan(nnz_lo.begin(), nnz_lo.end(), A_lo.row_offsets.begin()+1);	

	// count the total number of non-zeros of A_lo
	IndexType num_nz_lo = thrust::reduce(nnz_lo.begin(), nnz_lo.end());	

	// create row_offsets of upper triangle from nnz_up				
	A_up.row_offsets.resize(A.num_rows+1);
	thrust::fill(A_up.row_offsets.begin(), A_up.row_offsets.end(), 0);	
    thrust::inclusive_scan(nnz_up.begin(), nnz_up.end(), A_up.row_offsets.begin()+1);	

	// count the total number of non-zeros of A_up
	IndexType num_nz_up = thrust::reduce(nnz_up.begin(), nnz_up.end());

	// resize the num_entries
	A_lo.resize(A.num_rows, A.num_cols, num_nz_lo);
	A_up.resize(A.num_rows, A.num_cols, num_nz_up);

	// fill in the triangle A_lo and A_up
	triangular_fill<Matrix>(A, A_lo, A_up);
}

template <typename Matrix>
void extract_triangle(const Matrix& A, Matrix& A_lo, Matrix& A_up)
{
    CUSP_PROFILE_SCOPED();

    // dispatch on matrix format
    extract_triangle(A, A_lo, A_up, typename Matrix::format( ));
}

template <typename IndexArray>
void indices_to_nnz(const IndexArray& indices, IndexArray& nnz)
{
    CUSP_PROFILE_SCOPED();

    typedef typename IndexArray::value_type IndexType;
	
	IndexType n = nnz.size();
	IndexArray block_list(n);
    IndexArray block_count(n);

	// build temporary arrays
    typename IndexArray::iterator last_block_list_iter =  
    thrust::reduce_by_key(indices.begin(), indices.end(), thrust::make_constant_iterator<IndexType>(1), block_list.begin(), block_count.begin()).first;
	
	// gather block_count into bk from block_list (slots not written to will be 0)
    IndexType num_block_counted = thrust::distance(block_list.begin(), last_block_list_iter);
    thrust::scatter(block_count.begin(), block_count.begin() + num_block_counted, block_list.begin(), nnz.begin()); 
}

template <typename IndexArray>
void nnz_to_offsets(const IndexArray& nnz, IndexArray& offsets)
{
    CUSP_PROFILE_SCOPED();
	
	thrust::inclusive_scan(nnz.begin(), nnz.end(), offsets.begin()+1);
}

template <typename IndexArray>
void offsets_to_nnz(const IndexArray& offsets, IndexArray& nnz)
{
    CUSP_PROFILE_SCOPED();
    typedef typename IndexArray::value_type IndexType;
	
	thrust::transform(offsets.begin()+1, offsets.begin()+nnz.size()+1, offsets.begin(), nnz.begin(), thrust::minus<IndexType>());	
}

template <typename Matrix>
void sort_by_row_and_column(Matrix& mat, cusp::coo_format)
{
	sort_by_row_and_column(mat.row_indices, mat.column_indices, mat.values);
}

template <typename Matrix>
void sort_by_row_and_column(Matrix& mat, cusp::csr_format)
{
    typedef typename Matrix::index_type   IndexType;
    typedef typename Matrix::memory_space MemorySpace;
	
	cusp::array1d<IndexType, MemorySpace> row_indices_temp(mat.num_entries,0);	
	cusp::detail::offsets_to_indices(mat.row_offsets, row_indices_temp);
	
	sort_by_row_and_column(row_indices_temp, mat.column_indices, mat.values);
	
	cusp::detail::indices_to_offsets(row_indices_temp, mat.row_offsets);
}
	
template <typename Matrix>
void sort_by_row_and_column(Matrix& mat)
{
    CUSP_PROFILE_SCOPED();
	sort_by_row_and_column(mat, typename Matrix::format( ));
}	

// to_mkl: if true,  transform to one-based MKL matrix. 
//         if false, output the zero-based cusp csr matrix of host_memory
template <typename IndexType, typename ValueType, typename MemorySpace>
void msr_to_csr(const IndexType num_rows,
			 const IndexType num_cols,
			 const IndexType num_entries, 
			 const IndexType *ju,
			 const IndexType *jau,			 
			 const ValueType *au,
			 bool  to_mkl,
			 cusp::csr_matrix<IndexType,ValueType,MemorySpace>& A_csr, cusp::host_memory)
{
	// resize A_csr
	// the real number of entries must less or equal to the maxmum number of entries	
	IndexType num_entries_msr = jau[num_rows] - jau[0] + num_rows;
	assert(num_entries_msr <= num_entries);	
	A_csr.resize(num_rows, num_cols, num_entries_msr);	
	thrust::fill(A_csr.row_offsets.begin(), A_csr.row_offsets.end(), IndexType(0));	
	//thrust::fill(A_csr.column_indices.begin(), A_csr.column_indices.end(), IndexType(0));		
	//thrust::fill(A_csr.values.begin(), A_csr.values.end(), ValueType(0));	
	
	cusp::array1d<IndexType, MemorySpace> ju_(num_rows);
	cusp::array1d<IndexType, MemorySpace> jau_(num_entries_msr+10);	
	cusp::array1d<ValueType, MemorySpace> au_(num_entries_msr+10);	
	
	// temporary data, copy all data from ju_, jau_, au_
	// num_entries maybe larger than num_entries_msr
	// valid size of jau and au is num_entries_msr+1
	// copy more for safety
	thrust::copy(ju, ju+num_rows, ju_.begin());	
	thrust::copy(jau, jau+num_entries_msr+10, jau_.begin());	
	thrust::copy(au, au+num_entries_msr+10, au_.begin());	
	
	cusp::array1d<IndexType, MemorySpace> nnz(num_rows, 0);
	cusp::detail::offsets_to_nnz(jau_,nnz);
	// plus the diagonal elements
	thrust::transform(nnz.begin(), nnz.end(), nnz.begin(), cusp::detail::ADD<IndexType>(1));
	cusp::detail::nnz_to_offsets(nnz, A_csr.row_offsets);	
	
	// count the offsets of diagonl elements in each row
	cusp::array1d<IndexType, MemorySpace> nnz_lo(num_rows, 0);
	thrust::transform(ju_.begin(), ju_.begin()+num_rows, jau_.begin(), nnz_lo.begin(), thrust::minus<IndexType>());
	
	cusp::array1d<IndexType, MemorySpace> diagonal_offsets(num_rows, 0);
	thrust::transform(A_csr.row_offsets.begin(), A_csr.row_offsets.begin()+num_rows, nnz_lo.begin(), diagonal_offsets.begin(), thrust::plus<IndexType>());

	// permute the diagonal elements from the first [1:n] position of the au array(inverse) in MSR format to the position of CSR format
	thrust::scatter_if(thrust::make_transform_iterator(au_.begin(), cusp::detail::INV<ValueType>(1)),
					   thrust::make_transform_iterator(au_.begin()+num_rows, cusp::detail::INV<ValueType>(1)),
					   diagonal_offsets.begin(),
                       thrust::make_transform_iterator(
                                thrust::make_zip_iterator( thrust::make_tuple( diagonal_offsets.begin(), diagonal_offsets.begin()+1 ) ),
                                cusp::detail::empty_row_functor<IndexType>()),
					   A_csr.values.begin());	
	// one-based
	thrust::scatter_if(thrust::make_counting_iterator(1),
					   thrust::make_counting_iterator(num_rows+1),
					   diagonal_offsets.begin(),
                       thrust::make_transform_iterator(
                                thrust::make_zip_iterator( thrust::make_tuple( diagonal_offsets.begin(), diagonal_offsets.begin()+1 ) ),
                                cusp::detail::empty_row_functor<IndexType>()), 
					   A_csr.column_indices.begin());
	
	// permute the global au indices and jau indices from MSR to CSR	
	IndexType num_entries_without_diagonals = jau[num_rows] - jau[0];
	cusp::array1d<IndexType, MemorySpace> indices_permuted(num_entries_without_diagonals, 0);
	// extend the permutated global numbering with zero filling			
	thrust::scatter_if(thrust::make_counting_iterator(1), 
				    thrust::make_counting_iterator(num_rows+1),
					thrust::make_transform_iterator(ju_.begin(), cusp::detail::ADD<IndexType>(-jau[0])),
					thrust::make_transform_iterator(
                              thrust::make_zip_iterator( thrust::make_tuple( 
							  thrust::make_transform_iterator(ju_.begin(), cusp::detail::ADD<IndexType>(-jau[0])), 
							  thrust::make_transform_iterator(ju_.begin()+1, cusp::detail::ADD<IndexType>(-jau[0])) ) ),
                              cusp::detail::empty_row_functor<IndexType>()), 
					indices_permuted.begin());
	// extend
	thrust::inclusive_scan(indices_permuted.begin(), indices_permuted.end(), indices_permuted.begin(), thrust::maximum<IndexType>());  
	// add
	thrust::transform(indices_permuted.begin(), indices_permuted.end(), thrust::make_counting_iterator(0), indices_permuted.begin(), thrust::plus<IndexType>());	
	
	// permute by indices_permuted
	thrust::scatter(au_.begin()+jau[0]-1, au_.begin()+jau[num_rows]-1, indices_permuted.begin(), A_csr.values.begin());	
	thrust::scatter(jau_.begin()+jau[0]-1, jau_.begin()+jau[num_rows]-1, indices_permuted.begin(), A_csr.column_indices.begin());	
	
	if (to_mkl) {
	// transform from one-based to zero-based
	thrust::transform(A_csr.column_indices.begin(), A_csr.column_indices.end(), A_csr.column_indices.begin(), cusp::detail::ADD<IndexType>(-1));	
	
	// sort
	cusp::detail::sort_by_row_and_column(A_csr);
	
	// transform from zero-based to one-based
	thrust::transform(A_csr.row_offsets.begin(), A_csr.row_offsets.end(), A_csr.row_offsets.begin(), cusp::detail::ADD<IndexType>(1));	
	thrust::transform(A_csr.column_indices.begin(), A_csr.column_indices.end(), A_csr.column_indices.begin(), cusp::detail::ADD<IndexType>(1));	
	} else {
	// transform from one-based to zero-based
	thrust::transform(A_csr.column_indices.begin(), A_csr.column_indices.end(), A_csr.column_indices.begin(), cusp::detail::ADD<IndexType>(-1));	

	// sort
	cusp::detail::sort_by_row_and_column(A_csr);
	}
}

// TODO: fix the bug: thrust::system::system_error
template <typename IndexType, typename ValueType, typename MemorySpace>
void msr_to_csr(const IndexType num_rows,
			 const IndexType num_cols,
			 const IndexType num_entries, 
			 const IndexType *ju,
			 const IndexType *jau,			 
			 const ValueType *au,
			 bool  to_mkl,
			 cusp::csr_matrix<IndexType,ValueType,MemorySpace>& A_csr, cusp::device_memory)
{
	// resize A_csr
	// the real number of entries must less or equal to the maxmum number of entries	
	IndexType num_entries_msr = jau[num_rows] - jau[0] + num_rows;
	assert(num_entries_msr <= num_entries);	
	A_csr.resize(num_rows, num_cols, num_entries_msr);	
	thrust::fill(A_csr.row_offsets.begin(), A_csr.row_offsets.end(), IndexType(0));	
	//thrust::fill(A_csr.column_indices.begin(), A_csr.column_indices.end(), IndexType(0));		
	//thrust::fill(A_csr.values.begin(), A_csr.values.end(), ValueType(0));	
	
	cusp::array1d<IndexType, MemorySpace> ju_(num_rows);
	cusp::array1d<IndexType, MemorySpace> jau_(num_entries_msr+10);	
	cusp::array1d<ValueType, MemorySpace> au_(num_entries_msr+10);	
	thrust::copy(ju, ju+num_rows, ju_.begin());	
	thrust::copy(jau, jau+num_entries_msr+10, jau_.begin());	
	thrust::copy(au, au+num_entries_msr+10, au_.begin());	
	
	cusp::array1d<IndexType, MemorySpace> nnz(num_rows, 0);
	cusp::detail::offsets_to_nnz(jau_,nnz);
	// plus the diagonal elements
	thrust::transform(nnz.begin(), nnz.end(), nnz.begin(), cusp::detail::ADD<IndexType>(1));	
	cusp::detail::nnz_to_offsets(nnz, A_csr.row_offsets);	
	
	// count the offsets of diagonl elements
	cusp::array1d<IndexType, MemorySpace> nnz_lo(num_rows, 0);
	thrust::transform(ju_.begin(), ju_.begin()+num_rows, jau_.begin(), nnz_lo.begin(), thrust::minus<IndexType>());
	
	cusp::array1d<IndexType, MemorySpace> diagonal_offsets(num_rows, 0);
	thrust::transform(A_csr.row_offsets.begin(), A_csr.row_offsets.begin()+num_rows, nnz_lo.begin(), diagonal_offsets.begin(), thrust::plus<IndexType>());

	// permute the diagonal elements from the first [1:n] position of the au array(inverse) in MSR format to the position of CSR format
	thrust::scatter_if(thrust::make_transform_iterator(au_.begin(), cusp::detail::INV<ValueType>(1)),
					   thrust::make_transform_iterator(au_.begin()+num_rows, cusp::detail::INV<ValueType>(1)),
					   diagonal_offsets.begin(),
                       thrust::make_transform_iterator(
                                thrust::make_zip_iterator( thrust::make_tuple( diagonal_offsets.begin(), diagonal_offsets.begin()+1 ) ),
                                cusp::detail::empty_row_functor<IndexType>()),
					   A_csr.values.begin());	
	// one-based
	thrust::scatter_if(thrust::make_counting_iterator(1),
					   thrust::make_counting_iterator(num_rows+1),
					   diagonal_offsets.begin(),
                       thrust::make_transform_iterator(
                                thrust::make_zip_iterator( thrust::make_tuple( diagonal_offsets.begin(), diagonal_offsets.begin()+1 ) ),
                                cusp::detail::empty_row_functor<IndexType>()), 
					   A_csr.column_indices.begin());
	
	// permute the global au indices and jau indices from MSR to CSR	
	IndexType num_entries_without_diagonals = jau[num_rows] - jau[0];
	cusp::array1d<IndexType, MemorySpace> indices_permuted(num_entries_without_diagonals, 0);
	// extend the permutated global numbering with zero filling			
	thrust::scatter_if(thrust::make_counting_iterator(1), 
				    thrust::make_counting_iterator(num_rows+1),
					thrust::make_transform_iterator(ju_.begin(), cusp::detail::ADD<IndexType>(-jau[0])),
					thrust::make_transform_iterator(
                              thrust::make_zip_iterator( thrust::make_tuple( 
							  thrust::make_transform_iterator(ju_.begin(), cusp::detail::ADD<IndexType>(-jau[0])), 
							  thrust::make_transform_iterator(ju_.begin()+1, cusp::detail::ADD<IndexType>(-jau[0])) ) ),
                              cusp::detail::empty_row_functor<IndexType>()), 
					indices_permuted.begin());
	// extend
	thrust::inclusive_scan(indices_permuted.begin(), indices_permuted.begin()+num_entries_without_diagonals, indices_permuted.begin(), thrust::maximum<IndexType>());  
	// add
	thrust::transform(indices_permuted.begin(), indices_permuted.begin()+num_entries_without_diagonals, thrust::make_counting_iterator(0), indices_permuted.begin(), thrust::plus<IndexType>());	
	
	// permute by indices_permuted
	thrust::scatter(au_.begin()+jau[0]-1, au_.begin()+jau[num_rows]-1, indices_permuted.begin(), A_csr.values.begin());	
	thrust::scatter(jau_.begin()+jau[0]-1, jau_.begin()+jau[num_rows]-1, indices_permuted.begin(), A_csr.column_indices.begin());	
					
	// transform from one-based to zero-based
	thrust::transform(A_csr.column_indices.begin(), A_csr.column_indices.end(), A_csr.column_indices.begin(), cusp::detail::ADD<IndexType>(-1));

	// sort
	cusp::detail::sort_by_row_and_column(A_csr);
}

template <typename IndexType, typename ValueType, typename MatrixType>
void msr_to_csr(const IndexType num_rows,
			 const IndexType num_cols,
			 const IndexType num_entries, 
			 const IndexType *ju,
			 const IndexType *jau,			 
			 const ValueType *au,
			 bool  to_mkl,
			 MatrixType& A_csr)
{
    CUSP_PROFILE_SCOPED();	
    typedef typename MatrixType::memory_space MemorySpace;

	msr_to_csr(num_rows, num_cols, num_entries, ju, jau, au, to_mkl, A_csr, typename MatrixType::memory_space( ));
}

template <typename IndexType, typename ValueType, typename Matrix>
void setup_mkl_matrix(cusp::csr_matrix<IndexType,ValueType,cusp::host_memory>& dst, Matrix& src, cusp::coo_format)
{
    typedef typename Matrix::memory_space MemorySpace;	
	
	// convert COO row_indices to CSR row_offsets
	cusp::array1d<IndexType,MemorySpace> row_offsets_temp(src.num_rows+1);	
	cusp::detail::indices_to_offsets(src.row_indices, row_offsets_temp);
	
	// CSR format used by MKL DSS on host
	dst.resize(src.num_rows, src.num_cols, src.num_entries);
	// MKL is 1-based, transform row_offsets and column_indices to be plusing one
	thrust::copy(thrust::make_transform_iterator(row_offsets_temp.begin(), cusp::detail::ADD<IndexType>(1)),
	thrust::make_transform_iterator(row_offsets_temp.end(), cusp::detail::ADD<IndexType>(1)),
	dst.row_offsets.begin());
	thrust::copy(thrust::make_transform_iterator(src.column_indices.begin(), cusp::detail::ADD<IndexType>(1)),
	thrust::make_transform_iterator(src.column_indices.end(), cusp::detail::ADD<IndexType>(1)),
	dst.column_indices.begin());
	thrust::copy(src.values.begin(), src.values.end(), dst.values.begin());
}

template <typename IndexType, typename ValueType, typename Matrix>
void setup_mkl_matrix(cusp::csr_matrix<IndexType,ValueType,cusp::host_memory>& dst, Matrix& src, cusp::csr_format)
{
	// CSR format used by MKL DSS on host
	dst.resize(src.num_rows, src.num_cols, src.num_entries);
	// MKL is 1-based, transform row_offsets and column_indices to be plusing one
	thrust::copy(thrust::make_transform_iterator(src.row_offsets.begin(), cusp::detail::ADD<IndexType>(1)),
	thrust::make_transform_iterator(src.row_offsets.end(), cusp::detail::ADD<IndexType>(1)),
	dst.row_offsets.begin());
	thrust::copy(thrust::make_transform_iterator(src.column_indices.begin(), cusp::detail::ADD<IndexType>(1)),
	thrust::make_transform_iterator(src.column_indices.end(), cusp::detail::ADD<IndexType>(1)),
	dst.column_indices.begin());
	thrust::copy(src.values.begin(), src.values.end(), dst.values.begin());
}

// bool sort: if true, sort by row and column that needed by MKL
//            if the input matrix has already been sorted, the bool can be false
template <typename IndexType, typename ValueType, typename Matrix>
void setup_mkl_matrix(cusp::csr_matrix<IndexType,ValueType,cusp::host_memory>& dst, Matrix& src)
{
  CUSP_PROFILE_SCOPED();
  setup_mkl_matrix(dst, src, typename Matrix::format( ));
}

} // end namespace detail
} // end namespace cusp

