/*
 *  Copyright 2008-2009 NVIDIA Corporation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  Author: Li Luo 
 *  Email:  li.luo@siat.ac.cn 
 */

/*! \file multicoloring.h
 *  \brief Multicoloring of a matrix.
 */

#include <iostream>
#include <set>
#include <thrust/functional.h>
#include <thrust/transform.h>
#include <cusp/coo_matrix.h>
#include <cusp/print.h>


#pragma once

namespace cusp {
namespace graph {
namespace detail {

namespace host {

struct adjacency_matrix
{
    std::vector<int> degrees;
    std::vector<int> neighbours;

    adjacency_matrix() {}
    inline size_t num_vertices( ) const { return degrees.size( ); }
    inline void resize( int n, int num_neighbours ) 
    {
        degrees.resize( n );
        neighbours.resize( num_neighbours );
    }
};

struct edge
{
    int u, v;
    edge( int uu, int vv ) : u( uu ), v( vv ) {}
    bool operator < ( const edge &e ) const { return u < e.u || ( u == e.u && v < e.v ); }
};

template< typename Array >
bool dbg_coloring_is_valid( const adjacency_matrix &g, const Array &coloring )
{
    for( int u = 0, i = 0 ; u < g.num_vertices( ) ; ++u )
        for( int k = 0 ; k < g.degrees[u] ; ++k, ++i )
            if( coloring[u] == coloring[g.neighbours[i]] )
                return false;
    return true;
}

template< typename Matrix >
void build_adjacency_matrix( const Matrix &a, adjacency_matrix &g, cusp::coo_format )
{
    typedef std::set<edge> edge_set;
    typedef edge_set::iterator edge_set_iterator;
    edge_set edges;
    size_t order = std::max( a.num_rows, a.num_cols );
    for( int i = 0 ; i != a.num_entries ; ++i ) 
    {
        int u = a.row_indices[i], v = a.column_indices[i];
        if( u == v )
            continue;
        edges.insert( edge( u, v ) );
        edges.insert( edge( v, u ) );
    }
    g.resize( order, edges.size( ) );
    int u = 0, i = 0, degree = 0;
    for( edge_set_iterator it = edges.begin( ) ; it != edges.end( ) ; ++it )
    {
        if( it->u != u ) 
        {
            g.degrees[u++] = degree;
            degree = 0;
            assert( u == it->u );
        }
        g.neighbours[i++] = it->v;
        degree++;
    }
    g.degrees[u] = degree;
}

template< typename Matrix, typename Vector, typename Format >
void greedy_multicoloring( const Matrix &a, Vector &coloring, Format format )
{
    adjacency_matrix g;
    build_adjacency_matrix( a, g, format );
    std::fill( coloring.begin( ), coloring.end( ), std::numeric_limits<int>::max( ) );
    // For each vertex, find the color s.t. c = min{ k > 0 : color[v] != k, where (u,v) in E(A) }.
    coloring[0] = 0;
    int i = g.degrees[0];
    for( int u = 1, end = g.num_vertices( ) ; u < end ; ++u )
    {
        typedef std::set<int> color_set;
        typedef color_set::iterator color_set_it;
        color_set colors;
        bool done = false;
        for( int k = 0 ; !done && k < g.degrees[u] ; ++k )
        {
            assert( i+k < g.neighbours.size( ) );
            int v = g.neighbours[i+k];
            if( v < u )
                colors.insert( coloring[v] );
            done = v >= u;
        }
        i += g.degrees[u];

        // The set of colors is already sorted.
        int color = 0;
        color_set_it it = colors.begin( );
        for( ; it != colors.end( ) && *it == color ; ++it, ++color )
            ;
        assert( u < coloring.size( ) );
        coloring[u] = color;
    }
    //assert( dbg_coloring_is_valid( g, coloring ) );
}

} // namespace host

// naive 5pt red black coloring in 2D 
template <typename IndexType>
struct red_black_multicoloring_functor
{

    red_black_multicoloring_functor() {}

    template <typename Tuple>
    __host__ __device__
    IndexType operator()(const Tuple& t)
    {
        const IndexType num = thrust::get<0>(t);
        const IndexType m = thrust::get<1>(t); // number of columns
        const IndexType n = thrust::get<2>(t); // number of rows
        const IndexType dof = thrust::get<3>(t); // degree of freedom

		const IndexType nd = num / dof;  // scalar mesh point
		const IndexType j = nd / m;      // row index
		const IndexType i = nd % m;		 // col index
        return (j%2)^(i%2);
    }
};

// Run the computation on the host.
template< typename Matrix, typename Array1d >
void greedy_multicoloring( const Matrix &a, Array1d &coloring, cusp::host_memory )
{
    host::greedy_multicoloring( a, coloring, typename Matrix::format( ) );
}

// The device path redirects the computation to the host (for the moment).
template< typename Matrix, typename Array1d >
void greedy_multicoloring(const Matrix &a, Array1d &coloring, cusp::device_memory )
{
    typedef cusp::array1d<typename Array1d::value_type, cusp::host_memory> host_array1d;
	cusp::coo_matrix<int, typename Matrix::value_type, cusp::host_memory> h_a(a);

    host_array1d h_coloring(coloring.size(), 0);
    host::greedy_multicoloring( h_a, h_coloring, typename Matrix::format( ) );
	coloring = h_coloring;
}

template< typename Array1d, typename IndexType >
void build_parameters_from_coloring(Array1d &coloring, Array1d &p, Array1d &pt,
 cusp::array1d<IndexType, cusp::host_memory> &bk, cusp::array1d<IndexType, cusp::host_memory> &ptr,
 IndexType &max_block_size, const IndexType n, const IndexType num_coloring, cusp::device_memory)
{	
	p.resize(n);
	thrust::sequence(p.begin(), p.end());
	// sort so that entries of the same color are adjacent, and in ascending order
	// p[i] is the original natural number of i in sorted number
	thrust::stable_sort_by_key(coloring.begin(), coloring.end(), p.begin());
	
	//cusp::print(coloring);
	
	Array1d p_tmp(p);
	pt.resize(n);	
	thrust::sequence(pt.begin(), pt.end());
	// pt[i] is the objective number of i in original number
	thrust::stable_sort_by_key(p_tmp.begin(), p_tmp.end(), pt.begin());
	
	Array1d block_list(num_coloring);
    Array1d block_count(num_coloring);

	// build temporary arrays
    typename Array1d::iterator last_block_list_iter =  
    thrust::reduce_by_key(coloring.begin(), coloring.end(), thrust::make_constant_iterator<IndexType>(1), block_list.begin(), block_count.begin()).first;
	
	// gather block_count into bk from block_list (slots not written to will be 0)
	Array1d bk_temp(num_coloring);
    IndexType num_block_counted = thrust::distance(block_list.begin(), last_block_list_iter);
    thrust::scatter(block_count.begin(), block_count.begin() + num_block_counted, block_list.begin(), bk_temp.begin()); 
	// copy from device to host
	bk = bk_temp;
	
	// calculate the size of the largest block
    max_block_size = thrust::max_element(bk_temp.begin(), bk_temp.end())[0];

    // fill the block pointers 
	Array1d ptr_temp(num_coloring+1, 0);
    thrust::inclusive_scan(bk_temp.begin(), bk_temp.end(), ptr_temp.begin()+1);
	// copy from device to host	
	ptr = ptr_temp;
}

template< typename Array1d, typename IndexType >
void build_parameters_from_coloring(Array1d &coloring, Array1d &p, Array1d &pt,
 cusp::array1d<IndexType, cusp::host_memory> &bk, cusp::array1d<IndexType, cusp::host_memory> &ptr,
 IndexType &max_block_size, const IndexType n, const IndexType num_coloring, cusp::host_memory)
{	
	p.resize(n);
	thrust::sequence(p.begin(), p.end());
	// sort so that entries of the same color are adjacent, and in ascending order
	// p[i] is the original natural number of i in sorted number
	thrust::stable_sort_by_key(coloring.begin(), coloring.end(), p.begin());
	
	//cusp::print(coloring);
	
	Array1d p_tmp(p);
	pt.resize(n);	
	thrust::sequence(pt.begin(), pt.end());
	// pt[i] is the objective number of i in original number
	thrust::stable_sort_by_key(p_tmp.begin(), p_tmp.end(), pt.begin());
	
	Array1d block_list(num_coloring);
    Array1d block_count(num_coloring);

	// build temporary arrays
    typename Array1d::iterator last_block_list_iter =  
    thrust::reduce_by_key(coloring.begin(), coloring.end(), thrust::make_constant_iterator<IndexType>(1), block_list.begin(), block_count.begin()).first;
	
	// gather block_count into bk from block_list (slots not written to will be 0)
	bk.resize(num_coloring);
    IndexType num_block_counted = thrust::distance(block_list.begin(), last_block_list_iter);
    thrust::scatter(block_count.begin(), block_count.begin() + num_block_counted, block_list.begin(), bk.begin()); 
	
	// calculate the size of the largest block
    max_block_size = thrust::max_element(bk.begin(), bk.end())[0];

    // fill the block pointers 
	ptr.resize(num_coloring+1);
	thrust::fill(ptr.begin(), ptr.end(), 0);	
    thrust::inclusive_scan(bk.begin(), bk.end(), ptr.begin()+1);
}

} // namespace detail

template< typename Matrix, typename Array1d >
void greedy_multicoloring(const  Matrix &a, Array1d &coloring )
{
    CUSP_PROFILE_SCOPED();
	
    detail::greedy_multicoloring( a, coloring, typename Matrix::memory_space( ) );
}

// n: size of all, equals to the row numbr of matrix
// num_coloring: number of colors
template< typename Array1d, typename IndexType >
void get_size( const Array1d &coloring, IndexType &n, IndexType &num_coloring )
{
    CUSP_PROFILE_SCOPED();
	
    n = coloring.size();
	num_coloring = thrust::max_element(coloring.begin(), coloring.end())[0]+1;
	std::cout << "number of colors: " << num_coloring << std::endl;
}

template< typename Array1d, typename IndexType >
void build_parameters_from_coloring(Array1d &coloring, Array1d &p, Array1d &pt,
 cusp::array1d<IndexType, cusp::host_memory> &bk, cusp::array1d<IndexType, cusp::host_memory> &ptr,
 IndexType &max_block_size, const IndexType n, const IndexType num_coloring )
{	
    CUSP_PROFILE_SCOPED();
	detail::build_parameters_from_coloring(coloring, p, pt, bk, ptr, max_block_size, n, num_coloring, typename Array1d::memory_space( ));
}

template< typename Array1d, typename IndexType >
void red_black_multicoloring(Array1d &coloring, const IndexType m, const IndexType n, const IndexType dof)
{
    CUSP_PROFILE_SCOPED();	

	thrust::sequence(coloring.begin(), coloring.end());
	
    thrust::transform(thrust::make_zip_iterator(thrust::make_tuple(coloring.begin(), thrust::make_constant_iterator(m), thrust::make_constant_iterator(n), thrust::make_constant_iterator(dof))),
                      thrust::make_zip_iterator(thrust::make_tuple(coloring.end(),   thrust::make_constant_iterator(m), thrust::make_constant_iterator(n), thrust::make_constant_iterator(dof))),
                      coloring.begin(),
                      detail::red_black_multicoloring_functor<IndexType>());	
}

} // namespace graph
} // namespace cusp

