/*
 *  Copyright 2008-2009 NVIDIA Corporation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

#pragma once

#include <cusp/detail/config.h>

#include <cusp/blas.h>

#include <limits>
#include <iostream>
#include <iomanip>

// Classes to monitor iterative solver progress, check for convergence, etc.
// Follows the implementation of Iteration in the ITL:
//   http://www.osl.iu.edu/research/itl/doc/Iteration.html

namespace cusp
{
/*! \addtogroup iterative_solvers Iterative Solvers
 *  \addtogroup monitors Monitors
 *  \ingroup iterative_solvers
 *  \{
 */

/*! \p default_monitor : Implements standard convergence criteria
 * and reporting for iterative solvers.
 *
 * \tparam Real real-valued type (e.g. \c float or \c double).
 *
 *  The following code snippet demonstrates how to configure
 *  the \p default_monitor and use it with an iterative solver.
 *
 *  \code
 *  #include <cusp/csr_matrix.h>
 *  #include <cusp/monitor.h>
 *  #include <cusp/krylov/cg.h>
 *  #include <cusp/gallery/poisson.h>
 *  
 *  int main(void)
 *  {
 *      // create an empty sparse matrix structure (CSR format)
 *      cusp::csr_matrix<int, float, cusp::device_memory> A;
 *  
 *      // initialize matrix
 *      cusp::gallery::poisson5pt(A, 10, 10);
 *  
 *      // allocate storage for solution (x) and right hand side (b)
 *      cusp::array1d<float, cusp::device_memory> x(A.num_rows, 0);
 *      cusp::array1d<float, cusp::device_memory> b(A.num_rows, 1);
 *  
 *      // set stopping criteria:
 *      //  iteration_limit    = 100
 *      //  relative_tolerance = 1e-6
 *      cusp::default_monitor<float> monitor(b, 100, 1e-6);
 *  
 *      // solve the linear system A x = b
 *      cusp::krylov::cg(A, x, b, monitor);
 *  
 *      // report solver results
 *      if (monitor.converged())
 *      {
 *          std::cout << "Solver converged to " << monitor.relative_tolerance() << " relative tolerance";
 *          std::cout << " after " << monitor.iteration_count() << " iterations" << std::endl;
 *      }
 *      else
 *      {
 *          std::cout << "Solver reached iteration limit " << monitor.iteration_limit() << " before converging";
 *          std::cout << " to " << monitor.relative_tolerance() << " relative tolerance " << std::endl;
 *      }
 *  
 *      return 0;
 *  }
 *  \endcode
 *
 *  \see \p verbose_monitor
 *
 */
template <typename Real>
class default_monitor
{
    public:
    /*! Construct a \p default_monitor for a given right-hand-side \p b
     *
     *  The \p default_monitor terminates iteration when the residual norm
     *  satisfies the condition
     *       ||b - A x|| <= absolute_tolerance + relative_tolerance * ||b||
     *  or when the iteration limit is reached.
     *
     *  \param b right-hand-side of the linear system A x = b
     *  \param iteration_limit maximum number of solver iterations to allow
     *  \param relative_tolerance determines convergence criteria
     *  \param absolute_tolerance determines convergence criteria
     *
     *  \tparam VectorType vector
     */
    template <typename Vector>
    default_monitor(const Vector& b, size_t iteration_limit = 500, Real relative_tolerance = 1e-5, Real absolute_tolerance = 0)
        : b_norm(cusp::blas::nrm2(b)),
          r_norm(std::numeric_limits<Real>::max()),
          z_norm_0(std::numeric_limits<Real>::max()),
          z_norm(std::numeric_limits<Real>::max()),
          iteration_limit_(iteration_limit),
          iteration_count_(0),
          relative_tolerance_(relative_tolerance),
          absolute_tolerance_(absolute_tolerance)
    {}

    /*! increment the iteration count
     */
    void operator++(void) {  ++iteration_count_; } // prefix increment

    /*! applies convergence criteria to determine whether iteration is finished
     *
     *  \param r residual vector of the linear system (r = b - A x)
     *  \tparam Vector vector
     */
    template <typename Vector>
    bool finished(const Vector& r)
    {
		#ifdef PRECON_NORM
		z_norm = cusp::blas::nrm2(r);
		#else 
        r_norm = cusp::blas::nrm2(r);
		#endif
        
        return converged() || iteration_count() >= iteration_limit();
    }
   
    /*! whether the last tested residual satifies the convergence tolerance
     */
    bool converged() const
    {
		#ifdef PRECON_NORM
	    return _z_norm() <= tolerance2();	
		#else 
        return residual_norm() <= tolerance1();
		#endif
    }

    /*! Euclidean norm of last residual
     */
    Real residual_norm() const { return r_norm; }
	
    /*! Euclidean norm of last z
     */
    Real _z_norm() const { return z_norm; }	
	
    /*! Euclidean norm of last relative residual
     */
    Real relative_residual_norm() const 
	{
		#ifdef PRECON_NORM
		return z_norm / z_norm_0; 
		#else 	
		return r_norm / b_norm;
		#endif
	}	
	
	/*! Euclidean norm of the beginning z_0
     */
	template <typename Vector>
	void set_z_norm_0(const Vector& z_0) 
	{
        z_norm_0 = cusp::blas::nrm2(z_0);		
	}

    /*! number of iterations
     */
    size_t iteration_count() const { return iteration_count_; }

    /*! maximum number of iterations
     */
    size_t iteration_limit() const { return iteration_limit_; }

    /*! relative tolerance
     */
    Real relative_tolerance() const { return relative_tolerance_; }
    
    /*! absolute tolerance
     */
    Real absolute_tolerance() const { return absolute_tolerance_; }
   
    /*! tolerance1
     *
     *  //Equal to absolute_tolerance() + relative_tolerance() * ||b||
	 *  Equal to max(absolute_tolerance(),relative_tolerance() * ||b||)
     *
     */ 
    //Real tolerance1() const { return absolute_tolerance() + relative_tolerance() * b_norm; }  //!!!!!!!
	Real tolerance1() const { return std::max(absolute_tolerance(), relative_tolerance() * b_norm); }
	
    /*! tolerance2
     *
     *  Equal to max(absolute_tolerance(), relative_tolerance() * ||z0||)
     *
     */ 
    Real tolerance2() const { return std::max(absolute_tolerance(), relative_tolerance() * z_norm_0); }  //!!!!!!!

    protected:

    Real r_norm;
    Real b_norm;
    Real z_norm;
    Real z_norm_0;	
    Real relative_tolerance_;
    Real absolute_tolerance_;

    size_t iteration_limit_;
    size_t iteration_count_;
};


/*! \p default_monitor is similar to \p default monitor except that
 * it displays the solver status during iteration and reports a 
 * summary after iteration has stopped.
 *
 * \tparam Real real-valued type (e.g. \c float or \c double).
 *
 * \see \p default_monitor
 */
template <typename Real>
class verbose_monitor : public default_monitor<Real>
{
    typedef cusp::default_monitor<Real> super;

    public:
    /*! Construct a \p verbose_monitor for a given right-hand-side \p b
     *
     *  The \p verbose_monitor terminates iteration when the residual norm
     *  satisfies the condition
     *       ||b - A x|| <= absolute_tolerance + relative_tolerance * ||b||
     *  or when the iteration limit is reached.
     *
     *  \param b right-hand-side of the linear system A x = b
     *  \param iteration_limit maximum number of solver iterations to allow
     *  \param relative_tolerance determines convergence criteria
     *  \param absolute_tolerance determines convergence criteria
     *
     *  \tparam VectorType vector
     */
    template <typename Vector>
    verbose_monitor(const Vector& b, size_t iteration_limit = 500, Real relative_tolerance = 1e-5, Real absolute_tolerance = 0)
        : super(b, iteration_limit, relative_tolerance, absolute_tolerance)
    {
        std::cout << "Solver will continue until ";
		#ifdef PRECON_NORM
        std::cout << "residual norm " << super::tolerance2() << " or reaching ";
		#else
        std::cout << "residual norm " << super::tolerance1() << " or reaching ";
		#endif		
        std::cout << super::iteration_limit() << " iterations " << std::endl;
        std::cout << "  Iteration Number  | Residual Norm" << std::endl;
    }
    
    template <typename Vector>
    bool finished(const Vector& r)
    {
        super::r_norm = cusp::blas::nrm2(r);

        std::cout << "       "  << std::setw(10) << super::iteration_count();
		#ifdef PRECON_NORM
        std::cout << "       "  << std::setw(10) << std::scientific << super::_z_norm() << std::endl;		
		#else		
        std::cout << "       "  << std::setw(10) << std::scientific << super::residual_norm() << std::endl;
		#endif	
		
        if (super::converged())
        {
            std::cout << "Successfully converged after " << super::iteration_count() << " iterations." << std::endl;
            return true;
        }
        else if (super::iteration_count() >= super::iteration_limit())
        {
            std::cout << "Failed to converge after " << super::iteration_count() << " iterations." << std::endl;
            return true;
        }
        else
        {
            return false;
        }
    }
};
/*! \}
 */

} // end namespace cusp

