/*
 *  Author: Li Luo, modified from J. Cohen 
 *  Email:  li.luo@siat.ac.cn 
 */

/*! \file asm.h
 *  \brief Additive Schwarz preconditoner
 *  
 */
#pragma once

#include <cusp/detail/config.h>
#include <cusp/array1d.h>
#include <cusp/csr_matrix.h>
#include <cusp/linear_operator.h>
#include <cusp/detail/device/utils.h>
#include <cusp/detail/device/texture.h>
#include <cusp/precond/detail/asm_local_jacobi.inl> 
#include <cusp/detail/utils_.h> 

#include <thrust/iterator/constant_iterator.h>
#include <thrust/reduce.h>
#include <thrust/unique.h>
#include <thrust/scan.h>

#define  BLOCKDIM 256

namespace cusp
{
namespace precond
{

enum asm_type {
	basic,
	restrict
};

/*! \p additive_schwarz: preconditoner based on
 *  domain decomposition.  Caller must provide domains via a zero-based per-element index. 
    domains: size equals to A.num_row, each index indicates each node, and the array element is the domain ID which belongs to.
 *  This preconditioner will
 * "grow" them to satisfy the subdomains.  
 */

template <typename IndexType, typename ValueType, typename MemorySpace>
class additive_schwarz : public cusp::linear_operator<ValueType, MemorySpace, IndexType>
{
    public:

	bool initialized;
	bool same_mesh; // if true, domains remains the same and created once in constructor
	bool same_structure;	
	
	template<typename MatrixType, typename VectorType>
    additive_schwarz(bool same_mesh, bool same_structure, const MatrixType& A, 
	const VectorType& domains, const IndexType overlap=1);
	
	template<typename MatrixType, typename VectorType>
    void setup(const MatrixType& A, 
	const VectorType& domains, 
	asm_type type=basic, 
	const IndexType overlap=1, 
	const IndexType sweeps=1,
	const IndexType iter=1000, 
	const ValueType jacobi_tol=1e-10);	

    template <typename Array1, typename Array2>
    void operator()(const Array1& x, Array2& y);

    template <typename MatrixType, typename Array1, typename Array2>
    void presmooth(const MatrixType& A, const Array1& x, Array2& y);

    template <typename MatrixType, typename Array1, typename Array2>
    void postsmooth(const MatrixType& A, const Array1& x, Array2& y);
	
    template <bool UseCache, bool Initialize, bool UseOmega, typename Array1, typename Array2>
    void apply(const Array1& x, Array2& y);	
	
	IndexType  sweeps;
    IndexType  jacobi_iterations;
	ValueType  jacobi_tol;	
	asm_type   type;
	
    size_t    get_shmem_size() const { return _shmem_size; }
    size_t    get_kernel_global_mem_usage() const { //global memory
      // for each entry, 2 indices + rhs
      // + result vector (write)
      // + per-domain data (dj & domain_ptr*2 (start & end))
      // + all matrix data
      // ignore per-domain data since it should be small relative to everything else
      return _domain_global_indices.size()*(sizeof(IndexType)*2 + sizeof(ValueType)) +
             _n * sizeof(ValueType) +  
             _dj.size() * sizeof(IndexType) +
             _domain_pointers.size() * sizeof(IndexType) * 2 +
			 _domain_lambda_max.size() * sizeof(ValueType) +
             _Aj_column_indices.size() * sizeof(IndexType) +
             _Aj_values.size() * sizeof(ValueType) +
             _Aj_row_offsets.size() * sizeof(IndexType) +
			 new_x.size() * sizeof(ValueType) +
			 local_x.size() * sizeof(ValueType);
    }

    protected:

    size_t     	_shmem_size;	// size of shared memory used for GPU kernel
    IndexType  	_max_domain_storage; // max number of CSR entries per domain
    IndexType  	_num_domains;
    IndexType  	_max_domain_size;
    IndexType  	_n;
	IndexType 	cta_size;
	IndexType 	_tuple_count;
  
	// each domain has its own _dj, _domain_pointers, etc.
    cusp::array1d<IndexType,MemorySpace> _dj; // size of domain j
    cusp::array1d<IndexType,MemorySpace> _domain_pointers; // start index of domain j's indices in _domain_xxx_indices
    cusp::array1d<IndexType,MemorySpace> _domain_local_indices;  // local index of entry 
    cusp::array1d<IndexType,MemorySpace> _domain_global_indices; // global index of entry 
    cusp::array1d<ValueType,MemorySpace> _domain_lambda_max; // estimate of max eigenvalue per submatrix 
    cusp::array1d<IndexType,MemorySpace> _domains; // original partition
	cusp::array1d<IndexType,MemorySpace> _tuple_domain; // used in setup

	// these are stored in *local* indices, sign flipped if in overlap region
	cusp::array1d<IndexType, MemorySpace> _Aj_row_offsets;
    cusp::array1d<IndexType, MemorySpace> _Aj_column_indices;
    cusp::array1d<ValueType, MemorySpace> _Aj_values;

	// use simple jacobi iteration
	cusp::array1d<ValueType,MemorySpace> new_x;
	cusp::array1d<ValueType,MemorySpace> local_x;
};

} // end namespace precond
} // end namespace cusp

#include <cusp/precond/detail/asm.inl>

