/*
 *  Copyright 2008-2009 NVIDIA Corporation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

#include <cusp/blas.h>
#include <cusp/coo_matrix.h>
#include <cusp/csr_matrix.h>

#include <thrust/scan.h>
#include <thrust/reduce.h>

#define BLOCK_DIMX 16            // tile (and threadblock) size in x
#define BLOCK_DIMY 16            // tile (and threadblock) size in y

namespace cusp
{
namespace precond
{
namespace detail
{
template <typename IndexType>
__global__ void count_nnz(const IndexType mx, const IndexType my, const IndexType Mx, const IndexType My, 
					const IndexType ratioi, const IndexType ratioj, const IndexType dof, IndexType *entries_row)
{
	/* 2D global index for inner data(without ghost points) of this processor */
	IndexType ix = blockIdx.x * blockDim.x + threadIdx.x;
	IndexType iy = blockIdx.y * blockDim.y + threadIdx.y;
	IndexType i_c,j_c,k,row_offset;

	if(ix<mx && iy<my) {
		i_c = (ix/ratioi);    // coarse grid node to left of fine grid node 
        j_c = (iy/ratioj);    // coarse grid node below fine grid node 
		row_offset = iy*mx*dof+ix*dof;
		for (k=0; k<dof; k++) {		
		int entries_per_row=0;		
		// has one point of left and below; or we are right on it
		entries_per_row++;
        // has one point of right and below 
        if (i_c*ratioi != ix) {
		entries_per_row++;
        }
        // has one point of left and above 
        if (j_c*ratioj != iy) { 
		entries_per_row++;
        }
        // has one point of right and above 
        if (j_c*ratioj != iy && i_c*ratioi != ix) { 
		entries_per_row++;
        }
		entries_row[row_offset+k] += entries_per_row; 
		}
	}
}

template <typename IndexType, typename ValueType>
__global__ void formP_coo(const IndexType mx, const IndexType my, const IndexType Mx, const IndexType My, 
					const IndexType ratioi, const IndexType ratioj, const IndexType dof, 
					IndexType *row_pointers, IndexType *row_indices,
					IndexType *column_indices, ValueType *values)
{
	/* 2D global index for inner data(without ghost points) of this processor */
	IndexType ix = blockIdx.x * blockDim.x + threadIdx.x;
	IndexType iy = blockIdx.y * blockDim.y + threadIdx.y;
	IndexType i_c,j_c;
	IndexType row,col,row1,col1,k;
	ValueType x,y;

	if(ix<mx && iy<my) {
		i_c = (ix/ratioi);    /* coarse grid node to left of fine grid node */
        j_c = (iy/ratioj);    /* coarse grid node below fine grid node */
	    /* 
         Only include those interpolation points that are truly 
         nonzero. Note this is very important for final grid lines
         in x and y directions; since they have no right/top neighbors
         */		
        x  = ((float)(ix - i_c*ratioi))/((float)ratioi); /* (x-x_i)/dx */
        y  = ((float)(iy - j_c*ratioj))/((float)ratioj); /* (y-y_i)/dy */
		
		row1=iy*mx*dof+ix*dof;   /* convert fine mesh to local "natural" numbering, corresponding to row index of P */
		col1=j_c*Mx*dof+i_c*dof; /* convert fine mesh to local "natural" numbering, corresponding to row index of P */
		/* has one point of left and below; or we are right on it */
		for (k=0; k<dof; k++) {
		  row = row1+k; 
		  IndexType entries_per_row = 0;
		  IndexType row_offset = row_pointers[row];
		  /* has one point of left and below; or we are right on it */
		  {		  
		  IndexType nnz_row=row_offset;
		  if (ix<=(Mx-1)*ratioi && iy<=(My-1)*ratioj) {
		  values[nnz_row] = x*y - x - y + 1.0;  /* (1-x)(1-y) */
		  } else if (ix>(Mx-1)*ratioi && iy<=(My-1)*ratioj) {
		  values[nnz_row] = -x*y + x - y + 1.0;	/* (1+x)(1-y) */
		  } else if (ix<=(Mx-1)*ratioi && iy>(My-1)*ratioj) {
		  values[nnz_row] = -x*y - x + y + 1.0;	/* (1-x)(1+y) */	  
		  } else {
		  values[nnz_row] = x*y + x + y + 1.0;	/* (1+x)(1+y) */	
		  }	
		  col = col1+k;  	   
		  row_indices[nnz_row] = row; column_indices[nnz_row] = col; 
		  entries_per_row++;
		  }
        /* has one point of right and below */
          if (i_c*ratioi != ix) {
		  IndexType nnz_row=row_offset+entries_per_row;
		  if (ix<=(Mx-1)*ratioi && iy<=(My-1)*ratioj) {	
		  col = col1+dof+k;  // convert coarse mesh to local "natural" numbering, corresponding to column index of P 		  
		  values[nnz_row] = -x*y + x;  	   /* x(1-y) */	  
		  } else if (ix>(Mx-1)*ratioi && iy<=(My-1)*ratioj) {
		  col = col1-dof+k;		  
		  values[nnz_row] = x*y - x;	   /* -x(1-y) */
		  } else if (ix<=(Mx-1)*ratioi && iy>(My-1)*ratioj) {	  
		  col = col1+dof+k;			  
		  values[nnz_row] = x*y + x;	   /* x(1+y) */
		  } else {		  
		  col = col1-dof+k;
		  values[nnz_row] = -x*y - x;	   /* -x(1+y) */  
		  }
		  row_indices[nnz_row] = row; column_indices[nnz_row] = col;
		  entries_per_row++;
          }
        /* has one point of left and above */
          if (j_c*ratioj != iy) { 
		  IndexType nnz_row=row_offset+entries_per_row;
		  if (ix<=(Mx-1)*ratioi && iy<=(My-1)*ratioj) {	
		  col = col1+Mx*dof+k;  // convert coarse mesh to local "natural" numbering, corresponding to column index of P 
		  values[nnz_row] = -x*y + y;   /* (1-x)y */
		  } else if (ix>(Mx-1)*ratioi && iy<=(My-1)*ratioj) {
		  col = col1+Mx*dof+k;  // convert coarse mesh to local "natural" numbering, corresponding to column index of P 
		  values[nnz_row] = x*y + y;	   /* (1+x)y */  
		  } else if (ix<=(Mx-1)*ratioi && iy>(My-1)*ratioj) {
		  col = col1-Mx*dof+k;  // convert coarse mesh to local "natural" numbering, corresponding to column index of P 
		  values[nnz_row] = x*y - y;	   /* (1-x)(-y) */  
		  } else {	  
		  col = col1-Mx*dof+k;  // convert coarse mesh to local "natural" numbering, corresponding to column index of P 
		  values[nnz_row] = -x*y - y;   /* (1+x)(-y) */	 
		  }
		  row_indices[nnz_row] = row; column_indices[nnz_row] = col;
          entries_per_row++;
		  }
        /* has one point of right and above */
          if (j_c*ratioj != iy && i_c*ratioi != ix) {
		  IndexType nnz_row=row_offset+entries_per_row;		  
		  if (ix<=(Mx-1)*ratioi && iy<=(My-1)*ratioj) {		  
		  col = col1+(Mx+1)*dof+k;  // convert coarse mesh to local "natural" numbering, corresponding to column index of P
		  values[nnz_row] = x*y;
		  } else if (ix>(Mx-1)*ratioi && iy<=(My-1)*ratioj) {
		  col = col1+(Mx-1)*dof+k;  // convert coarse mesh to local "natural" numbering, corresponding to column index of P
		  values[nnz_row] = -x*y;		  
		  } else if (ix<=(Mx-1)*ratioi && iy>(My-1)*ratioj) {
		  col = col1+(1-Mx)*dof+k;  // convert coarse mesh to local "natural" numbering, corresponding to column index of P
		  values[nnz_row] = -x*y;	  
		  } else {
		  col = col1-(Mx+1)*dof+k;  // convert coarse mesh to local "natural" numbering, corresponding to column index of P
		  values[nnz_row] = x*y;		   /* (-x)(-y) */
		  }		   
		  row_indices[nnz_row] = row; column_indices[nnz_row] = col;
          entries_per_row++;
		  }
		}
	}
}

template <typename IndexType, typename ValueType>
void smooth_prolongator(cusp::coo_matrix<IndexType,ValueType,cusp::device_memory>& P,
				IndexType& mx_c, IndexType& my_c,
				const IndexType mx, const IndexType my,
				const IndexType ratioi, const IndexType ratioj, const IndexType dof)
{
	IndexType Mx=(mx/ratioi)+((mx%ratioi)?1:0), My=(my/ratioj)+((my%ratioj)?1:0);
	IndexType nnz;
	
	mx_c = Mx; my_c = My;
	printf("mx:%d, my:%d, Mx: %d, My: %d\n",mx,my,Mx,My);
	
	// entries_row count the non zero entry number in each row of P
	cusp::array1d<IndexType, cusp::device_memory> entries_row(mx*my*dof, 0);
	dim3 dimBlock(BLOCK_DIMX, BLOCK_DIMY);
    dim3 dimGrid((mx+BLOCK_DIMX-1) / BLOCK_DIMX, (my+BLOCK_DIMY-1) / BLOCK_DIMY);
	/* precount nonzeros of P */
	count_nnz<IndexType><<<dimGrid, dimBlock>>>(mx,my,Mx,My,ratioi,ratioj,dof,
									thrust::raw_pointer_cast(&*entries_row.begin()));	

	// each element of row_pointers	present the start pointer of the first element in this row, the location					
	cusp::array1d<IndexType, cusp::device_memory> row_pointers(mx*my*dof+1, 0);
    thrust::inclusive_scan(entries_row.begin(), entries_row.end(), row_pointers.begin()+1);

	// count the total non zeros number of P
	nnz = thrust::reduce(entries_row.begin(), entries_row.end());
		
	/* resize P */
	P.resize(mx*my*dof, Mx*My*dof, nnz); /* rows number equals to fine mesh dof, cols number equals to coarse mesh dof */

	formP_coo<IndexType,ValueType><<<dimGrid, dimBlock>>>(mx,my,Mx,My,ratioi,ratioj,dof,
								thrust::raw_pointer_cast(&*row_pointers.begin()),
								thrust::raw_pointer_cast(&*(P.row_indices).begin()),
								thrust::raw_pointer_cast(&*(P.column_indices).begin()),		
								thrust::raw_pointer_cast(&*(P.values).begin()));
}

template <typename IndexType, typename ValueType>
void smooth_prolongator(cusp::csr_matrix<IndexType,ValueType,cusp::host_memory>& P,
				IndexType& mx_c, IndexType& my_c,
				const IndexType mx, const IndexType my,
				const IndexType ratioi, const IndexType ratioj, const IndexType dof)
{
	IndexType Mx=(mx/ratioi)+((mx%ratioi)?1:0), My=(my/ratioj)+((my%ratioj)?1:0);	
	mx_c = Mx; my_c = My;
	printf("mx:%d, my:%d, Mx: %d, My: %d\n",mx,my,Mx,My);
	/* sequel */
	IndexType i,j,k,i_c,j_c,row_offset,nnz1,nnz2;
	IndexType col,col1;
	float x,y;
	cusp::array1d<IndexType, cusp::host_memory> entries_row(mx*my*dof, 0);
		
	// precount nonzeros of P
	for (j=0; j<my; j++) {
     for (i=0; i<mx; i++) {
		
		i_c = (i/ratioi);    // coarse grid node to left of fine grid node 
        j_c = (j/ratioj);    // coarse grid node below fine grid node 
		row_offset = j*mx*dof+i*dof;
		for (k=0; k<dof; k++) {		
		int entries_per_row=0;		
		// has one point of left and below; or we are right on it
		entries_per_row++;
        // has one point of right and below 
        if (i_c*ratioi != i) {
		entries_per_row++;
        }
        // has one point of left and above 
        if (j_c*ratioj != j) {
		entries_per_row++;
        }
        // has one point of right and above 
        if (j_c*ratioj != j && i_c*ratioi != i) { 
		entries_per_row++;
        }
		entries_row[row_offset+k] += entries_per_row; 
		}
        }
	 }
	 
	// count the total non zeros number of P
	nnz1 = thrust::reduce(entries_row.begin(), entries_row.end());
	
	// resize P 
	P.resize(mx*my*dof, Mx*My*dof, nnz1); // rows number equals to fine mesh dof, cols number equals to coarse mesh dof
	P.row_offsets[0] = 0; 
    thrust::inclusive_scan(entries_row.begin(), entries_row.end(), P.row_offsets.begin()+1); 
	
	nnz2 = 0;
	for (j=0; j<my; j++) {
     for (i=0; i<mx; i++) {
		i_c = (i/ratioi);    // coarse grid node to left of fine grid node 
        j_c = (j/ratioj);    // coarse grid node below fine grid node
		
         //Only those interpolation points which are truly 
         //nonzero are included. Note this is very important for the last several grid lines
         //in x and y directions; since they have no right/top neighbors
        x  = ((float)(i - i_c*ratioi))/((float)ratioi); // (x-x_i)/dx 
        y  = ((float)(j - j_c*ratioj))/((float)ratioj); // (y-y_i)/dy
		  
		col1=j_c*Mx*dof+i_c*dof;
	   for (k=0; k<dof; k++) {
		// has one point of left and below; or we are right on it
		{
		  if (i<=(Mx-1)*ratioi && j<=(My-1)*ratioj) {
		  P.values[nnz2] = x*y - x - y + 1.0;  /* (1-x)(1-y) */
		  } else if (i>(Mx-1)*ratioi && j<=(My-1)*ratioj) {
		  P.values[nnz2] = -x*y + x - y + 1.0;	/* (1+x)(1-y) */
		  } else if (i<=(Mx-1)*ratioi && j>(My-1)*ratioj) {
		  P.values[nnz2] = -x*y - x + y + 1.0;	/* (1-x)(1+y) */	  
		  } else {
		  P.values[nnz2] = x*y + x + y + 1.0;	/* (1+x)(1+y) */	
		  }	
		  col = col1+k;  // convert coarse mesh to local "natural" numbering, corresponding to column index of P 
		  P.column_indices[nnz2] = col;	
		  nnz2++;			  
		}

        // has one point of right and below 
        if (i_c*ratioi != i) {
		  if (i<=(Mx-1)*ratioi && j<=(My-1)*ratioj) {	
		  col = col1+dof+k;  // convert coarse mesh to local "natural" numbering, corresponding to column index of P 		  
		  P.values[nnz2] = -x*y + x;  	   /* x(1-y) */	  
		  } else if (i>(Mx-1)*ratioi && j<=(My-1)*ratioj) {
		  col = col1-dof+k;		  
		  P.values[nnz2] = x*y - x;		   /* -x(1-y) */
		  } else if (i<=(Mx-1)*ratioi && j>(My-1)*ratioj) {	  
		  col = col1+dof+k;			  
		  P.values[nnz2] = x*y + x;		   /* x(1+y) */
		  } else {		  
		  col = col1-dof+k;
		  P.values[nnz2] = -x*y - x;	   /* -x(1+y) */  
		  }
		  P.column_indices[nnz2] = col; 
		  nnz2++;	
		}

        // has one point of left and above 
        if (j_c*ratioj != j) {
		  if (i<=(Mx-1)*ratioi && j<=(My-1)*ratioj) {	
		  col = col1+Mx*dof+k;  // convert coarse mesh to local "natural" numbering, corresponding to column index of P 
		  P.values[nnz2] = -x*y + y;   /* (1-x)y */
		  } else if (i>(Mx-1)*ratioi && j<=(My-1)*ratioj) {
		  col = col1+Mx*dof+k;  // convert coarse mesh to local "natural" numbering, corresponding to column index of P 
		  P.values[nnz2] = x*y + y;	   /* (1+x)y */  
		  } else if (i<=(Mx-1)*ratioi && j>(My-1)*ratioj) {
		  col = col1-Mx*dof+k;  // convert coarse mesh to local "natural" numbering, corresponding to column index of P 
		  P.values[nnz2] = x*y - y;	   /* (1-x)(-y) */  
		  } else {	  
		  col = col1-Mx*dof+k;  // convert coarse mesh to local "natural" numbering, corresponding to column index of P 
		  P.values[nnz2] = -x*y - y;   /* (1+x)(-y) */	 
		  }
		  P.column_indices[nnz2] = col;		  
		  nnz2++;	
        }
		
        // has one point of right and above 
        if (j_c*ratioj != j && i_c*ratioi != i) { 
		  if (i<=(Mx-1)*ratioi && j<=(My-1)*ratioj) {		  
		  col = col1+(Mx+1)*dof+k;  // convert coarse mesh to local "natural" numbering, corresponding to column index of P
		  P.values[nnz2] = x*y;
		  } else if (i>(Mx-1)*ratioi && j<=(My-1)*ratioj) {
		  col = col1+(Mx-1)*dof+k;  // convert coarse mesh to local "natural" numbering, corresponding to column index of P
		  P.values[nnz2] = -x*y;		  
		  } else if (i<=(Mx-1)*ratioi && j>(My-1)*ratioj) {
		  col = col1+(1-Mx)*dof+k;  // convert coarse mesh to local "natural" numbering, corresponding to column index of P
		  P.values[nnz2] = -x*y;	  
		  } else {
		  col = col1-(Mx+1)*dof+k;  // convert coarse mesh to local "natural" numbering, corresponding to column index of P
		  P.values[nnz2] = x*y;		   /* (-x)(-y) */
		  }
		  P.column_indices[nnz2] = col; 
		  nnz2++;
        }
	   }
	 }
	}
	assert(nnz1 == nnz2);
}

} // end namespace detail
} // end namespace precond
} // end namespace cusp

