/*
 *  Author: Li Luo 
 *  Email:  li.luo@siat.ac.cn 
 */
/*! \file asm.h
 *  \brief Additive Schwarz preconditoner
 *  
 */

namespace cusp
{
namespace precond
{
namespace detail
{

template <typename IndexType>
struct count_domain_storage : public thrust::unary_function<IndexType, IndexType>
{
  const IndexType *domain_pointer;
  const IndexType *Aj_row_offsets;

  count_domain_storage(const IndexType *dp, const IndexType *Aj) : domain_pointer(dp), Aj_row_offsets(Aj) { }

  __host__ __device__ IndexType operator()(IndexType i) const {
    IndexType domain_start = domain_pointer[i];
    IndexType domain_end   = domain_pointer[i+1];
	// domain_start < domain_end (for example 5, 10: data within [Aj_row_offsets[10],  Aj_row_offsets[5]) represents the domain i) 
    return Aj_row_offsets[domain_end] - Aj_row_offsets[domain_start];
  }
};

template<typename IndexType>
__global__ void count_neighbor_regions(
  IndexType *       counts,
  const IndexType * tuple_domain,
  const IndexType * tuple_index,
  const IndexType * row_offsets,
  const IndexType * column_indices, 
  const IndexType * domains, 
  IndexType    tuple_count)
{
  const IndexType num_threads = blockDim.x * gridDim.x;
  IndexType tid = threadIdx.x + blockIdx.x * blockDim.x;
  IndexType output_slot = 0;

  for (IndexType i = tid; i < tuple_count; i+= num_threads) {
    IndexType this_idx = tuple_index[i];
    IndexType this_domain = tuple_domain[i];
 
    IndexType row_start = row_offsets[this_idx];    
    IndexType row_end   = row_offsets[this_idx+1];    

    for (IndexType j=row_start; j < row_end; j++) {
      IndexType nbr_idx = column_indices[j];

      // skip diagonal element
      if (nbr_idx == this_idx)
        continue;

      IndexType nbr_domain = domains[nbr_idx];
      if (nbr_domain != this_domain) {
        output_slot++;
      }
    }
    counts[i] = output_slot;
  }
}


template<typename IndexType>
__global__ void grow_neighbor_regions(
  IndexType *      new_tuple_domain,
  IndexType *      new_tuple_index,
  const IndexType * tuple_domain,
  const IndexType * tuple_index,
  const IndexType * row_offsets,
  const IndexType * column_indices, 
  const IndexType * domains, 
  const IndexType * counts, 
  IndexType    tuple_count)
{
  const IndexType num_threads = blockDim.x * gridDim.x;
  IndexType tid = threadIdx.x + blockIdx.x * blockDim.x;
  IndexType output_slot = 0;

  for (IndexType i = tid; i < tuple_count; i+= num_threads) {
    IndexType this_idx = tuple_index[i];
    IndexType this_domain = tuple_domain[i];
 
    IndexType row_start = row_offsets[this_idx];    
    IndexType row_end   = row_offsets[this_idx+1];    

    IndexType output_start = counts[i];

    for (IndexType j=row_start; j < row_end; j++) {
      IndexType nbr_idx = column_indices[j];

      // skip diagonal element
      if (nbr_idx == this_idx)
        continue;

      IndexType nbr_domain = domains[nbr_idx];
      if (nbr_domain != this_domain) {
        // write (nbr_index, domain) into output slot
        new_tuple_domain[output_start + output_slot] = this_domain;
        new_tuple_index [output_start + output_slot] = nbr_idx;
      
        output_slot++;
      }
    }
  }
}

template<typename IndexType>
__global__ void
submatrix_per_row_counts(
    IndexType *per_row_counts, 
    const IndexType *domain_pointers, 
    const IndexType *tuple_domain, 
    const IndexType *domain_global_indices, 
    const IndexType *A_row_offsets,
    const IndexType *A_column_indices, 
    IndexType tuple_count)
{
  const IndexType num_threads = blockDim.x * gridDim.x;
  IndexType tid = threadIdx.x + blockIdx.x * blockDim.x;
  if (tid >= tuple_count)
    return;  
  for (IndexType tuple_id = tid; tuple_id < tuple_count; tuple_id+= num_threads) {  

  IndexType global_row_id = domain_global_indices[tuple_id];
  IndexType this_domain = tuple_domain[tuple_id];

  IndexType row_start = A_row_offsets[global_row_id];
  IndexType row_end   = A_row_offsets[global_row_id+1];
  IndexType domain_start = domain_pointers[this_domain];
  IndexType domain_end   = domain_pointers[this_domain+1];

  IndexType num_matches = 0;

  IndexType row_ptr = row_start;
  IndexType tuple_ptr = domain_start;

  IndexType A_col_val = -1;
  IndexType domain_index_val = -1; 

  while (row_ptr < row_end && tuple_ptr < domain_end) {
    // read either of these values if needed
    if (A_col_val == -1)
      A_col_val = A_column_indices[row_ptr];
    if (domain_index_val == -1)
      domain_index_val = domain_global_indices[tuple_ptr];
    
    if (A_col_val == domain_index_val) {
      row_ptr++;
      tuple_ptr++;
      A_col_val = -1;
      domain_index_val = -1;
      // got a match
      num_matches++;
    }
    else if (A_col_val < domain_index_val) {
      A_col_val = -1;
      row_ptr++;
    }
    else {
      domain_index_val = -1;
      tuple_ptr++;
    }
  }

  per_row_counts[tuple_id] = num_matches;
  }
}

template<typename IndexType, typename ValueType>
__global__ void
submatrix_fill(
    IndexType *Aj_column_indices, 
    ValueType *Aj_values,
    IndexType *Aj_row_offsets, 
    const IndexType *domains, 
    const IndexType *domain_pointers, 
    const IndexType *tuple_domain, 
    const IndexType *domain_global_indices, 
    const IndexType *domain_local_indices, 
    const IndexType *A_row_offsets,
    const IndexType *A_column_indices, 
    const ValueType *A_values, 
    IndexType tuple_count)
{
  const IndexType num_threads = blockDim.x * gridDim.x;
  IndexType tid = threadIdx.x + blockIdx.x * blockDim.x;
  if (tid >= tuple_count)
    return;  
  for (IndexType tuple_id = tid; tuple_id < tuple_count; tuple_id+= num_threads) {  

  IndexType global_row_id = domain_global_indices[tuple_id];
  IndexType this_domain = tuple_domain[tuple_id];

  IndexType row_start = A_row_offsets[global_row_id];
  IndexType row_end   = A_row_offsets[global_row_id+1];
  IndexType domain_start = domain_pointers[this_domain];
  IndexType domain_end   = domain_pointers[this_domain+1];

  IndexType row_ptr = row_start;
  IndexType tuple_ptr = domain_start;

  IndexType A_col_val = -1;
  IndexType domain_index_val = -1; 

  IndexType output_slot = Aj_row_offsets[tuple_id];

  while (row_ptr < row_end && tuple_ptr < domain_end) {
    // read either of these values if needed
    if (A_col_val == -1)
      A_col_val = A_column_indices[row_ptr];
    if (domain_index_val == -1)
      domain_index_val = domain_global_indices[tuple_ptr];
    
    if (A_col_val == domain_index_val) {

      // got a match - write into output slot and advance to next slot
      Aj_values[output_slot] = A_values[row_ptr];
      Aj_column_indices[output_slot] = domain_local_indices[tuple_ptr];

      output_slot++;

      row_ptr++;
      tuple_ptr++;
      A_col_val = -1;
      domain_index_val = -1;
    }
    else if (A_col_val < domain_index_val) {
      A_col_val = -1;
      row_ptr++;
    }
    else { // if (domain_index_val < A_col_val)
      domain_index_val = -1;
      tuple_ptr++;
    }
  }
  }
}

template<typename IndexType, typename ValueType>
__global__ void
estimate_local_lamdba_max(
  ValueType *domain_lambda_max,
  const IndexType *dj,
  const IndexType *domain_pointers,
  const IndexType *Aj_row_offsets,
  const IndexType *Aj_column_indices,  
  const ValueType *Aj_values,
  const IndexType num_domains)
{
  extern __shared__ ValueType local_scratch[];

  for (IndexType j = blockIdx.x; j < num_domains; j += gridDim.x) {  // domain index
  
  IndexType i = threadIdx.x; // local row index
  IndexType domain_start = domain_pointers[j]; // start of the row in this domain
  IndexType domain_size = dj[j];  // size of this domain
  if (domain_size == 0) return;  
  	  
  // initialize shared memory
  local_scratch[i] = (ValueType)0;
	  
  for (IndexType k = i; k < domain_size; k += blockDim.x) { 
	    IndexType global_k      = domain_start + k;
		ValueType row_sum_abs   = 0;
        ValueType inv_diag_abs  = 1;
        IndexType row_start     = Aj_row_offsets[global_k]; // Aj_row_offsets is global, must use global index global_k
        IndexType row_end       = Aj_row_offsets[global_k+1];

    for (IndexType row_ptr = row_start; row_ptr < row_end; row_ptr++) {
      ValueType val = Aj_values[row_ptr];
      row_sum_abs  += fabs(val);   // sum of absolute values in the current row
      IndexType col_id = Aj_column_indices[row_ptr];
      if (col_id == k) 
		inv_diag_abs = ((ValueType)1) / fabs(Aj_values[row_ptr]);  // absolute value of diag element
    }
	ValueType row_dinv_sum_abs = row_sum_abs * inv_diag_abs;
	// reduction in each thread
	if (local_scratch[i] < row_dinv_sum_abs) {  
		local_scratch[i] = row_dinv_sum_abs;
	}	
  }
  __syncthreads();

  // cta-wide reduction among threads in this block
  IndexType width = asm_round_to_next_power_of_two(blockDim.x)/2; // find the cloest 2^n integer for blockdim and divide half
  while (width > 0) {
	if ((i < width) && (i+width < blockDim.x))
      if (local_scratch[i+width] > local_scratch[i])
        local_scratch[i] = local_scratch[i+width];
    width /= 2;
    __syncthreads();
  }

  if (i == 0)  // only one thread write the memory
    domain_lambda_max[j] = local_scratch[0];
 }	
}

template<typename IndexType, typename ValueType, typename LocalSolver>
__global__ void
//__launch_bounds__(1024)
run_local_solvers(
  ValueType       *x,
  ValueType       *local_x,  
  ValueType       *new_x,  
  const ValueType *rhs,
  const IndexType *dj,
  const IndexType *domains,
  const IndexType *domain_pointers,
  const IndexType *domain_global_indices,
  const IndexType *Aj_row_offsets,
  const IndexType *Aj_col_indices,
  const ValueType *Aj_values,
  asm_type   	  asm_type,
  const IndexType num_domains,
  typename   LocalSolver::params p
)
{
  extern __shared__ ValueType local_scratch[];

  for (IndexType j = blockIdx.x; j < num_domains; j += gridDim.x) {  // domain index
  
  IndexType i = threadIdx.x; // local row index
  IndexType domain_start = domain_pointers[j]; // start of the row in this domain
  IndexType domain_size = dj[j];  // size of this domain
  if (domain_size == 0) return;
  
  // solve linear system locally 
  LocalSolver local_solver(j, p);
  local_solver(i, domain_size, domain_start, local_x, new_x, Aj_row_offsets, Aj_col_indices, Aj_values, rhs, domain_global_indices, local_scratch);  // there are __syncthreads(); inside
  
  // assemble local solutions to global solution
    for (IndexType k = i; k < domain_size; k += blockDim.x) {
	    IndexType global_k = domain_start + k;
	    IndexType global_i = domain_global_indices[global_k];
		if (asm_type == basic) {
			// atomic is for IndexType, do the conversion
			// atomic operation: avoid conflict adding at overlapping points   
			cusp::detail::device::atomicAdd(&x[global_i], local_x[global_k]);
		} else if (asm_type == restrict) {
			// interior: key step for RAS, to implement R^0	
			if (domains[global_i] == j)	{
				x[global_i] = local_x[global_k];  // global_i is always the indice of x
			}	
	    }
    }
  }	
}


template<typename ValueType, typename IndexType>
__global__ void
//__launch_bounds__(1024)
fill_local_x(
  const ValueType *x,
  ValueType       *local_x,  
  const IndexType *dj,
  const IndexType *domains,
  const IndexType *domain_pointers,
  const IndexType *domain_global_indices,
  const IndexType num_domains)
{
  for (IndexType j = blockIdx.x; j < num_domains; j += gridDim.x) {  // domain index
  
  IndexType i = threadIdx.x; // local row index
  IndexType domain_start = domain_pointers[j]; // start of the row in this domain
  IndexType domain_size = dj[j];  // size of this domain
  if (domain_size == 0) return;
  
  // fill local x
    for (IndexType k = i; k < domain_size; k += blockDim.x) {
	    IndexType global_k = domain_start + k;
	    IndexType global_i = domain_global_indices[global_k];
		local_x[global_k]  = x[global_i];  // global_i is always the indice of x
    }
  }	
}

void pq_gen(const int np, int *p, int *q)
{
    int a = 1, b = np;

    for(int i = 2; i < floor(sqrt(double(np))) + 1; i++)
        if(np % i == 0)  
        {   
            a = i;
            b = np / i;
        }
    *p = a;    
    *q = b;
}

template<typename IndexType>
__global__ void set_domains(
IndexType *domains, 
const IndexType m, 
const IndexType n, 
const IndexType m_p, 
const IndexType n_q,
const IndexType dof
)
{
	IndexType mod_x = m % gridDim.x;
	IndexType mod_y = n % gridDim.y;
	IndexType nn = n_q + (mod_y > blockIdx.y);
	IndexType mm = m_p + (mod_x > blockIdx.x);    
	if(threadIdx.x<mm && threadIdx.y<nn)
	{
		IndexType domain_index = __mul24(blockIdx.y, gridDim.x) + blockIdx.x;	
		IndexType horizontal_offset;
		if(mod_x)
		{	if(blockIdx.x > mod_x - 1)
			{
				horizontal_offset = __mul24(mod_x, m_p + 1) + __mul24(blockIdx.x - mod_x, m_p) + threadIdx.x;
			}
			else{
				horizontal_offset = __mul24(blockIdx.x, m_p + 1) + threadIdx.x; 
			}
		}
		else{
			horizontal_offset = __mul24(blockIdx.x, m_p) + threadIdx.x;
		}
		
		IndexType vertical_offset;
		if(mod_y)
		{
			if(blockIdx.y > mod_y - 1)
			{
				vertical_offset = __mul24(mod_y, n_q + 1) + __mul24(blockIdx.y - mod_y, n_q) + threadIdx.y;
			}	
			else
			{
				vertical_offset = __mul24(blockIdx.y, n_q + 1) + threadIdx.y;
			}

		}
		else{
			vertical_offset = __mul24(blockIdx.y, n_q) + threadIdx.y;		
		}
		
		IndexType native_sort_index = __mul24(vertical_offset, __mul24(m,dof)) + __mul24(horizontal_offset,dof);
		
		domains[native_sort_index] = domain_index;
	}
}

} // end namespace detail


//  constructor 
template <typename IndexType, typename ValueType, typename MemorySpace>
template<typename MatrixType, typename VectorType>
additive_schwarz<IndexType,ValueType,MemorySpace>
    ::additive_schwarz(bool same_mesh, bool same_structure, const MatrixType& A_unsorted, const VectorType& domains, const IndexType overlap)
	: cta_size(BLOCKDIM)
{
  CUSP_PROFILE_SCOPED();

  jacobi_iterations = 2;
  jacobi_tol = 1e-10;
  type = (asm_type)1;
  
  this->same_mesh = same_mesh;
  this->same_structure = same_structure;	  
  if (this->same_mesh && this->same_structure) {
  // convert to tuple (domain, element)
  
  // allocate enough space for A.num_rows * max(degree of A)
  
  // repeat [overlap] times
  //   for each tuple (d, i):
  //     for each neighbor j of i
  //       if domain[j] != d, add (d, j) to tuple list
  
  //   remove copies from tuple_list
  
  // sort tuples by domain
  
  // segmented_scan (tuple_list) - get local index of each i within its domain d
  
  // segmented_reduction (tuple_list) - get per-domain counts
  
  // for each domain j (in parallel)
  //   form Aj = Pj A Pj^t
  //   append all Aj's into one large buffer
  
  //   (validate that shmem is large enough)

  //------------------------------------------------------------
  // first sort the entries by row & col, which is required by the matrix assembly routines below
  coo_matrix<IndexType, ValueType, MemorySpace> Acoo = A_unsorted;
  thrust::sort_by_key(thrust::make_zip_iterator(thrust::make_tuple(Acoo.row_indices.begin(), Acoo.column_indices.begin())),
                      thrust::make_zip_iterator(thrust::make_tuple(Acoo.row_indices.end(),   Acoo.column_indices.end())),
                      Acoo.values.begin());
  csr_matrix<IndexType, ValueType, MemorySpace> A = Acoo;

  // validate inputs
  if (A.num_rows != A.num_cols) 
    throw cusp::invalid_input_exception("input matrix is not square");

  if (A.num_rows != domains.size())  
    throw cusp::invalid_input_exception("domain array size does not match array dimension");

  if (overlap < 0)
    throw cusp::invalid_input_exception("overlap must be non-negative");

  if (((IndexType)-1) >= 0)
    throw cusp::invalid_input_exception("IndexType must be a signed type");

  _n = A.num_rows;
  _domains = domains;
  _num_domains = thrust::max_element(_domains.begin(), _domains.end())[0]+1;

  cusp::array1d<IndexType, MemorySpace> tuple_domain = _domains;
  cusp::array1d<IndexType, MemorySpace> tuple_index(_n); // finally the set of vertices we need when overlap > 0
  thrust::sequence(tuple_index.begin(), tuple_index.end());

  IndexType tuple_count = _n; //initialize to be _n, will be enlarged	
  for (IndexType sweep=0; sweep < overlap; sweep ++) {

    cusp::array1d<IndexType,MemorySpace> counts(tuple_count+1, 0);

    const size_t MAX_BLOCKS_count_neighbor_regions = cusp::detail::device::arch::max_active_blocks(detail::count_neighbor_regions<IndexType>, BLOCKDIM, (size_t) 0);
    const size_t NUM_BLOCKS_count_neighbor_regions = std::min<size_t>(MAX_BLOCKS_count_neighbor_regions, cusp::detail::device::DIVIDE_INTO(tuple_count, BLOCKDIM));	
    detail::count_neighbor_regions<IndexType><<<NUM_BLOCKS_count_neighbor_regions, BLOCKDIM>>>(
      thrust::raw_pointer_cast(&*(counts.begin()+1)),
      thrust::raw_pointer_cast(&*tuple_domain.begin()),
      thrust::raw_pointer_cast(&*tuple_index.begin()),
      thrust::raw_pointer_cast(&*A.row_offsets.begin()),
      thrust::raw_pointer_cast(&*A.column_indices.begin()), 
      thrust::raw_pointer_cast(&*_domains.begin()), 
      tuple_count);

    // convert per-entry counts into pointers for where each vertex places its entries
    thrust::inclusive_scan(counts.begin(), counts.end(), counts.begin());
    IndexType number_of_verts_to_add = counts[tuple_count];
    tuple_domain.resize(tuple_count + number_of_verts_to_add);
    tuple_index.resize(tuple_count + number_of_verts_to_add);

    // write new entries into end of list
    const size_t MAX_BLOCKS_grow_neighbor_regions = cusp::detail::device::arch::max_active_blocks(detail::grow_neighbor_regions<IndexType>, BLOCKDIM, (size_t) 0);
    const size_t NUM_BLOCKS_grow_neighbor_regions = std::min<size_t>(MAX_BLOCKS_grow_neighbor_regions, cusp::detail::device::DIVIDE_INTO(tuple_count, BLOCKDIM));	
    detail::grow_neighbor_regions<IndexType><<<NUM_BLOCKS_grow_neighbor_regions, BLOCKDIM>>>(
      thrust::raw_pointer_cast(&*(tuple_domain.begin()+tuple_count)),
      thrust::raw_pointer_cast(&*(tuple_index.begin()+tuple_count)),
      thrust::raw_pointer_cast(&*tuple_domain.begin()),
      thrust::raw_pointer_cast(&*tuple_index.begin()),
      thrust::raw_pointer_cast(&*A.row_offsets.begin()),
      thrust::raw_pointer_cast(&*A.column_indices.begin()), 
      thrust::raw_pointer_cast(&*_domains.begin()), 
      thrust::raw_pointer_cast(&*counts.begin()), 
      tuple_count);

    // sort tuples - this bit of cleverness courtesy Nathan.  This sorts in lexicographical ordering, but uses the fast radix sort
    thrust::stable_sort_by_key(tuple_index.begin(), tuple_index.end(), tuple_domain.begin());
    thrust::stable_sort_by_key(tuple_domain.begin(), tuple_domain.end(), tuple_index.begin());

    // remove duplicates
    cusp::array1d<IndexType, MemorySpace> new_tuple_domain(tuple_domain.size());
    cusp::array1d<IndexType, MemorySpace> new_tuple_index(tuple_index.size());

    thrust::zip_iterator<thrust::tuple<typename cusp::array1d<IndexType,MemorySpace>::iterator, typename cusp::array1d<IndexType,MemorySpace>::iterator> > last_iter = 
      thrust::unique_copy(thrust::make_zip_iterator(thrust::make_tuple(tuple_domain.begin(), tuple_index.begin())),
                          thrust::make_zip_iterator(thrust::make_tuple(tuple_domain.end(), tuple_index.end())),
                          thrust::make_zip_iterator(thrust::make_tuple(new_tuple_domain.begin(), new_tuple_index.begin())));

    // get number of newly added entries
    IndexType new_size = thrust::distance(thrust::make_zip_iterator(thrust::make_tuple(new_tuple_domain.begin(), new_tuple_index.begin())), last_iter);

    // truncate
    new_tuple_index.resize(new_size);
    new_tuple_domain.resize(new_size);

    // copy back to original vector
    tuple_index.swap(new_tuple_index);
    tuple_domain.swap(new_tuple_domain);

    tuple_count = new_size;
  }

  // sort so entries in the same domain are adjacent, and in ascending order 
  thrust::stable_sort_by_key(tuple_domain.begin(), tuple_domain.end(), tuple_index.begin());

  //tuple_domain a risized tuple list containing tuples that are enlarged to include neighbour nodes
  // a new tuple: tuple_index
  // tuple_domain are sorted to be tuple_index

  // two temp arrays used for later important array: domain_list/domain_count -> _dj -> _domain_pointers
  cusp::array1d<IndexType,MemorySpace> domain_list(tuple_count);// size is m if we have m domain, m <= tuple_count
  cusp::array1d<IndexType,MemorySpace> domain_count(tuple_count);

  typename cusp::array1d<IndexType,MemorySpace>::iterator last_domain_list_iter =  
    thrust::reduce_by_key(tuple_domain.begin(), tuple_domain.end(), thrust::make_constant_iterator<IndexType>(1), domain_list.begin(), domain_count.begin()).first;

  // gather domain_count into dj from domain_list (slots not written to will be 0)
  // TODO: for each i in [domain_list.begin(), last_domain_list) _dj[domain_list[i]] = domain_count[i]
  // scatter: i in the range [first, last), the value *i is assigned to output[*(map + (i - first))]
  _dj.resize(_num_domains);
  thrust::fill(_dj.begin(), _dj.end(), 0);	  
  IndexType num_domains_counted = thrust::distance(domain_list.begin(), last_domain_list_iter);
  thrust::scatter(domain_count.begin(), domain_count.begin() + num_domains_counted, domain_list.begin(), _dj.begin()); 

  // calculate the size of the largest domain
  _max_domain_size = thrust::max_element(_dj.begin(), _dj.end())[0];

  // fill the domain pointers (scan of domain_count) 
  _domain_pointers.resize(_num_domains+1);
  thrust::fill(_domain_pointers.begin(), _domain_pointers.end(), 0);	   
  thrust::inclusive_scan(_dj.begin(), _dj.end(), _domain_pointers.begin()+1);

  // segmented scan to get local indices, tuple_index has corresponding global index
  _domain_local_indices.resize(tuple_count);
  thrust::exclusive_scan_by_key(tuple_domain.begin(), tuple_domain.end(), thrust::make_constant_iterator<IndexType>(1), _domain_local_indices.begin());
  _domain_global_indices = tuple_index;

  //1.   (domain, index) pairs: tuple_domain, 
  //2.	 _domain_global_indices: see 538
  //3.	 _domain_pointers: array with size = domain num, elements are the index of tuple_domain
  //     (since has been sorted, element of this array is getting larger), help to compute row_offsets, see line 51 above
  //4.   per-domain counts: _dj
  // we can also figure out if an entry is in the overlap region or not (tuple (d,i) is overlap iff if domain[i] != d)
  //--------------------------------------------
  // build the local matrices Aj & represent them somehow...
  // CSR format: 2 steps
  // 1. get row_offsets:---
  // for each tuple (index,domain) = (i,j):
  //   count number of cols Ai are in Domain j (list merge is fast because per-row entries in sorted order, as are domain indices
  // scan (counts) -> row_offsets

  // NOTE: By selecting exactly the submatrix corresponding to a domain, we implicitly set zero-dirichelet boundary conditions.!!!!!!!!
  // However, when overlap > 0, transmission conditions are not needed.  For the case of no overlap, we would probably want
  // to implement something more sophisticated like Optimized RAS (i.e. Robin transmission conditions).
  cusp::array1d<IndexType,MemorySpace> per_row_counts(tuple_count, 0);

  const size_t MAX_BLOCKS_submatrix_per_row_counts = cusp::detail::device::arch::max_active_blocks(detail::submatrix_per_row_counts<IndexType>, BLOCKDIM, (size_t) 0);
  const size_t NUM_BLOCKS_submatrix_per_row_counts = std::min<size_t>(MAX_BLOCKS_submatrix_per_row_counts, cusp::detail::device::DIVIDE_INTO(tuple_count, BLOCKDIM));	  
  detail::submatrix_per_row_counts<IndexType><<<NUM_BLOCKS_submatrix_per_row_counts,BLOCKDIM>>>(
    thrust::raw_pointer_cast(&*per_row_counts.begin()), 
    thrust::raw_pointer_cast(&*_domain_pointers.begin()), 
    thrust::raw_pointer_cast(&*tuple_domain.begin()), 
    thrust::raw_pointer_cast(&*_domain_global_indices.begin()), 
    thrust::raw_pointer_cast(&*A.row_offsets.begin()),
    thrust::raw_pointer_cast(&*A.column_indices.begin()), 
    tuple_count);  

  // convert into offsets, get the total size, and allocate buffers
  _Aj_row_offsets.resize(tuple_count+1);
  thrust::fill(_Aj_row_offsets.begin(), _Aj_row_offsets.end(), 0);	  
  thrust::inclusive_scan(per_row_counts.begin(), per_row_counts.end(), _Aj_row_offsets.begin()+1);
  
  // 2. get col indices and values:---
  // allocate col indices & values 
  // for each tuple (index,domain) = (i,j):
  //   for each col c of Ai that is in Dj
  //     write Aic into local matrix
  //     write local index for i into local matrix  
  
  IndexType num_Aj_entries = _Aj_row_offsets[_Aj_row_offsets.size()-1];
  _Aj_values.resize(num_Aj_entries);
  _Aj_column_indices.resize(num_Aj_entries);

  // calculate max storage for Aj for all j
  _max_domain_storage = thrust::transform_reduce(
    thrust::make_counting_iterator<IndexType>(0), 
    thrust::make_counting_iterator<IndexType>(0)+_num_domains,
    detail::count_domain_storage<IndexType>(thrust::raw_pointer_cast(&*_domain_pointers.begin()), thrust::raw_pointer_cast(&*_Aj_row_offsets.begin())), 
    (IndexType)0, thrust::maximum<IndexType>());
  // function object count_domain_storage is defined at the beginning.
  
    cta_size    = std::min(BLOCKDIM, _max_domain_size);
    _shmem_size = cta_size * sizeof(ValueType);
	// _shmem_size = detail::asm_local_jacobi_smoother<false,false,false,ValueType,IndexType>::required_shmem_bytes(_max_domain_size, _max_domain_storage);
    printf("requested block size: %d, cta_size: %d, _shmem_size: %d\n", _max_domain_size, cta_size, _shmem_size);
  
	new_x.resize(_domain_global_indices.size());
	thrust::fill(new_x.begin(), new_x.end(),(ValueType)0);		
	local_x.resize(_domain_global_indices.size());
	thrust::fill(local_x.begin(), local_x.end(),(ValueType)0);	
	
	_tuple_domain = tuple_domain;
	_tuple_count = tuple_count;	
  } else if (this->same_mesh && !this->same_structure) {
    if (A_unsorted.num_rows != domains.size())  
      throw cusp::invalid_input_exception("domain array size does not match array dimension");  
	_domains = domains;
	_num_domains = thrust::max_element(_domains.begin(), _domains.end())[0]+1;  
  }

	initialized = true;			
}

//  constructor 
template <typename IndexType, typename ValueType, typename MemorySpace>
template<typename MatrixType, typename VectorType>
void additive_schwarz<IndexType,ValueType,MemorySpace>
    ::setup(const MatrixType& A_unsorted, 
	const VectorType& domains, 
	asm_type type_, 
	const IndexType overlap, 
	const IndexType sweeps_,
	const IndexType iter, 
	const ValueType jacobi_tol_)
{
  CUSP_PROFILE_SCOPED();

	if (!initialized) {
		 printf("additive_schwarz has not yet been created!\n");
		 exit(1);		
	}
		
	sweeps = sweeps_;
	jacobi_iterations = iter;
	jacobi_tol = jacobi_tol_;
	type = type_;

  // convert to tuple (domain, element)
  
  // allocate enough space for A.num_rows * max(degree of A)
  
  // repeat [overlap] times
  //   for each tuple (d, i):
  //     for each neighbor j of i
  //       if domain[j] != d, add (d, j) to tuple list
  
  //   remove copies from tuple_list
  
  // sort tuples by domain
  
  // segmented_scan (tuple_list) - get local index of each i within its domain d
  
  // segmented_reduction (tuple_list) - get per-domain counts
  
  // for each domain j (in parallel)
  //   form Aj = Pj A Pj^t
  //   append all Aj's into one large buffer
  
  //   (validate that shmem is large enough)

  //------------------------------------------------------------
  // first sort the entries by row & col, which is required by the matrix assembly routines below
  coo_matrix<IndexType, ValueType, MemorySpace> Acoo = A_unsorted;
  thrust::sort_by_key(thrust::make_zip_iterator(thrust::make_tuple(Acoo.row_indices.begin(), Acoo.column_indices.begin())),
                      thrust::make_zip_iterator(thrust::make_tuple(Acoo.row_indices.end(),   Acoo.column_indices.end())),
                      Acoo.values.begin());
  csr_matrix<IndexType, ValueType, MemorySpace> A = Acoo;
  
  if (!same_structure) {
  // validate inputs
  if (A.num_rows != A.num_cols) 
    throw cusp::invalid_input_exception("input matrix is not square");

  if (overlap < 0)
    throw cusp::invalid_input_exception("overlap must be non-negative");

  if (((IndexType)-1) >= 0)
    throw cusp::invalid_input_exception("IndexType must be a signed type");

  _n = A.num_rows;
  
  if (!same_mesh) {
  if (A.num_rows != domains.size())  
    throw cusp::invalid_input_exception("domain array size does not match array dimension");  
  _domains = domains;
  _num_domains = thrust::max_element(_domains.begin(), _domains.end())[0]+1;
  }
  
  cusp::array1d<IndexType, MemorySpace> tuple_domain = _domains;
  cusp::array1d<IndexType, MemorySpace> tuple_index(_n); // finally the set of vertices we need when overlap > 0
  thrust::sequence(tuple_index.begin(), tuple_index.end());

  IndexType tuple_count = _n; //initialize to be _n, will be enlarged	
  for (IndexType sweep=0; sweep < overlap; sweep ++) {

    cusp::array1d<IndexType,MemorySpace> counts(tuple_count+1, 0);

    const size_t MAX_BLOCKS_count_neighbor_regions = cusp::detail::device::arch::max_active_blocks(detail::count_neighbor_regions<IndexType>, BLOCKDIM, (size_t) 0);
    const size_t NUM_BLOCKS_count_neighbor_regions = std::min<size_t>(MAX_BLOCKS_count_neighbor_regions, cusp::detail::device::DIVIDE_INTO(tuple_count, BLOCKDIM));	
    detail::count_neighbor_regions<IndexType><<<NUM_BLOCKS_count_neighbor_regions, BLOCKDIM>>>(
      thrust::raw_pointer_cast(&*(counts.begin()+1)),
      thrust::raw_pointer_cast(&*tuple_domain.begin()),
      thrust::raw_pointer_cast(&*tuple_index.begin()),
      thrust::raw_pointer_cast(&*A.row_offsets.begin()),
      thrust::raw_pointer_cast(&*A.column_indices.begin()), 
      thrust::raw_pointer_cast(&*_domains.begin()), 
      tuple_count);

    // convert per-entry counts into pointers for where each vertex places its entries
    thrust::inclusive_scan(counts.begin(), counts.end(), counts.begin());
    IndexType number_of_verts_to_add = counts[tuple_count];
    tuple_domain.resize(tuple_count + number_of_verts_to_add);
    tuple_index.resize(tuple_count + number_of_verts_to_add);

    // write new entries into end of list
    const size_t MAX_BLOCKS_grow_neighbor_regions = cusp::detail::device::arch::max_active_blocks(detail::grow_neighbor_regions<IndexType>, BLOCKDIM, (size_t) 0);
    const size_t NUM_BLOCKS_grow_neighbor_regions = std::min<size_t>(MAX_BLOCKS_grow_neighbor_regions, cusp::detail::device::DIVIDE_INTO(tuple_count, BLOCKDIM));	
    detail::grow_neighbor_regions<IndexType><<<NUM_BLOCKS_grow_neighbor_regions, BLOCKDIM>>>(
      thrust::raw_pointer_cast(&*(tuple_domain.begin()+tuple_count)),
      thrust::raw_pointer_cast(&*(tuple_index.begin()+tuple_count)),
      thrust::raw_pointer_cast(&*tuple_domain.begin()),
      thrust::raw_pointer_cast(&*tuple_index.begin()),
      thrust::raw_pointer_cast(&*A.row_offsets.begin()),
      thrust::raw_pointer_cast(&*A.column_indices.begin()), 
      thrust::raw_pointer_cast(&*_domains.begin()), 
      thrust::raw_pointer_cast(&*counts.begin()), 
      tuple_count);

    // sort tuples - this bit of cleverness courtesy Nathan.  This sorts in lexicographical ordering, but uses the fast radix sort
    thrust::stable_sort_by_key(tuple_index.begin(), tuple_index.end(), tuple_domain.begin());
    thrust::stable_sort_by_key(tuple_domain.begin(), tuple_domain.end(), tuple_index.begin());

    // remove duplicates
    cusp::array1d<IndexType, MemorySpace> new_tuple_domain(tuple_domain.size());
    cusp::array1d<IndexType, MemorySpace> new_tuple_index(tuple_index.size());

    thrust::zip_iterator<thrust::tuple<typename cusp::array1d<IndexType,MemorySpace>::iterator, typename cusp::array1d<IndexType,MemorySpace>::iterator> > last_iter = 
      thrust::unique_copy(thrust::make_zip_iterator(thrust::make_tuple(tuple_domain.begin(), tuple_index.begin())),
                          thrust::make_zip_iterator(thrust::make_tuple(tuple_domain.end(), tuple_index.end())),
                          thrust::make_zip_iterator(thrust::make_tuple(new_tuple_domain.begin(), new_tuple_index.begin())));

    // get number of newly added entries
    IndexType new_size = thrust::distance(thrust::make_zip_iterator(thrust::make_tuple(new_tuple_domain.begin(), new_tuple_index.begin())), last_iter);

    // truncate
    new_tuple_index.resize(new_size);
    new_tuple_domain.resize(new_size);

    // copy back to original vector
    tuple_index.swap(new_tuple_index);
    tuple_domain.swap(new_tuple_domain);

    tuple_count = new_size;
  }

  // sort so entries in the same domain are adjacent, and in ascending order 
  thrust::stable_sort_by_key(tuple_domain.begin(), tuple_domain.end(), tuple_index.begin());

  //tuple_domain a risized tuple list containing tuples that are enlarged to include neighbour nodes
  // a new tuple: tuple_index
  // tuple_domain are sorted to be tuple_index

  // two temp arrays used for later important array: domain_list/domain_count -> _dj -> _domain_pointers
  cusp::array1d<IndexType,MemorySpace> domain_list(tuple_count);// size is m if we have m domain, m <= tuple_count
  cusp::array1d<IndexType,MemorySpace> domain_count(tuple_count);

  typename cusp::array1d<IndexType,MemorySpace>::iterator last_domain_list_iter =  
    thrust::reduce_by_key(tuple_domain.begin(), tuple_domain.end(), thrust::make_constant_iterator<IndexType>(1), domain_list.begin(), domain_count.begin()).first;

  // gather domain_count into dj from domain_list (slots not written to will be 0)
  // TODO: for each i in [domain_list.begin(), last_domain_list) _dj[domain_list[i]] = domain_count[i]
  // scatter: i in the range [first, last), the value *i is assigned to output[*(map + (i - first))]
  _dj.resize(_num_domains);
  thrust::fill(_dj.begin(), _dj.end(), 0);	  
  IndexType num_domains_counted = thrust::distance(domain_list.begin(), last_domain_list_iter);
  thrust::scatter(domain_count.begin(), domain_count.begin() + num_domains_counted, domain_list.begin(), _dj.begin()); 

  // calculate the size of the largest domain
  _max_domain_size = thrust::max_element(_dj.begin(), _dj.end())[0];

  // fill the domain pointers (scan of domain_count) 
  _domain_pointers.resize(_num_domains+1);
  thrust::fill(_domain_pointers.begin(), _domain_pointers.end(), 0);	   
  thrust::inclusive_scan(_dj.begin(), _dj.end(), _domain_pointers.begin()+1);

  // segmented scan to get local indices, tuple_index has corresponding global index
  _domain_local_indices.resize(tuple_count);
  thrust::exclusive_scan_by_key(tuple_domain.begin(), tuple_domain.end(), thrust::make_constant_iterator<IndexType>(1), _domain_local_indices.begin());
  _domain_global_indices = tuple_index;

  //1.   (domain, index) pairs: tuple_domain, 
  //2.	 _domain_global_indices: see 538
  //3.	 _domain_pointers: array with size = domain num, elements are the index of tuple_domain
  //     (since has been sorted, element of this array is getting larger), help to compute row_offsets, see line 51 above
  //4.   per-domain counts: _dj
  // we can also figure out if an entry is in the overlap region or not (tuple (d,i) is overlap iff if domain[i] != d)
  //--------------------------------------------
  // build the local matrices Aj & represent them somehow...
  // CSR format: 2 steps
  // 1. get row_offsets:---
  // for each tuple (index,domain) = (i,j):
  //   count number of cols Ai are in Domain j (list merge is fast because per-row entries in sorted order, as are domain indices
  // scan (counts) -> row_offsets

  // NOTE: By selecting exactly the submatrix corresponding to a domain, we implicitly set zero-dirichelet boundary conditions.!!!!!!!!
  // However, when overlap > 0, transmission conditions are not needed.  For the case of no overlap, we would probably want
  // to implement something more sophisticated like Optimized RAS (i.e. Robin transmission conditions).
  cusp::array1d<IndexType,MemorySpace> per_row_counts(tuple_count, 0);

  const size_t MAX_BLOCKS_submatrix_per_row_counts = cusp::detail::device::arch::max_active_blocks(detail::submatrix_per_row_counts<IndexType>, BLOCKDIM, (size_t) 0);
  const size_t NUM_BLOCKS_submatrix_per_row_counts = std::min<size_t>(MAX_BLOCKS_submatrix_per_row_counts, cusp::detail::device::DIVIDE_INTO(tuple_count, BLOCKDIM));	  
  detail::submatrix_per_row_counts<IndexType><<<NUM_BLOCKS_submatrix_per_row_counts,BLOCKDIM>>>(
    thrust::raw_pointer_cast(&*per_row_counts.begin()), 
    thrust::raw_pointer_cast(&*_domain_pointers.begin()), 
    thrust::raw_pointer_cast(&*tuple_domain.begin()), 
    thrust::raw_pointer_cast(&*_domain_global_indices.begin()), 
    thrust::raw_pointer_cast(&*A.row_offsets.begin()),
    thrust::raw_pointer_cast(&*A.column_indices.begin()), 
    tuple_count);  

  // convert into offsets, get the total size, and allocate buffers
  _Aj_row_offsets.resize(tuple_count+1);
  thrust::fill(_Aj_row_offsets.begin(), _Aj_row_offsets.end(), 0);	  
  thrust::inclusive_scan(per_row_counts.begin(), per_row_counts.end(), _Aj_row_offsets.begin()+1);
  
  // 2. get col indices and values:---
  // allocate col indices & values 
  // for each tuple (index,domain) = (i,j):
  //   for each col c of Ai that is in Dj
  //     write Aic into local matrix
  //     write local index for i into local matrix  
  
  IndexType num_Aj_entries = _Aj_row_offsets[_Aj_row_offsets.size()-1];
  _Aj_values.resize(num_Aj_entries);
  _Aj_column_indices.resize(num_Aj_entries);

  // calculate max storage for Aj for all j
  _max_domain_storage = thrust::transform_reduce(
    thrust::make_counting_iterator<IndexType>(0), 
    thrust::make_counting_iterator<IndexType>(0)+_num_domains,
    detail::count_domain_storage<IndexType>(thrust::raw_pointer_cast(&*_domain_pointers.begin()), thrust::raw_pointer_cast(&*_Aj_row_offsets.begin())), 
    (IndexType)0, thrust::maximum<IndexType>());
  // function object count_domain_storage is defined at the beginning.
  
    cta_size    = std::min(BLOCKDIM, _max_domain_size);
    _shmem_size = cta_size * sizeof(ValueType);
	// _shmem_size = detail::asm_local_jacobi_smoother<false,false,false,ValueType,IndexType>::required_shmem_bytes(_max_domain_size, _max_domain_storage);
    printf("requested block size: %d, cta_size: %d, _shmem_size: %d\n", _max_domain_size, cta_size, _shmem_size);
  
	new_x.resize(_domain_global_indices.size());
	thrust::fill(new_x.begin(), new_x.end(),(ValueType)0);		
	local_x.resize(_domain_global_indices.size());
	thrust::fill(local_x.begin(), local_x.end(),(ValueType)0);	
	
	_tuple_domain = tuple_domain;	
	_tuple_count = tuple_count;
  }

  // finally, fill in the submatrix Aj data
  const size_t MAX_BLOCKS_submatrix_fill = cusp::detail::device::arch::max_active_blocks(
  detail::submatrix_fill<IndexType, ValueType>, BLOCKDIM, (size_t) 0);
  const size_t NUM_BLOCKS_submatrix_fill = std::min<size_t>(MAX_BLOCKS_submatrix_fill, cusp::detail::device::DIVIDE_INTO(_tuple_count, BLOCKDIM));	  
  
  detail::submatrix_fill<IndexType, ValueType><<<NUM_BLOCKS_submatrix_fill,BLOCKDIM>>>(
    thrust::raw_pointer_cast(&*_Aj_column_indices.begin()), 
    thrust::raw_pointer_cast(&*_Aj_values.begin()), 
    thrust::raw_pointer_cast(&*_Aj_row_offsets.begin()), 
    thrust::raw_pointer_cast(&*_domains.begin()), 
    thrust::raw_pointer_cast(&*_domain_pointers.begin()), 
    thrust::raw_pointer_cast(&*_tuple_domain.begin()), 
    thrust::raw_pointer_cast(&*_domain_global_indices.begin()), 
    thrust::raw_pointer_cast(&*_domain_local_indices.begin()), 
    thrust::raw_pointer_cast(&*A.row_offsets.begin()),
    thrust::raw_pointer_cast(&*A.column_indices.begin()), 
    thrust::raw_pointer_cast(&*A.values.begin()), 
    _tuple_count);  
	
    // for each subblock, calculate the max_i(sum_j |Aij|) as estimate for largest eigenvalue
    _domain_lambda_max.resize(_num_domains);	
    const size_t MAX_BLOCKS_estimate_local_lamdba_max = cusp::detail::device::arch::max_active_blocks(
	detail::estimate_local_lamdba_max<IndexType, ValueType>,
	cta_size, (size_t) _shmem_size);
    const size_t NUM_BLOCKS_estimate_local_lamdba_max = std::min<size_t>(MAX_BLOCKS_estimate_local_lamdba_max, _num_domains);
	
    detail::estimate_local_lamdba_max<IndexType, ValueType>
    <<<NUM_BLOCKS_estimate_local_lamdba_max, cta_size, _shmem_size>>>(
      thrust::raw_pointer_cast(&*_domain_lambda_max.begin()), 
      thrust::raw_pointer_cast(&*_dj.begin()), 
      thrust::raw_pointer_cast(&*_domain_pointers.begin()), 
      thrust::raw_pointer_cast(&*_Aj_row_offsets.begin()),
      thrust::raw_pointer_cast(&*_Aj_column_indices.begin()),	  
      thrust::raw_pointer_cast(&*_Aj_values.begin()),
	  _num_domains);		
}

template <typename IndexType, typename ValueType, typename MemorySpace>
template<bool UseCache, bool Initialize, bool UseOmega, typename VectorType1, typename VectorType2>
    void additive_schwarz<IndexType,ValueType,MemorySpace>
    ::apply(const VectorType1& bb, VectorType2& xx)
{
    typename detail::asm_local_jacobi_smoother<UseCache, Initialize, UseOmega, ValueType, IndexType>::params p;
    p.iterations = jacobi_iterations;
	p.tol 		 = jacobi_tol;
	p.blockdim   = cta_size;
	p.l_inf_colsum_ptr = thrust::raw_pointer_cast(&*_domain_lambda_max.begin());

    const size_t MAX_BLOCKS = cusp::detail::device::arch::max_active_blocks(
	detail::run_local_solvers<IndexType, ValueType, detail::asm_local_jacobi_smoother<UseCache, Initialize, UseOmega, ValueType, IndexType> >,
	cta_size, (size_t) _shmem_size);
    const size_t NUM_BLOCKS = std::min<size_t>(MAX_BLOCKS, _num_domains);

	// use texture
	const ValueType *rhs = thrust::raw_pointer_cast(&*bb.begin());
    if (UseCache)
    bind_x(rhs);
	
    detail::run_local_solvers<IndexType, ValueType, detail::asm_local_jacobi_smoother<UseCache, Initialize, UseOmega, ValueType, IndexType> >
    <<<NUM_BLOCKS, cta_size, _shmem_size>>>(
      thrust::raw_pointer_cast(&*xx.begin()),
      thrust::raw_pointer_cast(&*local_x.begin()),	 
      thrust::raw_pointer_cast(&*new_x.begin()),	
      rhs,
      thrust::raw_pointer_cast(&*_dj.begin()),
      thrust::raw_pointer_cast(&*_domains.begin()),
      thrust::raw_pointer_cast(&*_domain_pointers.begin()),
      thrust::raw_pointer_cast(&*_domain_global_indices.begin()),	
      thrust::raw_pointer_cast(&*_Aj_row_offsets.begin()),
      thrust::raw_pointer_cast(&*_Aj_column_indices.begin()),
      thrust::raw_pointer_cast(&*_Aj_values.begin()),
	  type,
	  _num_domains,
      p);
	  
	// use texture
    if (UseCache)
    unbind_x(rhs);    
}

//! linear_operator which applies the preconditioner 
//! \param bb current RHS vector
//! \param xx vector which shall be smoothed
template <typename IndexType, typename ValueType, typename MemorySpace>
template<typename VectorType1, typename VectorType2>
    void additive_schwarz<IndexType,ValueType,MemorySpace>
    ::operator()(const VectorType1& bb, VectorType2& xx)
{
    CUSP_PROFILE_SCOPED();
    // invoke a kernel: 1 thread block per domain.
    // per cta:
    //   read corresponding submatrix & RHS into shmem / registers
    //   -sync-
    //   solve local system with 0-initial guess (jacobi iterations)
    //   -sync-
    //   for non-overlapped points, write solution

    cudaFuncCache old_cache_setting;
    cudaThreadGetCacheConfig(&old_cache_setting);
    cudaThreadSetCacheConfig(cudaFuncCachePreferShared);
	
	if(type == basic) {
	cusp::blas::fill(xx,(ValueType)0);
	}
	cusp::blas::fill(local_x,(ValueType)0);
	
	#ifdef CUSP_USE_TEXTURE_MEMORY
	apply<true, false, false>(bb, xx);
	#else
	apply<false, false, false>(bb, xx);	
	#endif
	
	cudaThreadSetCacheConfig(old_cache_setting);
}	

//! apply as presmooth
//! \param bb current RHS vector
//! \param xx vector which shall be smoothed
template <typename IndexType, typename ValueType, typename MemorySpace>
template<typename MatrixType, typename VectorType1, typename VectorType2>
    void additive_schwarz<IndexType,ValueType,MemorySpace>
    ::presmooth(const MatrixType& A, const VectorType1& bb, VectorType2& xx)
{
    CUSP_PROFILE_SCOPED();
    // invoke a kernel: 1 thread block per domain.
    // per cta:
    //   read corresponding submatrix & RHS into shmem / registers
    //   -sync-
    //   solve local system with 0-initial guess (jacobi iterations)
    //   -sync-
    //   for non-overlapped points, write solution

    cudaFuncCache old_cache_setting;
    cudaThreadGetCacheConfig(&old_cache_setting);
    cudaThreadSetCacheConfig(cudaFuncCachePreferShared);
	
	if(type == basic) {
	cusp::blas::fill(xx,(ValueType)0);
	}
	
	#ifdef CUSP_USE_TEXTURE_MEMORY
	apply<true, true, true>(bb, xx);
	#else
	apply<false, true, true>(bb, xx);	
	#endif
	
	VectorType1 residual(A.num_rows,ValueType(0));
	VectorType1 e(A.num_rows,ValueType(0));	
	for( int s = 0; s<sweeps-1; s++ )
	{
	// compute residual <- b - A*x
	cusp::multiply(A, xx, residual);
	cusp::blas::axpby(bb, residual, residual, ValueType(1), ValueType(-1));
	
	if(type == basic) {
	cusp::blas::fill(e,(ValueType)0);
	}

	#ifdef CUSP_USE_TEXTURE_MEMORY
	apply<true, true, true>(residual, e);
	#else
	apply<false, true, true>(residual, e);	
	#endif
	
	cusp::blas::axpby(xx, e, xx, ValueType(1), ValueType(1));									
	}		
	
	cudaThreadSetCacheConfig(old_cache_setting);
}	

//! apply as postsmooth
//! \param bb current RHS vector
//! \param xx vector which shall be smoothed
template <typename IndexType, typename ValueType, typename MemorySpace>
template<typename MatrixType, typename VectorType1, typename VectorType2>
    void additive_schwarz<IndexType,ValueType,MemorySpace>
    ::postsmooth(const MatrixType& A, const VectorType1& bb, VectorType2& xx)
{
    CUSP_PROFILE_SCOPED();
    // invoke a kernel: 1 thread block per domain.
    // per cta:
    //   read corresponding submatrix & RHS into shmem / registers
    //   -sync-
    //   solve local system with 0-initial guess (jacobi iterations)
    //   -sync-
    //   for non-overlapped points, write solution

    cudaFuncCache old_cache_setting;
    cudaThreadGetCacheConfig(&old_cache_setting);
    cudaThreadSetCacheConfig(cudaFuncCachePreferShared);

	VectorType1 residual(A.num_rows,ValueType(0));
	VectorType1 e(A.num_rows,ValueType(0));	
	for( int s = 0; s<sweeps; s++ )
	{
	// compute residual <- b - A*x
	cusp::multiply(A, xx, residual);
	cusp::blas::axpby(bb, residual, residual, ValueType(1), ValueType(-1));
	
	if(type == basic) {
	cusp::blas::fill(e,(ValueType)0);
	}

	// initialize the subdomain with e = D^{-1}*residual internally in apply() 
	#ifdef CUSP_USE_TEXTURE_MEMORY
	apply<true, true, true>(residual, e);
	#else
	apply<false, true, true>(residual, e);	
	#endif
	
	cusp::blas::axpby(xx, e, xx, ValueType(1), ValueType(1));									
	}		
	
	cudaThreadSetCacheConfig(old_cache_setting);
}	

} // end namespace precond
} // end namespace cusp

