
/*! \file asm_local_jacobi.inl
 *  \brief Inline file for local jacobi smoother class 
 */

namespace cusp
{
namespace precond
{
namespace detail
{

//! works for up to 16-bit numbers(int)
//! from http://bits.stephan-brumme.com/roundUpToNextPowerOfTwo.html
//! copied from http://graphics.stanford.edu/~seander/bithacks.html
//! See that web site for an explanation.
//! Original source claims this code is in the public domain.
__device__ int asm_round_to_next_power_of_two(int x)
{
    x--;
    x |= x >> 1;  // handle  2 bit numbers
    x |= x >> 2;  // handle  4 bit numbers
    x |= x >> 4;  // handle  8 bit numbers
    x |= x >> 8;  // handle 16 bit numbers
    return x+1;
}

template<bool UseCache, bool Initialize,  bool UseOmega, typename ValueType, typename IndexType>
struct asm_local_jacobi_smoother
{
  struct params {
    IndexType iterations;
	IndexType blockdim;
    ValueType tol;
    ValueType *l_inf_colsum_ptr;
  };

  IndexType iterations;
  IndexType blockdim;
  ValueType tol;
  ValueType omega;
   
  __device__ asm_local_jacobi_smoother(IndexType j, const params &p) {
    this->iterations   = p.iterations;
    this->blockdim     = p.blockdim;	
	this->tol 		   = p.tol;
	this->omega		   = UseOmega ? ValueType(4.0/3.0) / p.l_inf_colsum_ptr[j] : ValueType(1);
  }

  static size_t required_shmem_bytes(IndexType max_domain_size, IndexType max_domain_storage) {
    //return BLOCKDIM * sizeof(ValueType); 
	return 0;
  }

  __device__ void operator()(IndexType i, const IndexType domain_size, const IndexType domain_start,
      ValueType *local_x, ValueType *new_x, const IndexType *A_row_offsets, const IndexType *A_col_id, const ValueType *A_values, 
	  const ValueType *rhs, const IndexType *domain_global_indices, ValueType *local_scratch) {
	  
	// initialize x <- omega * D^-1 * b 
	if (Initialize) {
	  for (IndexType k = i; k < domain_size; k += blockdim) { 
	    IndexType global_k = domain_start + k;
	    IndexType global_i = domain_global_indices[global_k];
        IndexType row_start = A_row_offsets[global_k]; // A_row_offsets is global, must use global index global_k
        IndexType row_end   = A_row_offsets[global_k+1];
		ValueType inv_diag  = 1;		
		
        for (IndexType row_ptr = row_start; row_ptr < row_end; row_ptr++) { // traversal all nnzeros in this row
          IndexType col_id = A_col_id[row_ptr];
          if (col_id == k) 
            inv_diag = ((ValueType)1) / A_values[row_ptr];  // diag element
        }
		local_x[global_k]  = omega * inv_diag * fetch_x<UseCache>(global_i, rhs);
	  }	  
	}  
	  
    for (IndexType iteration = Initialize ? 1:0; iteration < this->iterations; iteration++) {
	  // initialize shared memory
	  local_scratch[i] = (ValueType)0;
      __syncthreads();
	  
	  // each thread handle k elements	  
	  for (IndexType k = i; k < domain_size; k += blockdim) { 
	    IndexType global_k = domain_start + k;
	    IndexType global_i = domain_global_indices[global_k];
        IndexType row_start = A_row_offsets[global_k]; // A_row_offsets is global, must use global index global_k
        IndexType row_end   = A_row_offsets[global_k+1];
        ValueType inv_diag  = 1;
        ValueType rhs_accum = 0;
		
        for (IndexType row_ptr = row_start; row_ptr < row_end; row_ptr++) { // traversal all nnzeros in this row
          IndexType col_id = A_col_id[row_ptr];
          if (col_id == k) 
            inv_diag = ((ValueType)1) / A_values[row_ptr];  // diag element
          else 
            rhs_accum += (A_values[row_ptr] * local_x[domain_start + col_id]); // other elements
        }
		// jacobi iteration	
        new_x[global_k] = ((ValueType)1 - omega) * local_x[global_k] + omega * inv_diag * (fetch_x<UseCache>(global_i, rhs) - rhs_accum); 

		ValueType error = new_x[global_k] - local_x[global_k];
		local_scratch[i] += error * error;
	  }
	  __syncthreads();

	  // now do a cta-wide reduction with '+' as the operator 
	  IndexType width = asm_round_to_next_power_of_two(blockdim)/2; // find the cloest 2^n integer for blockdim and divide half
	  while (width > 0) {
		if ((i < width) && (i+width < blockdim))
        local_scratch[i] += local_scratch[i+width];
		width /= 2;
		__syncthreads();
	  }

	  // converge criteria
	  if (local_scratch[0] <= tol*tol)  break;

	  // update	
	  for (IndexType k = i; k < domain_size; k += blockdim) {	
		IndexType global_k = domain_start + k;	 	  
        local_x[global_k]  = new_x[global_k];
	  }
    }
  }
};


} // end namespace detail 
} // end namespace precond 
} // end namespace cusp 

