/*
 *  Copyright 2008-2009 NVIDIA Corporation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

#include <cusp/blas.h>
#include <cusp/coo_matrix.h>
#include <cusp/csr_matrix.h>

#include <thrust/scan.h>
#include <thrust/reduce.h>

#define BLOCK_DIMX 16            // tile (and threadblock) size in x
#define BLOCK_DIMY 16            // tile (and threadblock) size in y

namespace cusp
{
namespace precond
{
namespace detail
{
template <typename IndexType>
__global__ void count_nnz(const IndexType mx, const IndexType my, const IndexType Mx, const IndexType My, const IndexType dof, IndexType *entries_row)
{
	/* 2D global index for inner data(without ghost points) of this processor */
	IndexType ix = blockIdx.x * blockDim.x + threadIdx.x;
	IndexType iy = blockIdx.y * blockDim.y + threadIdx.y;
	IndexType k,row_offset;

	if (ix<mx && iy<my) {
		row_offset = iy*mx*dof+ix*dof;
		for (k=0; k<dof; k++) {		
		if (ix == mx-1 && iy == my-1) {
		entries_row[row_offset+k] = 1; 		
		} else if (ix == mx-1 || iy == my-1) {
		entries_row[row_offset+k] = 2; 				
		} else {
		entries_row[row_offset+k] = 4; 
		}
		}
	}
}

template <typename IndexType, typename ValueType>
__global__ void formP_coo(const IndexType mx, const IndexType my, const IndexType Mx, const IndexType My, 
					const IndexType dof, 
					IndexType *row_pointers, IndexType *row_indices,
					IndexType *column_indices, ValueType *values)
{
	/* 2D global index for inner data(without ghost points) of this processor */
	IndexType ix = blockIdx.x * blockDim.x + threadIdx.x;
	IndexType iy = blockIdx.y * blockDim.y + threadIdx.y;
	IndexType i_c,j_c;
	IndexType row,col,row_base,col_base,k;
	ValueType x,y;

	if(ix<mx && iy<my) {		
		i_c = ix*(Mx-1)/(mx-1);    // coarse grid node to the left of fine grid node 
        j_c = iy*(My-1)/(my-1);    // coarse grid node below fine grid node 
		
	    /* 
         Only include those interpolation points that are truly 
         nonzero. Note this is very important for final grid lines
         in x and y directions; since they have no right/top neighbors
         */		
        x  = ((ValueType)ix/(ValueType)(mx-1))*(Mx-1)-i_c; // (x-x_i)/dx 
        y  = ((ValueType)iy/(ValueType)(my-1))*(My-1)-j_c; // (y-y_i)/dy
		
		row_base=iy*mx*dof+ix*dof;   /* convert fine mesh to local "natural" numbering, corresponding to row index of P */
		col_base=j_c*Mx*dof+i_c*dof; /* convert fine mesh to local "natural" numbering, corresponding to row index of P */
		/* has one point of left and below; or we are right on it */
		for (k=0; k<dof; k++) {
		  row = row_base+k; 
		  IndexType entries_per_row = 0;
		  IndexType row_offset = row_pointers[row];
		  /* has one point of left and below; or we are right on it */
		  {		  
		  IndexType nnz_row=row_offset;
		  values[nnz_row] = x*y - x - y + 1.0;  /* (1-x)(1-y) */
		  col = col_base+k;  	   
		  row_indices[nnz_row] = row; column_indices[nnz_row] = col; 
		  entries_per_row++;
		  }
          /* has one point of right and below */
		  if (ix != mx-1) { // right boundary does not have right point
		  IndexType nnz_row=row_offset+entries_per_row;
		  col = col_base+dof+k;  // convert coarse mesh to local "natural" numbering, corresponding to column index of P 		  
		  values[nnz_row] = -x*y + x;  	   /* x(1-y) */	  
		  row_indices[nnz_row] = row; column_indices[nnz_row] = col;
		  entries_per_row++;
          }
          /* has one point of left and above */
          if (iy != my-1) {
		  IndexType nnz_row=row_offset+entries_per_row;
		  col = col_base+Mx*dof+k;  // convert coarse mesh to local "natural" numbering, corresponding to column index of P 
		  values[nnz_row] = -x*y + y;   /* (1-x)y */
		  row_indices[nnz_row] = row; column_indices[nnz_row] = col;
          entries_per_row++;
		  }
        /* has one point of right and above */
          if (!(ix == mx-1 || iy == my-1)) { 
		  IndexType nnz_row=row_offset+entries_per_row;		  
		  col = col_base+(Mx+1)*dof+k;  // convert coarse mesh to local "natural" numbering, corresponding to column index of P
		  values[nnz_row] = x*y;
		  row_indices[nnz_row] = row; column_indices[nnz_row] = col;
          entries_per_row++;
		  }
		}
	}
}

template <typename IndexType, typename ValueType>
void bilinear_prolongator(cusp::coo_matrix<IndexType,ValueType,cusp::device_memory>& P,
				IndexType& mx_c, IndexType& my_c,
				const IndexType mx, const IndexType my,
				const IndexType ratioi, const IndexType ratioj, const IndexType dof)
{
	IndexType Mx=((mx-1)/ratioi)+1+((mx-1)%ratioi?1:0), My=((my-1)/ratioj)+1+((my-1)%ratioj?1:0);	
	IndexType nnz,nnz_all = ((mx-1)*(my-1)*4+(mx-1)*2+(my-1)*2+1)*dof;
	
	mx_c = Mx; my_c = My;
	printf("coarsening (%d, %d) to (%d, %d)\n",mx,my,Mx,My);	
	
	// entries_row count the non zero entry number in each row of P
	cusp::array1d<IndexType, cusp::device_memory> entries_row(mx*my*dof, 0);
	dim3 dimBlock(BLOCK_DIMX, BLOCK_DIMY);
    dim3 dimGrid((mx+BLOCK_DIMX-1) / BLOCK_DIMX, (my+BLOCK_DIMY-1) / BLOCK_DIMY);
	/* precount nonzeros of P */
	count_nnz<IndexType><<<dimGrid, dimBlock>>>(mx,my,Mx,My,dof,
									thrust::raw_pointer_cast(&*entries_row.begin()));	

	// each element of row_pointers	present the start pointer of the first element in this row, the location					
	cusp::array1d<IndexType, cusp::device_memory> row_pointers(mx*my*dof+1, 0);
    thrust::inclusive_scan(entries_row.begin(), entries_row.end(), row_pointers.begin()+1);

	// count the total non zeros number of P
	nnz = thrust::reduce(entries_row.begin(), entries_row.end());
	assert(nnz_all == nnz);
	
	/* resize P */
	P.resize(mx*my*dof, Mx*My*dof, nnz_all); /* rows number equals to fine mesh dof, cols number equals to coarse mesh dof */

	formP_coo<IndexType,ValueType><<<dimGrid, dimBlock>>>(mx,my,Mx,My,dof,
								thrust::raw_pointer_cast(&*row_pointers.begin()),
								thrust::raw_pointer_cast(&*(P.row_indices).begin()),
								thrust::raw_pointer_cast(&*(P.column_indices).begin()),		
								thrust::raw_pointer_cast(&*(P.values).begin()));
}

/* sequence implementation */
template <typename IndexType, typename ValueType>
void bilinear_prolongator(cusp::csr_matrix<IndexType,ValueType,cusp::host_memory>& P,
				IndexType& mx_c, IndexType& my_c,
				const IndexType mx, const IndexType my,
				const IndexType ratioi, const IndexType ratioj, const IndexType dof)
{
	IndexType i,j,k,i_c,j_c;
	IndexType row,col,col_base;
	IndexType nnz,nnz_all = ((mx-1)*(my-1)*4+(mx-1)*2+(my-1)*2+1)*dof;
	IndexType Mx=((mx-1)/ratioi)+1+((mx-1)%ratioi?1:0), My=((my-1)/ratioj)+1+((my-1)%ratioj?1:0);	
	ValueType x,y;	
	
	mx_c = Mx; my_c = My;
	printf("coarsening (%d, %d) to (%d, %d)\n",mx,my,Mx,My);	

	cusp::coo_matrix<IndexType,ValueType,cusp::host_memory> P_coo(mx*my*dof, Mx*My*dof, nnz_all);
	
	nnz = 0;
	for (j=0; j<my; j++) {
     for (i=0; i<mx; i++) {
		i_c = i*(Mx-1)/(mx-1);    // coarse grid node to the left of fine grid node 
        j_c = j*(My-1)/(my-1);    // coarse grid node below fine grid node 
		
         //Only those interpolation points which are truly 
         //nonzero are included. Note this is very important for the last several grid lines
         //in x and y directions; since they have no right/top neighbors
        x  = ((ValueType)i/(ValueType)(mx-1))*(Mx-1)-i_c; // (x-x_i)/dx 
        y  = ((ValueType)j/(ValueType)(my-1))*(My-1)-j_c; // (y-y_i)/dy
		  
		col_base=j_c*Mx*dof+i_c*dof;
	    for (k=0; k<dof; k++) {
		row = j*mx*dof+i*dof+k;
		
		// has one point of left and below; or we are right on it
		{
		  col = col_base+k;  // convert coarse mesh to local "natural" numbering, corresponding to column index of P_coo 		
		  P_coo.values[nnz] = x*y - x - y + 1.0;  /* (1-x)(1-y) */
		  P_coo.row_indices[nnz] = row;				  
		  P_coo.column_indices[nnz] = col;	
		  nnz++;			  
		}

        // has one point of right and below 
        if (i != mx-1) { // right boundary does not have right point
		  col = col_base+dof+k;  // convert coarse mesh to local "natural" numbering, corresponding to column index of P_coo 		  
		  P_coo.values[nnz] = -x*y + x;  	   /* x(1-y) */	 
		  P_coo.row_indices[nnz] = row;		  
		  P_coo.column_indices[nnz] = col; 
		  nnz++;	
		}

        // has one point of left and above 
        if (j != my-1) {
		  col = col_base+Mx*dof+k;  // convert coarse mesh to local "natural" numbering, corresponding to column index of P_coo 
		  P_coo.values[nnz] = -x*y + y;   /* (1-x)y */
		  P_coo.row_indices[nnz] = row;				  
		  P_coo.column_indices[nnz] = col;		  
		  nnz++;	
        }
		
        // has one point of right and above 
        if (!(i == mx-1 || j == my-1)) { 
		  col = col_base+(Mx+1)*dof+k;  // convert coarse mesh to local "natural" numbering, corresponding to column index of P_coo
		  P_coo.values[nnz] = x*y;
		  P_coo.row_indices[nnz] = row;				  
		  P_coo.column_indices[nnz] = col; 
		  nnz++;
        }
	   }
	 }
	}
	assert(nnz_all == nnz);
	P = P_coo;
}

} // end namespace detail
} // end namespace precond
} // end namespace cusp

