/*
 *  Author: Li Luo 
 *  Email:  li.luo@siat.ac.cn 
 */
 
#include <cusp/blas.h>
#include <cusp/elementwise.h>
#include <cusp/multiply.h>
#include <cusp/monitor.h>
#include <cusp/transpose.h>
#include <cusp/graph/maximal_independent_set.h>
#include <cusp/precond/diagonal.h>
#include <cusp/precond/aggregate.h>
#include <cusp/precond/bilinear.h>
#include <cusp/precond/strength.h>
#include <cusp/krylov/arnoldi.h>

#include <cusp/detail/format_utils.h>
#include <cusp/detail/utils_.h>
#include <cusp/detail/spectral_radius.h>

#include <thrust/extrema.h>
#include <thrust/transform.h>
#include <thrust/gather.h>
#include <thrust/reduce.h>

namespace cusp
{
namespace precond
{
namespace detail
{

template <typename MatrixType>
struct Dinv_A : public cusp::linear_operator<typename MatrixType::value_type, typename MatrixType::memory_space>
{
    const MatrixType& A;
    const cusp::precond::diagonal<typename MatrixType::value_type, typename MatrixType::memory_space> Dinv;

    Dinv_A(const MatrixType& A)
        : A(A), Dinv(A),
          cusp::linear_operator<typename MatrixType::value_type, typename MatrixType::memory_space>(A.num_rows, A.num_cols, A.num_entries + A.num_rows)
          {}

    template <typename Array1, typename Array2>
    void operator()(const Array1& x, Array2& y) const
    {
        cusp::multiply(A,x,y);
        cusp::multiply(Dinv,y,y);
    }
};

template <typename MatrixType>
double estimate_rho_Dinv_A(const MatrixType& A)
{
    typedef typename MatrixType::value_type   ValueType;
    typedef typename MatrixType::memory_space MemorySpace;

    Dinv_A<MatrixType> Dinv_A(A);

    return cusp::detail::ritz_spectral_radius(Dinv_A, 8);
}


template <typename T>
struct square : thrust::unary_function<T,T>
{
    __host__ __device__
    T operator()(const T& x) { return x * x; }
};

template <typename T>
struct sqrt_functor : thrust::unary_function<T,T>
{
    __host__ __device__
    T operator()(const T& x) { return sqrt(x); }
};

template <typename IndexType>
struct not_negative_one
{
  __host__ __device__
  bool operator()(const IndexType x)
  {
    return x != -1;
  }
};

template <typename Array1,
typename Array2,
typename MatrixType,
typename Array3>
void fit_candidates(const Array1& aggregates,
                    const Array2& B,
                    MatrixType& Q_,
                    Array3& R)
{
    typedef typename MatrixType::index_type IndexType;
    typedef typename MatrixType::value_type ValueType;
    typedef typename MatrixType::memory_space MemorySpace;

    CUSP_PROFILE_SCOPED();
    IndexType num_unaggregated = thrust::count(aggregates.begin(), aggregates.end(), -1.0);
    IndexType num_aggregates = *thrust::max_element(aggregates.begin(), aggregates.end()) + 1;

    cusp::coo_matrix<IndexType,ValueType,MemorySpace> Q;
    Q.resize(aggregates.size(), num_aggregates, aggregates.size()-num_unaggregated);
    R.resize(num_aggregates);

    // gather values into Q
    thrust::copy_if(thrust::make_zip_iterator(thrust::make_tuple(thrust::counting_iterator<IndexType>(0), aggregates.begin(), B.begin())),
                    thrust::make_zip_iterator(thrust::make_tuple(thrust::counting_iterator<IndexType>(aggregates.size()), aggregates.end(), B.end())),
                    aggregates.begin(),
                    thrust::make_zip_iterator(thrust::make_tuple(Q.row_indices.begin(), Q.column_indices.begin(), Q.values.begin())),
                    not_negative_one<IndexType>());

    // compute norm over each aggregate
    {
        // compute Qt
        cusp::coo_matrix<IndexType,ValueType,MemorySpace> Qt;
        cusp::transpose(Q, Qt);

        // compute sum of squares for each column of Q (rows of Qt)
        cusp::array1d<IndexType, MemorySpace> temp(num_aggregates);
        thrust::reduce_by_key(Qt.row_indices.begin(), Qt.row_indices.end(),
                              thrust::make_transform_iterator(Qt.values.begin(), square<ValueType>()),
                              temp.begin(),
                              R.begin());

        // compute square root of each column sum
        thrust::transform(R.begin(), R.end(), R.begin(), sqrt_functor<ValueType>());
    }

    // rescale columns of Q
    thrust::transform(Q.values.begin(), Q.values.end(),
                      thrust::make_permutation_iterator(R.begin(), Q.column_indices.begin()),
                      Q.values.begin(),
                      thrust::divides<ValueType>());

    // copy/convert Q to output matrix Q_
    Q_ = Q;
}

template <typename Matrix>
void setup_level_matrix(Matrix& dst, Matrix& src){dst.swap(src);}

template <typename Matrix1, typename Matrix2>
void setup_level_matrix(Matrix1& dst, Matrix2& src){dst = src;}

} // end namespace detail

template <typename IndexType, typename ValueType, typename MemorySpace, typename SmootherPolicyType>
template <typename MatrixType>
multigrid<IndexType,ValueType,MemorySpace,SmootherPolicyType>::
multigrid(bool same_structure_, const MatrixType& A, cusparseHandle_t& handle,
	cusp::precond::mg_type mgtype, size_t max_levels, bool red_black, 
	bool use_bilinear_iter, const IndexType dof, const IndexType fmx, const IndexType fmy, 
	const IndexType ratioi, const IndexType ratioj, 	
	const IndexType num_domain, bool use_AS_smoother, bool no_postsmooth_,	
	bool symmetric, bool positive_definite,
	const IndexType order_, const IndexType overlap,
	bool use_mkl_sv, cusp::relaxation::sparskit_ilu_type sparskit_ilutype,
	cusp::relaxation::ilupp_type ilupp_ilu_type, 
	const IndexType prepro_type, const IndexType precond_no,
	const IndexType ilupp_max_levels, const ValueType mem_factor,
	const ValueType threshold, const ValueType min_pivot, 
	bool max_fillin_is_inf, const IndexType ilupp_fill_in,
	bool begin_total_piv, bool detail_param,
	cusp::relaxation::petsc_factor_type petsc_factor_type, 
	const IndexType petsc_factor_fill_in, const ValueType petsc_factor_fill,
	bool allow_diagonal_fill, 
	cusp::relaxation::petsc_factor_mat_ordering_type mat_ordering_type,
	const ValueType petsc_factor_rtol, cusp::precond::coarse_solver_type coarse_solver_)
	: same_structure(same_structure_), no_postsmooth(no_postsmooth_), order(order_)
{
  CUSP_PROFILE_SCOPED();

  // initialize parameters
  {
  BI.use_bilinear_iter = use_bilinear_iter;  
  BI.dof = dof;
  BI.fmx = fmx;
  BI.fmy = fmy;
  BI.ratioi = ratioi;
  BI.ratioj = ratioj;

  AS.restricted = true;  
  AS.use_AS_smoother = use_AS_smoother;
  AS.num_domain = num_domain;
  AS.overlap = overlap;  
  
  MC_SSOR.max_colors = 10;
  MC_SSOR.use_SOR = true;
  MC_SSOR.use_Jacobi = true;
  MC_SSOR.red_black = red_black;

  SPARSKIT2_ILU.use_mkl_sv = use_mkl_sv;
  SPARSKIT2_ILU.sparskit_ilu_type = sparskit_ilutype;    
  SPARSKIT2_ILU.maxfil = 10;
  SPARSKIT2_ILU.lfil = 1;
  SPARSKIT2_ILU.droptol = 1e-4;
  SPARSKIT2_ILU.mbloc = A.num_rows;
  SPARSKIT2_ILU.permtol = 1e-3;  
  SPARSKIT2_ILU.alph = 1.0;     
  
  ILUPP.use_mkl_sv = use_mkl_sv;
  ILUPP.ilupp_ilu_type = ilupp_ilu_type;
  ILUPP.prepro_type = prepro_type;
  ILUPP.precond_no = precond_no;
  ILUPP.max_levels = ilupp_max_levels;
  ILUPP.mem_factor = mem_factor;
  ILUPP.threshold = threshold;
  ILUPP.min_pivot = min_pivot;
  ILUPP.max_fillin_is_inf = max_fillin_is_inf;
  ILUPP.fill_in = ilupp_fill_in;
  ILUPP.begin_total_piv = begin_total_piv;
  ILUPP.detail_param = detail_param;
  
  PETSC_FACTOR.use_mkl_sv = use_mkl_sv;
  PETSC_FACTOR.allow_diagonal_fill = allow_diagonal_fill;
  PETSC_FACTOR.type = petsc_factor_type;
  PETSC_FACTOR.mat_ordering_type = mat_ordering_type;
  PETSC_FACTOR.fill_in = petsc_factor_fill_in;
  PETSC_FACTOR.fill = petsc_factor_fill;
  PETSC_FACTOR.rtol = petsc_factor_rtol;
  
  MKL_ILU0.use_mkl_sv = use_mkl_sv;
  
  DSS.symmetric = symmetric;
  DSS.positive_definite = positive_definite;
  
  theta = 0.;
  sweeps = 1;
  apply_num = 1;
  apply_rtol = 1e-6;
  apply_atol = 1e-10;
  mg_type = mgtype;
  coarse_solver = coarse_solver_;
  setup_already = false;
  }
  
  // avoid reallocations which force matrix copies
  levels.reserve(20);
  SmootherPolicyType::level_smoother.reserve(20);

  levels.push_back(typename multigrid<IndexType,ValueType,MemorySpace,SmootherPolicyType>::level());
  levels.back().A_ = A; // copy
  cusp::detail::sort_by_row_and_column(levels.back().A_);  // sort by row and column  
  if (BI.use_bilinear_iter) {
  setup_already = true;
  // global grid size from the finest level  
  levels.back().mx = BI.fmx; 
  levels.back().my = BI.fmy;
  } else {
  levels.back().B.resize(A.num_rows);
  }

  // max_num_rows of the coarest level has not been considered
  while (levels.size() < max_levels)
  {
    extend_hierarchy(handle);
  }
	
  std::cout << "levels size:" << levels.size() << std::endl;
  
  if (coarse_solver == dss_lu) {
    DSS.opt  = MKL_DSS_DEFAULTS;
	DSS.sym  = DSS.symmetric ? MKL_DSS_SYMMETRIC : MKL_DSS_NON_SYMMETRIC;
	DSS.type = DSS.positive_definite ? MKL_DSS_POSITIVE_DEFINITE : MKL_DSS_INDEFINITE;	
	DSS.nrhs = 1;		
	DSS.error = dss_create(DSS.handle, DSS.opt);
	if (DSS.error != MKL_DSS_SUCCESS) { 
	std::cout << "dss_create error" << std::endl;	
	exit(1);
	}  
  }
  
  IndexType lc = levels.size() - 1;     
  if (coarse_solver == petsc_lu) {
    PETSC_FACTOR.PETSC_LU = new cusp::relaxation::petsc_factor<IndexType,ValueType,MemorySpace>(
	BI.use_bilinear_iter&&same_structure, levels[lc].A_, handle, false,
	cusp::relaxation::petsc_factor_lu, 0, 5.0,
	false, cusp::relaxation::petsc_factor_mat_natural, 1e-10);
  }
}

template <typename IndexType, typename ValueType, typename MemorySpace, typename SmootherPolicyType>
template <typename MatrixType>
void multigrid<IndexType,ValueType,MemorySpace,SmootherPolicyType>::
setup(const MatrixType& A, const ValueType theta_,
	const IndexType apply_num_, const ValueType apply_rtol_, const ValueType apply_atol_, 		
	const IndexType sweeps_,
	const IndexType max_colors, bool use_SOR, bool use_Jacobi,
	const IndexType jacobi_iter_, bool restricted, 
	const IndexType maxfil, const IndexType lfil, const ValueType droptol,
	const IndexType mbloc, const ValueType permtol, const ValueType alph)
{
  CUSP_PROFILE_SCOPED();

  // get parameters
  {
  apply_num = apply_num_;
  apply_rtol = apply_rtol_;
  apply_atol = apply_atol_;
  
  MC_SSOR.max_colors = max_colors;
  MC_SSOR.use_SOR = use_SOR;
  MC_SSOR.use_Jacobi = use_Jacobi;
  
  AS.jacobi_iter = jacobi_iter_;
  AS.restricted = restricted;
  
  SPARSKIT2_ILU.maxfil = maxfil;
  SPARSKIT2_ILU.lfil = lfil;
  SPARSKIT2_ILU.droptol = droptol;
  SPARSKIT2_ILU.mbloc = mbloc;
  SPARSKIT2_ILU.permtol = permtol;  
  SPARSKIT2_ILU.alph = alph; 

  theta = theta_;
  sweeps = sweeps_;  
  }
  
  if (!setup_already) {
  levels[0].A_ = A; // copy new matrix during each setup time
  }
  if (!BI.use_bilinear_iter) {
  thrust::fill(levels[0].B.begin(), levels[0].B.end(), ValueType(1.0));	  
  }

  // setup components on levels except the coarest
  IndexType level = 0;
  while (level+1 < levels.size())
  {
    setup_on_level(level);
	level++;
  }
	
  IndexType lc = levels.size() - 1;
  cusp::array2d<ValueType,cusp::host_memory> coarse_dense;
  ValueType *dense_array;
  IndexType	*pivot_array;
  _INTEGER_t nrows,ncols,nentries;
  _INTEGER_t *rowIndex,*columns;
  _DOUBLE_PRECISION_t *values_;
  switch (coarse_solver)
  { 
	case cusp_lu:   
	coarse_dense = cusp::array2d<ValueType,cusp::host_memory>(levels[lc].A_);
	CULU.LU = cusp::detail::lu_solver<ValueType, cusp::host_memory>(coarse_dense);
	CULU.temp_x.resize(levels[lc].A_.num_rows);
	break;
	case magma_lu:   
	// assign the coarse level to MAGMA for using dense LU solver on GPU 
	MLU.A_dense = levels[lc].A_;
	// get the raw pointer to be the input for MAGMA 
	dense_array = thrust::raw_pointer_cast(&*MLU.A_dense.values.begin());	
	MLU.n     = levels[lc].A_.num_rows;
	MLU.pivot.resize(MLU.n);
	pivot_array = thrust::raw_pointer_cast(&*MLU.pivot.begin());
	magma_xgetrf_gpu(MLU.n,MLU.n,dense_array,MLU.n,pivot_array,&MLU.info);
	break;	
	case dss_lu:   
	nrows = levels[lc].A_.num_rows;
	ncols = levels[lc].A_.num_cols;
	nentries = levels[lc].A_.num_entries;
	
	// setup matrix on host for MKL, redundant sort is unnecessary
	cusp::detail::setup_mkl_matrix(DSS.A_csr,levels[lc].A_);
	
	// get the raw host pointer to be used in DSS
	rowIndex  = thrust::raw_pointer_cast(&*DSS.A_csr.row_offsets.begin());
	columns   = thrust::raw_pointer_cast(&*DSS.A_csr.column_indices.begin());	
    values_   = thrust::raw_pointer_cast(&*DSS.A_csr.values.begin());	
	
	DSS.error = dss_define_structure(DSS.handle, DSS.sym, rowIndex, nrows, ncols, columns, nentries);
	if (DSS.error != MKL_DSS_SUCCESS) { 
	std::cout << "dss_define_structure error" << std::endl;	
	exit(1);
	}  
	DSS.error = dss_reorder(DSS.handle, DSS.opt, 0);
	if (DSS.error != MKL_DSS_SUCCESS) { 
	std::cout << "dss_reorder error" << std::endl;	
	exit(1);
	} 	
	DSS.error = dss_factor_real(DSS.handle, DSS.type, values_);
	if (DSS.error != MKL_DSS_SUCCESS) { 
	std::cout << "dss_factor_real error" << std::endl;	
	exit(1);
	}
    DSS.temp_x.resize(levels[lc].A_.num_rows); 	
	break;	
	case petsc_lu: 
	(PETSC_FACTOR.PETSC_LU)->setup(levels[lc].A_, sweeps);
	break;	  	
	default:	
	(PETSC_FACTOR.PETSC_LU)->setup(levels[lc].A_, sweeps);
	break;
  } 
  
  // Setup solve matrix for each level
  for( size_t lvl = 0; lvl < levels.size(); lvl++ ) {
    detail::setup_level_matrix( levels[lvl].A, levels[lvl].A_ );
  }	
  if (setup_already) { // enable updating the setup matrix
  setup_already = false;
  }
}

template <typename IndexType, typename ValueType, typename MemorySpace, typename SmootherPolicyType>
multigrid<IndexType,ValueType,MemorySpace,SmootherPolicyType>::~multigrid() {	
	// free context of DSS
	if (coarse_solver == dss_lu) {	
	DSS.error = dss_delete(DSS.handle, DSS.opt);
	if (DSS.error != MKL_DSS_SUCCESS) { 
	std::cout << "dss_delete error" << std::endl;	
	exit(1);
	} 		
	} 
	
	if (coarse_solver == petsc_lu) {
	delete PETSC_FACTOR.PETSC_LU;
	}
}

template <typename IndexType, typename ValueType, typename MemorySpace, typename SmootherPolicyType>
void multigrid<IndexType,ValueType,MemorySpace,SmootherPolicyType>::extend_hierarchy(cusparseHandle_t& handle)
{
  CUSP_PROFILE_SCOPED();
  
  SetupMatrixType RAP;  
  cusp::array1d<IndexType,MemorySpace>  domains; 
  IndexType mx_c, my_c; 
   
  if (BI.use_bilinear_iter) {
  SetupMatrixType P;
  detail::bilinear_prolongator(P,mx_c,my_c,levels.back().mx,levels.back().my,BI.ratioi,BI.ratioj,BI.dof);  
  
  if (AS.use_AS_smoother) {
  domains.resize(levels.back().A_.num_rows);  
  // Create parameter domains for additive schwarz smoothing
  int p, q, m, n, m_p, n_q;// p*q=nd, 2D domain partition   
  m=levels.back().mx;
  n=levels.back().my;

  cusp::precond::detail::pq_gen(AS.num_domain, &p, &q);
  m_p = m / p;
  n_q = n / q;
  dim3 grid(p, q, 1);
  dim3 block(16, 16, 1);
  cusp::precond::detail::set_domains<IndexType><<<grid,block>>>(thrust::raw_pointer_cast(&*domains.begin()), m, n, m_p, n_q, BI.dof);
  // restrict the size adapting to the mesh
  AS.num_domain = IndexType(AS.num_domain / (BI.ratioi * BI.ratioj));
  }
 
  // compute restriction operator (transpose of prolongator)
  SetupMatrixType R;
  cusp::transpose(P,R);

  // construct Galerkin product R*A*P
  levels.back().residual.resize(levels.back().A_.num_rows); 
  SetupMatrixType AP;
  cusp::multiply(levels.back().A_, P, AP);
  cusp::multiply(R, AP, RAP);
  
  detail::setup_level_matrix(levels.back().R, R);
  detail::setup_level_matrix(levels.back().P, P);    
  }
  
  SmootherPolicyType::create_smoother(BI.use_bilinear_iter, BI.use_bilinear_iter&&same_structure, MC_SSOR.red_black, levels.back().A_,
						levels.back().mx, levels.back().my, BI.dof, domains, order, 
						AS.overlap, handle, MKL_ILU0.use_mkl_sv, SPARSKIT2_ILU.sparskit_ilu_type, ILUPP.ilupp_ilu_type, 
						ILUPP.prepro_type, ILUPP.precond_no, ILUPP.max_levels, ILUPP.mem_factor, ILUPP.threshold, ILUPP.min_pivot, 
						ILUPP.max_fillin_is_inf, ILUPP.fill_in, ILUPP.begin_total_piv, ILUPP.detail_param,
						PETSC_FACTOR.type, PETSC_FACTOR.fill_in, PETSC_FACTOR.fill, PETSC_FACTOR.allow_diagonal_fill,
						PETSC_FACTOR.mat_ordering_type, PETSC_FACTOR.rtol);				
  
  levels.push_back(level());
  
  if (BI.use_bilinear_iter) { 
  levels.back().A_.swap(RAP);
  cusp::detail::sort_by_row_and_column(levels.back().A_);  
  levels.back().x.resize(levels.back().A_.num_rows);
  levels.back().b.resize(levels.back().A_.num_rows);
  levels.back().mx = mx_c;
  levels.back().my = my_c;  
  } 
}

template <typename IndexType, typename ValueType, typename MemorySpace, typename SmootherPolicyType>
void multigrid<IndexType,ValueType,MemorySpace,SmootherPolicyType>::setup_on_level(IndexType i)
{
  CUSP_PROFILE_SCOPED();

  // compute spectral radius of diag(C)^-1 * C
  ValueType rho_DinvA = detail::estimate_rho_Dinv_A(levels[i].A_);

  cusp::array1d<IndexType,MemorySpace>  domains; 
  cusp::array1d<ValueType,MemorySpace>  B_coarse;  
  
  if (!BI.use_bilinear_iter) {  
  SetupMatrixType P;
  // compute stength of connection matrix
  SetupMatrixType C;
  detail::symmetric_strength_of_connection(levels[i].A_, C, theta);
  // compute aggregates
  cusp::array1d<IndexType,MemorySpace> aggregates(C.num_rows,0);
  detail::standard_aggregation(C, aggregates);
  // compute tenative prolongator and coarse nullspace vector
  SetupMatrixType 				T;
  detail::fit_candidates(aggregates, levels[i].B, T, B_coarse);
  // compute prolongation operator
  detail::smooth_prolongator(levels[i].A_, T, P, ValueType(4.0/3.0), rho_DinvA);  // TODO if C != A then compute rho_Dinv_C
  if (AS.use_AS_smoother) { 
	// each aggregation services as a subdomain
    domains = aggregates;
  }
  
  // compute restriction operator (transpose of prolongator)
  SetupMatrixType R;
  cusp::transpose(P,R);
  
  detail::setup_level_matrix(levels[i].R, R);
  detail::setup_level_matrix(levels[i].P, P);  
  
  levels[i].residual.resize(levels[i].A_.num_rows);  
  }
  
  // construct Galerkin product R*A*P  
  SetupMatrixType RAP;  
  if (!setup_already) {
    SolveMatrixType RAP_temp;     
    SolveMatrixType AP;
	SolveMatrixType A = levels[i].A_;
    cusp::multiply(A, levels[i].P, AP);
    cusp::multiply(levels[i].R, AP, RAP_temp);
    RAP = RAP_temp;	
  }
  
  SmootherPolicyType::setup_smoother(i, levels[i].A_, rho_DinvA, domains, 
						sweeps, 
						MC_SSOR.max_colors, MC_SSOR.use_SOR, MC_SSOR.use_Jacobi,
						AS.overlap, AS.restricted, AS.jacobi_iter,
						SPARSKIT2_ILU.maxfil, SPARSKIT2_ILU.lfil, 
						SPARSKIT2_ILU.droptol, SPARSKIT2_ILU.mbloc,
						SPARSKIT2_ILU.permtol, SPARSKIT2_ILU.alph);  

  if (!setup_already) {
  levels[i+1].A_.swap(RAP);
  cusp::detail::sort_by_row_and_column(levels[i+1].A_); 
  }  
  if (!BI.use_bilinear_iter) {
  levels[i+1].x.resize(levels[i+1].A_.num_rows);
  levels[i+1].b.resize(levels[i+1].A_.num_rows);  
  levels[i+1].B.swap(B_coarse);
  }
}
    
template <typename IndexType, typename ValueType, typename MemorySpace, typename SmootherPolicyType>
template <typename Array1, typename Array2>
void multigrid<IndexType,ValueType,MemorySpace,SmootherPolicyType>::operator()(const Array1& b, Array2& x)
{
  CUSP_PROFILE_SCOPED();

  switch (mg_type)
  { 
	case multiplicative:   
    // perform 1 V-cycle
	use_post_first = false;
    multiplicative_(b, x, 0);
	break;
	case cascade:   
    // perform cascade mg with V-cycle
    cascade_(b, x);
	break;	
	case full:   
    // perform full mg with V-cycle
    full_(b, x);
	break;		
	default:	
    // perform 1 V-cycle
	use_post_first = false;
    multiplicative_(b, x, 0);
	break;
  } 
}

template <typename IndexType, typename ValueType, typename MemorySpace, typename SmootherPolicyType>
template <typename Array1, typename Array2>
void multigrid<IndexType,ValueType,MemorySpace,SmootherPolicyType>::solve(const Array1& b, Array2& x)
{
  CUSP_PROFILE_SCOPED();

  cusp::default_monitor<ValueType> monitor(b,apply_num,apply_rtol,apply_atol);  
}

template <typename IndexType, typename ValueType, typename MemorySpace, typename SmootherPolicyType>
template <typename Array1, typename Array2, typename Monitor>
void multigrid<IndexType,ValueType,MemorySpace,SmootherPolicyType>::solve(const Array1& b, Array2& x, Monitor& monitor )
{
  CUSP_PROFILE_SCOPED();

  const size_t n = levels[0].A.num_rows;

  // use simple iteration
  cusp::array1d<ValueType,MemorySpace> update(n);
  cusp::array1d<ValueType,MemorySpace> residual(n);

  // compute initial residual
  cusp::multiply(levels[0].A, x, residual);
  cusp::blas::axpby(b, residual, residual, ValueType(1.0), ValueType(-1.0));

  while(!monitor.finished(residual))
  {   
	  switch (mg_type)
	  { 
		case multiplicative:   
		// perform 1 V-cycle
		use_post_first = false;
		multiplicative_(residual, update, 0);
		break;
		case cascade:   
		// perform cascade with V-cycle
		cascade_(residual, update);
		break;	
		case full:   
		// perform full mg with V-cycle
		full_(residual, update);
		break;		
		default:	
		// perform 1 V-cycle
		use_post_first = false;
		multiplicative_(residual, update, 0);
		break;
	  } 
  
      // x += M * r
      cusp::blas::axpy(update, x, ValueType(1.0));

      // update residual
      cusp::multiply(levels[0].A, x, residual);
      cusp::blas::axpby(b, residual, residual, ValueType(1.0), ValueType(-1.0));
      ++monitor;
  }   
}

template <typename IndexType, typename ValueType, typename MemorySpace, typename SmootherPolicyType>
template <typename Array1, typename Array2>
void multigrid<IndexType,ValueType,MemorySpace,SmootherPolicyType>
::multiplicative_(const Array1& b, Array2& x, const size_t i)
{
  CUSP_PROFILE_SCOPED();

  ValueType *dense_array;
  IndexType *pivot_array;
  _DOUBLE_PRECISION_t *b_array,*x_array;
  ValueType *x_array_;
  if (i + 1 == levels.size())
  {
	  switch (coarse_solver)
	  { 
		case cusp_lu:   
		CULU.temp_b = b;
		CULU.LU(CULU.temp_b, CULU.temp_x);
		x = CULU.temp_x;
		break;
		case magma_lu:   
		dense_array = thrust::raw_pointer_cast(&*MLU.A_dense.values.begin());	
		pivot_array = thrust::raw_pointer_cast(&*MLU.pivot.begin());	
		x = b;
		x_array_     = thrust::raw_pointer_cast(&*x.begin());	
		magma_xgetrs_gpu(MagmaNoTrans, MLU.n, 1, dense_array, MLU.n, pivot_array, x_array_, MLU.n, &MLU.info);
		break;	
		case dss_lu:   
		DSS.temp_b = b;
		b_array = thrust::raw_pointer_cast(&*DSS.temp_b.begin());	
		x_array = thrust::raw_pointer_cast(&*DSS.temp_x.begin());	
		DSS.error = dss_solve_real(DSS.handle, DSS.opt, b_array, DSS.nrhs, x_array);	
		if (DSS.error != MKL_DSS_SUCCESS) { 
		std::cout << "dss_solve_real error" << std::endl;	
		exit(1);
		} 
		x = DSS.temp_x;	
		break;	
		case petsc_lu:
		cusp::multiply(*PETSC_FACTOR.PETSC_LU, b, x);
		break;
		default:	
		cusp::multiply(*PETSC_FACTOR.PETSC_LU, b, x);
		break;
	  } 
  }
  else
  {
    // presmooth
	if (use_post_first) { // used in full type mg
    SmootherPolicyType::postsmooth(levels[i].A, b, x, i);	
	use_post_first = false;
	} else {
    thrust::fill(x.begin(), x.end(), ValueType(0.0));
	// actually more advanced initial value is set inside presmooth()
    SmootherPolicyType::presmooth(levels[i].A, b, x, i);
	}
	
    // compute residual <- b - A*x
    cusp::multiply(levels[i].A, x, levels[i].residual);
    cusp::blas::axpby(b, levels[i].residual, levels[i].residual, ValueType(1.0), ValueType(-1.0));
	
    // restrict to coarse grid
    cusp::multiply(levels[i].R, levels[i].residual, levels[i + 1].b);
	
    // compute coarse grid solution
    multiplicative_(levels[i + 1].b, levels[i + 1].x, i + 1);

    // apply coarse grid correction 
    cusp::multiply(levels[i].P, levels[i + 1].x, levels[i].residual);
    cusp::blas::axpy(levels[i].residual, x, ValueType(1.0));

    // postsmooth
	if (!no_postsmooth) {
    SmootherPolicyType::postsmooth(levels[i].A, b, x, i);
	}
  }
}

template <typename IndexType, typename ValueType, typename MemorySpace, typename SmootherPolicyType>
template <typename Array1, typename Array2>
void multigrid<IndexType,ValueType,MemorySpace,SmootherPolicyType>
::cascade_(const Array1& b, Array2& x)
{
  CUSP_PROFILE_SCOPED();
  
  size_t j, lcoarest;
  ValueType *dense_array;
  IndexType *pivot_array;
  _DOUBLE_PRECISION_t *b_array,*x_array;
  ValueType *x_array_;  
  if (levels.size() == 1)
  {
    // coarsest grid solve
	  switch (coarse_solver)
	  { 
		case cusp_lu:   
		CULU.temp_b = b;
		CULU.LU(CULU.temp_b, CULU.temp_x);
		x = CULU.temp_x;
		break;
		case magma_lu:   
		dense_array = thrust::raw_pointer_cast(&*MLU.A_dense.values.begin());	
		pivot_array = thrust::raw_pointer_cast(&*MLU.pivot.begin());	
		x = b;
		x_array_     = thrust::raw_pointer_cast(&*x.begin());	
		magma_xgetrs_gpu(MagmaNoTrans, MLU.n, 1, dense_array, MLU.n, pivot_array, x_array_, MLU.n, &MLU.info);
		break;	
		case dss_lu:   
		DSS.temp_b = b;
		b_array = thrust::raw_pointer_cast(&*DSS.temp_b.begin());	
		x_array = thrust::raw_pointer_cast(&*DSS.temp_x.begin());	
		DSS.error = dss_solve_real(DSS.handle, DSS.opt, b_array, DSS.nrhs, x_array);	
		if (DSS.error != MKL_DSS_SUCCESS) { 
		std::cout << "dss_solve_real error" << std::endl;	
		exit(1);
		} 
		x = DSS.temp_x;	
		break;	
		case petsc_lu:
		cusp::multiply(*PETSC_FACTOR.PETSC_LU, b, x);
		break;
		default:	
		cusp::multiply(*PETSC_FACTOR.PETSC_LU, b, x);
		break;
	  } 
  }
  else
  {
    // restrict the RHS through all levels to coarsest  
    cusp::multiply(levels[0].R, b, levels[1].b);  	
    for (j=1; j<levels.size()-1; j++) {
    cusp::multiply(levels[j].R, levels[j].b, levels[j + 1].b);  
    }
  
    // way up through the levels
	lcoarest = levels.size()-1;
    // coarsest grid solve
    thrust::fill(levels[lcoarest].x.begin(), levels[lcoarest].x.end(), ValueType(0.0));	  
	  ValueType *dense_array;
	  IndexType *pivot_array;
	  _DOUBLE_PRECISION_t *b_array,*x_array;
	  ValueType *x_array_;
	  switch (coarse_solver)
	  { 
		case cusp_lu:   
		CULU.temp_b = levels[lcoarest].b;
		CULU.LU(CULU.temp_b, CULU.temp_x);
		levels[lcoarest].x = CULU.temp_x;
		break;
		case magma_lu:   
		dense_array = thrust::raw_pointer_cast(&*MLU.A_dense.values.begin());	
		pivot_array = thrust::raw_pointer_cast(&*MLU.pivot.begin());	
		levels[lcoarest].x = levels[lcoarest].b;
		x_array_     = thrust::raw_pointer_cast(&*levels[lcoarest].x.begin());	
		magma_xgetrs_gpu(MagmaNoTrans, MLU.n, 1, dense_array, MLU.n, pivot_array, x_array_, MLU.n, &MLU.info);
		break;	
		case dss_lu:   
		DSS.temp_b = levels[lcoarest].b;
		b_array = thrust::raw_pointer_cast(&*DSS.temp_b.begin());	
		x_array = thrust::raw_pointer_cast(&*DSS.temp_x.begin());	
		DSS.error = dss_solve_real(DSS.handle, DSS.opt, b_array, DSS.nrhs, x_array);	
		if (DSS.error != MKL_DSS_SUCCESS) { 
		std::cout << "dss_solve_real error" << std::endl;	
		exit(1);
		} 
		levels[lcoarest].x = DSS.temp_x;	
		break;	
		case petsc_lu:
		cusp::multiply(*PETSC_FACTOR.PETSC_LU, levels[lcoarest].b, levels[lcoarest].x);
		break;
		default:	
		cusp::multiply(*PETSC_FACTOR.PETSC_LU, levels[lcoarest].b, levels[lcoarest].x);
		break;
	  } 

	if (levels.size() == 2) {
    cusp::multiply(levels[lcoarest - 1].P, levels[lcoarest].x, x);		
	} else {
    cusp::multiply(levels[lcoarest - 1].P, levels[lcoarest].x, levels[lcoarest - 1].x);	
	
	// start with the second coarsest grid
	for (j=levels.size()-2; j>1; j--) {
	SmootherPolicyType::postsmooth(levels[j].A, levels[j].b, levels[j].x, j);	
    cusp::multiply(levels[j - 1].P, levels[j].x, levels[j - 1].x);	
	}
	SmootherPolicyType::postsmooth(levels[1].A, levels[1].b, levels[1].x, 1);	 	
    cusp::multiply(levels[0].P, levels[1].x, x);	
	}	
	// smooth the finest grid
	SmootherPolicyType::postsmooth(levels[0].A, b, x, 0);	  
  }
}

template <typename IndexType, typename ValueType, typename MemorySpace, typename SmootherPolicyType>
template <typename Array1, typename Array2>
void multigrid<IndexType,ValueType,MemorySpace,SmootherPolicyType>
::full_(const Array1& b, Array2& x)
{
  CUSP_PROFILE_SCOPED();
  
  size_t j, lcoarest;
  ValueType *dense_array;
  IndexType *pivot_array;
  _DOUBLE_PRECISION_t *b_array,*x_array;
  ValueType *x_array_;  
  if (levels.size() == 1)
  { 
    // coarsest grid solve
	  switch (coarse_solver)
	  { 
		case cusp_lu:   
		CULU.temp_b = b;
		CULU.LU(CULU.temp_b, CULU.temp_x);
		x = CULU.temp_x;
		break;
		case magma_lu:   
		dense_array = thrust::raw_pointer_cast(&*MLU.A_dense.values.begin());	
		pivot_array = thrust::raw_pointer_cast(&*MLU.pivot.begin());	
		x = b;
		x_array_     = thrust::raw_pointer_cast(&*x.begin());	
		magma_xgetrs_gpu(MagmaNoTrans, MLU.n, 1, dense_array, MLU.n, pivot_array, x_array_, MLU.n, &MLU.info);
		break;	
		case dss_lu:   
		DSS.temp_b = b;
		b_array = thrust::raw_pointer_cast(&*DSS.temp_b.begin());	
		x_array = thrust::raw_pointer_cast(&*DSS.temp_x.begin());	
		DSS.error = dss_solve_real(DSS.handle, DSS.opt, b_array, DSS.nrhs, x_array);	
		if (DSS.error != MKL_DSS_SUCCESS) { 
		std::cout << "dss_solve_real error" << std::endl;	
		exit(1);
		} 
		x = DSS.temp_x;	
		break;	
		case petsc_lu:
		cusp::multiply(*PETSC_FACTOR.PETSC_LU, b, x);
		break;
		default:	
		cusp::multiply(*PETSC_FACTOR.PETSC_LU, b, x);
		break;
	  } 
  }
  else
  {
    // restrict the RHS through all levels to coarsest  
    cusp::multiply(levels[0].R, b, levels[1].b);  	
    for (j=1; j<levels.size()-1; j++) {
    cusp::multiply(levels[j].R, levels[j].b, levels[j + 1].b);  
    }
  
    // way up through the levels
	lcoarest = levels.size()-1;
    // coarsest grid solve
    thrust::fill(levels[lcoarest].x.begin(), levels[lcoarest].x.end(), ValueType(0.0));	  	
	  ValueType *dense_array;
	  IndexType *pivot_array;
	  _DOUBLE_PRECISION_t *b_array,*x_array;
	  ValueType *x_array_;
	  switch (coarse_solver)
	  { 
		case cusp_lu:   
		CULU.temp_b = levels[lcoarest].b;
		CULU.LU(CULU.temp_b, CULU.temp_x);
		levels[lcoarest].x = CULU.temp_x;
		break;
		case magma_lu:   
		dense_array = thrust::raw_pointer_cast(&*MLU.A_dense.values.begin());	
		pivot_array = thrust::raw_pointer_cast(&*MLU.pivot.begin());	
		levels[lcoarest].x = levels[lcoarest].b;
		x_array_     = thrust::raw_pointer_cast(&*levels[lcoarest].x.begin());	
		magma_xgetrs_gpu(MagmaNoTrans, MLU.n, 1, dense_array, MLU.n, pivot_array, x_array_, MLU.n, &MLU.info);
		break;	
		case dss_lu:   
		DSS.temp_b = levels[lcoarest].b;
		b_array = thrust::raw_pointer_cast(&*DSS.temp_b.begin());	
		x_array = thrust::raw_pointer_cast(&*DSS.temp_x.begin());	
		DSS.error = dss_solve_real(DSS.handle, DSS.opt, b_array, DSS.nrhs, x_array);	
		if (DSS.error != MKL_DSS_SUCCESS) { 
		std::cout << "dss_solve_real error" << std::endl;	
		exit(1);
		} 
		levels[lcoarest].x = DSS.temp_x;	
		break;	
		case petsc_lu:
		cusp::multiply(*PETSC_FACTOR.PETSC_LU, levels[lcoarest].b, levels[lcoarest].x);
		break;
		default:	
		cusp::multiply(*PETSC_FACTOR.PETSC_LU, levels[lcoarest].b, levels[lcoarest].x);
		break;
	  } 
	
	if (levels.size() == 2) {
    cusp::multiply(levels[lcoarest - 1].P, levels[lcoarest].x, x);		
	} else {
    cusp::multiply(levels[lcoarest - 1].P, levels[lcoarest].x, levels[lcoarest - 1].x);	
	
	// start with the second coarsest grid
	for (j=levels.size()-2; j>1; j--) {	
    // perform 1 V-cycle
	use_post_first = true;
    multiplicative_(levels[j].b, levels[j].x, j);
    cusp::multiply(levels[j - 1].P, levels[j].x, levels[j - 1].x);	
	}
    multiplicative_(levels[1].b, levels[1].x, 1); 	
    cusp::multiply(levels[0].P, levels[1].x, x);	
	}		
    // perform 1 V-cycle
	use_post_first = true;
    multiplicative_(b, x, 0);	
  }
}

template <typename IndexType, typename ValueType, typename MemorySpace, typename SmootherPolicyType>
void multigrid<IndexType,ValueType,MemorySpace,SmootherPolicyType>
::print( void )
{
	IndexType num_levels = levels.size();

	std::cout << "\tNumber of Levels:\t" << num_levels << std::endl;
	std::cout << "\tOperator Complexity:\t" << operator_complexity() << std::endl;
	std::cout << "\tGrid Complexity:\t" << grid_complexity() << std::endl;
	std::cout << "\tlevel\tunknowns\tnonzeros:\t" << std::endl;

	IndexType nnz = 0;

	for(size_t index = 0; index < levels.size(); index++)
		nnz += levels[index].A.num_entries;

	for(size_t index = 0; index < levels.size(); index++)
  {
		double percent = (double)levels[index].A.num_entries / nnz;
		std::cout << "\t" << index << "\t" << levels[index].A.num_cols << "\t\t" \
              << levels[index].A.num_entries << " \t[" << 100*percent << "%]" \
              << std::endl;
	}
} 

template <typename IndexType, typename ValueType, typename MemorySpace, typename SmootherPolicyType>
double multigrid<IndexType,ValueType,MemorySpace,SmootherPolicyType>
::operator_complexity( void )
{
	size_t nnz = 0;

	for(size_t index = 0; index < levels.size(); index++)
		nnz += levels[index].A.num_entries;

	return (double) nnz / (double) levels[0].A.num_entries;
} 

template <typename IndexType, typename ValueType, typename MemorySpace,typename SmootherPolicyType>
double multigrid<IndexType,ValueType,MemorySpace,SmootherPolicyType>
::grid_complexity( void )
{
	size_t unknowns = 0;
	for(size_t index = 0; index < levels.size(); index++)
		unknowns += levels[index].A.num_rows;

	return (double) unknowns / (double) levels[0].A.num_rows;
} 

} // end namespace precond
} // end namespace cusp

