/*
 *  Copyright 2008-2009 NVIDIA Corporation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

/*! \file pamg.inl
 *  \brief Inline file for pamg.h
 */

#include <cusp/blas.h>
#include <cusp/multiply.h>
#include <cusp/transpose.h>

#include <cusp/csr_matrix.h> // TODO remove this when generic spmv works for COO matrices


namespace cusp
{
namespace precond
{

namespace detail
{

//! kernel to fill restriction operator matrix
//! WARNING: this is currently a sequential code on the GPU which takes about 0.5% of the runtime
template <typename IndexType, typename ValueType>
__global__ void 
fill_restriction_operator(
  IndexType* R_row_indices, 
  IndexType* R_column_indices, 
  ValueType* R_values, 
  int* paths, 
  int* path_lengths, 
  int path_count, 
  int path_size, 
  int* row_visited, 
  int num_rows)
{
    int entry=0;

    // set entries for the covered paths / lines
    for (int i = 0; i<path_count;i++)
    {
        for (int j=0;j<path_lengths[i];j++)
        {
            R_row_indices[entry] = i; R_column_indices[entry] = paths[i*path_size+j]; R_values[entry] = 1;
            entry++;
        }
    }

    // set entries for the uncovered unknowns (identity)
    int non_covered_nodes_count=0;
    for (int i=0;i<num_rows;i++)
    {
        int row=row_visited[i];
        if (row==0)
        {
            R_row_indices[entry] = path_count+non_covered_nodes_count; R_column_indices[entry] = i; R_values[entry] = 1;
            non_covered_nodes_count++;
            entry++;
        }
    }
}

} // end namespace detail



//! constructor
template <typename IndexType, typename ValueType, typename MemorySpace>
pamg<IndexType,ValueType,MemorySpace>::pamg(
  const cusp::csr_matrix<IndexType,ValueType,MemorySpace>& A, 
  const pamg_parameters<IndexType,ValueType> &_params,
  ValueType _alpha,
  int _verbosity)  :
    params(_params),
    verbosity(_verbosity), 
    alpha(_alpha), 
    amg_solver(0)
{
    init(A);
}


//! intialization routine
template <typename IndexType, typename ValueType, typename MemorySpace>
void pamg<IndexType,ValueType,MemorySpace>::init(const cusp::csr_matrix<IndexType,ValueType,MemorySpace>& A)
{
    // reserve space for the given level count
    levels.reserve(params.level_count);

    // insert finest level into the level hierarchy
    levels.push_back(typename pamg<IndexType,ValueType,MemorySpace>::level());
    levels.back().A = A; // copy & convert to hyb

    double cutoff = params.use_smoothed_aggregation ? .5 : .95;

    bool extend_succeeded = true;

    // set up multilevel hierarchy
    do
    {
        extend_succeeded = extend_hierarchy();        
    } // stop if number of unknowns <= coarsening_unknowns_threshold, unknown reduction is smaller than the cutoff, or the maximum level_count is reached, or we failed
    while (extend_succeeded &&
           ((int)levels.back().A.num_rows > params.coarsening_unknowns_threshold) &&
           (((double)(levels.back().A.num_rows)/(double)(levels[levels.size()-2].A.num_rows))<cutoff) &&
           ((int)levels.size()<=params.level_count));

    // if we failed because reduction is less than cutoff, and SA is enabled, we want to back out of this level and rewind one level
    if (extend_succeeded && 
        levels.size() > 1 && 
        params.use_smoothed_aggregation && 
        ((double)(levels.back().A.num_rows)/(double)(levels[levels.size()-2].A.num_rows))>=cutoff) {
      //undo extend_hierarchy by one      
      if (verbosity > 0)
        printf("[pamg] un-extending hierarchy, removing level %d\n", (int)levels.size()-1);
          
      delete levels.back().cover;
      delete levels.back().line_smoother;
      levels.pop_back();
    }
    
//    levels[0].cover->visualize_path_on_3D_uniform_grid("graph_cover.out", (int)sqrt((ValueType)A.num_rows), (int)sqrt((ValueType)A.num_rows), 1);

    if (params.use_smoothed_aggregation) {
        cusp::coo_matrix<IndexType, ValueType, MemorySpace> Alast_coo = levels.back().A;
        amg_solver = new cusp::precond::smoothed_aggregation<IndexType, ValueType, MemorySpace>(Alast_coo);
    }

    if (verbosity>0) {
        printf("[pamg] using %d levels with %d unknowns on the coarsest grid.\n",(int)levels.size(),(int)levels.back().A.num_rows);
        for (int i=1; i < (int)levels.size(); i++) {
          printf("[pamg] %d: %d rows (%.1f %%)\n", i, (int)levels[i].A.num_rows, 100 * ((float)levels[i].A.num_rows) / ((float)levels[i-1].A.num_rows)); 
        }
    }
}

//! destructor
template <typename IndexType, typename ValueType, typename MemorySpace>
pamg<IndexType,ValueType,MemorySpace>::~pamg()
{
    for (int i=0;i<(int)levels.size();i++)
    {
        delete levels[i].cover;
        delete levels[i].line_smoother;
    }
    delete amg_solver;
}

//! set up one level of the multi-level hierarchy
template <typename IndexType, typename ValueType, typename MemorySpace>
bool pamg<IndexType,ValueType,MemorySpace>::extend_hierarchy(void)
{
    // create path cover for the current level (taking care of the potentially given custom path options
    // creating up to 2x the coverage seems to be a good heuristic here
    int smoother_path_size = params.max_path_length;
    int smoother_path_count = 2*levels.back().A.num_rows / smoother_path_size;

    bool done = false;
    while (!done) {
      cusp::graph::path_cover_parameters cov_params;
      cov_params.path_count = smoother_path_count;
      cov_params.path_size = smoother_path_size; 
      cov_params.num_serialization_steps = params.covering_num_serialization_steps;
      cov_params.min_abs_weight = params.covering_min_abs_weight;
      cov_params.num_merging_steps = params.covering_num_merging_steps;

      levels.back().cover = new cusp::graph::path_cover<IndexType, ValueType, MemorySpace>(levels.back().A, cov_params, verbosity);

      int num_at_next_level = levels.back().A.num_rows - levels.back().cover->covered_row_count + levels.back().cover->actual_path_count;
      float reduction_fraction = ((float)num_at_next_level) / ((float)levels.back().A.num_rows);
      done = true;

      if (reduction_fraction > .5f && smoother_path_size > 4 && smoother_path_count > 2) {
        smoother_path_count *= 2;
        smoother_path_size /= 2;
        delete levels.back().cover;
        levels.back().cover = 0;
        if (verbosity > 0)
          printf("[pamg] retrying at level %d\n", (int)levels.size());
        done = false;
      }
    } // end while

    // if the path coverage failed entirely, then return false without finishing this level
    if (levels.back().cover->actual_path_count == 0) {
      if (verbosity>0)
        printf("[pamg] path_cover failed, extend_hierarchy failed\n");
      return false;
    }

    // create line smoother for the current level
    levels.back().line_smoother = new cusp::relaxation::line<IndexType,ValueType,MemorySpace>(levels.back().A, *levels.back().cover, params.smoother_sweep_count, alpha);

    cusp::graph::path_cover<IndexType, ValueType, MemorySpace> &current_cover = *(levels.back().cover);

    int coarse_node_count = (levels.back().A.num_rows-current_cover.covered_row_count)+current_cover.actual_path_count;

    if (verbosity>0)
        printf("[pamg] number of coarse nodes %d\n",coarse_node_count);

    // allocate restriction operator matrix
    coo_matrix<IndexType,ValueType,MemorySpace> Rtmp(coarse_node_count,levels.back().A.num_rows,levels.back().A.num_rows);

    // get some raw pointers for kernel launch
    IndexType* Rtmp_row_indices_d = thrust::raw_pointer_cast(&Rtmp.row_indices[0]);
    IndexType* Rtmp_column_indices_d = thrust::raw_pointer_cast(&Rtmp.column_indices[0]);
    ValueType* Rtmp_values_d = thrust::raw_pointer_cast(&Rtmp.values[0]);
    int* paths_d = thrust::raw_pointer_cast(&current_cover.paths[0]);
    int* path_lengths_d = thrust::raw_pointer_cast(&current_cover.path_lengths[0]);
    int* row_visited_d = thrust::raw_pointer_cast(&current_cover.row_visited[0]);

    // fill the restriction operator (currently sequential on the GPU!)
    detail::fill_restriction_operator<<<1,1>>>(Rtmp_row_indices_d, Rtmp_column_indices_d, Rtmp_values_d, paths_d, path_lengths_d, current_cover.actual_path_count, current_cover.params.path_size, row_visited_d, levels.back().A.num_rows);

    // convert coo matrix to hyb format
    cusp::hyb_matrix<IndexType,ValueType,MemorySpace> R=Rtmp;

    // compute prolongation operator (transpose of restriction)
    cusp::hyb_matrix<IndexType,ValueType,MemorySpace> P;
    cusp::transpose(R,P);

    // construct Galerkin product R*A*P i.e. the coarse system
    cusp::hyb_matrix<IndexType,ValueType,MemorySpace> RAP;
    {
        cusp::hyb_matrix<IndexType,ValueType,MemorySpace> AP;
        cusp::multiply(levels.back().A, P, AP);
        cusp::multiply(R, AP, RAP);
    }

    // put the computed restriction/prolongation operators into the current level
    levels.back().R.swap(R);
    levels.back().P.swap(P);

    // create next level
    levels.push_back(level());
    levels.back().A.swap(RAP);

    return true;
}


//! linear operator which applies the preconditioner
//! \param input vector
//! \param output vector
template <typename IndexType, typename ValueType, typename MemorySpace>
template <typename Array1, typename Array2>
void pamg<IndexType,ValueType,MemorySpace>::operator()(const Array1& x, Array2& y) const
{
    // perform 1 V-cycle
    thrust::fill(y.begin(), y.end(), typename Array1::value_type(0));
    _solve(x, y, 0);
}

//! actual preconditioner application
template <typename IndexType, typename ValueType, typename MemorySpace>
void pamg<IndexType,ValueType,MemorySpace>
::_solve(const cusp::array1d<ValueType,MemorySpace>& b,
               cusp::array1d<ValueType,MemorySpace>& x,
         const int i) const
{
    if (i + 1 == (int)levels.size()) // coarsest level solve done by a Jacobi-preconditioned CG solver iterated to a residual of 1e-17
    {
        if (amg_solver) {
            amg_solver->_solve(b, x, 0);
            amg_solver->_solve(b, x, 0);
        }
        else {

            cusp::array1d<ValueType,MemorySpace> temp_b(b);

           // set stopping criteria (todo: experiment with this limit - can we get away with much higher tolerance?)
           cusp::default_monitor<ValueType> monitor(x, 10000, (ValueType)1e-5);

           // setup preconditioner
           cusp::precond::diagonal<ValueType, MemorySpace> M(levels.back().A);

            // solve
            cusp::krylov::cg(levels.back().A, x, temp_b, monitor,M);
        }
    }
   else // smoothing, restriction, coarse solve, prolongation, smoothing for non-coarsest levels
   {
        // temporary variables for residual and coarse level variables
        cusp::array1d<ValueType,MemorySpace> residual(levels[i].P.num_rows);
        cusp::array1d<ValueType,MemorySpace> coarse_b(levels[i].P.num_cols);
        cusp::array1d<ValueType,MemorySpace> coarse_x(levels[i].P.num_cols, 0);

        // presmooth
        (*(levels[i].line_smoother))(b,x);

        // compute residual <- b - A*x
        cusp::multiply(levels[i].A, x, residual);

        cusp::blas::axpby(b, residual, residual, (ValueType)1.0, (ValueType)-1.0);

        // restrict to coarse grid
        cusp::multiply(levels[i].R, residual, coarse_b);

        // compute coarse grid solution
        _solve(coarse_b, coarse_x, i + 1);
        //_solve(coarse_b, coarse_x, i + 1);

        // apply coarse grid correction
        cusp::multiply(levels[i].P, coarse_x, residual);
        cusp::blas::axpy(residual, x, (ValueType)1.0f);


        // postsmooth
        (*(levels[i].line_smoother))(b,x);
    }
}

//! method that uses the multi-level scheme as full iterative solver
//! \param b RHS vector
//! \param x solution vector
//! \param tol stopping threshold (absolute l2 norm of residual)
template <typename IndexType, typename ValueType, typename MemorySpace>
void pamg<IndexType,ValueType,MemorySpace>
::solve(const cusp::array1d<ValueType,MemorySpace>& b,
               cusp::array1d<ValueType,MemorySpace>& x,
               const ValueType tol) const
{
    cusp::array1d<ValueType, MemorySpace> residual(b.size());
    cusp::array1d<ValueType, MemorySpace> Ax(b.size());
    cusp::array1d<ValueType, MemorySpace> Aresid(b.size());

    int iter = 0;
    ValueType error = 0;
    ValueType last_error = 0;

    do {
        _solve(b, x, 0);

        cusp::multiply(levels[0].A, x, Ax);
        cusp::blas::axpby(Ax, b, residual, (ValueType)1.0, (ValueType)-1.0);
        cusp::multiply(levels[0].A, residual, Aresid);
        error = sqrt(blas::dotc(residual,Aresid));

        if (verbosity>0) {
          std::cout << "[pamg] iter:" << iter++ << "res_norm: " << error << "(" << error / last_error << "x)" << std::endl;
        }

        last_error = error;
    } while (error > tol);


    if (verbosity) 
      printf("[pamg] converged to a residual of %e after %d iterations.\n", error, iter);
    
    
    /*
    cusp::array1d<ValueType, MemorySpace> x_tmp(b.size(),0.0);

    // compute residual <- b - A*x
    cusp::multiply(levels[0].A, x, residual);
    cusp::blas::axpby(b, residual, residual, (ValueType)1.0, (ValueType)-1.0);

    int iter=0;

    // residual computation
    ValueType res_norm = sqrt(blas::dotc(residual,residual));

    // main loop
    while(res_norm>tol)
    {
        // solve for current iteration
        _solve(residual,x_tmp,0);
        cusp::blas::axpby(x, x_tmp, x, 1.0f, 1.0f);
        thrust::fill(x_tmp.begin(), x_tmp.end(), 0.0);

        iter++;

        // compute residual
        cusp::multiply(levels[0].A, x, residual);
        cusp::blas::axpby(b, residual, residual, 1.0f, -1.0f);
        res_norm = sqrt(blas::dotc(residual,residual));

        if (verbosity>0)
            printf("[pamg] iter: %d\t res_norm:%e\n", iter, res_norm);
    }

    if (verbosity) 
      printf("[pamg] converged to a residual of %e after %d iterations.\n", res_norm, iter);
*/
}

//! helper function to get the level count
template <typename IndexType, typename ValueType, typename MemorySpace>
int pamg<IndexType,ValueType,MemorySpace>
::get_level_count() const
 {
    return levels.size();
 }


} // end namespace precond
} // end namespace cusp

