/*
 *  Author: Li Luo 
 *  Email:  li.luo@siat.ac.cn 
 */
#include <cusp/detail/spectral_radius.h>

namespace cusp
{
namespace precond
{

template <typename SmootherType>
smoother_type_base<SmootherType>::~smoother_type_base() 
{ 
  while (!level_smoother.empty()) {
    delete level_smoother.back();
    level_smoother.pop_back();
  }
}

template <typename SmootherType>
template <typename MatrixType, typename VectorType1, typename VectorType2>
void smoother_type_base<SmootherType>::presmooth(const MatrixType& A, const VectorType1& b, VectorType2& x, const size_t i) 
{
  level_smoother[i]->presmooth(A, b, x);
}

template <typename SmootherType>
template <typename MatrixType, typename VectorType1, typename VectorType2>
void smoother_type_base<SmootherType>::postsmooth(const MatrixType& A, const VectorType1& b, VectorType2& x, const size_t i) 
{
  level_smoother[i]->postsmooth(A, b, x);
}

template<typename IndexType, typename ValueType, typename MemoryType>
template <typename MatrixType, typename VectorType>
void jacobi_policy<IndexType,ValueType,MemoryType>::create_smoother(bool same_mesh, bool same_structure, bool red_black, const MatrixType& A,
						const IndexType mx, const IndexType my, const IndexType dof, 
						const VectorType& domains, const IndexType order, 
						const IndexType overlap, cusparseHandle_t& handle, bool use_mkl_sv,
						cusp::relaxation::sparskit_ilu_type sparskit_ilu_type,
						cusp::relaxation::ilupp_type ilupp_ilu_type, 
						const IndexType prepro_type, const IndexType precond_no,
						const IndexType max_levels, const ValueType mem_factor,
						const ValueType threshold, const ValueType min_pivot, 
						bool max_fillin_is_inf, const IndexType ilupp_fill_in,
						bool begin_total_piv, bool detail_param,
						cusp::relaxation::petsc_factor_type petsc_factor_type, 
						const IndexType petsc_factor_fill_in, const ValueType petsc_factor_fill,
						bool allow_diagonal_fill, 
						cusp::relaxation::petsc_factor_mat_ordering_type mat_ordering_type,
						const ValueType petsc_factor_rtol)
{
  this->level_smoother.push_back(new cusp::relaxation::jacobi<ValueType, MemoryType>(same_structure, A));
}

template<typename IndexType, typename ValueType, typename MemoryType>
template <typename MatrixType, typename VectorType>
void jacobi_policy<IndexType,ValueType,MemoryType>::setup_smoother(size_t i, const MatrixType& A,
						const ValueType rho_DinvA, const VectorType& domains, 
						const IndexType sweeps,
						const IndexType max_colors, bool SOR, bool Jacobi,
						const IndexType overlap, bool restricted, const IndexType jacobi_iter,
						const IndexType maxfil, const IndexType lfil, 
						const ValueType droptol, const IndexType mbloc,
						const ValueType permtol, const ValueType alph)
{
  //  4/3 * 1/rho is a good default, where rho is the spectral radius of D^-1(A)
  ValueType omega = ValueType(4.0/3.0) / rho_DinvA;
  //ValueType omega = 1.0;
  this->level_smoother[i]->setup(A, omega, sweeps);
}

template<typename IndexType, typename ValueType, typename MemoryType>
template <typename MatrixType, typename VectorType>
void mc_ssor_policy<IndexType,ValueType,MemoryType>::create_smoother(bool same_mesh, bool same_structure, bool red_black, const MatrixType& A, 
						const IndexType mx, const IndexType my, const IndexType dof,
						const VectorType& domains, const IndexType order, 
						const IndexType overlap, cusparseHandle_t& handle, bool use_mkl_sv,
						cusp::relaxation::sparskit_ilu_type sparskit_ilu_type,
						cusp::relaxation::ilupp_type ilupp_ilu_type, 
						const IndexType prepro_type, const IndexType precond_no,
						const IndexType max_levels, const ValueType mem_factor,
						const ValueType threshold, const ValueType min_pivot, 
						bool max_fillin_is_inf, const IndexType ilupp_fill_in,
						bool begin_total_piv, bool detail_param,
						cusp::relaxation::petsc_factor_type petsc_factor_type, 
						const IndexType petsc_factor_fill_in, const ValueType petsc_factor_fill,
						bool allow_diagonal_fill, 
						cusp::relaxation::petsc_factor_mat_ordering_type mat_ordering_type,
						const ValueType petsc_factor_rtol)
{
  this->level_smoother.push_back(new cusp::relaxation::mc_ssor<IndexType, ValueType, MemoryType>(same_structure, red_black, A, mx, my, dof));
}

template<typename IndexType, typename ValueType, typename MemoryType>
template <typename MatrixType, typename VectorType>
void mc_ssor_policy<IndexType,ValueType,MemoryType>::setup_smoother(size_t i, const MatrixType& A,
						const ValueType rho_DinvA, const VectorType& domains, 
						const IndexType sweeps,
						const IndexType max_colors, bool SOR, bool Jacobi,
						const IndexType overlap, bool restricted, const IndexType jacobi_iter,
						const IndexType maxfil, const IndexType lfil, 
						const ValueType droptol, const IndexType mbloc,
						const ValueType permtol, const ValueType alph)
{
  //  4/3 * 1/rho is a good default, where rho is the spectral radius of D^-1(A)
  ValueType omega = ValueType(4.0/3.0) / rho_DinvA;
  //ValueType omega = 1.0;
  this->level_smoother[i]->setup(A, omega, max_colors, sweeps, SOR, Jacobi);
}

template<typename IndexType, typename ValueType, typename MemoryType>
template <typename MatrixType, typename VectorType>
void mkl_ilu0_policy<IndexType,ValueType,MemoryType>::create_smoother(bool same_mesh, bool same_structure, bool red_black, const MatrixType& A, 
						const IndexType mx, const IndexType my, const IndexType dof,
						const VectorType& domains, const IndexType order, 
						const IndexType overlap, cusparseHandle_t& handle, bool use_mkl_sv,
						cusp::relaxation::sparskit_ilu_type sparskit_ilu_type,
						cusp::relaxation::ilupp_type ilupp_ilu_type, 
						const IndexType prepro_type, const IndexType precond_no,
						const IndexType max_levels, const ValueType mem_factor,
						const ValueType threshold, const ValueType min_pivot, 
						bool max_fillin_is_inf, const IndexType ilupp_fill_in,
						bool begin_total_piv, bool detail_param,
						cusp::relaxation::petsc_factor_type petsc_factor_type, 
						const IndexType petsc_factor_fill_in, const ValueType petsc_factor_fill,
						bool allow_diagonal_fill, 
						cusp::relaxation::petsc_factor_mat_ordering_type mat_ordering_type,
						const ValueType petsc_factor_rtol)
{
  this->level_smoother.push_back(new cusp::relaxation::mkl_ilu0<IndexType, ValueType, MemoryType>(same_structure, A, handle, use_mkl_sv));
}

template<typename IndexType, typename ValueType, typename MemoryType>
template <typename MatrixType, typename VectorType>
void mkl_ilu0_policy<IndexType,ValueType,MemoryType>::setup_smoother(size_t i, const MatrixType& A,
						const ValueType rho_DinvA, const VectorType& domains, 
						const IndexType sweeps,
						const IndexType max_colors, bool SOR, bool Jacobi,
						const IndexType overlap, bool restricted, const IndexType jacobi_iter,
						const IndexType maxfil, const IndexType lfil, 
						const ValueType droptol, const IndexType mbloc,
						const ValueType permtol, const ValueType alph)
{
  this->level_smoother[i]->setup(A, sweeps);
}

template<typename IndexType, typename ValueType, typename MemoryType>
template <typename MatrixType, typename VectorType>
void sparskit2_ilu_policy<IndexType,ValueType,MemoryType>::create_smoother(bool same_mesh, bool same_structure, bool red_black, const MatrixType& A, 
						const IndexType mx, const IndexType my, const IndexType dof,
						const VectorType& domains, const IndexType order, 
						const IndexType overlap, cusparseHandle_t& handle, bool use_mkl_sv,
						cusp::relaxation::sparskit_ilu_type sparskit_ilu_type,
						cusp::relaxation::ilupp_type ilupp_ilu_type, 
						const IndexType prepro_type, const IndexType precond_no,
						const IndexType max_levels, const ValueType mem_factor,
						const ValueType threshold, const ValueType min_pivot, 
						bool max_fillin_is_inf, const IndexType ilupp_fill_in,
						bool begin_total_piv, bool detail_param,
						cusp::relaxation::petsc_factor_type petsc_factor_type, 
						const IndexType petsc_factor_fill_in, const ValueType petsc_factor_fill,
						bool allow_diagonal_fill, 
						cusp::relaxation::petsc_factor_mat_ordering_type mat_ordering_type,
						const ValueType petsc_factor_rtol)
{
  this->level_smoother.push_back(new cusp::relaxation::sparskit2_ilu<IndexType, ValueType, MemoryType>(same_structure, A, handle, use_mkl_sv, sparskit_ilu_type));
}

template<typename IndexType, typename ValueType, typename MemoryType>
template <typename MatrixType, typename VectorType>
void sparskit2_ilu_policy<IndexType,ValueType,MemoryType>::setup_smoother(size_t i, const MatrixType& A,
						const ValueType rho_DinvA, const VectorType& domains, 
						const IndexType sweeps,
						const IndexType max_colors, bool SOR, bool Jacobi,
						const IndexType overlap, bool restricted, const IndexType jacobi_iter,
						const IndexType maxfil, const IndexType lfil, 
						const ValueType droptol, const IndexType mbloc,
						const ValueType permtol, const ValueType alph)
{
  this->level_smoother[i]->setup(A, maxfil, lfil, droptol, mbloc, permtol, alph, sweeps);
}

template<typename IndexType, typename ValueType, typename MemoryType>
template <typename MatrixType, typename VectorType>
void ilupp_policy<IndexType,ValueType,MemoryType>::create_smoother(bool same_mesh, bool same_structure, bool red_black, const MatrixType& A, 
						const IndexType mx, const IndexType my, const IndexType dof,
						const VectorType& domains, const IndexType order, 
						const IndexType overlap, cusparseHandle_t& handle, bool use_mkl_sv,
						cusp::relaxation::sparskit_ilu_type sparskit_ilu_type,
						cusp::relaxation::ilupp_type ilupp_ilu_type, 
						const IndexType prepro_type, const IndexType precond_no,
						const IndexType max_levels, const ValueType mem_factor,
						const ValueType threshold, const ValueType min_pivot, 
						bool max_fillin_is_inf, const IndexType ilupp_fill_in,
						bool begin_total_piv, bool detail_param,
						cusp::relaxation::petsc_factor_type petsc_factor_type, 
						const IndexType petsc_factor_fill_in, const ValueType petsc_factor_fill,
						bool allow_diagonal_fill, 
						cusp::relaxation::petsc_factor_mat_ordering_type mat_ordering_type,
						const ValueType petsc_factor_rtol)
{
  this->level_smoother.push_back(new cusp::relaxation::ilupp<IndexType, ValueType, MemoryType>(same_structure, A, handle, use_mkl_sv, ilupp_ilu_type, 
						prepro_type, precond_no, max_levels, mem_factor, threshold, min_pivot, max_fillin_is_inf, ilupp_fill_in, begin_total_piv, detail_param));
}

template<typename IndexType, typename ValueType, typename MemoryType>
template <typename MatrixType, typename VectorType>
void ilupp_policy<IndexType,ValueType,MemoryType>::setup_smoother(size_t i, const MatrixType& A,
						const ValueType rho_DinvA, const VectorType& domains, 
						const IndexType sweeps,
						const IndexType max_colors, bool SOR, bool Jacobi,
						const IndexType overlap, bool restricted, const IndexType jacobi_iter,
						const IndexType maxfil, const IndexType lfil, 
						const ValueType droptol, const IndexType mbloc,
						const ValueType permtol, const ValueType alph)
{
  this->level_smoother[i]->setup(A, sweeps);
}

template<typename IndexType, typename ValueType, typename MemoryType>
template <typename MatrixType, typename VectorType>
void petsc_factor_policy<IndexType,ValueType,MemoryType>::create_smoother(bool same_mesh, bool same_structure, bool red_black, const MatrixType& A, 
						const IndexType mx, const IndexType my, const IndexType dof,
						const VectorType& domains, const IndexType order, 
						const IndexType overlap, cusparseHandle_t& handle, bool use_mkl_sv,
						cusp::relaxation::sparskit_ilu_type sparskit_ilu_type,
						cusp::relaxation::ilupp_type ilupp_ilu_type, 
						const IndexType prepro_type, const IndexType precond_no,
						const IndexType max_levels, const ValueType mem_factor,
						const ValueType threshold, const ValueType min_pivot, 
						bool max_fillin_is_inf, const IndexType ilupp_fill_in,
						bool begin_total_piv, bool detail_param,
						cusp::relaxation::petsc_factor_type petsc_factor_type, 
						const IndexType petsc_factor_fill_in, const ValueType petsc_factor_fill,
						bool allow_diagonal_fill, 
						cusp::relaxation::petsc_factor_mat_ordering_type mat_ordering_type,
						const ValueType petsc_factor_rtol)
{
  this->level_smoother.push_back(new cusp::relaxation::petsc_factor<IndexType, ValueType, MemoryType>(same_structure, A, handle, use_mkl_sv, petsc_factor_type, 
						petsc_factor_fill_in,petsc_factor_fill,allow_diagonal_fill,mat_ordering_type,petsc_factor_rtol));
}

template<typename IndexType, typename ValueType, typename MemoryType>
template <typename MatrixType, typename VectorType>
void petsc_factor_policy<IndexType,ValueType,MemoryType>::setup_smoother(size_t i, const MatrixType& A,
						const ValueType rho_DinvA, const VectorType& domains, 
						const IndexType sweeps,
						const IndexType max_colors, bool SOR, bool Jacobi,
						const IndexType overlap, bool restricted, const IndexType jacobi_iter,
						const IndexType maxfil, const IndexType lfil, 
						const ValueType droptol, const IndexType mbloc,
						const ValueType permtol, const ValueType alph)
{
  this->level_smoother[i]->setup(A, sweeps);
}

template<typename IndexType, typename ValueType, typename MemoryType>
template <typename MatrixType, typename VectorType>
void polynomial_policy<IndexType,ValueType,MemoryType>::create_smoother(bool same_mesh, bool same_structure, bool red_black, const MatrixType& A, 
						const IndexType mx, const IndexType my, const IndexType dof,
						const VectorType& domains, const IndexType order, 
						const IndexType overlap, cusparseHandle_t& handle, bool use_mkl_sv,
						cusp::relaxation::sparskit_ilu_type sparskit_ilu_type,
						cusp::relaxation::ilupp_type ilupp_ilu_type, 
						const IndexType prepro_type, const IndexType precond_no,
						const IndexType max_levels, const ValueType mem_factor,
						const ValueType threshold, const ValueType min_pivot, 
						bool max_fillin_is_inf, const IndexType ilupp_fill_in,
						bool begin_total_piv, bool detail_param,
						cusp::relaxation::petsc_factor_type petsc_factor_type, 
						const IndexType petsc_factor_fill_in, const ValueType petsc_factor_fill,
						bool allow_diagonal_fill, 
						cusp::relaxation::petsc_factor_mat_ordering_type mat_ordering_type,
						const ValueType petsc_factor_rtol)
{
  this->level_smoother.push_back(new cusp::relaxation::polynomial<ValueType, MemoryType>());
}

template<typename IndexType, typename ValueType, typename MemoryType>
template <typename MatrixType, typename VectorType>
void polynomial_policy<IndexType,ValueType,MemoryType>::setup_smoother(size_t i, const MatrixType& A,
						const ValueType rho_DinvA, const VectorType& domains, 
						const IndexType sweeps,
						const IndexType max_colors, bool SOR, bool Jacobi,
						const IndexType overlap, bool restricted, const IndexType jacobi_iter,
						const IndexType maxfil, const IndexType lfil, 
						const ValueType droptol, const IndexType mbloc,
						const ValueType permtol, const ValueType alph)
{
  cusp::array1d<ValueType,cusp::host_memory> coeff;
  ValueType rho = cusp::detail::ritz_spectral_radius_symmetric(A, 8);
  cusp::relaxation::detail::chebyshev_polynomial_coefficients(rho,coeff);
  this->level_smoother[i]->setup(coeff,sweeps);
}

template<typename IndexType, typename ValueType, typename MemoryType>
template <typename MatrixType, typename VectorType>
void kpz_polynomial_policy<IndexType,ValueType,MemoryType>::create_smoother(bool same_mesh, bool same_structure, bool red_black, const MatrixType& A, 
						const IndexType mx, const IndexType my, const IndexType dof,
						const VectorType& domains, const IndexType order, 
						const IndexType overlap, cusparseHandle_t& handle, bool use_mkl_sv,
						cusp::relaxation::sparskit_ilu_type sparskit_ilu_type,
						cusp::relaxation::ilupp_type ilupp_ilu_type, 
						const IndexType prepro_type, const IndexType precond_no,
						const IndexType max_levels, const ValueType mem_factor,
						const ValueType threshold, const ValueType min_pivot, 
						bool max_fillin_is_inf, const IndexType ilupp_fill_in,
						bool begin_total_piv, bool detail_param,
						cusp::relaxation::petsc_factor_type petsc_factor_type, 
						const IndexType petsc_factor_fill_in, const ValueType petsc_factor_fill,
						bool allow_diagonal_fill, 
						cusp::relaxation::petsc_factor_mat_ordering_type mat_ordering_type,
						const ValueType petsc_factor_rtol)
{
  this->level_smoother.push_back(new cusp::relaxation::kpz_polynomial<ValueType, MemoryType>(A, order));
}

template<typename IndexType, typename ValueType, typename MemoryType>
template <typename MatrixType, typename VectorType>
void hierarchical_as_policy<IndexType,ValueType,MemoryType>::create_smoother(bool same_mesh, bool same_structure, bool red_black, const MatrixType& A, 
						const IndexType mx, const IndexType my, const IndexType dof,
						const VectorType& domains, const IndexType order, 
						const IndexType overlap, cusparseHandle_t& handle, bool use_mkl_sv,
						cusp::relaxation::sparskit_ilu_type sparskit_ilu_type,
						cusp::relaxation::ilupp_type ilupp_ilu_type, 
						const IndexType prepro_type, const IndexType precond_no,
						const IndexType max_levels, const ValueType mem_factor,
						const ValueType threshold, const ValueType min_pivot, 
						bool max_fillin_is_inf, const IndexType ilupp_fill_in,
						bool begin_total_piv, bool detail_param,
						cusp::relaxation::petsc_factor_type petsc_factor_type, 
						const IndexType petsc_factor_fill_in, const ValueType petsc_factor_fill,
						bool allow_diagonal_fill, 
						cusp::relaxation::petsc_factor_mat_ordering_type mat_ordering_type,
						const ValueType petsc_factor_rtol)
{
  this->level_smoother.push_back(new cusp::relaxation::hierarchical_as<IndexType, ValueType, MemoryType>(same_mesh, same_structure, A, domains, overlap));
}

template<typename IndexType, typename ValueType, typename MemoryType>
template <typename MatrixType, typename VectorType>
void hierarchical_as_policy<IndexType,ValueType,MemoryType>::setup_smoother(size_t i, const MatrixType& A,
						const ValueType rho_DinvA, const VectorType& domains, 
						const IndexType sweeps,
						const IndexType max_colors, bool SOR, bool Jacobi,
						const IndexType overlap, bool restricted, const IndexType jacobi_iter,
						const IndexType maxfil, const IndexType lfil, 
						const ValueType droptol, const IndexType mbloc,
						const ValueType permtol, const ValueType alph)
{
  this->level_smoother[i]->setup(A, domains, restricted, overlap, sweeps, jacobi_iter);
}

} // end namespace precond
} // end namespace cusp

