 /*
 *  Author: Li Luo 
 *  Email:  li.luo@siat.ac.cn 
 */
/*! \file multigrid.h
 *  \brief geometric and algebraic multigrid preconditioner.
 *  
 */

#pragma once

#include <cusp/detail/config.h>

#include <vector>
#include <cusp/linear_operator.h>

#include <cusp/coo_matrix.h>
#include <cusp/csr_matrix.h>
#include <cusp/hyb_matrix.h>
#include <cusp/precond/smoother_type.h>
#include <cusp/detail/lu.h>
#include <cusp/precond/magma_overload.h>
#include <mkl_dss.h>

#if defined(MKL_ILP64)
#define MKL_INT long long
#else
#define MKL_INT int
#endif

namespace cusp
{
namespace precond
{

/*! \addtogroup preconditioners Preconditioners
 *  \ingroup preconditioners
 *  \{
 */

enum mg_type {
	multiplicative,
	cascade,
	full
};
 
enum coarse_solver_type {
	cusp_lu,
	magma_lu,
	dss_lu,
	petsc_lu
};
 
template <typename IndexType, typename ValueType, typename MemorySpace>
struct mg_container
{
};

template <typename IndexType, typename ValueType>
struct mg_container<IndexType,ValueType,cusp::host_memory>
{
    // use CSR on host
    typedef typename cusp::csr_matrix<IndexType,ValueType,cusp::host_memory> setup_type;
    typedef typename cusp::csr_matrix<IndexType,ValueType,cusp::host_memory> solve_type;
};

template <typename IndexType, typename ValueType>
struct mg_container<IndexType,ValueType,cusp::device_memory>
{
    // use COO for setup, use csr for solve
    typedef typename cusp::coo_matrix<IndexType,ValueType,cusp::device_memory> setup_type;
    typedef typename cusp::csr_matrix<IndexType,ValueType,cusp::device_memory> solve_type;
};


/* multilevel additive Schwarz preconditoner */
template <typename IndexType, typename ValueType, typename MemorySpace, typename SmootherPolicyType=jacobi_policy<IndexType,ValueType,MemorySpace> >
class multigrid : public cusp::linear_operator<ValueType, MemorySpace, IndexType>, 
                             public SmootherPolicyType
{

    typedef typename mg_container<IndexType,ValueType,MemorySpace>::setup_type SetupMatrixType;
    typedef typename mg_container<IndexType,ValueType,MemorySpace>::solve_type SolveMatrixType;

    struct level
    {
		IndexType	    mx,my; // grid size for bilinear interpolation
        SetupMatrixType A_; // matrix, cost no memory since using swap();
        SolveMatrixType R;  // restriction operator
        SolveMatrixType A;  // matrix
        SolveMatrixType P;  // prolongation operator
        cusp::array1d<ValueType,MemorySpace> B;               // near-nullspace candidates
        cusp::array1d<ValueType,MemorySpace> x;               // per-level solution
        cusp::array1d<ValueType,MemorySpace> b;               // per-level rhs
        cusp::array1d<ValueType,MemorySpace> residual;        // per-level residual
    };

    std::vector<level> levels;
    
	// bilinear interpolation context
	struct BILINEAR_ITER
	{	
	bool use_bilinear_iter;
	IndexType dof,fmx,fmy;
	IndexType ratioi,ratioj;
	};
	BILINEAR_ITER BI;
	
	// coarse solver
	struct CUSP_LU
	{	
    cusp::detail::lu_solver<ValueType, cusp::host_memory> LU;
	cusp::array1d<ValueType,cusp::host_memory> temp_b;
	cusp::array1d<ValueType,cusp::host_memory> temp_x;
	};
	CUSP_LU CULU;
	
	// MAGMA context
	struct MAGMA_LU
	{
	cusp::array2d<ValueType,cusp::device_memory,cusp::column_major> A_dense;
	cusp::array1d<IndexType,cusp::host_memory> pivot;
	IndexType n;
	IndexType info;
	};
	MAGMA_LU MLU;

	// MKL DSS context
	struct MKL_DSS
	{	
	cusp::csr_matrix<_INTEGER_t,_DOUBLE_PRECISION_t,cusp::host_memory> A_csr;
	cusp::array1d<_DOUBLE_PRECISION_t,cusp::host_memory> temp_b;
	cusp::array1d<_DOUBLE_PRECISION_t,cusp::host_memory> temp_x;
	_MKL_DSS_HANDLE_t handle;
	_INTEGER_t error;	
	_INTEGER_t nrhs;
	MKL_INT opt;
	MKL_INT sym;
	MKL_INT type;
	bool symmetric;
	bool positive_definite;
	};
	MKL_DSS DSS;

	// multi-coloring SSOR smoother context
	struct MC_SSOR_SMOOTH
	{	
	bool red_black;	
	bool use_SOR;
	bool use_Jacobi;
	IndexType max_colors;
	IndexType mx,my,dof;
	};
	MC_SSOR_SMOOTH MC_SSOR;
	
	// additive Schwarz method context
	struct ADDITIVE_SCHWARZ_SMOOTH
	{		
	bool use_AS_smoother;
	bool restricted;	
	IndexType jacobi_iter;
	IndexType num_domain;
	IndexType overlap;	
	};
	ADDITIVE_SCHWARZ_SMOOTH AS;
	
	// MKL ilu0 context	
	struct MKL_ILU0_SMOOTH
	{		
	bool use_mkl_sv;
	};
	MKL_ILU0_SMOOTH MKL_ILU0;	
	
	// SPARSKIT2 ilu context	
	struct SPARSKIT2_ILU_SMOOTH
	{		
	bool use_mkl_sv;
	cusp::relaxation::sparskit_ilu_type sparskit_ilu_type;
	IndexType maxfil;
	IndexType lfil;
	ValueType droptol;
	IndexType mbloc;
	ValueType permtol;
	ValueType alph;
	};
	SPARSKIT2_ILU_SMOOTH SPARSKIT2_ILU;		

	// ILUPP ilu context	
	struct ILUPP_SMOOTH
	{		
	bool use_mkl_sv;
	cusp::relaxation::ilupp_type ilupp_ilu_type;
	IndexType prepro_type;
	IndexType precond_no;
	IndexType max_levels; 
	ValueType mem_factor;
	ValueType threshold; 
	ValueType min_pivot;
	bool max_fillin_is_inf; 
	IndexType fill_in;
	bool begin_total_piv;
	bool detail_param;
	};
	ILUPP_SMOOTH ILUPP;	

	// PETSc petsc_factor context
	struct PETSC_FACTOR_SMOOTH
	{		
	bool use_mkl_sv;
	bool allow_diagonal_fill;
	cusp::relaxation::petsc_factor_type type;
	cusp::relaxation::petsc_factor_mat_ordering_type mat_ordering_type;	
	IndexType fill_in;
	ValueType fill;
	ValueType rtol;
	cusp::relaxation::petsc_factor<IndexType,ValueType,MemorySpace>* PETSC_LU;
	};
	PETSC_FACTOR_SMOOTH PETSC_FACTOR;		
	
	
	bool same_structure;
	bool no_postsmooth;
	bool use_post_first;
	bool setup_already; // avoid constructing twice for the hierarchies
	cusp::precond::mg_type mg_type;	
	cusp::precond::coarse_solver_type coarse_solver;
	IndexType sweeps;
	IndexType order;
	IndexType apply_num;
    ValueType theta;
	ValueType apply_rtol;
	ValueType apply_atol;	
	
    public:

    template <typename MatrixType>
    multigrid(bool same_structure, const MatrixType& A, cusparseHandle_t& handle, 
	cusp::precond::mg_type mgtype=(cusp::precond::mg_type)0, size_t max_levels=100, bool red_black=false, 
	bool use_bilinear_iter=false, const IndexType dof=1, const IndexType fmx=100, const IndexType fmy=100, 
	const IndexType ratioi=2, const IndexType ratioj=2, 	
	const IndexType num_domain=16, bool use_AS_smoother=false, bool no_postsmooth=false,	
	bool symmetric=false, bool positive_definite=false,
	const IndexType order=3, const IndexType overlap=1,
	bool use_mkl_sv=false, cusp::relaxation::sparskit_ilu_type sparskit_ilutype=(cusp::relaxation::sparskit_ilu_type)6,
	cusp::relaxation::ilupp_type ilupp_ilu_type=(cusp::relaxation::ilupp_type)2, 
	const IndexType prepro_type=0, const IndexType precond_no=0,
	const IndexType ilupp_max_levels=100, const ValueType mem_factor=3.0,
	const ValueType threshold=1000.0, const ValueType min_pivot=0.01, 
	bool max_fillin_is_inf=true, const IndexType ilupp_fill_in=100,
	bool begin_total_piv=true, bool detail_param=false, 
	cusp::relaxation::petsc_factor_type petsc_factor_type=(cusp::relaxation::petsc_factor_type)1, 
	const IndexType petsc_factor_fill_in=0, const ValueType petsc_factor_fill=5.0,
	bool allow_diagonal_fill=false, 
	cusp::relaxation::petsc_factor_mat_ordering_type mat_ordering_type=(cusp::relaxation::petsc_factor_mat_ordering_type)0,
	const ValueType petsc_factor_rtol=1.e-10, cusp::precond::coarse_solver_type coarse_solver=(cusp::precond::coarse_solver_type)3);

    template <typename MatrixType>
    void setup(const MatrixType& A, const ValueType theta=0,
	const IndexType apply_num=1, const ValueType apply_rtol=1e-6, const ValueType apply_atol=0, 		
	const IndexType sweeps=3,
	const IndexType max_colors=10, bool use_SOR=false, bool use_Jacobi=true,
	const IndexType jacobi_iter = 10, bool restricted=false, 
	const IndexType maxfil=10, const IndexType lfil=0, const ValueType droptol=0.0001,
	const IndexType mbloc=0, const ValueType permtol=0.01, const ValueType alph=1.0);
	
    ~multigrid();
    
    template <typename Array1, typename Array2>
    void operator()(const Array1& x, Array2& y);

    template <typename Array1, typename Array2>
    void solve(const Array1& b, Array2& x);

    template <typename Array1, typename Array2, typename Monitor>
    void solve(const Array1& b, Array2& x, Monitor& monitor);
	
    template <typename Array1, typename Array2>
    void full_(const Array1& b, Array2& x);	

    template <typename Array1, typename Array2>
    void cascade_(const Array1& b, Array2& x);
	
    template <typename Array1, typename Array2>
    void multiplicative_(const Array1& b, Array2& x, const size_t i);
	
    void print( void );

    double operator_complexity( void );

    double grid_complexity( void );
	
    protected:

    void extend_hierarchy(cusparseHandle_t& handle);

    void setup_on_level(IndexType level);
};

} // end namespace precond
} // end namespace cusp

#include <cusp/precond/detail/multigrid.inl>

