/*
 *  Copyright 2008-2009 NVIDIA Corporation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

/*! \file pamg.h
 *  \brief path aggregation multigrid preconditioner.
 *  
 */

#pragma once

#include <cusp/detail/config.h>

#include <cusp/linear_operator.h>
#include <cusp/csr_matrix.h>
#include <cusp/graph/path_cover.h>
#include <cusp/relaxation/line_relaxation.h>
#include <cusp/krylov/cg.h>
#include <cusp/monitor.h>
#include <cusp/precond/diagonal.h>
#include <cusp/precond/smoothed_aggregation.h>

namespace cusp
{
namespace precond
{

template <typename IndexType, typename ValueType>
struct pamg_parameters
{
  IndexType level_count;
  IndexType smoother_sweep_count;
  IndexType coarsening_unknowns_threshold;
  IndexType max_path_length;
  bool 		use_smoothed_aggregation;
  IndexType covering_num_serialization_steps;
  ValueType covering_min_abs_weight;
  IndexType covering_num_merging_steps;
  
  pamg_parameters() {
    level_count = 15;
    smoother_sweep_count = 3;
    coarsening_unknowns_threshold = 100;
    max_path_length = 256;
    use_smoothed_aggregation = true;
    covering_num_serialization_steps = 5;
    covering_min_abs_weight = .1;
    covering_num_merging_steps = 4;
  }
};

/*! \addtogroup preconditioners Preconditioners
 *  \ingroup preconditioners
 *  \{
 */

/*! \p pamg : path aggregation multigrid preconditioner
 *
 *
 */
template <typename IndexType, typename ValueType, typename MemorySpace>
class pamg : public cusp::linear_operator<ValueType, MemorySpace, IndexType>
{
    struct level
    {
        cusp::hyb_matrix<IndexType,ValueType,MemorySpace> R;  // restriction operator
        cusp::hyb_matrix<IndexType,ValueType,MemorySpace> A;  // matrix
        cusp::hyb_matrix<IndexType,ValueType,MemorySpace> P;  // prolongation operator

        cusp::graph::path_cover<IndexType,ValueType,MemorySpace> *cover;
        cusp::relaxation::line<IndexType,ValueType,MemorySpace> *line_smoother;

        level() : cover(0), line_smoother(0) { }
    };

    std::vector<level> levels;
    cusp::precond::smoothed_aggregation<IndexType, ValueType, MemorySpace>* amg_solver;
    pamg_parameters<IndexType,ValueType> params;
    int verbosity;
    ValueType alpha;

    public:

    //! constructor
    pamg(const cusp::csr_matrix<IndexType,ValueType,MemorySpace>& A, const pamg_parameters<IndexType,ValueType> &_params=pamg_parameters<IndexType,ValueType>(), ValueType _alpha=0.0, int _verbosity=0);

    //! intialization routine
    void init(const cusp::csr_matrix<IndexType,ValueType,MemorySpace>& A);

    //! destructor
    ~pamg();

    //! linear operator which applies the preconditioner
    //! \param input vector
    //! \param output vector
    template <typename Array1, typename Array2>
    void operator()(const Array1& x, Array2& y) const;

    //! method that uses the multi-level scheme as full iterative solver
    //! \param b RHS vector
    //! \param x solution vector
    //! \param tol stopping threshold (absolute l2 norm of residual)
    void solve(const cusp::array1d<ValueType,MemorySpace>& b, cusp::array1d<ValueType,MemorySpace>& x,
             const ValueType tol) const;

    //! helper function to get the level count
    int get_level_count() const;

    protected:

    //! set up one level of the multi-level hierarchy
    bool extend_hierarchy(void);

    //! actual preconditioner application
    void _solve(const cusp::array1d<ValueType,MemorySpace>& b,
                      cusp::array1d<ValueType,MemorySpace>& x,
                const int i) const;
};

} // end namespace precond
} // end namespace cusp

#include <cusp/precond/detail/pamg.inl>

