/*
 *  Author: Li Luo 
 *  Email:  li.luo@siat.ac.cn 
 */

#pragma once

#include <cusp/detail/config.h>

#include <vector> 

#include <cusp/relaxation/jacobi_.h>
#include <cusp/relaxation/mc_ssor.h>
#include <cusp/relaxation/hierarchical_as.h>
#include <cusp/relaxation/mkl_ilu0.h>
#include <cusp/relaxation/sparskit2_ilu.h>
#include <cusp/relaxation/ilupp.h>
#include <cusp/relaxation/petsc_factor.h>
#include <cusp/relaxation/polynomial_sweeps.h>
#include <cusp/relaxation/kpz_polynomial.h>



namespace cusp
{
namespace precond
{
template <typename SmootherType>
class smoother_type_base
{
protected:
    std::vector<SmootherType *> level_smoother;

    virtual ~smoother_type_base();
    template <typename MatrixType, typename VectorType1, typename VectorType2>
    void presmooth(const MatrixType& A, const VectorType1& b, VectorType2& x, const size_t i);

    template <typename MatrixType, typename VectorType1, typename VectorType2>
    void postsmooth(const MatrixType& A, const VectorType1& b, VectorType2& x, const size_t i);
};

template<typename IndexType, typename ValueType, typename MemoryType>
class jacobi_policy : public smoother_type_base<cusp::relaxation::jacobi<ValueType, MemoryType> >
{
    typedef smoother_type_base<cusp::relaxation::jacobi<ValueType, MemoryType> > base_type;
protected:

    template <typename MatrixType, typename VectorType>
    void create_smoother(bool same_mesh, bool same_structure, bool red_black, const MatrixType& A, 
						const IndexType mx, const IndexType my, const IndexType dof,
						const VectorType& domains, const IndexType order, 
						const IndexType overlap, cusparseHandle_t& handle, bool use_mkl_sv,
						cusp::relaxation::sparskit_ilu_type sparskit_ilu_type,
						cusp::relaxation::ilupp_type ilupp_ilu_type, 
						const IndexType prepro_type, const IndexType precond_no,
						const IndexType max_levels, const ValueType mem_factor,
						const ValueType threshold, const ValueType min_pivot, 
						bool max_fillin_is_inf, const IndexType ilupp_fill_in,
						bool begin_total_piv, bool detail_param,
						cusp::relaxation::petsc_factor_type petsc_factor_type, 
						const IndexType petsc_factor_fill_in, const ValueType petsc_factor_fill,
						bool allow_diagonal_fill, 
						cusp::relaxation::petsc_factor_mat_ordering_type mat_ordering_type,
						const ValueType petsc_factor_rtol);

    template <typename MatrixType, typename VectorType>
    void setup_smoother(size_t i, const MatrixType& A,
						const ValueType rho_DinvA, const VectorType& domains, 
						const IndexType sweeps,
						const IndexType max_colors, bool SOR, bool Jacobi,
						const IndexType overlap, bool restricted, const IndexType jacobi_iter,
						const IndexType maxfil, const IndexType lfil, 
						const ValueType droptol, const IndexType mbloc,
						const ValueType permtol, const ValueType alph);
};

template<typename IndexType, typename ValueType, typename MemoryType>
class mc_ssor_policy : public smoother_type_base<cusp::relaxation::mc_ssor<IndexType, ValueType, MemoryType> >
{
    typedef smoother_type_base<cusp::relaxation::mc_ssor<IndexType, ValueType, MemoryType> > base_type;
protected:

    template <typename MatrixType, typename VectorType>
    void create_smoother(bool same_mesh, bool same_structure, bool red_black, const MatrixType& A, 
						const IndexType mx, const IndexType my, const IndexType dof,	
						const VectorType& domains, const IndexType order, 
						const IndexType overlap, cusparseHandle_t& handle, bool use_mkl_sv,
						cusp::relaxation::sparskit_ilu_type sparskit_ilu_type,
						cusp::relaxation::ilupp_type ilupp_ilu_type, 
						const IndexType prepro_type, const IndexType precond_no,
						const IndexType max_levels, const ValueType mem_factor,
						const ValueType threshold, const ValueType min_pivot, 
						bool max_fillin_is_inf, const IndexType ilupp_fill_in,
						bool begin_total_piv, bool detail_param,
						cusp::relaxation::petsc_factor_type petsc_factor_type, 
						const IndexType petsc_factor_fill_in, const ValueType petsc_factor_fill,
						bool allow_diagonal_fill, 
						cusp::relaxation::petsc_factor_mat_ordering_type mat_ordering_type,
						const ValueType petsc_factor_rtol);

    template <typename MatrixType, typename VectorType>
    void setup_smoother(size_t i, const MatrixType& A,
						const ValueType rho_DinvA, const VectorType& domains, 
						const IndexType sweeps,
						const IndexType max_colors, bool SOR, bool Jacobi,
						const IndexType overlap, bool restricted, const IndexType jacobi_iter,
						const IndexType maxfil, const IndexType lfil, 
						const ValueType droptol, const IndexType mbloc,
						const ValueType permtol, const ValueType alph);
};

template<typename IndexType, typename ValueType, typename MemoryType>
class mkl_ilu0_policy : public smoother_type_base<cusp::relaxation::mkl_ilu0<IndexType, ValueType, MemoryType> >
{
    typedef smoother_type_base<cusp::relaxation::mkl_ilu0<IndexType, ValueType, MemoryType> > base_type;
protected:

    template <typename MatrixType, typename VectorType>
    void create_smoother(bool same_mesh, bool same_structure, bool red_black, const MatrixType& A, 
						const IndexType mx, const IndexType my, const IndexType dof,	
						const VectorType& domains, const IndexType order, 
						const IndexType overlap, cusparseHandle_t& handle, bool use_mkl_sv,
						cusp::relaxation::sparskit_ilu_type sparskit_ilu_type,
						cusp::relaxation::ilupp_type ilupp_ilu_type, 
						const IndexType prepro_type, const IndexType precond_no,
						const IndexType max_levels, const ValueType mem_factor,
						const ValueType threshold, const ValueType min_pivot, 
						bool max_fillin_is_inf, const IndexType ilupp_fill_in,
						bool begin_total_piv, bool detail_param,
						cusp::relaxation::petsc_factor_type petsc_factor_type, 
						const IndexType petsc_factor_fill_in, const ValueType petsc_factor_fill,
						bool allow_diagonal_fill, 
						cusp::relaxation::petsc_factor_mat_ordering_type mat_ordering_type,
						const ValueType petsc_factor_rtol);

    template <typename MatrixType, typename VectorType>
    void setup_smoother(size_t i, const MatrixType& A,
						const ValueType rho_DinvA, const VectorType& domains, 
						const IndexType sweeps,
						const IndexType max_colors, bool SOR, bool Jacobi,
						const IndexType overlap, bool restricted, const IndexType jacobi_iter,
						const IndexType maxfil, const IndexType lfil, 
						const ValueType droptol, const IndexType mbloc,
						const ValueType permtol, const ValueType alph);
};

template<typename IndexType, typename ValueType, typename MemoryType>
class sparskit2_ilu_policy : public smoother_type_base<cusp::relaxation::sparskit2_ilu<IndexType, ValueType, MemoryType> >
{
    typedef smoother_type_base<cusp::relaxation::sparskit2_ilu<IndexType, ValueType, MemoryType> > base_type;
protected:

    template <typename MatrixType, typename VectorType>
    void create_smoother(bool same_mesh, bool same_structure, bool red_black, const MatrixType& A, 
						const IndexType mx, const IndexType my, const IndexType dof,	
						const VectorType& domains, const IndexType order, 
						const IndexType overlap, cusparseHandle_t& handle, bool use_mkl_sv,
						cusp::relaxation::sparskit_ilu_type sparskit_ilu_type,
						cusp::relaxation::ilupp_type ilupp_ilu_type, 
						const IndexType prepro_type, const IndexType precond_no,
						const IndexType max_levels, const ValueType mem_factor,
						const ValueType threshold, const ValueType min_pivot, 
						bool max_fillin_is_inf, const IndexType ilupp_fill_in,
						bool begin_total_piv, bool detail_param,
						cusp::relaxation::petsc_factor_type petsc_factor_type, 
						const IndexType petsc_factor_fill_in, const ValueType petsc_factor_fill,
						bool allow_diagonal_fill, 
						cusp::relaxation::petsc_factor_mat_ordering_type mat_ordering_type,
						const ValueType petsc_factor_rtol);

    template <typename MatrixType, typename VectorType>
    void setup_smoother(size_t i, const MatrixType& A,
						const ValueType rho_DinvA, const VectorType& domains, 
						const IndexType sweeps,
						const IndexType max_colors, bool SOR, bool Jacobi,
						const IndexType overlap, bool restricted, const IndexType jacobi_iter,
						const IndexType maxfil, const IndexType lfil, 
						const ValueType droptol, const IndexType mbloc,
						const ValueType permtol, const ValueType alph);
};

template<typename IndexType, typename ValueType, typename MemoryType>
class ilupp_policy : public smoother_type_base<cusp::relaxation::ilupp<IndexType, ValueType, MemoryType> >
{
    typedef smoother_type_base<cusp::relaxation::ilupp<IndexType, ValueType, MemoryType> > base_type;
protected:

    template <typename MatrixType, typename VectorType>
    void create_smoother(bool same_mesh, bool same_structure, bool red_black, const MatrixType& A, 
						const IndexType mx, const IndexType my, const IndexType dof,	
						const VectorType& domains, const IndexType order, 
						const IndexType overlap, cusparseHandle_t& handle, bool use_mkl_sv,
						cusp::relaxation::sparskit_ilu_type sparskit_ilu_type,
						cusp::relaxation::ilupp_type ilupp_ilu_type, 
						const IndexType prepro_type, const IndexType precond_no,
						const IndexType max_levels, const ValueType mem_factor,
						const ValueType threshold, const ValueType min_pivot, 
						bool max_fillin_is_inf, const IndexType ilupp_fill_in,
						bool begin_total_piv, bool detail_param,
						cusp::relaxation::petsc_factor_type petsc_factor_type, 
						const IndexType petsc_factor_fill_in, const ValueType petsc_factor_fill,
						bool allow_diagonal_fill, 
						cusp::relaxation::petsc_factor_mat_ordering_type mat_ordering_type,
						const ValueType petsc_factor_rtol);

    template <typename MatrixType, typename VectorType>
    void setup_smoother(size_t i, const MatrixType& A,
						const ValueType rho_DinvA, const VectorType& domains, 
						const IndexType sweeps,
						const IndexType max_colors, bool SOR, bool Jacobi,
						const IndexType overlap, bool restricted, const IndexType jacobi_iter,
						const IndexType maxfil, const IndexType lfil, 
						const ValueType droptol, const IndexType mbloc,
						const ValueType permtol, const ValueType alph);
};

template<typename IndexType, typename ValueType, typename MemoryType>
class petsc_factor_policy : public smoother_type_base<cusp::relaxation::petsc_factor<IndexType, ValueType, MemoryType> >
{
    typedef smoother_type_base<cusp::relaxation::petsc_factor<IndexType, ValueType, MemoryType> > base_type;
protected:

    template <typename MatrixType, typename VectorType>
    void create_smoother(bool same_mesh, bool same_structure, bool red_black, const MatrixType& A, 
						const IndexType mx, const IndexType my, const IndexType dof,	
						const VectorType& domains, const IndexType order, 
						const IndexType overlap, cusparseHandle_t& handle, bool use_mkl_sv,
						cusp::relaxation::sparskit_ilu_type sparskit_ilu_type,
						cusp::relaxation::ilupp_type ilupp_ilu_type, 
						const IndexType prepro_type, const IndexType precond_no,
						const IndexType max_levels, const ValueType mem_factor,
						const ValueType threshold, const ValueType min_pivot, 
						bool max_fillin_is_inf, const IndexType ilupp_fill_in,
						bool begin_total_piv, bool detail_param,
						cusp::relaxation::petsc_factor_type petsc_factor_type, 
						const IndexType petsc_factor_fill_in, const ValueType petsc_factor_fill,
						bool allow_diagonal_fill, 
						cusp::relaxation::petsc_factor_mat_ordering_type mat_ordering_type,
						const ValueType petsc_factor_rtol);

    template <typename MatrixType, typename VectorType>
    void setup_smoother(size_t i, const MatrixType& A,
						const ValueType rho_DinvA, const VectorType& domains, 
						const IndexType sweeps,
						const IndexType max_colors, bool SOR, bool Jacobi,
						const IndexType overlap, bool restricted, const IndexType jacobi_iter,
						const IndexType maxfil, const IndexType lfil, 
						const ValueType droptol, const IndexType mbloc,
						const ValueType permtol, const ValueType alph);
};

template<typename IndexType, typename ValueType, typename MemoryType>
class polynomial_policy : public smoother_type_base<cusp::relaxation::polynomial<ValueType, MemoryType> >
{
    typedef smoother_type_base<cusp::relaxation::jacobi<ValueType, MemoryType> > base_type;
protected:

    template <typename MatrixType, typename VectorType>
    void create_smoother(bool same_mesh, bool same_structure, bool red_black, const MatrixType& A, 
						const IndexType mx, const IndexType my, const IndexType dof,	
						const VectorType& domains, const IndexType order, 
						const IndexType overlap, cusparseHandle_t& handle, bool use_mkl_sv,
						cusp::relaxation::sparskit_ilu_type sparskit_ilu_type,
						cusp::relaxation::ilupp_type ilupp_ilu_type, 
						const IndexType prepro_type, const IndexType precond_no,
						const IndexType max_levels, const ValueType mem_factor,
						const ValueType threshold, const ValueType min_pivot, 
						bool max_fillin_is_inf, const IndexType ilupp_fill_in,
						bool begin_total_piv, bool detail_param,
						cusp::relaxation::petsc_factor_type petsc_factor_type, 
						const IndexType petsc_factor_fill_in, const ValueType petsc_factor_fill,
						bool allow_diagonal_fill, 
						cusp::relaxation::petsc_factor_mat_ordering_type mat_ordering_type,
						const ValueType petsc_factor_rtol);

    template <typename MatrixType, typename VectorType>
    void setup_smoother(size_t i, const MatrixType& A,
						const ValueType rho_DinvA, const VectorType& domains, 
						const IndexType sweeps,
						const IndexType max_colors, bool SOR, bool Jacobi,
						const IndexType overlap, bool restricted, const IndexType jacobi_iter,
						const IndexType maxfil, const IndexType lfil, 
						const ValueType droptol, const IndexType mbloc,
						const ValueType permtol, const ValueType alph);
};

template<typename IndexType, typename ValueType, typename MemoryType>
class kpz_polynomial_policy : public smoother_type_base<cusp::relaxation::kpz_polynomial<ValueType, MemoryType> >
{
    typedef smoother_type_base<cusp::relaxation::jacobi<ValueType, MemoryType> > base_type;
protected:

    template <typename MatrixType, typename VectorType>
    void create_smoother(bool same_mesh, bool same_structure, bool red_black, const MatrixType& A, 
						const IndexType mx, const IndexType my, const IndexType dof,	
						const VectorType& domains, const IndexType order, 
						const IndexType overlap, cusparseHandle_t& handle, bool use_mkl_sv,
						cusp::relaxation::sparskit_ilu_type sparskit_ilu_type,
						cusp::relaxation::ilupp_type ilupp_ilu_type, 
						const IndexType prepro_type, const IndexType precond_no,
						const IndexType max_levels, const ValueType mem_factor,
						const ValueType threshold, const ValueType min_pivot, 
						bool max_fillin_is_inf, const IndexType ilupp_fill_in,
						bool begin_total_piv, bool detail_param,
						cusp::relaxation::petsc_factor_type petsc_factor_type, 
						const IndexType petsc_factor_fill_in, const ValueType petsc_factor_fill,
						bool allow_diagonal_fill, 
						cusp::relaxation::petsc_factor_mat_ordering_type mat_ordering_type,
						const ValueType petsc_factor_rtol);

    template <typename MatrixType, typename VectorType>
    void setup_smoother(size_t i, const MatrixType& A,
						const ValueType rho_DinvA, const VectorType& domains, 
						const IndexType sweeps,
						const IndexType max_colors, bool SOR, bool Jacobi,
						const IndexType overlap, bool restricted, const IndexType jacobi_iter,
						const IndexType maxfil, const IndexType lfil, 
						const ValueType droptol, const IndexType mbloc,
						const ValueType permtol, const ValueType alph);

public:

    kpz_polynomial_policy() { kpz_degree = 3; }
    IndexType kpz_degree;
};

template<typename IndexType, typename ValueType, typename MemoryType>
class hierarchical_as_policy : public smoother_type_base<cusp::relaxation::hierarchical_as<IndexType, ValueType, MemoryType> >
{
    typedef smoother_type_base<cusp::relaxation::hierarchical_as<IndexType, ValueType, MemoryType> > base_type;
protected:

    template <typename MatrixType, typename VectorType>
    void create_smoother(bool same_mesh, bool same_structure, bool red_black, const MatrixType& A, 
						const IndexType mx, const IndexType my, const IndexType dof,	
						const VectorType& domains, const IndexType order, 
						const IndexType overlap, cusparseHandle_t& handle, bool use_mkl_sv,
						cusp::relaxation::sparskit_ilu_type sparskit_ilu_type,
						cusp::relaxation::ilupp_type ilupp_ilu_type, 
						const IndexType prepro_type, const IndexType precond_no,
						const IndexType max_levels, const ValueType mem_factor,
						const ValueType threshold, const ValueType min_pivot, 
						bool max_fillin_is_inf, const IndexType ilupp_fill_in,
						bool begin_total_piv, bool detail_param,
						cusp::relaxation::petsc_factor_type petsc_factor_type, 
						const IndexType petsc_factor_fill_in, const ValueType petsc_factor_fill,
						bool allow_diagonal_fill, 
						cusp::relaxation::petsc_factor_mat_ordering_type mat_ordering_type,
						const ValueType petsc_factor_rtol);

    template <typename MatrixType, typename VectorType>
    void setup_smoother(size_t i, const MatrixType& A,
						const ValueType rho_DinvA, const VectorType& domains, 
						const IndexType sweeps,
						const IndexType max_colors, bool SOR, bool Jacobi,
						const IndexType overlap, bool restricted, const IndexType jacobi_iter,
						const IndexType maxfil, const IndexType lfil, 
						const ValueType droptol, const IndexType mbloc,
						const ValueType permtol, const ValueType alph);
};

} // end namespace precond
} // end namespace cusp

#include <cusp/precond/detail/smoother_type.inl>

