/*
 *  Author: Li Luo 
 *  Email:  li.luo@siat.ac.cn  
*/

/*! \file chebyshev_polynomial.inl
 *  \brief Inline file for chebyshev_polynomial.h
 */

#include <cusp/multiply.h>
#include <cusp/detail/format_utils.h>

#include <math.h>

namespace cusp
{
namespace relaxation
{
namespace detail
{
	template <typename ValueType>
	void chebyshev_polynomial_coefficients( const ValueType rho, 
						cusp::array1d<ValueType,cusp::host_memory>& coefficients,
						const int order = 3,
						const ValueType scale = 1.0) 
	{	
		coefficients.resize(11);
		/* define least squares coefficients for polys ; evalue range [0,4] */
		/* Saad (1985) */
		switch(order) {
		case 0:
			coefficients[0] = 1.0;
			break;
		case 1:
			coefficients[0] = 5.0;      coefficients[1] = -1.0;
			break;
		case 2:
			coefficients[0] = 14.0;     coefficients[1] = -7.0;     coefficients[2] = 1.0;
			break;
		case 3:
			coefficients[0] = 30.0;     coefficients[1] = -27.0;    coefficients[2] = 9.0; 
			coefficients[3] = -1.0;
			break;
		case 4:
			coefficients[0] = 55.0;     coefficients[1] = -77.0;    coefficients[2] = 44.0;
			coefficients[3] = -11.0;    coefficients[4] = 1.0;
			break;
		case 5:
			coefficients[0] = 91.0;     coefficients[1] = -182.0;   coefficients[2] = 156.0;
			coefficients[3] = -65.0;    coefficients[4] = 13.0;     coefficients[5] = -1.0;
			break;
		case 6:
			coefficients[0] = 140.0;    coefficients[1] = -378.0;  coefficients[2] = 450.0;
			coefficients[3] = -275.0;   coefficients[4] = 90.0;    coefficients[5] = -15.0;
			coefficients[6] = 1.0;
			break;
		case 7:
			coefficients[0] = 204.0;    coefficients[1] = -714.0;  coefficients[2] = 1122.0;
			coefficients[3] = -935.0;   coefficients[4] = 442.0;   coefficients[5] = -119.0;
			coefficients[6] = 17.0;     coefficients[7] = -1.0;
			break;
		case 8:
			coefficients[0] = 285.0;    coefficients[1] = -1254.0; coefficients[2] = 2508.0;
			coefficients[3] = -2717.0;  coefficients[4] = 1729.0;  coefficients[5] = -665.0;
			coefficients[6] = 152.0;    coefficients[7] = -19.0;   coefficients[8] = 1.0;
			break;
		case 9:
			coefficients[0] = 385.0;    coefficients[1] = -2079.0; coefficients[2] = 5148.0;
			coefficients[3] = -7007.0;  coefficients[4] = 5733.0;  coefficients[5] = -2940.0;
			coefficients[6] = 952.0;    coefficients[7] = -189.0;  coefficients[8] = 21.0;
			coefficients[9] = -1.0;
			break;
		case 10:
			coefficients[0] = 506.0;    coefficients[1] = -3289.0; coefficients[2] = 9867.0;
			coefficients[3] = -16445.0; coefficients[4] = 16744.0; coefficients[5] = -10948.0;
			coefficients[6] = 4692.0;   coefficients[7] = -1311.0; coefficients[8] = 230.0;
			coefficients[9] = -23.0;    coefficients[10] = 1.0;
			break;
		}		

		// rescaled to original expression 
		//cusp::blas::scal(coefficients, 4.0/(3.0+2.0*order));
		ValueType temp, scaling;
		// transform coefficients to ecue interc [0,b]
		// TODO: ChaoYang excate diagonal vector to judge whether all elements are positive or negative
		// TODO: ChaoYang rho = -rho if matrix diagonal elements are not all positive
		temp = 4.0 / rho;
		scaling = 1.0;
		for( size_t i = 0; i <= order; i++ ) {
			coefficients[i] *= scaling;
			scaling *= temp;
		}
		// rescale to make c[order]=1.0, different from Chao Yang PC_poly
		//ValueType scale_factor = scale/coefficients[0];
		//cusp::blas::scal(coefficients, scale_factor);
	}
}

// constructor
template <typename ValueType, typename MemorySpace>
    chebyshev_polynomial<ValueType,MemorySpace>
    ::chebyshev_polynomial()
    {
    }

template <typename ValueType, typename MemorySpace>
template<typename VectorType>
    chebyshev_polynomial<ValueType,MemorySpace>
    ::chebyshev_polynomial(const VectorType& coefficients, const int order_)
    {
        CUSP_PROFILE_SCOPED();
		
		default_coefficients = coefficients;
		order = order_;
		sweeps = 1;		
		initialized = true;		
    }

template <typename ValueType, typename MemorySpace>
    void chebyshev_polynomial<ValueType,MemorySpace>
    ::setup(const int sweeps_)
    {
        CUSP_PROFILE_SCOPED();
		
		if (!initialized) {
		  printf("chebyshev_polynomial has not yet been created!\n");
		  exit(1);		
		}
		
		sweeps = sweeps_;
    }
	
// linear_operator
template <typename ValueType, typename MemorySpace>
template<typename MatrixType, typename VectorType1, typename VectorType2>
    void chebyshev_polynomial<ValueType,MemorySpace>
    ::operator()(const MatrixType& A, const VectorType1& b, VectorType2& x) const
    {
        CUSP_PROFILE_SCOPED();

        chebyshev_polynomial<ValueType,MemorySpace>::operator()(A,b,x,default_coefficients);
    }

template <typename ValueType, typename MemorySpace>
template<typename MatrixType, typename VectorType1, typename VectorType2>
    void chebyshev_polynomial<ValueType,MemorySpace>
    ::presmooth(const MatrixType& A, const VectorType1& b, VectorType2& x)
    {
        CUSP_PROFILE_SCOPED();
        
		const VectorType1& residual = b;

        x = residual;
		ValueType scale_factor = default_coefficients[order];
		cusp::blas::scal(x, scale_factor);

		cusp::array1d<ValueType,MemorySpace> mult(A.num_rows);

		for( int i= order-1; i>=0; i-- )
		{
			cusp::multiply(A, x, mult);
			scale_factor = default_coefficients[i];
			cusp::blas::axpby(mult, residual, x, ValueType(1), scale_factor);
		}
		
		cusp::array1d<ValueType,MemorySpace> res(A.num_rows);
		for( int s = 0; s<sweeps-1; s++ )
		{	
			// compute res <- b - A*x
			cusp::multiply(A, x, res);
			cusp::blas::axpby(b, res, res, ValueType(1), ValueType(-1));

			scale_factor = default_coefficients[order];
			cusp::array1d<ValueType,MemorySpace> h(res);
			cusp::blas::scal(h, scale_factor);

			for( int i= order-1; i>=0; i-- )
			{
				cusp::multiply(A, h, mult);
				scale_factor = default_coefficients[i];
				cusp::blas::axpby(mult, res, h, ValueType(1), scale_factor);
			}
			
			cusp::blas::axpby(x, h, x, ValueType(1), ValueType(1));
		}	
    }

template <typename ValueType, typename MemorySpace>
template<typename MatrixType, typename VectorType1, typename VectorType2>
    void chebyshev_polynomial<ValueType,MemorySpace>
    ::postsmooth(const MatrixType& A, const VectorType1& b, VectorType2& x)
    {
        CUSP_PROFILE_SCOPED();

        chebyshev_polynomial<ValueType,MemorySpace>::operator()(A,b,x,default_coefficients);
    }

// override default coefficients
template <typename ValueType, typename MemorySpace>
template<typename MatrixType, typename VectorType1, typename VectorType2, typename VectorType3>
    void chebyshev_polynomial<ValueType,MemorySpace>
    ::operator()(const MatrixType& A, const VectorType1& b, VectorType2& x, VectorType3& coefficients) const
    {
		cusp::array1d<ValueType,MemorySpace> residual(A.num_rows);
		for( int s = 0; s<sweeps; s++ )
		{
			// compute residual <- b - A*x
			cusp::multiply(A, x, residual);
			cusp::blas::axpby(b, residual, residual, ValueType(1), ValueType(-1));

			ValueType scale_factor = coefficients[order];
			cusp::array1d<ValueType,MemorySpace> h(residual);
			cusp::blas::scal(h, scale_factor);

			cusp::array1d<ValueType,MemorySpace> mult(A.num_rows);

			for( int i= order-1; i>=0; i-- )
			{
				cusp::multiply(A, h, mult);
				scale_factor = coefficients[i];
				cusp::blas::axpby(mult, residual, h, ValueType(1), scale_factor);
			}

			cusp::blas::axpby(x, h, x, ValueType(1), ValueType(1));
		}	
    }

} // end namespace relaxation
} // end namespace cusp

