/*
 *  Copyright 2008-2009 NVIDIA Corporation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

/*! \file hierarchical_jacobi.inl
 *  \brief Inline file for hierarchical_jacobi.h
 */

namespace cusp
{
namespace relaxation
{

// constructor
template <typename IndexType, typename ValueType, typename MemorySpace>
hierarchical_jacobi<IndexType, ValueType, MemorySpace>::hierarchical_jacobi() :
  _initialized(false)
{
}

template <typename IndexType, typename ValueType, typename MemorySpace>
hierarchical_jacobi<IndexType, ValueType, MemorySpace>::hierarchical_jacobi(
  const cusp::csr_matrix<IndexType,ValueType,MemorySpace>& A, 
  const cusp::array1d<IndexType,MemorySpace> &domains,
  int inner_iterations,
  int outer_iterations,
  ValueType omega) :
    local_solver(A, domains, 1, false, cusp::precond::solver_jacobi),
    _initialized(false)
{
  local_solver.jacobi_iterations = inner_iterations;
  local_solver.jacobi_omega = omega;
  local_solver.schwarz_iterations = outer_iterations;
  _initialized = true;
}


template <typename IndexType, typename ValueType, typename MemorySpace>
void 
hierarchical_jacobi<IndexType, ValueType, MemorySpace>::initialize(
  const cusp::csr_matrix<IndexType,ValueType,MemorySpace>& A, 
  const cusp::array1d<IndexType,MemorySpace> &domains, 
  int inner_iterations, 
  int outer_iterations,
  ValueType omega)
{
  if (_initialized) {
    throw cusp::runtime_exception("hierarchical_jacobi::initialize cannot be called multiple times");
  }

  local_solver.initialize(A, domains, 1, false, cusp::precond::solver_jacobi);
  local_solver.jacobi_iterations = inner_iterations;
  local_solver.jacobi_omega = omega;
  local_solver.schwarz_iterations = outer_iterations;
  _initialized = true;
}



template <typename IndexType, typename ValueType, typename MemorySpace>
template<typename MatrixType, typename VectorType1, typename VectorType2>
    void hierarchical_jacobi<IndexType,ValueType,MemorySpace>
    ::operator()(const MatrixType &A, const VectorType1& b, VectorType2& x)
{
  if (!_initialized) {
    throw cusp::runtime_exception("hierarchical_jacobi::operator() - not initialized");
  }

  local_solver.relax(b, x);
}



template <typename IndexType, typename ValueType, typename MemorySpace>
template<typename MatrixType, typename VectorType1, typename VectorType2>
    void hierarchical_jacobi<IndexType,ValueType,MemorySpace>
    ::presmooth(const MatrixType &A, const VectorType1& b, VectorType2& x)
{
  if (!_initialized) {
    throw cusp::runtime_exception("hierarchical_jacobi::operator() - not initialized");
  }

  thrust::fill(x.begin(), x.end(), (typename VectorType2::value_type)0);
  local_solver.relax(b, x);
}



template <typename IndexType, typename ValueType, typename MemorySpace>
template<typename MatrixType, typename VectorType1, typename VectorType2>
    void hierarchical_jacobi<IndexType,ValueType,MemorySpace>
    ::postsmooth(const MatrixType &A, const VectorType1& b, VectorType2& x)
{
  if (!_initialized) {
    throw cusp::runtime_exception("hierarchical_jacobi::operator() - not initialized");
  }

  local_solver.relax(b, x);
}



} // end namespace relaxation
} // end namespace cusp

