/*
 *  Author: Li Luo 
 *  Email:  li.luo@siat.ac.cn 
 */

/*! \file ilupp.inl
 *  \brief Inline file for ilupp.h
 */
#include <cusp/fact.h>
#include <cusp/blas.h>
#include <cusp/detail/utils_.h>
#include <thrust/transform.h>
#include <cusp/multiply.h>
#include <cusp/permutation.h>
#include <performance/timer.h>
#include <cassert>

namespace cusp
{
namespace relaxation
{

template <typename IndexType, typename ValueType, typename MemorySpace>
    void ilupp<IndexType,ValueType,MemorySpace>
    ::init_parameter(const IndexType prepro_type, const ValueType threshold, const IndexType fill_in)
	{
		  assert(prepro_type >= 0);
		  assert(threshold > 0.);

		  iluplusplus::preprocessing_sequence ilupp_preproc_;			  
		  // set preprocessing_type
		  switch (prepro_type) {
			case 0  : ilupp_preproc_.set_normalize(); break;
			case 1  : ilupp_preproc_.set_PQ(); break;
			case 2  : ilupp_preproc_.set_MAX_WEIGHTED_MATCHING_ORDERING(); break;
			case 3  : ilupp_preproc_.set_MAX_WEIGHTED_MATCHING_ORDERING_UNIT_DIAG(); break;
			case 4  : ilupp_preproc_.set_MAX_WEIGHTED_MATCHING_ORDERING_DD_MOV_COR_IM(); break;
			case 5  : ilupp_preproc_.set_MAX_WEIGHTED_MATCHING_ORDERING_UNIT_DIAG_DD_MOV_COR_IM(); break;
			case 6  : ilupp_preproc_.set_SPARSE_FIRST_MAX_WEIGHTED_MATCHING_ORDERING(); break;
			case 7  : ilupp_preproc_.set_SPARSE_FIRST_MAX_WEIGHTED_MATCHING_ORDERING_UNIT_DIAG(); break;
			case 8  : ilupp_preproc_.set_SPARSE_FIRST_MAX_WEIGHTED_MATCHING_ORDERING_DD_MOV_COR_IM(); break;
			case 9  : ilupp_preproc_.set_SPARSE_FIRST_MAX_WEIGHTED_MATCHING_ORDERING_UNIT_DIAG_DD_MOV_COR_IM(); break;
			case 10 : ilupp_preproc_.set_NORM_MAX_WEIGHTED_MATCHING_ORDERING(); break;
			case 11 : ilupp_preproc_.set_NORM_MAX_WEIGHTED_MATCHING_ORDERING_DD_MOV_COR_IM(); break;
			case 12 : ilupp_preproc_.set_SPARSE_FIRST(); break;
			case 13 : ilupp_preproc_.set_MAX_WEIGHTED_MATCHING_ORDERING_PQ(); break;
			case 14 : ilupp_preproc_.set_NORM_MAX_WEIGHTED_MATCHING_ORDERING_PQ(); break;
			case 15 : ilupp_preproc_.set_MAX_WEIGHTED_MATCHING_ORDERING_MOVE_CORNER(); break;
			case 16 : ilupp_preproc_.set_NORM_MAX_WEIGHTED_MATCHING_ORDERING_MOVE_CORNER(); break;
			case 17 : ilupp_preproc_.set_MAX_WEIGHTED_MATCHING_ORDERING_MOVE_CORNER_IM(); break;
			case 18 : ilupp_preproc_.set_NORM_MAX_WEIGHTED_MATCHING_ORDERING_MOVE_CORNER_IM(); break;
			case 19 : ilupp_preproc_.set_MAX_WEIGHTED_MATCHING_ORDERING_WGT_MOV_COR(); break;
			case 20 : ilupp_preproc_.set_NORM_MAX_WEIGHTED_MATCHING_ORDERING_WGT_MOV_COR(); break;
			case 21 : ilupp_preproc_.set_MAX_WEIGHTED_MATCHING_ORDERING_WGT_MOV_COR_IM(); break;
			case 22 : ilupp_preproc_.set_NORM_MAX_WEIGHTED_MATCHING_ORDERING_WGT_MOV_COR_IM(); break;
			case 23 : ilupp_preproc_.set_MAX_WEIGHTED_MATCHING_ORDERING_WGT2_MOV_COR(); break;
			case 24 : ilupp_preproc_.set_NORM_MAX_WEIGHTED_MATCHING_ORDERING_WGT2_MOV_COR(); break;
			case 25 : ilupp_preproc_.set_MAX_WEIGHTED_MATCHING_ORDERING_WGT2_MOV_COR_IM(); break;
			case 26 : ilupp_preproc_.set_NORM_MAX_WEIGHTED_MATCHING_ORDERING_WGT2_MOV_COR_IM(); break;
			case 27 : ilupp_preproc_.set_MAX_WEIGHTED_MATCHING_ORDERING_SYM_PQ(); break;
			case 28 : ilupp_preproc_.set_NORM_MAX_WEIGHTED_MATCHING_ORDERING_SYM_PQ(); break;
			case 29 : ilupp_preproc_.set_MAX_WEIGHTED_MATCHING_ORDERING_SYMB_MOVE_CORNER(); break;
			case 30 : ilupp_preproc_.set_NORM_MAX_WEIGHTED_MATCHING_ORDERING_SYMB_MOVE_CORNER(); break;
			case 31 : ilupp_preproc_.set_MAX_WEIGHTED_MATCHING_ORDERING_SYMB_MOVE_CORNER_IM(); break;
			case 32 : ilupp_preproc_.set_NORM_MAX_WEIGHTED_MATCHING_ORDERING_SYMB_MOVE_CORNER_IM(); break;
			case 33 : ilupp_preproc_.set_MAX_WEIGHTED_MATCHING_ORDERING_SP_MOVE_CORNER(); break;
			case 34 : ilupp_preproc_.set_NORM_MAX_WEIGHTED_MATCHING_ORDERING_SP_MOVE_CORNER(); break;
			case 35 : ilupp_preproc_.set_MAX_WEIGHTED_MATCHING_ORDERING_SP_MOVE_CORNER_IM(); break;
			case 36 : ilupp_preproc_.set_NORM_MAX_WEIGHTED_MATCHING_ORDERING_SP_MOVE_CORNER_IM(); break;

			default: ilupp_preproc_.set_normalize(); break;
		  }

		  // set precond_parameters for ILU++
		  this->ilupp_param_.init(ilupp_preproc_, 0, ""); // ACHTUNG: setzt default-parameters

		   // set_threshold(tau) sets dropping threshold to 10^{-tau} and fill in
		  this->ilupp_param_.set_threshold(threshold); 
		  this->ilupp_param_.set_fill_in(fill_in);		  
	}
	
template <typename IndexType, typename ValueType, typename MemorySpace>
    void ilupp<IndexType,ValueType,MemorySpace>
    ::init_parameter_multilevel(const IndexType prepro_type, const IndexType precond_no,
                  const IndexType max_levels, const ValueType mem_factor,
                  const ValueType threshold, const ValueType min_pivot, 
				  bool max_fillin_is_inf, const IndexType fill_in, bool begin_total_piv)
	{
		  assert(prepro_type >= 0);
		  assert(precond_no >= 0);
		  assert(max_levels >= 1);
		  assert(mem_factor > 0.);
		  assert(threshold > 0.);
		  assert(min_pivot >= 0.);	

		  iluplusplus::preprocessing_sequence ilupp_preproc_;			  
		  // set preprocessing_type
		  switch (prepro_type) {
			case 0  : ilupp_preproc_.set_normalize(); break;
			case 1  : ilupp_preproc_.set_PQ(); break;
			case 2  : ilupp_preproc_.set_MAX_WEIGHTED_MATCHING_ORDERING(); break;
			case 3  : ilupp_preproc_.set_MAX_WEIGHTED_MATCHING_ORDERING_UNIT_DIAG(); break;
			case 4  : ilupp_preproc_.set_MAX_WEIGHTED_MATCHING_ORDERING_DD_MOV_COR_IM(); break;
			case 5  : ilupp_preproc_.set_MAX_WEIGHTED_MATCHING_ORDERING_UNIT_DIAG_DD_MOV_COR_IM(); break;
			case 6  : ilupp_preproc_.set_SPARSE_FIRST_MAX_WEIGHTED_MATCHING_ORDERING(); break;
			case 7  : ilupp_preproc_.set_SPARSE_FIRST_MAX_WEIGHTED_MATCHING_ORDERING_UNIT_DIAG(); break;
			case 8  : ilupp_preproc_.set_SPARSE_FIRST_MAX_WEIGHTED_MATCHING_ORDERING_DD_MOV_COR_IM(); break;
			case 9  : ilupp_preproc_.set_SPARSE_FIRST_MAX_WEIGHTED_MATCHING_ORDERING_UNIT_DIAG_DD_MOV_COR_IM(); break;
			case 10 : ilupp_preproc_.set_NORM_MAX_WEIGHTED_MATCHING_ORDERING(); break;
			case 11 : ilupp_preproc_.set_NORM_MAX_WEIGHTED_MATCHING_ORDERING_DD_MOV_COR_IM(); break;
			case 12 : ilupp_preproc_.set_SPARSE_FIRST(); break;
			case 13 : ilupp_preproc_.set_MAX_WEIGHTED_MATCHING_ORDERING_PQ(); break;
			case 14 : ilupp_preproc_.set_NORM_MAX_WEIGHTED_MATCHING_ORDERING_PQ(); break;
			case 15 : ilupp_preproc_.set_MAX_WEIGHTED_MATCHING_ORDERING_MOVE_CORNER(); break;
			case 16 : ilupp_preproc_.set_NORM_MAX_WEIGHTED_MATCHING_ORDERING_MOVE_CORNER(); break;
			case 17 : ilupp_preproc_.set_MAX_WEIGHTED_MATCHING_ORDERING_MOVE_CORNER_IM(); break;
			case 18 : ilupp_preproc_.set_NORM_MAX_WEIGHTED_MATCHING_ORDERING_MOVE_CORNER_IM(); break;
			case 19 : ilupp_preproc_.set_MAX_WEIGHTED_MATCHING_ORDERING_WGT_MOV_COR(); break;
			case 20 : ilupp_preproc_.set_NORM_MAX_WEIGHTED_MATCHING_ORDERING_WGT_MOV_COR(); break;
			case 21 : ilupp_preproc_.set_MAX_WEIGHTED_MATCHING_ORDERING_WGT_MOV_COR_IM(); break;
			case 22 : ilupp_preproc_.set_NORM_MAX_WEIGHTED_MATCHING_ORDERING_WGT_MOV_COR_IM(); break;
			case 23 : ilupp_preproc_.set_MAX_WEIGHTED_MATCHING_ORDERING_WGT2_MOV_COR(); break;
			case 24 : ilupp_preproc_.set_NORM_MAX_WEIGHTED_MATCHING_ORDERING_WGT2_MOV_COR(); break;
			case 25 : ilupp_preproc_.set_MAX_WEIGHTED_MATCHING_ORDERING_WGT2_MOV_COR_IM(); break;
			case 26 : ilupp_preproc_.set_NORM_MAX_WEIGHTED_MATCHING_ORDERING_WGT2_MOV_COR_IM(); break;
			case 27 : ilupp_preproc_.set_MAX_WEIGHTED_MATCHING_ORDERING_SYM_PQ(); break;
			case 28 : ilupp_preproc_.set_NORM_MAX_WEIGHTED_MATCHING_ORDERING_SYM_PQ(); break;
			case 29 : ilupp_preproc_.set_MAX_WEIGHTED_MATCHING_ORDERING_SYMB_MOVE_CORNER(); break;
			case 30 : ilupp_preproc_.set_NORM_MAX_WEIGHTED_MATCHING_ORDERING_SYMB_MOVE_CORNER(); break;
			case 31 : ilupp_preproc_.set_MAX_WEIGHTED_MATCHING_ORDERING_SYMB_MOVE_CORNER_IM(); break;
			case 32 : ilupp_preproc_.set_NORM_MAX_WEIGHTED_MATCHING_ORDERING_SYMB_MOVE_CORNER_IM(); break;
			case 33 : ilupp_preproc_.set_MAX_WEIGHTED_MATCHING_ORDERING_SP_MOVE_CORNER(); break;
			case 34 : ilupp_preproc_.set_NORM_MAX_WEIGHTED_MATCHING_ORDERING_SP_MOVE_CORNER(); break;
			case 35 : ilupp_preproc_.set_MAX_WEIGHTED_MATCHING_ORDERING_SP_MOVE_CORNER_IM(); break;
			case 36 : ilupp_preproc_.set_NORM_MAX_WEIGHTED_MATCHING_ORDERING_SP_MOVE_CORNER_IM(); break;

			default: ilupp_preproc_.set_normalize(); break;
		  }

		  // set precond_parameters for ILU++
		  this->ilupp_param_.init(ilupp_preproc_, precond_no, ""); // ACHTUNG: setzt default-parameters

		   // set_threshold(tau) sets dropping threshold to 10^{-tau}
		  this->ilupp_param_.set_threshold(threshold);

		  // set minimal pivot
		  this->ilupp_param_.set_MEM_FACTOR(mem_factor);

		  if (detail_param) {
		  this->ilupp_param_.set_MAX_FILLIN_IS_INF(max_fillin_is_inf);	// default is true		  
		  this->ilupp_param_.set_fill_in(fill_in);		  
		  this->ilupp_param_.set_MIN_PIVOT(min_pivot);
		  this->ilupp_param_.set_MAX_LEVELS(max_levels);	
		  this->ilupp_param_.set_BEGIN_TOTAL_PIV(begin_total_piv);  // default is true
		  }
	}
	
// constructor
template <typename IndexType, typename ValueType, typename MemorySpace>
    ilupp<IndexType,ValueType,MemorySpace>
    ::ilupp()
    {
    }

template <typename IndexType, typename ValueType, typename MemorySpace>
template<typename MatrixType>
    ilupp<IndexType,ValueType,MemorySpace>
    ::ilupp(bool same_structure, const MatrixType& A,
			cusparseHandle_t& handle, bool use_mkl_sv_, ilupp_type type_, 
			const IndexType prepro_type, const IndexType precond_no,
            const IndexType max_levels, const ValueType mem_factor,
            const ValueType threshold_, const ValueType min_pivot, 
			bool max_fillin_is_inf, const IndexType fill_in_,
			bool begin_total_piv, bool detail_param_)
        : handle(handle), use_mkl_sv(use_mkl_sv_), type(type_), detail_param(detail_param_)
    {
        CUSP_PROFILE_SCOPED();

		sweeps  = 1;
		fill_in = fill_in_;
		threshold = threshold_;
		this->same_structure = same_structure;			
		if (this->same_structure) {
		nrow = A.num_rows;
		if (type == ilupp_iluc || type == ilupp_ilut) {
		/*-------------------MKL routines begin----------------------*/		
		if (use_mkl_sv) {
		temp_mkl_r.resize(A.num_rows);
		temp_mkl_y.resize(A.num_rows);
		temp_mkl_x.resize(A.num_rows);

		ipr_h.resize(A.num_rows);
		pc_h.resize(A.num_rows);	
		ipc_h.resize(A.num_rows);		
		dr_h.resize(A.num_rows);
		dc_h.resize(A.num_rows);
		pb_h.resize(A.num_rows);
		px_h.resize(A.num_rows);		
		}
		/*-------------------MKL routines end----------------------*/
		}
		}
		
		if (type == ilupp_iluc || type == ilupp_ilut) {
		/*-------------------cuSPARSE routines begin----------------------*/			
		if (!use_mkl_sv) {		
		// create a matrix description for the lower
		cusparseCreateMatDescr(&descrM_lo); 

		cusparseSetMatType (descrM_lo, CUSPARSE_MATRIX_TYPE_TRIANGULAR);
		cusparseSetMatIndexBase (descrM_lo, CUSPARSE_INDEX_BASE_ZERO);
		// NOTE! CUSPARSE_DIAG_TYPE_UNIT seems not right
		//cusparseSetMatDiagType (descrM_lo, CUSPARSE_DIAG_TYPE_UNIT);
		cusparseSetMatFillMode (descrM_lo, CUSPARSE_FILL_MODE_LOWER);	

		// create the analysis info object for the lower	
		cusparseCreateSolveAnalysisInfo(&Analy_lo);
		
		// create a matrix description for the upper
		cusparseCreateMatDescr(&descrM_up); 
		cusparseSetMatType (descrM_up, CUSPARSE_MATRIX_TYPE_TRIANGULAR);
		cusparseSetMatIndexBase (descrM_up, CUSPARSE_INDEX_BASE_ZERO);
		cusparseSetMatDiagType (descrM_up, CUSPARSE_DIAG_TYPE_NON_UNIT);
		cusparseSetMatFillMode (descrM_up, CUSPARSE_FILL_MODE_UPPER);	
			
		// create the analysis info object for the upper	
		cusparseCreateSolveAnalysisInfo(&Analy_up);
		
		if (this->same_structure) {		
		temp_y.resize(A.num_rows);	
		ipr_d.resize(A.num_rows);
		pc_d.resize(A.num_rows);		
		ipc_d.resize(A.num_rows);			
		dr_d.resize(A.num_rows);
		dc_d.resize(A.num_rows);
		scaled_d.resize(A.num_rows);		
		pb_d.resize(A.num_rows);
		px_d.resize(A.num_rows);		
		}		
		}
		/*-------------------cuSPARSE routines end----------------------*/	
		// initialize parameters for iluc and ilut
		init_parameter(prepro_type, threshold, fill_in);		
		} else if (type == multilevel_ilu) {
		// initialize parameters for multilevel_ilu
		init_parameter_multilevel(prepro_type, precond_no,
						max_levels, mem_factor,
						threshold, min_pivot,
						max_fillin_is_inf, fill_in, begin_total_piv);
		}
		initialized = true;
	}		
	
template <typename IndexType, typename ValueType, typename MemorySpace>
template<typename MatrixType>
    void ilupp<IndexType,ValueType,MemorySpace>
    ::setup(const MatrixType& A_, const IndexType sweeps_)
    {
        CUSP_PROFILE_SCOPED();
		
		if (!initialized) {
		  printf("ilupp has not yet been created!\n");
		  exit(1);		
		}
		
		cusp::csr_matrix<IndexType, ValueType, MemorySpace> A = A_;  // reformat to CSR
		cusp::detail::sort_by_row_and_column(A);  // sort by row and column  
		
		sweeps  = sweeps_;
	
		if (!same_structure) {
		nrow = A.num_rows;	
		if (type == ilupp_iluc || type == ilupp_ilut) {				
		if (use_mkl_sv) {
			temp_mkl_r.resize(A.num_rows);
			temp_mkl_y.resize(A.num_rows);
			temp_mkl_x.resize(A.num_rows);
			ipr_h.resize(A.num_rows);
			pc_h.resize(A.num_rows);
			ipc_h.resize(A.num_rows);				
			dr_h.resize(A.num_rows);	
			dc_h.resize(A.num_rows);
			pb_h.resize(A.num_rows);
			px_h.resize(A.num_rows);			
		} else {
			temp_y.resize(A.num_rows);	
			ipr_d.resize(A.num_rows);
			pc_d.resize(A.num_rows);
			ipc_d.resize(A.num_rows);				
			dr_d.resize(A.num_rows);
			dc_d.resize(A.num_rows);
			scaled_d.resize(A.num_rows);				
			pb_d.resize(A.num_rows);
			px_d.resize(A.num_rows);			
		}	
		}	
		}
		
		// copy to host
		cusp::csr_matrix<IndexType,ValueType,cusp::host_memory> A_host(A);				
		IndexType *h_row  = thrust::raw_pointer_cast(&*A_host.row_offsets.begin());
		IndexType *h_col  = thrust::raw_pointer_cast(&*A_host.column_indices.begin());	
		ValueType *h_val  = thrust::raw_pointer_cast(&*A_host.values.begin());	
		
		iluplusplus::orientation_type orientation = iluplusplus::ROW;	
		
		if (type == ilupp_iluc || type == ilupp_ilut) {
		iluplusplus::matrix A_temp,L,U;	
		iluplusplus::vector Dr0,Dc0;
		iluplusplus::index_list pr0,pc0,ipr0,ipc0;		
		IndexType number_rows = A.num_rows;
		IndexType number_cols = A.num_cols;	
		IndexType number_nz   = A.num_entries;	
		IndexType number_rows_lo,number_cols_lo,number_nz_lo;	
		IndexType number_rows_up,number_cols_up,number_nz_up;	
		
		// interchange to make iluplusplus matrix
		A_temp.interchange(h_val,h_col,h_row,number_rows,number_cols,number_nz,orientation);	
		// perform preprocess
		iluplusplus::matrix A_preprocess = A_temp;
		A_preprocess.preprocess(ilupp_param_,pr0,pc0,ipr0,ipc0,Dr0,Dc0);
		if (type == ilupp_iluc) {
		// construct preconditioner
		iluplusplus::iluc iluc_(A_preprocess, fill_in, threshold);
		// get lower triangle 
		L = iluc_.extract_left_matrix();	
		// get upper triangle 		
		U = iluc_.extract_right_matrix();	
		} else {
		// construct preconditioner		
		iluplusplus::ilut ilut_(A_preprocess, fill_in, threshold);
		// get lower triangle 
		L = ilut_.extract_left_matrix();	
		// get upper triangle 		
		U = ilut_.extract_right_matrix();	
		}
		// interchange back
		// notice that we can not only use one iluplusplus matrix to handle the interchange,
		// two matrices A_temp and A_preprocess must be defined		
		A_temp.interchange(h_val,h_col,h_row,number_rows,number_cols,number_nz,orientation);
		
		// setup A_host by processed matrix for A_mkl and A_p later
		IndexType* h_row_temp  = (IndexType *)malloc((number_rows+1) * sizeof(IndexType));
		IndexType* h_col_temp  = (IndexType *)malloc(number_nz * sizeof(IndexType));
		ValueType* h_val_temp  = (ValueType *)malloc(number_nz * sizeof(ValueType));
		A_preprocess.interchange(h_val_temp,h_col_temp,h_row_temp,number_rows,number_cols,number_nz,orientation);			
		//A_host.resize(number_rows,number_cols,number_nz);
		thrust::copy(h_row_temp, h_row_temp+number_rows+1, A_host.row_offsets.begin());
		thrust::copy(h_col_temp, h_col_temp+number_nz, A_host.column_indices.begin());
		thrust::copy(h_val_temp, h_val_temp+number_nz, A_host.values.begin());		
		A_preprocess.interchange(h_val_temp,h_col_temp,h_row_temp,number_rows,number_cols,number_nz,orientation);			
		delete h_row_temp;
		delete h_col_temp;
		delete h_val_temp;		
		
		// get CSR format from L	
		number_rows_lo = L.rows(); number_cols_lo = L.columns(); number_nz_lo = L.non_zeroes();	
		IndexType* h_lo_row  = (IndexType *)malloc((number_rows_lo+1) * sizeof(IndexType));
		IndexType* h_lo_col  = (IndexType *)malloc(number_nz_lo * sizeof(IndexType));
		ValueType* h_lo_val  = (ValueType *)malloc(number_nz_lo * sizeof(ValueType));
		L.interchange(h_lo_val,h_lo_col,h_lo_row,number_rows_lo,number_cols_lo,number_nz_lo,orientation);		
		
		// get CSR format from U				
		number_rows_up = U.rows(); number_cols_up = U.columns(); number_nz_up = U.non_zeroes();		
		IndexType* h_up_row  = (IndexType *)malloc((number_rows_up+1) * sizeof(IndexType));
		IndexType* h_up_col  = (IndexType *)malloc(number_nz_up * sizeof(IndexType));
		ValueType* h_up_val  = (ValueType *)malloc(number_nz_up * sizeof(ValueType));
		U.interchange(h_up_val,h_up_col,h_up_row,number_rows_up,number_cols_up,number_nz_up,orientation);		

		// get permutation and scale coefficient from ILU++
		IndexType* ipr0_ = ipr0.get_indices();
		IndexType* pc0_  = pc0.get_indices();		
		IndexType* ipc0_ = ipc0.get_indices();			
		ValueType* Dr0_  = Dr0.get_data();
		ValueType* Dc0_  = Dc0.get_data();		
		
		if (use_mkl_sv) {
		thrust::copy(ipr0_, ipr0_+A.num_rows, ipr_h.begin());		
		thrust::copy(pc0_, pc0_+A.num_rows, pc_h.begin());
		thrust::copy(ipc0_, ipc0_+A.num_rows, ipc_h.begin());		
		thrust::copy(Dr0_, Dr0_+A.num_rows, dr_h.begin());	
		thrust::copy(Dc0_, Dc0_+A.num_rows, dc_h.begin());	
		
		// setup preprocessed matrix for MKL
		cusp::detail::setup_mkl_matrix(A_mkl, A_host);	
		
		A_lo_mkl.resize(number_rows_lo,number_cols_lo,number_nz_lo);
		thrust::copy(thrust::make_transform_iterator(h_lo_row, cusp::detail::ADD<IndexType>(1)),
		thrust::make_transform_iterator(h_lo_row+number_rows_lo+1, cusp::detail::ADD<IndexType>(1)),
		A_lo_mkl.row_offsets.begin());
		thrust::copy(thrust::make_transform_iterator(h_lo_col, cusp::detail::ADD<IndexType>(1)),
		thrust::make_transform_iterator(h_lo_col+number_nz_lo, cusp::detail::ADD<IndexType>(1)),
		A_lo_mkl.column_indices.begin());
		thrust::copy(h_lo_val, h_lo_val+number_nz_lo, A_lo_mkl.values.begin());
		
		A_up_mkl.resize(number_rows_up,number_cols_up,number_nz_up);
		thrust::copy(thrust::make_transform_iterator(h_up_row, cusp::detail::ADD<IndexType>(1)),
		thrust::make_transform_iterator(h_up_row+number_rows_up+1, cusp::detail::ADD<IndexType>(1)),
		A_up_mkl.row_offsets.begin());
		thrust::copy(thrust::make_transform_iterator(h_up_col, cusp::detail::ADD<IndexType>(1)),
		thrust::make_transform_iterator(h_up_col+number_nz_up, cusp::detail::ADD<IndexType>(1)),
		A_up_mkl.column_indices.begin());
		thrust::copy(h_up_val, h_up_val+number_nz_up, A_up_mkl.values.begin());	
		} else {		
		thrust::copy(ipr0_, ipr0_+A.num_rows, ipr_d.begin());		
		thrust::copy(pc0_, pc0_+A.num_rows, pc_d.begin());	
		thrust::copy(ipc0_, ipc0_+A.num_rows, ipc_d.begin());			
		thrust::copy(Dr0_, Dr0_+A.num_rows, dr_d.begin());	
		thrust::copy(Dc0_, Dc0_+A.num_rows, dc_d.begin());	
		
		// setup preprocessed matrix
		A_p = A_host;
		
		A_lo.resize(number_rows_lo,number_cols_lo,number_nz_lo);
		thrust::copy(h_lo_row, h_lo_row+number_rows_lo+1, A_lo.row_offsets.begin());
		thrust::copy(h_lo_col, h_lo_col+number_nz_lo, A_lo.column_indices.begin());
		thrust::copy(h_lo_val, h_lo_val+number_nz_lo, A_lo.values.begin());
		
		A_up.resize(number_rows_up,number_cols_up,number_nz_up);
		thrust::copy(h_up_row, h_up_row+number_rows_up+1, A_up.row_offsets.begin());
		thrust::copy(h_up_col, h_up_col+number_nz_up, A_up.column_indices.begin());
		thrust::copy(h_up_val, h_up_val+number_nz_up, A_up.values.begin());
		
		// Perform the analysis for the lower
		cusparseDcsrsv_analysis(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, A.num_rows, descrM_lo,
		thrust::raw_pointer_cast(&A_lo.values[0]),
		thrust::raw_pointer_cast(&A_lo.row_offsets[0]),
		thrust::raw_pointer_cast(&A_lo.column_indices[0]),	
		Analy_lo);
		// Perform the analysis for the upper
		cusparseDcsrsv_analysis(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, A.num_rows, descrM_up,
		thrust::raw_pointer_cast(&A_up.values[0]),
		thrust::raw_pointer_cast(&A_up.row_offsets[0]),
		thrust::raw_pointer_cast(&A_up.column_indices[0]),	
		Analy_up);		
		}
		
		// interchange back
		L.interchange(h_lo_val,h_lo_col,h_lo_row,number_rows_lo,number_cols_lo,number_nz_lo,orientation);		
		U.interchange(h_up_val,h_up_col,h_up_row,number_rows_up,number_cols_up,number_nz_up,orientation);	
				
		// free dynamic allocation
		delete h_lo_val;
		delete h_lo_col;
		delete h_lo_row;
		delete h_up_val;
		delete h_up_col;
		delete h_up_row;
		
		} else if (type == multilevel_ilu) {
		multilevel_ilu_ = new iluplusplus::multilevel_preconditioner(); 
		IndexType number_rows = nrow;
		IndexType number_nz = A.num_entries;	
		multilevel_ilu_->setup(h_val, h_col, h_row, number_rows, number_nz, orientation, ilupp_param_);		
		}
    }

// destructor	
template <typename IndexType, typename ValueType, typename MemorySpace>
    ilupp<IndexType,ValueType,MemorySpace>
    ::~ilupp()
    {
		if (type == ilupp_iluc || type == ilupp_ilut) {	
		if (!use_mkl_sv) {		
		// destroy cuSPARSE context
		cusparseDestroyMatDescr(descrM_lo);
		cusparseDestroySolveAnalysisInfo(Analy_lo);  
		cusparseDestroyMatDescr(descrM_up);
		cusparseDestroySolveAnalysisInfo(Analy_up);
		}	
		} else if (type == multilevel_ilu) {
		delete multilevel_ilu_;
		}
    }

// linear_operator
template <typename IndexType, typename ValueType, typename MemorySpace>
template <typename Array1, typename Array2>
    void ilupp<IndexType,ValueType,MemorySpace>
    ::operator()(const Array1& b, Array2& x)
    {
        CUSP_PROFILE_SCOPED();
		
		if (type == ilupp_iluc || type == ilupp_ilut) {
		if (use_mkl_sv) {
		// only CPU memory is needed, copy back from device
		temp_mkl_r = b;
		// scale due to preprocess
		cusp::blas::xmy(temp_mkl_r.begin(), temp_mkl_r.end(),
		thrust::make_transform_iterator(dr_h.begin(), cusp::detail::INV<ValueType>(1)), 
		temp_mkl_r.begin());	
		// permute due to preprocess		
		cusp::permute_array(ipr_h,temp_mkl_r,pb_h);		
	
		// get the raw host pointer to be used in MKL
		ValueType *temp_r = thrust::raw_pointer_cast(&*pb_h.begin());		
		ValueType *temp_y = thrust::raw_pointer_cast(&*temp_mkl_y.begin());		
		ValueType *temp_x = thrust::raw_pointer_cast(&*px_h.begin());		
		IndexType *h_lo_row  = thrust::raw_pointer_cast(&*A_lo_mkl.row_offsets.begin());
		IndexType *h_lo_col  = thrust::raw_pointer_cast(&*A_lo_mkl.column_indices.begin());	
		ValueType *h_lo_val  = thrust::raw_pointer_cast(&*A_lo_mkl.values.begin());	
		IndexType *h_up_row  = thrust::raw_pointer_cast(&*A_up_mkl.row_offsets.begin());
		IndexType *h_up_col  = thrust::raw_pointer_cast(&*A_up_mkl.column_indices.begin());	
		ValueType *h_up_val  = thrust::raw_pointer_cast(&*A_up_mkl.values.begin());			
		cvar1='L';
		cvar='N';
		cvar2='U';
		mkl_dcsrtrsv(&cvar1,&cvar,&cvar2,&nrow,h_lo_val,h_lo_row,h_lo_col,temp_r,temp_y);
		cvar1='U';
		cvar='N';
		cvar2='N';
		mkl_dcsrtrsv(&cvar1,&cvar,&cvar2,&nrow,h_up_val,h_up_row,h_up_col,temp_y,temp_x);
		// permute back
		cusp::permute_array(pc_h,px_h,temp_mkl_x);
		// scale the result to device
		cusp::blas::xmy(temp_mkl_x.begin(), temp_mkl_x.end(),
		thrust::make_transform_iterator(dc_h.begin(), cusp::detail::INV<ValueType>(1)), 
		temp_mkl_x.begin());
		// copy to device
		x = temp_mkl_x;
		} else {
		// scale due to preprocess		
		cusp::blas::xmy(b.begin(), b.end(),
		thrust::make_transform_iterator(dr_d.begin(), cusp::detail::INV<ValueType>(1)), 
		scaled_d.begin());
		// permute due to preprocess		
		cusp::permute_array(ipr_d,scaled_d,pb_d);						
		
		ValueType *temp_y_ = thrust::raw_pointer_cast(&*temp_y.begin());		
		const ValueType *b_ = thrust::raw_pointer_cast(&*pb_d.begin());	
		ValueType *x_ = thrust::raw_pointer_cast(&*px_d.begin());	
        // Forward Solve: L y = r
        cusparseDcsrsv_solve(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, nrow, 1.0, descrM_lo, 
									thrust::raw_pointer_cast(&A_lo.values[0]),
									thrust::raw_pointer_cast(&A_lo.row_offsets[0]),
									thrust::raw_pointer_cast(&A_lo.column_indices[0]),	
									Analy_lo, b_, temp_y_);    
        // Back Substitution: U x = y
        cusparseDcsrsv_solve(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, nrow, 1.0, descrM_up, 
									thrust::raw_pointer_cast(&A_up.values[0]),
									thrust::raw_pointer_cast(&A_up.row_offsets[0]),
									thrust::raw_pointer_cast(&A_up.column_indices[0]),	
									Analy_up, temp_y_, x_); 
											
		// permute back
		cusp::permute_array(pc_d,px_d,scaled_d);	
		// scale the result to device
		cusp::blas::xmy(scaled_d.begin(), scaled_d.end(),
		thrust::make_transform_iterator(dc_d.begin(), cusp::detail::INV<ValueType>(1)), 
		x.begin());
        }
		} else if (type == multilevel_ilu) {
		// only CPU memory is needed, copy back from device
		temp_mkl_x = b;	
		ValueType *temp_x = thrust::raw_pointer_cast(&*temp_mkl_x.begin());	
		multilevel_ilu_->apply_preconditioner(temp_x, nrow);	
		// copy the result to device
		x = temp_mkl_x;
		}
    }

template <typename IndexType, typename ValueType, typename MemorySpace>
template<typename MatrixType, typename VectorType1, typename VectorType2>
    void ilupp<IndexType,ValueType,MemorySpace>
    ::presmooth(const MatrixType&A, const VectorType1& b, VectorType2& x)
    {
        CUSP_PROFILE_SCOPED();
			
		if (type == ilupp_iluc || type == ilupp_ilut) {		
		if (use_mkl_sv) {
		// only CPU memory is needed, copy back from device
		temp_mkl_r = b;
		// scale due to preprocess
		cusp::blas::xmy(temp_mkl_r.begin(), temp_mkl_r.end(),
		thrust::make_transform_iterator(dr_h.begin(), cusp::detail::INV<ValueType>(1)), 
		temp_mkl_r.begin());
		// permute due to preprocess		
		cusp::permute_array(ipr_h,temp_mkl_r,pb_h);

		// get the raw host pointer to be used in MKL
		ValueType *temp_r = thrust::raw_pointer_cast(&*pb_h.begin());		
		ValueType *temp_y = thrust::raw_pointer_cast(&*temp_mkl_y.begin());		
		ValueType *temp_x = thrust::raw_pointer_cast(&*px_h.begin());				
		IndexType *h_lo_row  = thrust::raw_pointer_cast(&*A_lo_mkl.row_offsets.begin());
		IndexType *h_lo_col  = thrust::raw_pointer_cast(&*A_lo_mkl.column_indices.begin());	
		ValueType *h_lo_val  = thrust::raw_pointer_cast(&*A_lo_mkl.values.begin());	
		IndexType *h_up_row  = thrust::raw_pointer_cast(&*A_up_mkl.row_offsets.begin());
		IndexType *h_up_col  = thrust::raw_pointer_cast(&*A_up_mkl.column_indices.begin());	
		ValueType *h_up_val  = thrust::raw_pointer_cast(&*A_up_mkl.values.begin());		
		cvar1='L';
		cvar='N';
		cvar2='U';
		mkl_dcsrtrsv(&cvar1,&cvar,&cvar2,&nrow,h_lo_val,h_lo_row,h_lo_col,temp_r,temp_y);
		cvar1='U';
		cvar='N';
		cvar2='N';
		mkl_dcsrtrsv(&cvar1,&cvar,&cvar2,&nrow,h_up_val,h_up_row,h_up_col,temp_y,temp_x);		
		cusp::array1d<ValueType,cusp::host_memory> res(A.num_rows);
		cusp::array1d<ValueType,cusp::host_memory> e(A.num_rows);
		ValueType *res_ = thrust::raw_pointer_cast(&*res.begin());	
		ValueType *e_ = thrust::raw_pointer_cast(&*e.begin());	
		IndexType *h_row  = thrust::raw_pointer_cast(&*A_mkl.row_offsets.begin());
		IndexType *h_col  = thrust::raw_pointer_cast(&*A_mkl.column_indices.begin());	
		ValueType *h_val = thrust::raw_pointer_cast(&*A_mkl.values.begin());			
		for (IndexType s = 1; s<this->sweeps; s++)
		{	
			// compute res <- b - A*x
			cvar='N';			
			mkl_dcsrgemv(&cvar,&nrow,h_val,h_row,h_col,temp_x, res_);
			cblas_daxpy(nrow,ValueType(-1), temp_r, MKL_INT(1), res_, MKL_INT(1));
			cblas_dscal(nrow,ValueType(-1),res_,MKL_INT(1));		
			cvar1='L';
			cvar='N';
			cvar2='U';
			mkl_dcsrtrsv(&cvar1,&cvar,&cvar2,&nrow,h_lo_val,h_lo_row,h_lo_col,res_,temp_y);
			cvar1='U';
			cvar='N';
			cvar2='N';
			mkl_dcsrtrsv(&cvar1,&cvar,&cvar2,&nrow,h_up_val,h_up_row,h_up_col,temp_y,e_);	
			cblas_daxpy(nrow,ValueType(1), e_, MKL_INT(1), temp_x, MKL_INT(1));	
		}		
		// permute back
		cusp::permute_array(pc_h,px_h,temp_mkl_x);
		// scale the result to device
		cusp::blas::xmy(temp_mkl_x.begin(), temp_mkl_x.end(),
		thrust::make_transform_iterator(dc_h.begin(), cusp::detail::INV<ValueType>(1)), 
		temp_mkl_x.begin());
		// copy to device
		x = temp_mkl_x;
		} else {
		// scale due to preprocess		
		cusp::blas::xmy(b.begin(), b.end(),
		thrust::make_transform_iterator(dr_d.begin(), cusp::detail::INV<ValueType>(1)), 
		scaled_d.begin());
		// permute due to preprocess		
		cusp::permute_array(ipr_d,scaled_d,pb_d);	
		
		ValueType *temp_y_ = thrust::raw_pointer_cast(&*temp_y.begin());		
		const ValueType *b_ = thrust::raw_pointer_cast(&*pb_d.begin());	
		ValueType *x_ = thrust::raw_pointer_cast(&*px_d.begin());			
        // Forward Solve: L y = r
        cusparseDcsrsv_solve(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, nrow, 1.0, descrM_lo, 
									thrust::raw_pointer_cast(&A_lo.values[0]),
									thrust::raw_pointer_cast(&A_lo.row_offsets[0]),
									thrust::raw_pointer_cast(&A_lo.column_indices[0]),	
									Analy_lo, b_, temp_y_);    
        // Back Substitution: U x = y
        cusparseDcsrsv_solve(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, nrow, 1.0, descrM_up, 
									thrust::raw_pointer_cast(&A_up.values[0]),
									thrust::raw_pointer_cast(&A_up.row_offsets[0]),
									thrust::raw_pointer_cast(&A_up.column_indices[0]),	
									Analy_up, temp_y_, x_); 
		VectorType1 res(A.num_rows);
		VectorType2 e(A.num_rows);
		ValueType *res_ = thrust::raw_pointer_cast(&*res.begin());	
		ValueType *e_ = thrust::raw_pointer_cast(&*e.begin());	
		for (IndexType s = 1; s<this->sweeps; s++)
		{	
			// compute res <- b - A*x
			cusp::multiply(A_p, px_d, res);
			cusp::blas::axpby(pb_d, res, res, ValueType(1), ValueType(-1));			
			// Forward Solve: L y = r
			cusparseDcsrsv_solve(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, nrow, 1.0, descrM_lo, 
										thrust::raw_pointer_cast(&A_lo.values[0]),
										thrust::raw_pointer_cast(&A_lo.row_offsets[0]),
										thrust::raw_pointer_cast(&A_lo.column_indices[0]),	
										Analy_lo, res_, temp_y_);    
			// Back Substitution: U x = y
			cusparseDcsrsv_solve(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, nrow, 1.0, descrM_up, 
										thrust::raw_pointer_cast(&A_up.values[0]),
										thrust::raw_pointer_cast(&A_up.row_offsets[0]),
										thrust::raw_pointer_cast(&A_up.column_indices[0]),	
										Analy_up, temp_y_, e_); 
			cusp::blas::axpby(px_d, e, px_d, ValueType(1), ValueType(1));				
		}
		// permute back
		cusp::permute_array(pc_d,px_d,scaled_d);	
		// scale the result to device
		cusp::blas::xmy(scaled_d.begin(), scaled_d.end(),
		thrust::make_transform_iterator(dc_d.begin(), cusp::detail::INV<ValueType>(1)), 
		x.begin());		
        }
		} else if (type == multilevel_ilu) {
		// only CPU memory is needed, copy back from device
		temp_mkl_x = b;	
		ValueType *temp_x = thrust::raw_pointer_cast(&*temp_mkl_x.begin());	
		multilevel_ilu_->apply_preconditioner(temp_x, nrow);	
		// copy the result to device
		x = temp_mkl_x;
		
		VectorType1 res(A.num_rows);
		VectorType2 e(A.num_rows);		
		for (IndexType s = 1; s<this->sweeps; s++)
		{	
			// compute res <- b - A*x
			cusp::multiply(A, x, res); // no need to be permuted, permuted in apply_preconditioner
			cusp::blas::axpby(b, res, res, ValueType(1), ValueType(-1));	
			// copy from device to host			
			temp_mkl_x = res;
			ValueType *temp_x = thrust::raw_pointer_cast(&*temp_mkl_x.begin());	
			multilevel_ilu_->apply_preconditioner(temp_x, nrow);	
			// copy the result to device
			e = temp_mkl_x;			
			cusp::blas::axpby(x, e, x, ValueType(1), ValueType(1));									
		}			
		}
    }

template <typename IndexType, typename ValueType, typename MemorySpace>
template<typename MatrixType, typename VectorType1, typename VectorType2>
    void ilupp<IndexType,ValueType,MemorySpace>
    ::postsmooth(const MatrixType& A, const VectorType1& b, VectorType2& x)
    {
        CUSP_PROFILE_SCOPED();
			
		if (type == ilupp_iluc || type == ilupp_ilut) {	
		if (use_mkl_sv) {
		// only CPU memory is needed, copy back from device
		temp_mkl_r = b;
		// scale due to preprocess
		cusp::blas::xmy(temp_mkl_r.begin(), temp_mkl_r.end(),
		thrust::make_transform_iterator(dr_h.begin(), cusp::detail::INV<ValueType>(1)), 
		temp_mkl_r.begin());
		// permute due to preprocess		
		cusp::permute_array(ipr_h,temp_mkl_r,pb_h);	
		
		// only CPU memory is needed, copy back from device		
		temp_mkl_x = x;		
		// scale 
		cusp::blas::xmy(temp_mkl_x.begin(), temp_mkl_x.end(), dc_h.begin(), temp_mkl_x.begin());
		// permute 
		cusp::permute_array(ipc_h,temp_mkl_x,px_h);
		
		// get the raw host pointer to be used in MKL
		ValueType *temp_r = thrust::raw_pointer_cast(&*pb_h.begin());		
		ValueType *temp_y = thrust::raw_pointer_cast(&*temp_mkl_y.begin());		
		ValueType *temp_x = thrust::raw_pointer_cast(&*px_h.begin());		
		IndexType *h_lo_row  = thrust::raw_pointer_cast(&*A_lo_mkl.row_offsets.begin());
		IndexType *h_lo_col  = thrust::raw_pointer_cast(&*A_lo_mkl.column_indices.begin());	
		ValueType *h_lo_val  = thrust::raw_pointer_cast(&*A_lo_mkl.values.begin());	
		IndexType *h_up_row  = thrust::raw_pointer_cast(&*A_up_mkl.row_offsets.begin());
		IndexType *h_up_col  = thrust::raw_pointer_cast(&*A_up_mkl.column_indices.begin());	
		ValueType *h_up_val  = thrust::raw_pointer_cast(&*A_up_mkl.values.begin());			
		IndexType *h_row  = thrust::raw_pointer_cast(&*A_mkl.row_offsets.begin());
		IndexType *h_col  = thrust::raw_pointer_cast(&*A_mkl.column_indices.begin());	
		ValueType *h_val  = thrust::raw_pointer_cast(&*A_mkl.values.begin());
		cusp::array1d<ValueType,cusp::host_memory> res(A.num_rows);
		cusp::array1d<ValueType,cusp::host_memory> e(A.num_rows);
		ValueType *res_ = thrust::raw_pointer_cast(&*res.begin());	
		ValueType *e_ = thrust::raw_pointer_cast(&*e.begin());			
		for (IndexType s = 0; s<this->sweeps; s++)
		{	
			// compute res <- b - A*x
			cvar='N';			
			mkl_dcsrgemv(&cvar,&nrow,h_val,h_row,h_col,temp_x, res_);
			cblas_daxpy(nrow,ValueType(-1), temp_r, MKL_INT(1), res_, MKL_INT(1));
			cblas_dscal(nrow,ValueType(-1),res_,MKL_INT(1));
			cvar1='L';
			cvar='N';
			cvar2='U';
			mkl_dcsrtrsv(&cvar1,&cvar,&cvar2,&nrow,h_lo_val,h_lo_row,h_lo_col,res_,temp_y);
			cvar1='U';
			cvar='N';
			cvar2='N';
			mkl_dcsrtrsv(&cvar1,&cvar,&cvar2,&nrow,h_up_val,h_up_row,h_up_col,temp_y,e_);	
			cblas_daxpy(nrow,ValueType(1), e_, MKL_INT(1), temp_x, MKL_INT(1));	
		}		
		// permute back
		cusp::permute_array(pc_h,px_h,temp_mkl_x);
		// scale the result to device
		cusp::blas::xmy(temp_mkl_x.begin(), temp_mkl_x.end(),
		thrust::make_transform_iterator(dc_h.begin(), cusp::detail::INV<ValueType>(1)), 
		temp_mkl_x.begin());
		// copy to device
		x = temp_mkl_x;
		} else {
		// scale due to preprocess		
		cusp::blas::xmy(b.begin(), b.end(),
		thrust::make_transform_iterator(dr_d.begin(), cusp::detail::INV<ValueType>(1)), 
		scaled_d.begin());
		// permute due to preprocess		
		cusp::permute_array(ipr_d,scaled_d,pb_d);	
		
		// scale 
		cusp::blas::xmy(x.begin(), x.end(), dc_d.begin(), scaled_d.begin());
		// permute 
		cusp::permute_array(ipc_d,scaled_d,px_d);

		ValueType *temp_y_ = thrust::raw_pointer_cast(&*temp_y.begin());		
		VectorType1 res(A.num_rows);
		VectorType2 e(A.num_rows);
		ValueType *res_ = thrust::raw_pointer_cast(&*res.begin());	
		ValueType *e_ = thrust::raw_pointer_cast(&*e.begin());	
		for(IndexType s = 0; s<this->sweeps; s++)
		{	
			// compute res <- b - A*x
			cusp::multiply(A_p, px_d, res);
			cusp::blas::axpby(pb_d, res, res, ValueType(1), ValueType(-1));		
			// Forward Solve: L y = r
			cusparseDcsrsv_solve(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, nrow, 1.0, descrM_lo, 
										thrust::raw_pointer_cast(&A_lo.values[0]),
										thrust::raw_pointer_cast(&A_lo.row_offsets[0]),
										thrust::raw_pointer_cast(&A_lo.column_indices[0]),	
										Analy_lo, res_, temp_y_);    
			// Back Substitution: U x = y
			cusparseDcsrsv_solve(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, nrow, 1.0, descrM_up, 
										thrust::raw_pointer_cast(&A_up.values[0]),
										thrust::raw_pointer_cast(&A_up.row_offsets[0]),
										thrust::raw_pointer_cast(&A_up.column_indices[0]),	
										Analy_up, temp_y_, e_); 
			cusp::blas::axpby(px_d, e, px_d, ValueType(1), ValueType(1));									
		}	
		// permute back
		cusp::permute_array(pc_d,px_d,scaled_d);	
		// scale the result to device
		cusp::blas::xmy(scaled_d.begin(), scaled_d.end(),
		thrust::make_transform_iterator(dc_d.begin(), cusp::detail::INV<ValueType>(1)), 
		x.begin());			
        }
		} else if (type == multilevel_ilu) {
		VectorType1 res(A.num_rows);
		VectorType2 e(A.num_rows);		
		for (IndexType s = 0; s<this->sweeps; s++)
		{	
			// compute res <- b - A*x
			cusp::multiply(A, x, res);
			cusp::blas::axpby(b, res, res, ValueType(1), ValueType(-1));
			// copy from device to host
			temp_mkl_x = res;
			ValueType *temp_x = thrust::raw_pointer_cast(&*temp_mkl_x.begin());	
			multilevel_ilu_->apply_preconditioner(temp_x, nrow);	
			// copy the result to device
			e = temp_mkl_x;			
			cusp::blas::axpby(x, e, x, ValueType(1), ValueType(1));									
		}			
		}
    }

} // end namespace relaxation
} // end namespace cusp

