/*
 *  Copyright 2008-2009 NVIDIA Corporation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

/*! \file kpz_polynomial.inl
 *  \brief Inline file for kpz_polynomial.h
 */

#include <cusp/multiply.h>
#include <cusp/transpose.h>
#include <cusp/blas.h>

#include <thrust/functional.h>
#include <thrust/transform.h>

namespace cusp
{
namespace relaxation
{
namespace detail
{

template <typename ValueType, typename IndexType>
struct row_sum_abs_functor  : public thrust::unary_function<IndexType, ValueType>
{
    ValueType *row_data;
    IndexType *row_ptr; 

    row_sum_abs_functor(ValueType *data, IndexType *ptr) : row_data(data), row_ptr(ptr) {}

    __host__ __device__
    ValueType operator()(IndexType r)
    {
        IndexType row_start = row_data[r];
        IndexType row_end   = row_data[r+1];

        ValueType v = 0;
        for (IndexType row_i = row_start; row_i < row_end; row_i++) {
          ValueType data = row_data[row_i];
          v += (data < 0) ? -data : data;
        }
        return v;
    }
};

} // end namespace detail


// constructor
template <typename ValueType, typename MemorySpace>
template<typename MatrixType>
    kpz_polynomial<ValueType,MemorySpace>
    ::kpz_polynomial(const MatrixType& A, int poly_order) : order(poly_order), mu((ValueType)4)
    {
      // calculate l_inf_colsum
      typename cusp::csr_matrix<typename MatrixType::index_type, typename MatrixType::value_type, MemorySpace> Acsr;

      // transpose A into csr format (since we want col sums)
      cusp::transpose(A, Acsr);

      // compute max of the row sums - expressed as a transform_reduce on the row indices
      l_inf_colsum = thrust::transform_reduce(
        thrust::make_counting_iterator<typename MatrixType::index_type>(0), 
        thrust::make_counting_iterator<typename MatrixType::index_type>(0)+A.num_rows,
        detail::row_sum_abs_functor<typename MatrixType::value_type, typename MatrixType::index_type>(
          thrust::raw_pointer_cast(&*Acsr.values.begin()),
          thrust::raw_pointer_cast(&*Acsr.row_offsets.begin())),
        (ValueType)0, 
        thrust::maximum<ValueType>());
    }

// linear_operator
template <typename ValueType, typename MemorySpace>
template<typename MatrixType, typename VectorType1, typename VectorType2>
    void kpz_polynomial<ValueType,MemorySpace>
    ::operator()(const MatrixType& A, const VectorType1& b, VectorType2& x) 
    {
        ValueType smax = l_inf_colsum;
        ValueType smin = smax / mu;

        ValueType smu0 = 1/smax;
        ValueType smu1 = 1/smin;
        ValueType skappa = (ValueType)sqrt(smax/smin);
        ValueType delta = (skappa - 1) / (skappa + 1);
        ValueType beta = (ValueType)(sqrt(smu0) + sqrt(smu1));
        beta = beta * beta;
        ValueType chi = (ValueType) ((4*smu0*smu1) / beta);
      
        cusp::array1d<ValueType,MemorySpace> r(A.num_rows), v0(A.num_rows), v(A.num_rows), sn(A.num_rows);
        cusp::array1d<ValueType,MemorySpace> Ax(A.num_rows);

        // r = b - Ax
        cusp::multiply(A, x, r);
        cusp::blas::axpby(b, r, r, (ValueType)1, (ValueType)-1);


        // v0 = (smu0+smu1)/2 * r
        cusp::blas::copy(r, v0);
        cusp::blas::scal(v0, (ValueType) ((smu0+smu1)/2));

        // v = (sqrt(smu0) + sqrt(smu1))^2/2 * r - smu0*smu1*A*r
        cusp::multiply(A, r, v);
        cusp::blas::axpby(r,v,v, (ValueType) (beta/2), -smu0*smu1);
      
        for (int i=2; i <= order; i++) {
          // sn = chi *(r-A*v) + delta*delta*(v-v0);
          cusp::multiply(A, v, sn);
          cusp::blas::axpby(r, sn, sn, (ValueType)1, (ValueType)-1);
          cusp::blas::axpbypcz(sn, v, v0, sn, chi, delta*delta, -delta*delta);

          // v0 <- v
          // v  <- v + sn
          cusp::blas::copy(v, v0);
          cusp::blas::axpy(sn, v, (ValueType)1);
        }

        // x = x + v
        cusp::blas::axpy(v, x, (ValueType)1);
 

    }

template <typename ValueType, typename MemorySpace>
template<typename MatrixType, typename VectorType1, typename VectorType2>
    void kpz_polynomial<ValueType,MemorySpace>
    ::presmooth(const MatrixType& A, const VectorType1& b, VectorType2& x) 
    {
      thrust::fill(x.begin(), x.end(), (typename VectorType2::value_type)0);
      (*this)(A, b, x);
    }
 
template <typename ValueType, typename MemorySpace>
template<typename MatrixType, typename VectorType1, typename VectorType2>
    void kpz_polynomial<ValueType,MemorySpace>
    ::postsmooth(const MatrixType& A, const VectorType1& b, VectorType2& x) 
    {
      (*this)(A, b, x);
    }
 
} // end namespace relaxation
} // end namespace cusp

