/*
 *  Copyright 2010 NVIDIA Corporation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

/*! \file line_relaxation.inl
 *  \brief Inline file for line_relaxation.h
 */

#include <cusp/multiply.h>
#include <cusp/detail/format_utils.h>
#include <cusp/io/matrix_market.h>
#include <cusp/elementwise.h>

#include <thrust/functional.h>
#include <thrust/transform.h>
#include <thrust/iterator/zip_iterator.h>

namespace cusp
{
namespace relaxation
{
namespace detail
{

//! kernel to set up the tridiagonal system for the pcr kernel
template<typename IndexType, typename ValueType>
__global__
void
tridiagonal_system_setup(
        ValueType *a, ValueType *b,  ValueType *c,
        const IndexType* lines, const IndexType* line_ends, const int line_count, int line_size,
        int max_path_size, ValueType alpha, int* row_offsets, int* column_indices,
        ValueType* values)
{
    // one thread per path/line
    int line_number = blockDim.x * blockIdx.x + threadIdx.x;

    int current_diagonal_pos;
    int current_line_pos;
    int current_row;
    int previous_row;
    int next_row;
    int i;

    if (line_number>=line_count)
        return;

    // loop over this line
    for (int line_offset=0;line_offset<line_ends[line_number];line_offset++)
    {
        current_line_pos=line_number*line_size+line_offset;
        current_diagonal_pos=line_number*max_path_size+line_offset;
        current_row = lines[current_line_pos];

        // set center diagonal value
        i=row_offsets[current_row];
        while (column_indices[i]!=current_row)        // I assume the existence of a diagonal entry, which might be a problem
        {
            i++;
        }
        b[current_diagonal_pos]= values[i];
        b[current_diagonal_pos]+=alpha;

        // set west diagonal value
        if (line_offset-1>=0)
        {
            previous_row = lines[current_line_pos-1];
            i = row_offsets[current_row];
            while (column_indices[i]!=previous_row)
            {
                i++;
            }
            a[current_diagonal_pos] = values[i];
        }

        // set east diagonal value
        if (line_offset+1<line_ends[line_number])
        {
            next_row = lines[current_line_pos+1];
            i=row_offsets[current_row];
            while (column_indices[i]!=next_row)
            {
                i++;
            }
            c[current_diagonal_pos] = values[i];
        }

    }

}

//! kernel that sets up a ellpack matrix containing the non-permuted tridiagonal matrix
template<typename IndexType, typename ValueType>
__global__ void
tridiagonal_matrix_setup(
        IndexType *tridiag_column_indices, ValueType *tridiag_values, int tridiag_values_num_rows,
        const int *lines, const int *line_ends , const int line_count, const int line_size,
                const int max_path_size, const int *row_offsets, const int *column_indices,
                const ValueType * values)
{
    // one thread per line / path
    int line_number = blockDim.x * blockIdx.x + threadIdx.x;

    int current_line_pos;
    int current_row;
    int previous_row;
    int next_row;
    int tridiag_row_entry_count;
    int current_column_index;

    if (line_number>=line_count)
        return;

    // loop over current line
    for (int line_offset=0;line_offset<line_ends[line_number];line_offset++)
    {
        current_line_pos=line_number*line_size+line_offset;
        current_row = lines[current_line_pos];
        previous_row = (line_offset>0) ? lines[current_line_pos-1] : -1;
        next_row = (line_offset<line_ends[line_number]-1) ? lines[current_line_pos+1] : -1;

        tridiag_row_entry_count=0;

        // loop over current matrix row
        for (int i=row_offsets[current_row];i<row_offsets[current_row+1];i++)
        {
            // find column of the current entry
            current_column_index=column_indices[i];

            // set entries
            if ((current_column_index==current_row)||(current_column_index==previous_row)||(current_column_index==next_row))
            {
                tridiag_column_indices[current_row+tridiag_values_num_rows*tridiag_row_entry_count]=current_column_index;
                tridiag_values[current_row+tridiag_values_num_rows*tridiag_row_entry_count]=values[i];
                tridiag_row_entry_count++;
            }
        }
    }
}

//! method that applies a pointwise relaxation to non-covered unknowns in the system
template <typename ValueType>
__global__ void  apply_jacobi(
        ValueType *y,
        const int * row_visited, const int* row_offsets,
        const int * column_indices, const ValueType * values, const int num_rows, ValueType alpha, ValueType *x_old, const ValueType *rhs)
{
    // one thread per matrix row
    int current_row = blockDim.x * blockIdx.x + threadIdx.x;

    ValueType diag,r;

    if (current_row<num_rows)
    {
        // do it only on non-covered unknowns
        if (row_visited[current_row]==0)
        {
            r=rhs[current_row];
            for (int i=row_offsets[current_row];i<row_offsets[current_row+1];i++)
            {
                if (column_indices[i]==current_row)
                {
                    diag=values[i];
                }
                else
                {
                    r-=values[i]*x_old[column_indices[i]];
                }
            }
            y[current_row]=(r+alpha*x_old[current_row])/diag;
        }

    }
}


//! direct solver for the tridiagonal systems including the reordering of the RHS
//! this code is based on the PCR kernel of the cudpp project developed by Yao Zhang
template <typename ValueType>
__global__ void
pcr_kernel_with_reordering(ValueType *y, const int *lines, const int *line_ends , const int line_count, const int line_size,
                const int max_path_size, int grid_size_x, const ValueType* rhs, const ValueType * d_a, const ValueType * d_b, const ValueType * d_c, ValueType *d_d, ValueType *d_x)
{
    // one line / linear system  per block
    int line_number = blockIdx.x+blockIdx.y*grid_size_x;

    int current_line_pos;
    int current_row;

    if (line_number>=line_count)
        return;

    // setup of index variables for the linear solver
    int thid = threadIdx.x;
    int blid = blockIdx.x+blockIdx.y*grid_size_x;
    int delta = 1;
    const unsigned int systemSize = blockDim.x;
    int iteration = lround(log2(ValueType(systemSize/2.0)));

    // setup of shared memory decomposition
    extern __shared__ char shared[];

    ValueType* a = (ValueType*)shared;
    ValueType* b = (ValueType*)&a[systemSize+1];//
    ValueType* c = (ValueType*)&b[systemSize+1];
    ValueType* d = (ValueType*)&c[systemSize+1];
    ValueType* x = (ValueType*)&d[systemSize+1];

    // read three non-zero diagonals into shmem
    a[thid] = d_a[thid + blid * systemSize];
    b[thid] = d_b[thid + blid * systemSize];
    c[thid] = d_c[thid + blid * systemSize];
    d[thid] = 0;

    // read RHS while reordering it
    if (thid<line_ends[line_number])
    {
        current_line_pos=line_number*line_size+thid;
        current_row = lines[current_line_pos];

        d[thid]=rhs[current_row];
    }

    // failback for only one unknown per system which was not supported by the PCR kernel
    if (iteration==-1)
    {
        d_x[thid+blid*systemSize]=d[thid]/d_b[thid+blid*systemSize];
        return;
    }

    __syncthreads();

    //parallel cyclic reduction as done in the CUDPP PCR kernel
    for (int j = 0; j <iteration; j++)
    {
        int i = thid;

        int iRight = i+delta;
        if (iRight >= systemSize) iRight = systemSize - 1;

        int iLeft = i-delta;
        if (iLeft < 0) iLeft = 0;

        ValueType tmp1 = a[i] / b[iLeft];
        ValueType tmp2 = c[i] / b[iRight];
        ValueType tmp3 = d[iRight];
        ValueType tmp4 = d[iLeft];
        __syncthreads();

        b[i] = b[i]-c[iLeft]*tmp1-a[iRight]*tmp2;
        d[i] = d[i] - tmp4*tmp1-tmp3*tmp2;
        tmp3 = -a[iLeft];
        tmp4 = -c[iRight];
        __syncthreads();
        a[i] = tmp3*tmp1;
        c[i] = tmp4*tmp2;

        __syncthreads();


        delta *=2;
    }

    if (thid < delta)
    {

        int addr1 = thid;
        int addr2 = thid+delta;
        ValueType tmp3 = b[addr2]*b[addr1]-c[addr1]*a[addr2];
        x[addr1] = (b[addr2]*d[addr1]-c[addr1]*d[addr2])/tmp3;
        x[addr2] = (d[addr2]*b[addr1]-d[addr1]*a[addr2])/tmp3;
    }

    __syncthreads();

    // reorder back the results to their original position in the matrix
    if (thid<line_ends[line_number])
    {
        current_line_pos=line_number*line_size+thid;
        current_row = lines[current_line_pos];

        y[current_row]=x[thid];
    }
}



} // end namespace detail

//! initialization procedure
template <typename IndexType, typename ValueType, typename MemorySpace>
void line <IndexType,ValueType,MemorySpace>
    ::init(const csr_matrix<IndexType, ValueType, MemorySpace>& A)
{
    IndexType *row_offsets_d, *column_indices_d;
    const IndexType *lines_d, *line_ends_d;
    ValueType* values_d;
    ValueType *a_d, *b_d, *c_d;

    // get raw pointers for cuda kernel calls
    row_offsets_d=thrust::raw_pointer_cast(&A_mat.row_offsets[0]);
    column_indices_d=thrust::raw_pointer_cast(&A_mat.column_indices[0]);
    values_d=thrust::raw_pointer_cast(&A_mat.values[0]);
    lines_d=thrust::raw_pointer_cast(&cover.paths[0]);
    line_ends_d=thrust::raw_pointer_cast(&cover.path_lengths[0]);

    // setting tridiagonal systems sizes and fill it with reasonable values
    a.resize(cover.max_path_size*cover.actual_path_count, 0.0);
    b.resize(cover.max_path_size*cover.actual_path_count, 1.0);
    c.resize(cover.max_path_size*cover.actual_path_count, 0.0);

    // get raw pointers for cuda kernel calls
    a_d=thrust::raw_pointer_cast(&a[0]);
    b_d=thrust::raw_pointer_cast(&b[0]);
    c_d=thrust::raw_pointer_cast(&c[0]);

    // set parameters for kernel calls
    int block_count=512;
    int thread_count=(cover.actual_path_count+511)/512;

    // set up tridiagonal system
    detail::tridiagonal_system_setup<IndexType,ValueType><<<block_count,thread_count>>>(a_d, b_d, c_d,
            lines_d, line_ends_d, cover.actual_path_count, cover.params.path_size, cover.max_path_size, alpha,
            row_offsets_d, column_indices_d, values_d);

    // compute the matrix A_rest which allows to compute the RHS of the tridiagonal system by simple spmv operations
    // this might be still inefficient but does not take to much percentage of the computation time
    cusp::ell_matrix<IndexType, ValueType, MemorySpace> A_tridiag_tmp(A.num_rows,A.num_cols,cover.covered_row_count*3-cover.actual_path_count*2,3);
    IndexType *A_tridiag_tmp_column_indices_d=thrust::raw_pointer_cast(&A_tridiag_tmp.column_indices.values[0]);
    ValueType *A_tridiag_tmp_values_d=thrust::raw_pointer_cast(&A_tridiag_tmp.values.values[0]);
    thrust::fill(A_tridiag_tmp.column_indices.values.begin(), A_tridiag_tmp.column_indices.values.end(), -1);
    thrust::fill(A_tridiag_tmp.values.values.begin(), A_tridiag_tmp.values.values.end(), (ValueType)-1.0);

    detail::tridiagonal_matrix_setup<IndexType,ValueType><<<block_count,thread_count>>>(A_tridiag_tmp_column_indices_d,
            A_tridiag_tmp_values_d,    A_tridiag_tmp.column_indices.num_rows, lines_d, line_ends_d , cover.actual_path_count,
            cover.params.path_size, cover.max_path_size, row_offsets_d, column_indices_d, values_d);
    cusp::csr_matrix<IndexType, ValueType, MemorySpace> A_rest_tmp(A);
    cusp::subtract(A,A_tridiag_tmp,A_rest_tmp);
    A_rest=A_rest_tmp;
}

//! constructor
template <typename IndexType, typename ValueType, typename MemorySpace>
    line <IndexType,ValueType,MemorySpace>
    ::line(const csr_matrix<IndexType, ValueType, MemorySpace>& A, const cusp::graph::path_cover<IndexType, ValueType, MemorySpace>& _cover, int _sweep_count, ValueType _alpha)
     : A_mat(A), alpha(_alpha), cover(_cover), sweep_count(_sweep_count)
{
    init(A);
}

//! linear_operator which applies the line relaxation
//! \param bb current RHS vector
//! \param xx vector which shall be smoothed
template <typename IndexType, typename ValueType, typename MemorySpace>
template<typename VectorType1, typename VectorType2>
    void line<IndexType,ValueType,MemorySpace>
    ::operator()(const VectorType1& bb, VectorType2& xx) const
{
    const IndexType *row_offsets_d;
    const IndexType *column_indices_d;
    const IndexType *row_visited_d, *lines_d, *line_ends_d;
    const ValueType *a_d, *b_d, *c_d;
    const ValueType *values_d;

    // get raw pointers for cuda kernel calls
    row_offsets_d=thrust::raw_pointer_cast(&A_mat.row_offsets[0]);
    column_indices_d=thrust::raw_pointer_cast(&A_mat.column_indices[0]);
    values_d=thrust::raw_pointer_cast(&A_mat.values[0]);
    row_visited_d=thrust::raw_pointer_cast(&cover.row_visited[0]);
    lines_d=thrust::raw_pointer_cast(&cover.paths[0]);
    line_ends_d=thrust::raw_pointer_cast(&cover.path_lengths[0]);
    a_d=thrust::raw_pointer_cast(&a[0]);
    b_d=thrust::raw_pointer_cast(&b[0]);
    c_d=thrust::raw_pointer_cast(&c[0]);

    // setup of some temporary vectors including raw pointers
    // TODO: reduce amount of temporary arrays
    thrust::device_vector<ValueType> r(cover.max_path_size*cover.actual_path_count, 1.0);
    thrust::device_vector<ValueType> x(cover.max_path_size*cover.actual_path_count, 0.0);
    ValueType* r_d=thrust::raw_pointer_cast(&r[0]);
    ValueType* x_d=thrust::raw_pointer_cast(&x[0]);
    cusp::array1d<ValueType, MemorySpace> res(A_mat.num_rows, (ValueType)0.0);
    cusp::array1d<ValueType, MemorySpace> Ax(A_mat.num_rows, (ValueType)0.0);
    cusp::array1d<ValueType, MemorySpace> y(A_mat.num_rows, (ValueType)0.0);

    //-----------------------------------------------------------------
    // compute residual
    //-----------------------------------------------------------------
    // Ax <- A_mat x
    cusp::multiply(A_mat, xx, Ax);

    // res <- b - A*x
    blas::axpby(bb, Ax, res, ValueType(1), ValueType(-1));

    //-----------------------------------------------------------------
    // compute y = M^-1 * res, with M^-1 our line solver preconditioner
    //-----------------------------------------------------------------

    // allocate intermediate vector for several line relaxation sweeps
    cusp::array1d<ValueType, MemorySpace> x_old(A_mat.num_rows, (ValueType)0.0);

    // get raw pointer for kernel calls
    const ValueType* rhs_d = thrust::raw_pointer_cast(&res[0]);
    ValueType* y_d=thrust::raw_pointer_cast(&y[0]);
    ValueType* x_old_d = thrust::raw_pointer_cast(&x_old[0]);

    // another set of temporary variables
    // TODO: this should be reduced in the future for a better RAM usage
    cusp::array1d<ValueType, MemorySpace> x_tmp1(x_old.size());
    cusp::array1d<ValueType, MemorySpace> x_tmp2(x_old.size());
    ValueType *x_tmp1_d = thrust::raw_pointer_cast(&x_tmp1[0]);

    // compute RHS for the PCR kernel solve
    cusp::multiply(A_rest, x_old, x_tmp1);
    blas::axpby(res, x_tmp1, x_tmp1, ValueType(1), ValueType(-1));

    // cleaning output temporary output vector
    thrust::fill(y.begin(),y.end(),0.0);

    // initialization for the tridiagonal solver
    int systemSize=cover.max_path_size;
    int numSystems=cover.actual_path_count;
    
    dim3 grid;
    dim3 threads(systemSize);
    grid.x=(int)ceil(sqrt((double)numSystems));
    grid.y=(numSystems+grid.x-1)/grid.x;

    // direct solver for the tridiagonal systems including the reordering of the RHS which is given in x_tmp1
    detail::pcr_kernel_with_reordering<<< grid, threads,(systemSize+1)*5*sizeof(ValueType)>>>(y_d, lines_d, line_ends_d , cover.actual_path_count, cover.params.path_size,
                    cover.max_path_size, grid.x, x_tmp1_d, a_d, b_d, c_d, r_d, x_d);


    // applies Jacobi relaxation to those rows that are not set do "visited" in row_visited and fill results
    // in the remaining unset values of y
    detail::apply_jacobi<<<(A_mat.num_rows+511)/512,512>>>(y_d, row_visited_d, row_offsets_d,
            column_indices_d, values_d, A_mat.num_rows, alpha, x_old_d, rhs_d);

    // do further sweeps
    for (int i=0;i<sweep_count-1;i++)
    {
        // update current loop vector
        x_old=y;

        // set up rhs
        cusp::multiply(A_rest, x_old, x_tmp1);
        blas::axpby(res, x_tmp1, x_tmp1, ValueType(1), ValueType(-1));

        // initialize output vector
        thrust::fill(y.begin(),y.end(),0.0);

        // perform tridiagonal solver
        detail::pcr_kernel_with_reordering<<< grid, threads,(systemSize+1)*5*sizeof(ValueType)>>>(y_d, lines_d, line_ends_d , cover.actual_path_count, cover.params.path_size,
                        cover.max_path_size, grid.x, x_tmp1_d, a_d, b_d, c_d, r_d, x_d);

        // fill in remaining uncovered values by point relaxation
        detail::apply_jacobi<<<(A_mat.num_rows+511)/512,512>>>(y_d, row_visited_d, row_offsets_d,
                column_indices_d, values_d, A_mat.num_rows, alpha, x_old_d, rhs_d);
    }

    //-----------------------------------------------------------------
    // xx <- y + xx;
    //-----------------------------------------------------------------
    blas::axpy(y, xx, ValueType(1));
}


} // end namespace relaxation
} // end namespace cusp

