/*
 *  Author: Li Luo 
 *  Email:  li.luo@siat.ac.cn 
 */

/*! \file mc_ssor.inl
 *  \brief Inline file for mc_ssor.h
 */

#include <cusp/multiply.h>
#include <cusp/blas.h>
#include <cusp/detail/format_utils.h>

#include <thrust/functional.h>
#include <thrust/transform.h>
#include <thrust/iterator/zip_iterator.h>

#include <cusp/graph/multicoloring.h>
#include <cusp/permutation.h>
#include <cusp/relaxation/mcssor_csr_vector.h>
#include <performance/timer.h>

namespace cusp
{
namespace relaxation
{
namespace detail
{

template <typename ValueType>
struct mc_ssor_presmooth_functor
{
    ValueType omega;

    mc_ssor_presmooth_functor(ValueType omega) : omega(omega) {}

    __host__ __device__
    ValueType operator()(const ValueType& b, const ValueType& d) const
    {
        return omega * b / d;
    }
};

template <typename ValueType>
struct mc_ssor_postsmooth_functor
{
    ValueType omega;

    mc_ssor_postsmooth_functor(ValueType omega) : omega(omega) {}

    template <typename Tuple>
    __host__ __device__
    ValueType operator()(const Tuple& t)
    {
        const ValueType x = thrust::get<0>(t);
        const ValueType d = thrust::get<1>(t);
        const ValueType b = thrust::get<2>(t);
        const ValueType y = thrust::get<3>(t);

        return x + omega * (b - y) / d;
    }
};

template <typename ValueType>
struct mc_ssor_operator_functor
{
    ValueType omega;

    mc_ssor_operator_functor(ValueType omega) : omega(omega) {}

    template <typename Tuple>
    __host__ __device__
    ValueType operator()(const Tuple& t)
    {
        const ValueType d = thrust::get<0>(t);
        const ValueType b = thrust::get<1>(t);
        const ValueType y = thrust::get<2>(t);

        return omega * (b - y) / d;
    }
};

template <typename ValueType>
struct mc_ssor_scale_functor
{
    ValueType omega;

    mc_ssor_scale_functor(ValueType omega) : omega(omega) {}

    template <typename Tuple>
    __host__ __device__
    ValueType operator()(const Tuple& t)
    {
        const ValueType d = thrust::get<0>(t);
        const ValueType x = thrust::get<1>(t);

        return omega * x / d;
    }
};

// get number of entries for each color
template <typename IndexType, typename Array1d >
    void get_block_entries_offsets(const Array1d& row_offsets, const cusp::array1d<IndexType, cusp::host_memory>& ptr, cusp::array1d<IndexType, cusp::host_memory>& entries_offsets, cusp::device_memory)
    {
		cusp::array1d<IndexType, cusp::device_memory> ptr_temp(ptr);
		cusp::array1d<IndexType, cusp::device_memory> entries_offsets_temp(ptr_temp.size());

		const IndexType *row_offsets_ptr = thrust::raw_pointer_cast(&row_offsets[0]);	
		const IndexType *p_ptr = thrust::raw_pointer_cast(&ptr_temp[0]);
		IndexType *entries_offsets_ptr = thrust::raw_pointer_cast(&entries_offsets_temp[0]);
		// get offsets of global entries indice
		cusp::detail::subscript<IndexType, IndexType>((IndexType)ptr_temp.size(), p_ptr, row_offsets_ptr, entries_offsets_ptr, typename Array1d::memory_space( ));
		// copy from device to host
		entries_offsets = entries_offsets_temp;
	}	

template <typename IndexType, typename Array1d >
    void get_block_entries_offsets(const Array1d& row_offsets, const cusp::array1d<IndexType, cusp::host_memory>& ptr, cusp::array1d<IndexType, cusp::host_memory>& entries_offsets, cusp::host_memory)
    {
		entries_offsets.resize(ptr.size());
		const IndexType *row_offsets_ptr = thrust::raw_pointer_cast(&row_offsets[0]);	
		const IndexType *p_ptr = thrust::raw_pointer_cast(&ptr[0]);
		IndexType *entries_offsets_ptr = thrust::raw_pointer_cast(&entries_offsets[0]);	
		// get offsets of global entries indice
		cusp::detail::subscript<IndexType, IndexType>((IndexType)ptr.size(), p_ptr, row_offsets_ptr, entries_offsets_ptr, typename Array1d::memory_space( ));	
	}
} // end namespace detail


// constructor
template <typename IndexType, typename ValueType, typename MemorySpace>
    mc_ssor<IndexType,ValueType,MemorySpace>
    ::mc_ssor() : default_omega(0.0), max_colors(10), sweeps(1), SOR(false), Jacobi(true)
    {
    }
	
// get number of entries for each color
template <typename IndexType, typename Array1d >
    void get_block_entries_offsets(const Array1d& row_offsets, const cusp::array1d<IndexType, cusp::host_memory>& ptr, cusp::array1d<IndexType, cusp::host_memory>& entries_offsets)
    {
        CUSP_PROFILE_SCOPED();
		detail::get_block_entries_offsets<IndexType>(row_offsets, ptr, entries_offsets, typename Array1d::memory_space( ));
	}	
	
template <typename IndexType, typename ValueType, typename MemorySpace>
template<typename MatrixType>
    mc_ssor<IndexType,ValueType,MemorySpace>
    ::mc_ssor(bool same_structure, bool red_black, const MatrixType& A_, const IndexType mx, const IndexType my, const IndexType dof)
    {
        CUSP_PROFILE_SCOPED();

		default_omega = 1.0;
		max_colors = 10;
		sweeps = 1;
		SOR = true;
		Jacobi = true;
		this->same_structure = same_structure;
		this->red_black = red_black;
		this->mx = mx;
		this->my = my;
		this->dof = dof;
		
		if (this->same_structure) {
		cusp::coo_matrix<IndexType, ValueType, MemorySpace> A = A_;
		cusp::array1d<IndexType, MemorySpace> coloring(A.num_rows,0);
		
		// perform coloring
		timer color;			
		if (this->red_black) {
		assert(A.num_rows==mx*my*dof);
		cusp::graph::red_black_multicoloring(coloring, mx, my, dof);		
		} else {
		cusp::graph::greedy_multicoloring(A, coloring);		
		}
		std::cout << "color in " << color.seconds_elapsed() << " s " << std::endl;
		
		// build parameters according to the coloring
		cusp::graph::get_size(coloring, n, num_coloring);
		cusp::graph::build_parameters_from_coloring(coloring, p, pt, bk, ptr, max_block_size, n, num_coloring);
		
		// row_indices --> nnz
		cusp::array1d<IndexType, MemorySpace> nnz(A.num_rows);
		cusp::detail::indices_to_nnz(A.row_indices, nnz);
		
		// nnz --> permuted nnz
		cusp::array1d<IndexType, MemorySpace> p_nnz(A.num_rows);	
		cusp::permute_array(pt, nnz, p_nnz);
		
		// permuted nnz --> permuted row_offsets
		cusp::array1d<IndexType, MemorySpace> p_row_offsets(A.num_rows+1,0);
		cusp::detail::nnz_to_offsets(p_nnz, p_row_offsets);	
		
		// get number of entries for each color		
		get_block_entries_offsets<IndexType>(p_row_offsets, ptr, entries_offsets);
		
		// initialize
		p_b.resize(A.num_rows);
		p_x.resize(A.num_rows);
		y.resize(A.num_rows);
		thrust::fill(y.begin(), y.end(), (ValueType)0);			
		}
		
		initialized = true;
    }

template <typename IndexType, typename ValueType, typename MemorySpace>
template<typename MatrixType>
    void mc_ssor<IndexType,ValueType,MemorySpace>
    ::setup(const MatrixType& A_,
	const ValueType omega, const IndexType max_colors_,
	const IndexType sweeps_, bool SOR_, bool Jacobi_)
    {
        CUSP_PROFILE_SCOPED();

		if (!initialized) {
		  printf("mc_ssor has not yet been created!\n");
		  exit(1);		
		}
		
		cusp::coo_matrix<IndexType, ValueType, MemorySpace> A = A_;
		default_omega = omega;
		max_colors = max_colors_;
		sweeps = sweeps_;
		SOR = SOR_;
		Jacobi = Jacobi_;
		
		if (!same_structure) {
		cusp::array1d<IndexType, MemorySpace> coloring(A.num_rows,0);
		
		// perform coloring
		timer color;			
		if (red_black) {
		assert(A.num_rows==mx*my*dof);
		cusp::graph::red_black_multicoloring(coloring, mx, my, dof);		
		} else {
		cusp::graph::greedy_multicoloring(A, coloring);		
		}
		std::cout << "color in " << color.seconds_elapsed() << " s " << std::endl;
		
		// build parameters according to the coloring
		cusp::graph::get_size(coloring, n, num_coloring);
		cusp::graph::build_parameters_from_coloring(coloring, p, pt, bk, ptr, max_block_size, n, num_coloring);
		
		// row_indices --> nnz
		cusp::array1d<IndexType, MemorySpace> nnz(A.num_rows);
		cusp::detail::indices_to_nnz(A.row_indices, nnz);
		
		// nnz --> permuted nnz
		cusp::array1d<IndexType, MemorySpace> p_nnz(A.num_rows);	
		cusp::permute_array(pt, nnz, p_nnz);
		
		// permuted nnz --> permuted row_offsets
		cusp::array1d<IndexType, MemorySpace> p_row_offsets(A.num_rows+1,0);
		cusp::detail::nnz_to_offsets(p_nnz, p_row_offsets);	
		
		// get number of entries for each color		
		get_block_entries_offsets<IndexType>(p_row_offsets, ptr, entries_offsets);
		
		// initialize
		p_b.resize(A.num_rows);
		p_x.resize(A.num_rows);
		y.resize(A.num_rows);
		thrust::fill(y.begin(), y.end(), (ValueType)0);			
		}		
		
		// permute matrix in coloring sort
		cusp::coo_matrix<IndexType, ValueType, MemorySpace> ptAp_temp;
		/*
		cusp::coo_matrix<IndexType, ValueType, MemorySpace> p_matrix, pt_matrix;		
		cusp::build_permutation_matrix(p, p_matrix);
		cusp::build_permutation_matrix(pt, pt_matrix);	
		cusp::compute_ptAp(A, pt_matrix, p_matrix, ptAp_temp);
		*/
		cusp::compute_ptAp(A, pt, ptAp_temp);
	
		// convert COO to CSR format
		ptAp = ptAp_temp;

        // extract the main A_diagonal
        cusp::detail::extract_diagonal(A, A_diagonal);		
		
        // extract the main ptAp_diagonal
        cusp::detail::extract_diagonal(ptAp, ptAp_diagonal);
    }
	
// linear_operator
template <typename IndexType, typename ValueType, typename MemorySpace>
template <typename Array1, typename Array2>
    void mc_ssor<IndexType,ValueType,MemorySpace>
    ::operator()(const Array1& b, Array2& x)
    {
        mc_ssor<IndexType,ValueType,MemorySpace>::operator()(b,x,default_omega);
    }

// override default omega
template <typename IndexType, typename ValueType, typename MemorySpace>
template <typename Array1, typename Array2>
    void mc_ssor<IndexType,ValueType,MemorySpace>
    ::operator()(const Array1& b, Array2& x, ValueType omega)
    {
        CUSP_PROFILE_SCOPED();
		#if 0  
		/* SPMV version implementation of Heuline, triangle Lo and Up are applied for preconditioning step "Mz=r". 
		   Using one more "if" inside the SPMV kernels, maybe low efficient. Though the x is full vector, the kernel only
		   compute those elements belong to the triangule, so other elements result in zero. */
		
		// permute array forward	
		cusp::permute_array(pt,b,p_b);	
		
		// forward SOR
		for( IndexType k = 0; k < num_coloring; ++k ) {		
		IndexType num = bk[k];
		IndexType off_k1 = ptr[k+1];
		IndexType off_k  = ptr[k];
		IndexType row_offsets_k = entries_offsets[k];
		
		temp.resize(num);		

		IndexType block_entries = entries_offsets[k+1] - row_offsets_k;
		cusp::relaxation::triangular_forward_spmv_csr<IndexType, ValueType>(block_entries, num, row_offsets_k, off_k,
		thrust::raw_pointer_cast(&ptAp.row_offsets[off_k]),
		thrust::raw_pointer_cast(&ptAp.column_indices[row_offsets_k]),
        thrust::raw_pointer_cast(&ptAp.values[row_offsets_k]),
        thrust::raw_pointer_cast(&y[0]),
        thrust::raw_pointer_cast(&temp[0]),
		typename Array1::memory_space( ));

        thrust::transform(thrust::make_zip_iterator(thrust::make_tuple(ptAp_diagonal.begin()+off_k, p_b.begin()+off_k, temp.begin())),
                          thrust::make_zip_iterator(thrust::make_tuple(ptAp_diagonal.begin()+off_k1, p_b.begin()+off_k1, temp.end())),
                          y.begin()+off_k,
                          detail::mc_ssor_operator_functor<ValueType>(omega));
		}

        thrust::transform(thrust::make_zip_iterator(thrust::make_tuple(ptAp_diagonal.begin(), y.begin())),
                          thrust::make_zip_iterator(thrust::make_tuple(ptAp_diagonal.end(), y.end())),
                          y.begin(),
                          detail::mc_ssor_scale_functor<ValueType>(omega));
						  
		// backward SOR
		for( IndexType k = num_coloring-1; k > -1; --k ) {		
		IndexType num = bk[k];
		IndexType off_k1 = ptr[k+1];
		IndexType off_k  = ptr[k];
		IndexType row_offsets_k = entries_offsets[k];
		
		temp.resize(num);		

		IndexType block_entries = entries_offsets[k+1] - row_offsets_k;
		cusp::relaxation::triangular_backward_spmv_csr<IndexType, ValueType>(block_entries, num, row_offsets_k, off_k1-1,
		thrust::raw_pointer_cast(&ptAp.row_offsets[off_k]),
		thrust::raw_pointer_cast(&ptAp.column_indices[row_offsets_k]),
        thrust::raw_pointer_cast(&ptAp.values[row_offsets_k]),
        thrust::raw_pointer_cast(&p_x[0]),
        thrust::raw_pointer_cast(&temp[0]),
		typename Array1::memory_space( ));

        thrust::transform(thrust::make_zip_iterator(thrust::make_tuple(ptAp_diagonal.begin()+off_k, y.begin()+off_k, temp.begin())),
                          thrust::make_zip_iterator(thrust::make_tuple(ptAp_diagonal.begin()+off_k1, y.begin()+off_k1, temp.end())),
                          p_x.begin()+off_k,
                          detail::mc_ssor_operator_functor<ValueType>(omega));
		}
		cusp::blas::scal(p_x, 1.0/(2.0-omega));

		// permute array backward
		cusp::permute_array(p,p_x,x);
		#else 
		/* Still apply the Richardson sweep version for "Mz=r", in which the SPMV is implemented on the whole matrix A and full x. 
		We used to apply this method for smoothing in multigrid method. */
		
		// permute array forward	
		cusp::permute_array(pt,b,p_b);
		
        // x <- omega * D^-1 * b 
        thrust::transform(p_b.begin(), p_b.end(),
                          ptAp_diagonal.begin(),
                          p_x.begin(),
                          detail::mc_ssor_presmooth_functor<ValueType>(default_omega));			
		
        for (IndexType s = 0; s < sweeps; s++) {
		// forward SOR
		for( IndexType k = 0; k < num_coloring; ++k ) {		
		IndexType num = bk[k];
		IndexType off_k1 = ptr[k+1];
		IndexType off_k  = ptr[k];
		IndexType row_offsets_k = entries_offsets[k];
		
		temp.resize(num);		

		IndexType block_entries = entries_offsets[k+1] - row_offsets_k;
		cusp::relaxation::block_spmv_csr<IndexType, ValueType>(block_entries, num, row_offsets_k,
		thrust::raw_pointer_cast(&ptAp.row_offsets[off_k]),
		thrust::raw_pointer_cast(&ptAp.column_indices[row_offsets_k]),
        thrust::raw_pointer_cast(&ptAp.values[row_offsets_k]),
        thrust::raw_pointer_cast(&p_x[0]),
        thrust::raw_pointer_cast(&temp[0]),
		typename Array1::memory_space( ));
		
        thrust::transform(thrust::make_zip_iterator(thrust::make_tuple(p_x.begin()+off_k, ptAp_diagonal.begin()+off_k, p_b.begin()+off_k, temp.begin())),
                          thrust::make_zip_iterator(thrust::make_tuple(p_x.begin()+off_k1, ptAp_diagonal.begin()+off_k1, p_b.begin()+off_k1, temp.end())),
                          p_x.begin()+off_k,
                          detail::mc_ssor_postsmooth_functor<ValueType>(default_omega));				  
		}
		if(!SOR) {
		// backward SOR
		for( IndexType k = num_coloring-1; k > -1; --k ) {		
		IndexType num = bk[k];
		IndexType off_k1 = ptr[k+1];
		IndexType off_k  = ptr[k];
		IndexType row_offsets_k = entries_offsets[k];
		
		temp.resize(num);		

		IndexType block_entries = entries_offsets[k+1] - row_offsets_k;
		cusp::relaxation::block_spmv_csr<IndexType, ValueType>(block_entries, num, row_offsets_k,
		thrust::raw_pointer_cast(&ptAp.row_offsets[off_k]),
		thrust::raw_pointer_cast(&ptAp.column_indices[row_offsets_k]),
        thrust::raw_pointer_cast(&ptAp.values[row_offsets_k]),
        thrust::raw_pointer_cast(&p_x[0]),
        thrust::raw_pointer_cast(&temp[0]),
		typename Array1::memory_space( ));

        thrust::transform(thrust::make_zip_iterator(thrust::make_tuple(p_x.begin()+off_k, ptAp_diagonal.begin()+off_k, p_b.begin()+off_k, temp.begin())),
                          thrust::make_zip_iterator(thrust::make_tuple(p_x.begin()+off_k1, ptAp_diagonal.begin()+off_k1, p_b.begin()+off_k1, temp.end())),
                          p_x.begin()+off_k,
                          detail::mc_ssor_postsmooth_functor<ValueType>(default_omega));
		}
		}
		}
		// permute array backward
		cusp::permute_array(p,p_x,x);
		#endif
    }

template <typename IndexType, typename ValueType, typename MemorySpace>
template<typename MatrixType, typename VectorType1, typename VectorType2>
    void mc_ssor<IndexType,ValueType,MemorySpace>
    ::presmooth(const MatrixType&A, const VectorType1& b, VectorType2& x)
    {
        CUSP_PROFILE_SCOPED();

		if (Jacobi && num_coloring>=max_colors) 
		{ 
        // x <- omega * D^-1 * b 
        thrust::transform(b.begin(), b.end(),
                          A_diagonal.begin(),
                          x.begin(),
                          detail::mc_ssor_presmooth_functor<ValueType>(default_omega));
						  
		temp.resize(x.size());

        for (IndexType s=1; s < sweeps; s++) {

          // y <- A*x
          cusp::multiply(A, x, temp);
        
          // x <- x + omega * D^-1 * (b - y)
          thrust::transform(thrust::make_zip_iterator(thrust::make_tuple(x.begin(), A_diagonal.begin(), b.begin(), temp.begin())),
                            thrust::make_zip_iterator(thrust::make_tuple(x.end(),   A_diagonal.end(),   b.end(),   temp.end())),
                            x.begin(),
                            detail::mc_ssor_postsmooth_functor<ValueType>(default_omega));
        }		
		} else {
		// permute array forward	
		cusp::permute_array(pt,b,p_b);	

        // x <- omega * D^-1 * b 
        thrust::transform(p_b.begin(), p_b.end(),
                          ptAp_diagonal.begin(),
                          p_x.begin(),
                          detail::mc_ssor_presmooth_functor<ValueType>(default_omega));		
		
        for (IndexType s = 1; s < sweeps; s++) {
		// forward SOR
		for( IndexType k = 0; k < num_coloring; ++k ) {		
		IndexType num = bk[k];
		IndexType off_k1 = ptr[k+1];
		IndexType off_k  = ptr[k];
		IndexType row_offsets_k = entries_offsets[k];
		
		temp.resize(num);		

		IndexType block_entries = entries_offsets[k+1] - row_offsets_k;
		cusp::relaxation::block_spmv_csr<IndexType, ValueType>(block_entries, num, row_offsets_k,
		thrust::raw_pointer_cast(&ptAp.row_offsets[off_k]),
		thrust::raw_pointer_cast(&ptAp.column_indices[row_offsets_k]),
        thrust::raw_pointer_cast(&ptAp.values[row_offsets_k]),
        thrust::raw_pointer_cast(&p_x[0]),
        thrust::raw_pointer_cast(&temp[0]),
		typename VectorType1::memory_space( ));

        thrust::transform(thrust::make_zip_iterator(thrust::make_tuple(p_x.begin()+off_k, ptAp_diagonal.begin()+off_k, p_b.begin()+off_k, temp.begin())),
                          thrust::make_zip_iterator(thrust::make_tuple(p_x.begin()+off_k1, ptAp_diagonal.begin()+off_k1, p_b.begin()+off_k1, temp.end())),
                          p_x.begin()+off_k,
                          detail::mc_ssor_postsmooth_functor<ValueType>(default_omega));
		}
		if(!SOR) {
		// backward SOR
		for( IndexType k = num_coloring-1; k > -1; --k ) {		
		IndexType num = bk[k];
		IndexType off_k1 = ptr[k+1];
		IndexType off_k  = ptr[k];
		IndexType row_offsets_k = entries_offsets[k];
		
		temp.resize(num);		

		IndexType block_entries = entries_offsets[k+1] - row_offsets_k;
		cusp::relaxation::block_spmv_csr<IndexType, ValueType>(block_entries, num, row_offsets_k,
		thrust::raw_pointer_cast(&ptAp.row_offsets[off_k]),
		thrust::raw_pointer_cast(&ptAp.column_indices[row_offsets_k]),
        thrust::raw_pointer_cast(&ptAp.values[row_offsets_k]),
        thrust::raw_pointer_cast(&p_x[0]),
        thrust::raw_pointer_cast(&temp[0]),
		typename VectorType1::memory_space( ));
		
        thrust::transform(thrust::make_zip_iterator(thrust::make_tuple(p_x.begin()+off_k, ptAp_diagonal.begin()+off_k, p_b.begin()+off_k, temp.begin())),
                          thrust::make_zip_iterator(thrust::make_tuple(p_x.begin()+off_k1, ptAp_diagonal.begin()+off_k1, p_b.begin()+off_k1, temp.end())),
                          p_x.begin()+off_k,
                          detail::mc_ssor_postsmooth_functor<ValueType>(default_omega));
		}
		}
		}
		// permute array backward
		cusp::permute_array(p,p_x,x);
		}
    }

template <typename IndexType, typename ValueType, typename MemorySpace>
template<typename MatrixType, typename VectorType1, typename VectorType2>
    void mc_ssor<IndexType,ValueType,MemorySpace>
    ::postsmooth(const MatrixType& A, const VectorType1& b, VectorType2& x)
    {
        CUSP_PROFILE_SCOPED();
		
		if (Jacobi && num_coloring>=max_colors) 
		{ 
		temp.resize(x.size());
        for (IndexType s=0; s < sweeps; s++) {

          // y <- A*x
          cusp::multiply(A, x, temp);
        
          // x <- x + omega * D^-1 * (b - y)
          thrust::transform(thrust::make_zip_iterator(thrust::make_tuple(x.begin(), A_diagonal.begin(), b.begin(), temp.begin())),
                            thrust::make_zip_iterator(thrust::make_tuple(x.end(),   A_diagonal.end(),   b.end(),   temp.end())),
                            x.begin(),
                            detail::mc_ssor_postsmooth_functor<ValueType>(default_omega));
        }			
		} else {
		// permute array forward	
		cusp::permute_array(pt,b,p_b);
		cusp::permute_array(pt,x,p_x);		

        for (IndexType s = 0; s < sweeps; s++) {
		// forward SOR
		for( IndexType k = 0; k < num_coloring; ++k ) {		
		IndexType num = bk[k];
		IndexType off_k1 = ptr[k+1];
		IndexType off_k  = ptr[k];
		IndexType row_offsets_k = entries_offsets[k];
		
		temp.resize(num);		

		IndexType block_entries = entries_offsets[k+1] - row_offsets_k;
		cusp::relaxation::block_spmv_csr<IndexType, ValueType>(block_entries, num, row_offsets_k,
		thrust::raw_pointer_cast(&ptAp.row_offsets[off_k]),
		thrust::raw_pointer_cast(&ptAp.column_indices[row_offsets_k]),
        thrust::raw_pointer_cast(&ptAp.values[row_offsets_k]),
        thrust::raw_pointer_cast(&p_x[0]),
        thrust::raw_pointer_cast(&temp[0]),
		typename VectorType1::memory_space( ));
		
        thrust::transform(thrust::make_zip_iterator(thrust::make_tuple(p_x.begin()+off_k, ptAp_diagonal.begin()+off_k, p_b.begin()+off_k, temp.begin())),
                          thrust::make_zip_iterator(thrust::make_tuple(p_x.begin()+off_k1, ptAp_diagonal.begin()+off_k1, p_b.begin()+off_k1, temp.end())),
                          p_x.begin()+off_k,
                          detail::mc_ssor_postsmooth_functor<ValueType>(default_omega));				  
		}
		if(!SOR) {
		// backward SOR
		for( IndexType k = num_coloring-1; k > -1; --k ) {		
		IndexType num = bk[k];
		IndexType off_k1 = ptr[k+1];
		IndexType off_k  = ptr[k];
		IndexType row_offsets_k = entries_offsets[k];
		
		temp.resize(num);		

		IndexType block_entries = entries_offsets[k+1] - row_offsets_k;
		cusp::relaxation::block_spmv_csr<IndexType, ValueType>(block_entries, num, row_offsets_k,
		thrust::raw_pointer_cast(&ptAp.row_offsets[off_k]),
		thrust::raw_pointer_cast(&ptAp.column_indices[row_offsets_k]),
        thrust::raw_pointer_cast(&ptAp.values[row_offsets_k]),
        thrust::raw_pointer_cast(&p_x[0]),
        thrust::raw_pointer_cast(&temp[0]),
		typename VectorType1::memory_space( ));

        thrust::transform(thrust::make_zip_iterator(thrust::make_tuple(p_x.begin()+off_k, ptAp_diagonal.begin()+off_k, p_b.begin()+off_k, temp.begin())),
                          thrust::make_zip_iterator(thrust::make_tuple(p_x.begin()+off_k1, ptAp_diagonal.begin()+off_k1, p_b.begin()+off_k1, temp.end())),
                          p_x.begin()+off_k,
                          detail::mc_ssor_postsmooth_functor<ValueType>(default_omega));
		}
		}
		}
		// permute array backward
		cusp::permute_array(p,p_x,x);
		}
    }

} // end namespace relaxation
} // end namespace cusp

