/*
 *  Author: Li Luo 
 *  Email:  li.luo@siat.ac.cn 
 */

/*! \file mkl_ilu0.inl
 *  \brief Inline file for mkl_ilu0.h
 */
#include <cusp/fact.h>
#include <cusp/detail/utils_.h>
#include <thrust/transform.h>
#include <cusp/multiply.h>

namespace cusp
{
namespace relaxation
{

// constructor
template <typename IndexType, typename ValueType, typename MemorySpace>
    mkl_ilu0<IndexType,ValueType,MemorySpace>
    ::mkl_ilu0()
    {
    }

template <typename IndexType, typename ValueType, typename MemorySpace>
template<typename MatrixType>
    mkl_ilu0<IndexType,ValueType,MemorySpace>
    ::mkl_ilu0(bool same_structure, const MatrixType& A, cusparseHandle_t& handle, bool use_mkl_sv_)
        : handle(handle), use_mkl_sv(use_mkl_sv_)
    {
        CUSP_PROFILE_SCOPED();
		
		/*-------------------MKL routines begin----------------------*/		
		// initialize the parameters for ilu0		
		ipar = (MKL_INT *) malloc (128*sizeof(MKL_INT));
		dpar = (ValueType *) malloc (128*sizeof(ValueType));		
		ipar[1] = 6;
		ipar[5] = 1;
		ipar[6] = 1;	
		ipar[30] = 1;
		dpar[30] = 1.E-20;
		dpar[31] = 1.E-16;

		sweeps = 1;
		this->same_structure = same_structure;		
		if (this->same_structure) {
			ivar = A.num_rows;
			val.resize(A.num_entries);
			A_fact_mkl.resize(A.num_rows, A.num_cols, A.num_entries);		
			if (use_mkl_sv) {
				temp_mkl_r.resize(A.num_rows);
				temp_mkl_y.resize(A.num_rows);
				temp_mkl_x.resize(A.num_rows);
			}
		}
		/*-------------------MKL routines end----------------------*/

		/*-------------------cuSPARSE routines begin----------------------*/			
		if (!use_mkl_sv) {		
		// create a matrix description for the lower
		status = cusparseCreateMatDescr(&descrM_lo); 

		cusparseSetMatType (descrM_lo, CUSPARSE_MATRIX_TYPE_TRIANGULAR);
		cusparseSetMatIndexBase (descrM_lo, CUSPARSE_INDEX_BASE_ZERO);
		// NOTE! CUSPARSE_DIAG_TYPE_UNIT seems not right
		//cusparseSetMatDiagType (descrM_lo, CUSPARSE_DIAG_TYPE_UNIT);
		cusparseSetMatFillMode (descrM_lo, CUSPARSE_FILL_MODE_LOWER);	

		// create the analysis info object for the lower	
		status = cusparseCreateSolveAnalysisInfo(&Analy_lo);
		
		// create a matrix description for the upper
		status = cusparseCreateMatDescr(&descrM_up); 
		cusparseSetMatType (descrM_up, CUSPARSE_MATRIX_TYPE_TRIANGULAR);
		cusparseSetMatIndexBase (descrM_up, CUSPARSE_INDEX_BASE_ZERO);
		cusparseSetMatDiagType (descrM_up, CUSPARSE_DIAG_TYPE_NON_UNIT);
		cusparseSetMatFillMode (descrM_up, CUSPARSE_FILL_MODE_UPPER);	
			
		// create the analysis info object for the upper	
		status = cusparseCreateSolveAnalysisInfo(&Analy_up);
		
		if (this->same_structure) {
		temp_y.resize(A.num_rows);	
		}
		}
		/*-------------------cuSPARSE routines end----------------------*/		
		
		initialized = true;
	}		
	
template <typename IndexType, typename ValueType, typename MemorySpace>
template<typename MatrixType>
    void mkl_ilu0<IndexType,ValueType,MemorySpace>
    ::setup(const MatrixType& A_, const IndexType sweeps_)
    {
        CUSP_PROFILE_SCOPED();
		
		if (!initialized) {
		  printf("mkl_ilu0 has not yet been created!\n");
		  exit(1);		
		}
		
		cusp::csr_matrix<IndexType, ValueType, MemorySpace> A = A_;
		cusp::detail::sort_by_row_and_column(A);  // sort by row and column  		
		sweeps = sweeps_;
		
		if (!same_structure) {
		ivar = A.num_rows;
		val.resize(A.num_entries);
		A_fact_mkl.resize(A.num_rows, A.num_cols, A.num_entries);		
		if (use_mkl_sv) {
			temp_mkl_r.resize(A.num_rows);
			temp_mkl_y.resize(A.num_rows);
			temp_mkl_x.resize(A.num_rows);
		} else {
			temp_y.resize(A.num_rows);			
		}
		}
		
		/*-------------------MKL routines begin----------------------*/			
		// MKL is one-based, transform row_offsets and column_indices to be plusing one
		thrust::copy(thrust::make_transform_iterator(A.row_offsets.begin(), cusp::detail::ADD<IndexType>(1)),
		thrust::make_transform_iterator(A.row_offsets.end(), cusp::detail::ADD<IndexType>(1)),
		A_fact_mkl.row_offsets.begin());
		thrust::copy(thrust::make_transform_iterator(A.column_indices.begin(), cusp::detail::ADD<IndexType>(1)),
		thrust::make_transform_iterator(A.column_indices.end(), cusp::detail::ADD<IndexType>(1)),
		A_fact_mkl.column_indices.begin());
		thrust::copy(A.values.begin(), A.values.end(), val.begin());

		// get the raw host pointer to be used in MKL
		IndexType *h_row  = thrust::raw_pointer_cast(&*A_fact_mkl.row_offsets.begin());
		IndexType *h_col  = thrust::raw_pointer_cast(&*A_fact_mkl.column_indices.begin());	
		ValueType *h_val = thrust::raw_pointer_cast(&*val.begin());			
		ValueType *values = thrust::raw_pointer_cast(&*A_fact_mkl.values.begin());		
		
		// call MKL routine dcsrilu0
		dcsrilu0(&ivar, h_val, h_row, h_col, values, ipar, dpar, &ierr);
		if (ierr!=0)
		{
		  printf("Preconditioner dcsrilu0 has returned the ERROR code %d\n", ierr);
		  exit(1);
		}	
			
		cusp::csr_matrix<IndexType, ValueType, MemorySpace> A_fact;			
		if (!use_mkl_sv) {
		create_fact_from_mkl(A, values, A_fact);
		//create_fact_from_mkl(IndexType(A.num_rows), IndexType(A.num_cols), h_row, h_col, values, A_fact);			
		
		cusp::detail::extract_triangle(A_fact, A_lo, A_up);
		/*-------------------MKL routines end----------------------*/
		
		/*-------------------cuSPARSE routines begin----------------------*/			
		// Perform the analysis for the lower
		status = cusparseDcsrsv_analysis(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, A.num_rows, descrM_lo,
		thrust::raw_pointer_cast(&A_lo.values[0]),
		thrust::raw_pointer_cast(&A_lo.row_offsets[0]),
		thrust::raw_pointer_cast(&A_lo.column_indices[0]),	
		Analy_lo);
		
		// Perform the analysis for the upper
		status = cusparseDcsrsv_analysis(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, A.num_rows, descrM_up,
		thrust::raw_pointer_cast(&A_up.values[0]),
		thrust::raw_pointer_cast(&A_up.row_offsets[0]),
		thrust::raw_pointer_cast(&A_up.column_indices[0]),	
		Analy_up);
		}	
		/*-------------------cuSPARSE routines end----------------------*/		
    }

// destructor	
template <typename IndexType, typename ValueType, typename MemorySpace>
    mkl_ilu0<IndexType,ValueType,MemorySpace>
    ::~mkl_ilu0()
    {
		// free parameters
		delete ipar;
		delete dpar;

		if (!use_mkl_sv) {		
		// destroy cuSPARSE context
		status = cusparseDestroyMatDescr(descrM_lo);
		status = cusparseDestroySolveAnalysisInfo(Analy_lo);  
		status = cusparseDestroyMatDescr(descrM_up);
		status = cusparseDestroySolveAnalysisInfo(Analy_up);
		}	
    }

// linear_operator
template <typename IndexType, typename ValueType, typename MemorySpace>
template <typename Array1, typename Array2>
    void mkl_ilu0<IndexType,ValueType,MemorySpace>
    ::operator()(const Array1& b, Array2& x)
    {
        CUSP_PROFILE_SCOPED();
		
		if (use_mkl_sv) {
		// only CPU memory is needed, copy back from device
		temp_mkl_r = b;
		// get the raw host pointer to be used in MKL
		ValueType *temp_r = thrust::raw_pointer_cast(&*temp_mkl_r.begin());		
		ValueType *temp_y = thrust::raw_pointer_cast(&*temp_mkl_y.begin());		
		ValueType *temp_x = thrust::raw_pointer_cast(&*temp_mkl_x.begin());		
		IndexType *h_row  = thrust::raw_pointer_cast(&*A_fact_mkl.row_offsets.begin());
		IndexType *h_col  = thrust::raw_pointer_cast(&*A_fact_mkl.column_indices.begin());	
		ValueType *values = thrust::raw_pointer_cast(&*A_fact_mkl.values.begin());		
		cvar1='L';
		cvar='N';
		cvar2='U';
		mkl_dcsrtrsv(&cvar1,&cvar,&cvar2,&ivar,values,h_row,h_col,temp_r,temp_y);
		cvar1='U';
		cvar='N';
		cvar2='N';
		mkl_dcsrtrsv(&cvar1,&cvar,&cvar2,&ivar,values,h_row,h_col,temp_y,temp_x);
		// copy the result to device
		x = temp_mkl_x;
		} else {
		ValueType *temp_y_ = thrust::raw_pointer_cast(&*temp_y.begin());		
		const ValueType *b_ = thrust::raw_pointer_cast(&*b.begin());	
		ValueType *x_ = thrust::raw_pointer_cast(&*x.begin());		
        // Forward Solve: L y = r
        status = cusparseDcsrsv_solve(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, A_fact_mkl.num_rows, 1.0, descrM_lo, 
									thrust::raw_pointer_cast(&A_lo.values[0]),
									thrust::raw_pointer_cast(&A_lo.row_offsets[0]),
									thrust::raw_pointer_cast(&A_lo.column_indices[0]),	
									Analy_lo, b_, temp_y_);    
        // Back Substitution: U x = y
        status = cusparseDcsrsv_solve(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, A_fact_mkl.num_rows, 1.0, descrM_up, 
									thrust::raw_pointer_cast(&A_up.values[0]),
									thrust::raw_pointer_cast(&A_up.row_offsets[0]),
									thrust::raw_pointer_cast(&A_up.column_indices[0]),	
									Analy_up, temp_y_, x_);  
        }
    }

template <typename IndexType, typename ValueType, typename MemorySpace>
template<typename MatrixType, typename VectorType1, typename VectorType2>
    void mkl_ilu0<IndexType,ValueType,MemorySpace>
    ::presmooth(const MatrixType&A, const VectorType1& b, VectorType2& x)
    {
        CUSP_PROFILE_SCOPED();
			
		if (use_mkl_sv) {
		// only CPU memory is needed, copy back from device
		temp_mkl_r = b;

		// get the raw host pointer to be used in MKL
		ValueType *temp_r = thrust::raw_pointer_cast(&*temp_mkl_r.begin());		
		ValueType *temp_y = thrust::raw_pointer_cast(&*temp_mkl_y.begin());		
		ValueType *temp_x = thrust::raw_pointer_cast(&*temp_mkl_x.begin());		
		IndexType *h_row  = thrust::raw_pointer_cast(&*A_fact_mkl.row_offsets.begin());
		IndexType *h_col  = thrust::raw_pointer_cast(&*A_fact_mkl.column_indices.begin());	
		ValueType *h_val = thrust::raw_pointer_cast(&*val.begin());					
		ValueType *values = thrust::raw_pointer_cast(&*A_fact_mkl.values.begin());		
		cvar1='L';
		cvar='N';
		cvar2='U';
		mkl_dcsrtrsv(&cvar1,&cvar,&cvar2,&ivar,values,h_row,h_col,temp_r,temp_y);
		cvar1='U';
		cvar='N';
		cvar2='N';
		mkl_dcsrtrsv(&cvar1,&cvar,&cvar2,&ivar,values,h_row,h_col,temp_y,temp_x);

		cusp::array1d<ValueType,cusp::host_memory> res(A.num_rows);
		cusp::array1d<ValueType,cusp::host_memory> e(A.num_rows);
		ValueType *res_ = thrust::raw_pointer_cast(&*res.begin());	
		ValueType *e_ = thrust::raw_pointer_cast(&*e.begin());			
		for(IndexType s = 1; s<this->sweeps; s++)
		{	
			// compute res <- b - A*x
			cvar='N';			
			mkl_dcsrgemv(&cvar,&ivar,h_val,h_row,h_col,temp_x, res_);
			cblas_daxpy(ivar,ValueType(-1), temp_r, MKL_INT(1), res_, MKL_INT(1));
			cblas_dscal(ivar,ValueType(-1),res_,MKL_INT(1));
			
			cvar1='L';
			cvar='N';
			cvar2='U';
			mkl_dcsrtrsv(&cvar1,&cvar,&cvar2,&ivar,values,h_row,h_col,res_,temp_y);
			cvar1='U';
			cvar='N';
			cvar2='N';
			mkl_dcsrtrsv(&cvar1,&cvar,&cvar2,&ivar,values,h_row,h_col,temp_y,e_);	
			
			cblas_daxpy(ivar,ValueType(1), e_, MKL_INT(1), temp_x, MKL_INT(1));	
		}		
		
		// copy the result to device
		x = temp_mkl_x;
		} else {
		ValueType *temp_y_ = thrust::raw_pointer_cast(&*temp_y.begin());		
		const ValueType *b_ = thrust::raw_pointer_cast(&*b.begin());	
		ValueType *x_ = thrust::raw_pointer_cast(&*x.begin());		
        // Forward Solve: L y = r
        status = cusparseDcsrsv_solve(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, A.num_rows, 1.0, descrM_lo, 
									thrust::raw_pointer_cast(&A_lo.values[0]),
									thrust::raw_pointer_cast(&A_lo.row_offsets[0]),
									thrust::raw_pointer_cast(&A_lo.column_indices[0]),	
									Analy_lo, b_, temp_y_);    
        // Back Substitution: U x = y
        status = cusparseDcsrsv_solve(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, A.num_rows, 1.0, descrM_up, 
									thrust::raw_pointer_cast(&A_up.values[0]),
									thrust::raw_pointer_cast(&A_up.row_offsets[0]),
									thrust::raw_pointer_cast(&A_up.column_indices[0]),	
									Analy_up, temp_y_, x_);  
		
		VectorType1 res(A.num_rows);
		VectorType2 e(A.num_rows);
		ValueType *res_ = thrust::raw_pointer_cast(&*res.begin());	
		ValueType *e_ = thrust::raw_pointer_cast(&*e.begin());	
		for(IndexType s = 1; s<this->sweeps; s++)
		{	
			// compute res <- b - A*x
			cusp::multiply(A, x, res);
			cusp::blas::axpby(b, res, res, ValueType(1), ValueType(-1));	
		
			// Forward Solve: L y = r
			status = cusparseDcsrsv_solve(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, A.num_rows, 1.0, descrM_lo, 
										thrust::raw_pointer_cast(&A_lo.values[0]),
										thrust::raw_pointer_cast(&A_lo.row_offsets[0]),
										thrust::raw_pointer_cast(&A_lo.column_indices[0]),	
										Analy_lo, res_, temp_y_);    
			// Back Substitution: U x = y
			status = cusparseDcsrsv_solve(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, A.num_rows, 1.0, descrM_up, 
										thrust::raw_pointer_cast(&A_up.values[0]),
										thrust::raw_pointer_cast(&A_up.row_offsets[0]),
										thrust::raw_pointer_cast(&A_up.column_indices[0]),	
										Analy_up, temp_y_, e_); 
										
			cusp::blas::axpby(x, e, x, ValueType(1), ValueType(1));									
		}		
        }		
    }

template <typename IndexType, typename ValueType, typename MemorySpace>
template<typename MatrixType, typename VectorType1, typename VectorType2>
    void mkl_ilu0<IndexType,ValueType,MemorySpace>
    ::postsmooth(const MatrixType& A, const VectorType1& b, VectorType2& x)
    {
        CUSP_PROFILE_SCOPED();
			
		if (use_mkl_sv) {
		// only CPU memory is needed, copy back from device
		temp_mkl_r = b;
		temp_mkl_x = x;
		
		// get the raw host pointer to be used in MKL
		ValueType *temp_r = thrust::raw_pointer_cast(&*temp_mkl_r.begin());		
		ValueType *temp_y = thrust::raw_pointer_cast(&*temp_mkl_y.begin());		
		ValueType *temp_x = thrust::raw_pointer_cast(&*temp_mkl_x.begin());		
		IndexType *h_row  = thrust::raw_pointer_cast(&*A_fact_mkl.row_offsets.begin());
		IndexType *h_col  = thrust::raw_pointer_cast(&*A_fact_mkl.column_indices.begin());
		ValueType *h_val = thrust::raw_pointer_cast(&*val.begin());					
		ValueType *values = thrust::raw_pointer_cast(&*A_fact_mkl.values.begin());		

		cusp::array1d<ValueType,cusp::host_memory> res(A.num_rows);
		cusp::array1d<ValueType,cusp::host_memory> e(A.num_rows);
		ValueType *res_ = thrust::raw_pointer_cast(&*res.begin());	
		ValueType *e_ = thrust::raw_pointer_cast(&*e.begin());			
		for(IndexType s = 0; s<this->sweeps; s++)
		{	
			// compute res <- b - A*x
			cvar='N';			
			mkl_dcsrgemv(&cvar,&ivar,h_val,h_row,h_col,temp_x, res_);
			cblas_daxpy(ivar,ValueType(-1), temp_r, MKL_INT(1), res_, MKL_INT(1));
			cblas_dscal(ivar,ValueType(-1),res_,MKL_INT(1));
			
			cvar1='L';
			cvar='N';
			cvar2='U';
			mkl_dcsrtrsv(&cvar1,&cvar,&cvar2,&ivar,values,h_row,h_col,res_,temp_y);
			cvar1='U';
			cvar='N';
			cvar2='N';
			mkl_dcsrtrsv(&cvar1,&cvar,&cvar2,&ivar,values,h_row,h_col,temp_y,e_);	
			
			cblas_daxpy(ivar,ValueType(1), e_, MKL_INT(1), temp_x, MKL_INT(1));	
		}		
		
		// copy the result to device
		x = temp_mkl_x;
		} else {
		ValueType *temp_y_ = thrust::raw_pointer_cast(&*temp_y.begin());		
		VectorType1 res(A.num_rows);
		VectorType2 e(A.num_rows);
		ValueType *res_ = thrust::raw_pointer_cast(&*res.begin());	
		ValueType *e_ = thrust::raw_pointer_cast(&*e.begin());	
		for(IndexType s = 0; s<this->sweeps; s++)
		{	
			// compute res <- b - A*x
			cusp::multiply(A, x, res);
			cusp::blas::axpby(b, res, res, ValueType(1), ValueType(-1));	
		
			// Forward Solve: L y = r
			status = cusparseDcsrsv_solve(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, A.num_rows, 1.0, descrM_lo, 
										thrust::raw_pointer_cast(&A_lo.values[0]),
										thrust::raw_pointer_cast(&A_lo.row_offsets[0]),
										thrust::raw_pointer_cast(&A_lo.column_indices[0]),	
										Analy_lo, res_, temp_y_);    
			// Back Substitution: U x = y
			status = cusparseDcsrsv_solve(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, A.num_rows, 1.0, descrM_up, 
										thrust::raw_pointer_cast(&A_up.values[0]),
										thrust::raw_pointer_cast(&A_up.row_offsets[0]),
										thrust::raw_pointer_cast(&A_up.column_indices[0]),	
										Analy_up, temp_y_, e_); 
										
			cusp::blas::axpby(x, e, x, ValueType(1), ValueType(1));									
		}		
        }		
    }

} // end namespace relaxation
} // end namespace cusp

