/*
 *  Author: Li Luo 
 *  Email:  li.luo@siat.ac.cn 
 */

/*! \file petsc_factor.inl
 *  \brief Inline file for petsc_factor.h
 */
#include <cusp/fact.h>
#include <cusp/blas.h>
#include <cusp/detail/utils_.h>
#include <thrust/transform.h>
#include <cusp/multiply.h>
#include <cusp/permutation.h>
#include <performance/timer.h>

namespace cusp
{
namespace relaxation
{

template <typename IndexType, typename ValueType, typename MemorySpace>
template <typename MatrixType, typename VectorType>
PetscErrorCode petsc_factor<IndexType,ValueType,MemorySpace>
    ::setup_petsc_lower(Mat Factor, MatrixType& lower, VectorType& ipr)
{
	PetscErrorCode ierr;  	
	PetscInt i, nz, nzLower, offset, rowOffset, n = Factor->rmap->n;
	PetscInt *AiLo, *AjLo;
	PetscScalar *AALo;
	Mat_SeqAIJ        *a = (Mat_SeqAIJ*)Factor->data;
	const PetscInt    *ai = a->i,*aj = a->j,*vi;
	const MatScalar   *aa = a->a,*v;
	IS isrow = a->row;
	const PetscInt   *r;
	VectorType pr;
  
    PetscFunctionBegin;	
      // first figure out the number of nonzeros in the lower triangular matrix including 1's on the diagonal
      nzLower=n+ai[n]-ai[1];
      
      // Allocate Space for the lower triangular matrix 	  
      ierr = PetscMalloc((n+1)*sizeof(PetscInt),&AiLo);CHKERRQ(ierr);
      ierr = PetscMalloc(nzLower*sizeof(PetscInt),&AjLo);CHKERRQ(ierr);	  
      ierr = PetscMalloc(nzLower*sizeof(PetscScalar),&AALo);CHKERRQ(ierr);	  	
      
      // fill the lower triangular matrix
      AiLo[0]=(PetscInt) 0;
      AiLo[n]=nzLower;
      AjLo[0]=(PetscInt) 0;
      AALo[0]=(MatScalar) 1.0;
      v    = aa;
      vi   = aj;
      offset=1;
      rowOffset=1;
      for (i=1; i<n; i++) {
		nz  = ai[i+1] - ai[i];
		// additional 1 for the term on the diagonal
		AiLo[i]=rowOffset;
		rowOffset+=nz+1;

		ierr = PetscMemcpy(&(AjLo[offset]), vi, nz*sizeof(PetscInt));CHKERRQ(ierr);
		ierr = PetscMemcpy(&(AALo[offset]), v, nz*sizeof(MatScalar));CHKERRQ(ierr);
		
		offset+=nz;
		AjLo[offset]=(PetscInt) i;
		AALo[offset]=(MatScalar) 1.0;
		offset+=1;
		
		v  += nz;
		vi += nz;
      }    

    lower.resize(n,n,nzLower);	  
	if (use_mkl_sv) {  
	// MKL is one-based, transform row_offsets and column_indices to be plusing one
	thrust::copy(thrust::make_transform_iterator(AiLo, cusp::detail::ADD<PetscInt>(1)),
	thrust::make_transform_iterator(AiLo+n+1, cusp::detail::ADD<PetscInt>(1)),
	lower.row_offsets.begin());
	thrust::copy(thrust::make_transform_iterator(AjLo, cusp::detail::ADD<PetscInt>(1)),
	thrust::make_transform_iterator(AjLo+nzLower, cusp::detail::ADD<PetscInt>(1)),
	lower.column_indices.begin());
	thrust::copy(AALo, AALo+nzLower, lower.values.begin());	
	} else {
    // The Lower triangular matrix
    lower.row_offsets.assign(AiLo,AiLo+n+1);
    lower.column_indices.assign(AjLo,AjLo+nzLower);
    lower.values.assign(AALo,AALo+nzLower);
      
    // cusparse analysis 
	cusparseDcsrsv_analysis(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, n, descrM_lo,
	thrust::raw_pointer_cast(&lower.values[0]),
	thrust::raw_pointer_cast(&lower.row_offsets[0]),
	thrust::raw_pointer_cast(&lower.column_indices[0]),	
	Analy_lo);
	}

	// free the temporary memory
    ierr = PetscFree(AiLo);CHKERRQ(ierr); 
    ierr = PetscFree(AjLo);CHKERRQ(ierr); 
    ierr = PetscFree(AALo);CHKERRQ(ierr);   	
	
	// get the inverse permutation from PETSc ordering 
	ierr = ISGetIndices(isrow,&r);CHKERRQ(ierr);
	pr.assign(&r[0], &r[0]+n);
    ierr = ISRestoreIndices(isrow,&r);CHKERRQ(ierr); 
	ipr.resize(pr.size());
	cusp::get_inverse_permutation(pr, ipr);	
	
	// if true, enhance efficiency without permutation
	ierr = ISIdentity(isrow,&row_identity);CHKERRQ(ierr);
	
    PetscFunctionReturn(0);	  	
}

template <typename IndexType, typename ValueType, typename MemorySpace>
template <typename MatrixType, typename VectorType>
PetscErrorCode petsc_factor<IndexType,ValueType,MemorySpace>
    ::setup_petsc_upper(Mat Factor, MatrixType& upper, VectorType& pc, VectorType& ipc)
{
	PetscErrorCode ierr;  	
	PetscInt i, nz, nzUpper, offset, n = Factor->rmap->n;
    PetscInt *AiUp, *AjUp;
    PetscScalar *AAUp;
	Mat_SeqAIJ        *a = (Mat_SeqAIJ*)Factor->data;
    const PetscInt    *aj = a->j,*adiag = a->diag,*vi;
    const MatScalar   *aa = a->a,*v;
	IS iscol = a->icol;
	const PetscInt   *c;

    PetscFunctionBegin;	
      // next, figure out the number of nonzeros in the upper triangular matrix
      nzUpper = adiag[0]-adiag[n];
      
      // Allocate Space for the upper triangular matrix
      ierr = PetscMalloc((n+1)*sizeof(PetscInt),&AiUp);CHKERRQ(ierr);
      ierr = PetscMalloc(nzUpper*sizeof(PetscInt),&AjUp);CHKERRQ(ierr);	  
      ierr = PetscMalloc(nzUpper*sizeof(PetscScalar),&AAUp);CHKERRQ(ierr);	
      
      // Fill the upper triangular matrix
      AiUp[0]=(PetscInt) 0;
      AiUp[n]=nzUpper;
      offset = nzUpper;
      for (i=n-1; i>=0; i--) {
		v   = aa + adiag[i+1] + 1;
		vi  = aj + adiag[i+1] + 1;
		
		// number of elements NOT on the diagonal
		nz = adiag[i] - adiag[i+1]-1;
		
		// decrement the offset
		offset -= (nz+1);
		
		// first, set the diagonal elements
		AjUp[offset] = (PetscInt) i;
		AAUp[offset] = 1./v[nz];
		AiUp[i] = AiUp[i+1] - (nz+1);
		
		ierr = PetscMemcpy(&(AjUp[offset+1]), vi, nz*sizeof(PetscInt));CHKERRQ(ierr);
		ierr = PetscMemcpy(&(AAUp[offset+1]), v, nz*sizeof(MatScalar));CHKERRQ(ierr);
      }  
	  
      upper.resize(n,n,nzUpper);
	  if (use_mkl_sv) {  
	  // MKL is one-based, transform row_offsets and column_indices to be plusing one
	  thrust::copy(thrust::make_transform_iterator(AiUp, cusp::detail::ADD<PetscInt>(1)),
	  thrust::make_transform_iterator(AiUp+n+1, cusp::detail::ADD<PetscInt>(1)),
	  upper.row_offsets.begin());
	  thrust::copy(thrust::make_transform_iterator(AjUp, cusp::detail::ADD<PetscInt>(1)),
	  thrust::make_transform_iterator(AjUp+nzUpper, cusp::detail::ADD<PetscInt>(1)),
	  upper.column_indices.begin());
	  thrust::copy(AAUp, AAUp+nzUpper, upper.values.begin());	
	  } else {	  
      // The Upper triangular matrix
      upper.row_offsets.assign(AiUp,AiUp+n+1);
      upper.column_indices.assign(AjUp,AjUp+nzUpper);
      upper.values.assign(AAUp,AAUp+nzUpper);
    
      // cusparse analysis 
	  cusparseDcsrsv_analysis(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, n, descrM_up,
	  thrust::raw_pointer_cast(&upper.values[0]),
	  thrust::raw_pointer_cast(&upper.row_offsets[0]),
	  thrust::raw_pointer_cast(&upper.column_indices[0]),	
	  Analy_up);
	  }
	  // free the temporary memory	  
      ierr = PetscFree(AiUp);CHKERRQ(ierr); 
      ierr = PetscFree(AjUp);CHKERRQ(ierr); 
      ierr = PetscFree(AAUp);CHKERRQ(ierr);  	
	
	  // get the inverse permutation from PETSc ordering 
	  ierr = ISGetIndices(iscol,&c);CHKERRQ(ierr);
	  pc.assign(&c[0], &c[0]+n);
      ierr = ISRestoreIndices(iscol,&c);CHKERRQ(ierr); 
	  ipc.resize(pc.size());
	  cusp::get_inverse_permutation(pc, ipc);	

	  // if true, enhance efficiency without permutation	  
	  ierr = ISIdentity(iscol,&col_identity);CHKERRQ(ierr);	  
	  
    PetscFunctionReturn(0);	  	
}

	
// constructor
template <typename IndexType, typename ValueType, typename MemorySpace>
    petsc_factor<IndexType,ValueType,MemorySpace>
    ::petsc_factor()
    {
    }

template <typename IndexType, typename ValueType, typename MemorySpace>
template<typename MatrixType>
    petsc_factor<IndexType,ValueType,MemorySpace>
    ::petsc_factor(bool same_structure, const MatrixType& A,
			cusparseHandle_t& handle, bool use_mkl_sv_, petsc_factor_type type_, 
			const IndexType fill_in_, const ValueType fill_, bool allow_diagonal_fill_, 
			petsc_factor_mat_ordering_type mat_ordering_type_,
			const ValueType rtol_)
        : handle(handle), use_mkl_sv(use_mkl_sv_), type(type_), mat_ordering_type(mat_ordering_type_)
    {
        CUSP_PROFILE_SCOPED();

		sweeps  = 1;
		fill_in = fill_in_;
		fill 	= fill_;
		allow_diagonal_fill = allow_diagonal_fill_;
		rtol 	= rtol_;
		
		this->same_structure = same_structure;			
		if (this->same_structure) {
		nrow = A.num_rows;
		/*-------------------MKL routines begin----------------------*/		
		if (use_mkl_sv) {
		temp_mkl_r.resize(A.num_rows);
		temp_mkl_y.resize(A.num_rows);
		temp_mkl_x.resize(A.num_rows);

		ipr_h.resize(A.num_rows);
		pc_h.resize(A.num_rows);	
		ipc_h.resize(A.num_rows);	
		}
		/*-------------------MKL routines end----------------------*/
		}
		
		/*-------------------cuSPARSE routines begin----------------------*/			
		if (!use_mkl_sv) {		
		// create a matrix description for the lower
		cusparseCreateMatDescr(&descrM_lo); 

		cusparseSetMatType (descrM_lo, CUSPARSE_MATRIX_TYPE_TRIANGULAR);
		cusparseSetMatIndexBase (descrM_lo, CUSPARSE_INDEX_BASE_ZERO);
		// NOTE! CUSPARSE_DIAG_TYPE_UNIT seems not right
		//cusparseSetMatDiagType (descrM_lo, CUSPARSE_DIAG_TYPE_UNIT);
		cusparseSetMatFillMode (descrM_lo, CUSPARSE_FILL_MODE_LOWER);	

		// create the analysis info object for the lower	
		cusparseCreateSolveAnalysisInfo(&Analy_lo);
		
		// create a matrix description for the upper
		cusparseCreateMatDescr(&descrM_up); 
		cusparseSetMatType (descrM_up, CUSPARSE_MATRIX_TYPE_TRIANGULAR);
		cusparseSetMatIndexBase (descrM_up, CUSPARSE_INDEX_BASE_ZERO);
		cusparseSetMatDiagType (descrM_up, CUSPARSE_DIAG_TYPE_NON_UNIT);
		cusparseSetMatFillMode (descrM_up, CUSPARSE_FILL_MODE_UPPER);	
			
		// create the analysis info object for the upper	
		cusparseCreateSolveAnalysisInfo(&Analy_up);
		
		if (this->same_structure) {		
		temp_y.resize(A.num_rows);	
		
		ipr_d.resize(A.num_rows);
		pc_d.resize(A.num_rows);		
		ipc_d.resize(A.num_rows);						
		}		
		}
		/*-------------------cuSPARSE routines end----------------------*/		

		initialized = true;
	}		
	
template <typename IndexType, typename ValueType, typename MemorySpace>
template<typename MatrixType>
    PetscErrorCode petsc_factor<IndexType,ValueType,MemorySpace>
    ::setup(const MatrixType& A, const IndexType sweeps_)
    {	
		PetscFunctionBegin;	
		
        CUSP_PROFILE_SCOPED();
		
		if (!initialized) {
		  printf("petsc_factor has not yet been created!\n");
		  exit(1);		
		}
			
		CUSPMATRIX     *A_cusp = new CUSPMATRIX(A);
		cusp::detail::sort_by_row_and_column(*A_cusp);  // sort by row and column  	
		
		sweeps  = sweeps_;
	
		if (!same_structure) {
		nrow = A.num_rows;	
		if (use_mkl_sv) {
			temp_mkl_r.resize(A.num_rows);
			temp_mkl_y.resize(A.num_rows);
			temp_mkl_x.resize(A.num_rows);
			
			ipr_h.resize(A.num_rows);
			pc_h.resize(A.num_rows);
			ipc_h.resize(A.num_rows);						
		} else {
			temp_y.resize(A.num_rows);	
			
			ipr_d.resize(A.num_rows);
			pc_d.resize(A.num_rows);
			ipc_d.resize(A.num_rows);						
		}	
		}

		/*-------------------PETSc routines begin----------------------*/				
		PetscErrorCode ierr;  	
		Mat            A_CPU,Factor; 
		PC			   pc;	
		
		// routine added by Li Luo for PETSc usage
		ierr = MatCUSPCreateFromGPU(&A_CPU, A_cusp);CHKERRQ(ierr);  
  
		// create PETSc PC context 
		ierr = PCCreate(PETSC_COMM_SELF,&pc);CHKERRQ(ierr);  
		if (type == petsc_factor_ilu) {
		ierr = PCSetType(pc,PCILU);CHKERRQ(ierr);  
		ierr = PCFactorSetLevels(pc,(PetscInt)fill_in);CHKERRQ(ierr); 
		if (allow_diagonal_fill) {
		ierr = PCFactorSetAllowDiagonalFill(pc);CHKERRQ(ierr);
		}
		} else if (type == petsc_factor_lu) {
		ierr = PCSetType(pc,PCLU);CHKERRQ(ierr);  		
		}

		ierr = PCFactorSetFill(pc,(PetscReal)fill);CHKERRQ(ierr); 
		  switch (mat_ordering_type) {
			case petsc_factor_mat_natural  : 
			ierr = PCFactorSetMatOrderingType(pc,MATORDERINGNATURAL);CHKERRQ(ierr); 	
			break;
			case petsc_factor_mat_nd  : 
			ierr = PCFactorSetMatOrderingType(pc,MATORDERINGND);CHKERRQ(ierr); 	
			break;
			case petsc_factor_mat_lwd  : 
			ierr = PCFactorSetMatOrderingType(pc,MATORDERING1WD);CHKERRQ(ierr); 	
			break;
			case petsc_factor_mat_rcm  : 
			ierr = PCFactorSetMatOrderingType(pc,MATORDERINGRCM);CHKERRQ(ierr); 	
			break;			
			case petsc_factor_mat_qmd  : 
			ierr = PCFactorSetMatOrderingType(pc,MATORDERINGQMD);CHKERRQ(ierr); 	
			break;				
			default: 
			ierr = PCFactorSetMatOrderingType(pc,MATORDERINGNATURAL);CHKERRQ(ierr); 
			break;
		  }		
		ierr = PCFactorReorderForNonzeroDiagonal(pc,(PetscReal)rtol);CHKERRQ(ierr); 
		
		ierr = PCSetOperators(pc,A_CPU,A_CPU,DIFFERENT_NONZERO_PATTERN);CHKERRQ(ierr); 	
		// factorized by PETSc' own approaches
		ierr = PCSetUp(pc);CHKERRQ(ierr); 
		// get factor from PC
		ierr = PCFactorGetMatrix(pc,&Factor);CHKERRQ(ierr); // need not to be destroyed 
		
		if (use_mkl_sv) {
		  ierr = setup_petsc_lower(Factor, A_lo_mkl, ipr_h);CHKERRQ(ierr);
		  ierr = setup_petsc_upper(Factor, A_up_mkl, pc_h, ipc_h);CHKERRQ(ierr);
		} else {	
		  ierr = setup_petsc_lower(Factor, A_lo, ipr_d);CHKERRQ(ierr);
		  ierr = setup_petsc_upper(Factor, A_up, pc_d, ipc_d);CHKERRQ(ierr);
		}	  
		
	    // destroy
		ierr = PCDestroy(&pc);CHKERRQ(ierr);	
		ierr = MatDestroy(&A_CPU);CHKERRQ(ierr);	
		/*-------------------PETSc routines end----------------------*/	 
		
		if (row_identity && col_identity) {
		cusp::detail::setup_mkl_matrix(A_mkl,*A_cusp);			
		} else {
		// permute matrix by permutation indices from PETSc
		if (use_mkl_sv)	{ 
		cusp::coo_matrix<IndexType, ValueType, cusp::host_memory> ptAqt_temp, A_coo(*A_cusp);	
		cusp::compute_ptAqt(A_coo, ipr_h, pc_h, ptAqt_temp);	
		// setup matrix on host for MKL, redundant sort is unnecessary
		cusp::detail::setup_mkl_matrix(A_mkl,ptAqt_temp);

		pb_h.resize(A.num_rows);
		px_h.resize(A.num_rows);			
		} else {
		cusp::coo_matrix<IndexType, ValueType, MemorySpace> ptAqt_temp, A_coo(*A_cusp);	
		cusp::compute_ptAqt(A_coo, ipr_d, pc_d, ptAqt_temp);
		// convert to CSR format
		A_p = ptAqt_temp;	

		pb_d.resize(A.num_rows);
		px_d.resize(A.num_rows);			
		}	
		}
		delete A_cusp;	
		PetscFunctionReturn(0);	  			
    }

// destructor	
template <typename IndexType, typename ValueType, typename MemorySpace>
    petsc_factor<IndexType,ValueType,MemorySpace>
    ::~petsc_factor()
    {
		if (!use_mkl_sv) {		
		// destroy cuSPARSE context
		cusparseDestroyMatDescr(descrM_lo);
		cusparseDestroySolveAnalysisInfo(Analy_lo);  
		cusparseDestroyMatDescr(descrM_up);
		cusparseDestroySolveAnalysisInfo(Analy_up);
		}	
    }

// apply without permutation
template <typename IndexType, typename ValueType, typename MemorySpace>
template <typename Array1, typename Array2>
    void petsc_factor<IndexType,ValueType,MemorySpace>
    ::apply_natural(const Array1& b, Array2& x)
    {
        CUSP_PROFILE_SCOPED();
		
		if (use_mkl_sv) {
		// only CPU memory is needed, copy back from device
		temp_mkl_r = b;
		// get the raw host pointer to be used in MKL
		ValueType *temp_r = thrust::raw_pointer_cast(&*temp_mkl_r.begin());		
		ValueType *temp_y = thrust::raw_pointer_cast(&*temp_mkl_y.begin());		
		ValueType *temp_x = thrust::raw_pointer_cast(&*temp_mkl_x.begin());		
		IndexType *h_lo_row  = thrust::raw_pointer_cast(&*A_lo_mkl.row_offsets.begin());
		IndexType *h_lo_col  = thrust::raw_pointer_cast(&*A_lo_mkl.column_indices.begin());	
		ValueType *h_lo_val  = thrust::raw_pointer_cast(&*A_lo_mkl.values.begin());	
		IndexType *h_up_row  = thrust::raw_pointer_cast(&*A_up_mkl.row_offsets.begin());
		IndexType *h_up_col  = thrust::raw_pointer_cast(&*A_up_mkl.column_indices.begin());	
		ValueType *h_up_val  = thrust::raw_pointer_cast(&*A_up_mkl.values.begin());			
		cvar1='L';
		cvar='N';
		cvar2='U';
		mkl_dcsrtrsv(&cvar1,&cvar,&cvar2,&nrow,h_lo_val,h_lo_row,h_lo_col,temp_r,temp_y);
		cvar1='U';
		cvar='N';
		cvar2='N';
		mkl_dcsrtrsv(&cvar1,&cvar,&cvar2,&nrow,h_up_val,h_up_row,h_up_col,temp_y,temp_x);
		// copy to device
		x = temp_mkl_x;
		} else {				
		ValueType *temp_y_ = thrust::raw_pointer_cast(&*temp_y.begin());		
		const ValueType *b_ = thrust::raw_pointer_cast(&*b.begin());	
		ValueType *x_ = thrust::raw_pointer_cast(&*x.begin());	
        // Forward Solve: L y = r
        cusparseDcsrsv_solve(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, nrow, 1.0, descrM_lo, 
									thrust::raw_pointer_cast(&A_lo.values[0]),
									thrust::raw_pointer_cast(&A_lo.row_offsets[0]),
									thrust::raw_pointer_cast(&A_lo.column_indices[0]),	
									Analy_lo, b_, temp_y_);    
        // Back Substitution: U x = y
        cusparseDcsrsv_solve(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, nrow, 1.0, descrM_up, 
									thrust::raw_pointer_cast(&A_up.values[0]),
									thrust::raw_pointer_cast(&A_up.row_offsets[0]),
									thrust::raw_pointer_cast(&A_up.column_indices[0]),	
									Analy_up, temp_y_, x_); 
        }
    }
	
// apply with permutation
template <typename IndexType, typename ValueType, typename MemorySpace>
template <typename Array1, typename Array2>
    void petsc_factor<IndexType,ValueType,MemorySpace>
    ::apply_reorder(const Array1& b, Array2& x)
    {
        CUSP_PROFILE_SCOPED();
		
		if (use_mkl_sv) {
		// only CPU memory is needed, copy back from device
		temp_mkl_r = b;
		// permute due to preprocess		
		cusp::permute_array(ipr_h,temp_mkl_r,pb_h);		
	
		// get the raw host pointer to be used in MKL
		ValueType *temp_r = thrust::raw_pointer_cast(&*pb_h.begin());		
		ValueType *temp_y = thrust::raw_pointer_cast(&*temp_mkl_y.begin());		
		ValueType *temp_x = thrust::raw_pointer_cast(&*px_h.begin());		
		IndexType *h_lo_row  = thrust::raw_pointer_cast(&*A_lo_mkl.row_offsets.begin());
		IndexType *h_lo_col  = thrust::raw_pointer_cast(&*A_lo_mkl.column_indices.begin());	
		ValueType *h_lo_val  = thrust::raw_pointer_cast(&*A_lo_mkl.values.begin());	
		IndexType *h_up_row  = thrust::raw_pointer_cast(&*A_up_mkl.row_offsets.begin());
		IndexType *h_up_col  = thrust::raw_pointer_cast(&*A_up_mkl.column_indices.begin());	
		ValueType *h_up_val  = thrust::raw_pointer_cast(&*A_up_mkl.values.begin());			
		cvar1='L';
		cvar='N';
		cvar2='U';
		mkl_dcsrtrsv(&cvar1,&cvar,&cvar2,&nrow,h_lo_val,h_lo_row,h_lo_col,temp_r,temp_y);
		cvar1='U';
		cvar='N';
		cvar2='N';
		mkl_dcsrtrsv(&cvar1,&cvar,&cvar2,&nrow,h_up_val,h_up_row,h_up_col,temp_y,temp_x);
		// permute back
		cusp::permute_array(ipc_h,px_h,temp_mkl_x);
		// copy to device
		x = temp_mkl_x;
		} else {
		// permute due to preprocess		
		cusp::permute_array(ipr_d,b,pb_d);						
		
		ValueType *temp_y_ = thrust::raw_pointer_cast(&*temp_y.begin());		
		const ValueType *b_ = thrust::raw_pointer_cast(&*pb_d.begin());	
		ValueType *x_ = thrust::raw_pointer_cast(&*px_d.begin());	
        // Forward Solve: L y = r
        cusparseDcsrsv_solve(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, nrow, 1.0, descrM_lo, 
									thrust::raw_pointer_cast(&A_lo.values[0]),
									thrust::raw_pointer_cast(&A_lo.row_offsets[0]),
									thrust::raw_pointer_cast(&A_lo.column_indices[0]),	
									Analy_lo, b_, temp_y_);    
        // Back Substitution: U x = y
        cusparseDcsrsv_solve(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, nrow, 1.0, descrM_up, 
									thrust::raw_pointer_cast(&A_up.values[0]),
									thrust::raw_pointer_cast(&A_up.row_offsets[0]),
									thrust::raw_pointer_cast(&A_up.column_indices[0]),	
									Analy_up, temp_y_, x_); 
											
		// permute back
		cusp::permute_array(ipc_d,px_d,x);	
        }
    }

// linear_operator
template <typename IndexType, typename ValueType, typename MemorySpace>
template <typename Array1, typename Array2>
    void petsc_factor<IndexType,ValueType,MemorySpace>
    ::operator()(const Array1& b, Array2& x)
    {
        CUSP_PROFILE_SCOPED();
		
		if (row_identity && col_identity) {
		apply_natural(b, x);		
		} else {
		apply_reorder(b, x);
		}
    }	

template <typename IndexType, typename ValueType, typename MemorySpace>
template<typename MatrixType, typename VectorType1, typename VectorType2>
    void petsc_factor<IndexType,ValueType,MemorySpace>
    ::presmooth_natural(const MatrixType&A, const VectorType1& b, VectorType2& x)
    {
        CUSP_PROFILE_SCOPED();
			
		if (use_mkl_sv) {
		// only CPU memory is needed, copy back from device
		temp_mkl_r = b;
		// get the raw host pointer to be used in MKL
		ValueType *temp_r = thrust::raw_pointer_cast(&*temp_mkl_r.begin());		
		ValueType *temp_y = thrust::raw_pointer_cast(&*temp_mkl_y.begin());		
		ValueType *temp_x = thrust::raw_pointer_cast(&*temp_mkl_x.begin());				
		IndexType *h_lo_row  = thrust::raw_pointer_cast(&*A_lo_mkl.row_offsets.begin());
		IndexType *h_lo_col  = thrust::raw_pointer_cast(&*A_lo_mkl.column_indices.begin());	
		ValueType *h_lo_val  = thrust::raw_pointer_cast(&*A_lo_mkl.values.begin());	
		IndexType *h_up_row  = thrust::raw_pointer_cast(&*A_up_mkl.row_offsets.begin());
		IndexType *h_up_col  = thrust::raw_pointer_cast(&*A_up_mkl.column_indices.begin());	
		ValueType *h_up_val  = thrust::raw_pointer_cast(&*A_up_mkl.values.begin());		
		cvar1='L';
		cvar='N';
		cvar2='U';
		mkl_dcsrtrsv(&cvar1,&cvar,&cvar2,&nrow,h_lo_val,h_lo_row,h_lo_col,temp_r,temp_y);
		cvar1='U';
		cvar='N';
		cvar2='N';
		mkl_dcsrtrsv(&cvar1,&cvar,&cvar2,&nrow,h_up_val,h_up_row,h_up_col,temp_y,temp_x);		
		cusp::array1d<ValueType,cusp::host_memory> res(A.num_rows);
		cusp::array1d<ValueType,cusp::host_memory> e(A.num_rows);
		ValueType *res_ = thrust::raw_pointer_cast(&*res.begin());	
		ValueType *e_ = thrust::raw_pointer_cast(&*e.begin());	
		IndexType *h_row  = thrust::raw_pointer_cast(&*A_mkl.row_offsets.begin());
		IndexType *h_col  = thrust::raw_pointer_cast(&*A_mkl.column_indices.begin());	
		ValueType *h_val = thrust::raw_pointer_cast(&*A_mkl.values.begin());			
		for (IndexType s = 1; s<this->sweeps; s++)
		{	
			// compute res <- b - A*x
			cvar='N';			
			mkl_dcsrgemv(&cvar,&nrow,h_val,h_row,h_col,temp_x, res_);
			cblas_daxpy(nrow,ValueType(-1), temp_r, MKL_INT(1), res_, MKL_INT(1));
			cblas_dscal(nrow,ValueType(-1),res_,MKL_INT(1));		
			cvar1='L';
			cvar='N';
			cvar2='U';
			mkl_dcsrtrsv(&cvar1,&cvar,&cvar2,&nrow,h_lo_val,h_lo_row,h_lo_col,res_,temp_y);
			cvar1='U';
			cvar='N';
			cvar2='N';
			mkl_dcsrtrsv(&cvar1,&cvar,&cvar2,&nrow,h_up_val,h_up_row,h_up_col,temp_y,e_);	
			cblas_daxpy(nrow,ValueType(1), e_, MKL_INT(1), temp_x, MKL_INT(1));	
		}		
		// copy to device
		x = temp_mkl_x;
		} else {
		ValueType *temp_y_ = thrust::raw_pointer_cast(&*temp_y.begin());		
		const ValueType *b_ = thrust::raw_pointer_cast(&*b.begin());	
		ValueType *x_ = thrust::raw_pointer_cast(&*x.begin());			
        // Forward Solve: L y = r
        cusparseDcsrsv_solve(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, nrow, 1.0, descrM_lo, 
									thrust::raw_pointer_cast(&A_lo.values[0]),
									thrust::raw_pointer_cast(&A_lo.row_offsets[0]),
									thrust::raw_pointer_cast(&A_lo.column_indices[0]),	
									Analy_lo, b_, temp_y_);    
        // Back Substitution: U x = y
        cusparseDcsrsv_solve(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, nrow, 1.0, descrM_up, 
									thrust::raw_pointer_cast(&A_up.values[0]),
									thrust::raw_pointer_cast(&A_up.row_offsets[0]),
									thrust::raw_pointer_cast(&A_up.column_indices[0]),	
									Analy_up, temp_y_, x_); 
		VectorType1 res(A.num_rows);
		VectorType2 e(A.num_rows);
		ValueType *res_ = thrust::raw_pointer_cast(&*res.begin());	
		ValueType *e_ = thrust::raw_pointer_cast(&*e.begin());	
		for (IndexType s = 1; s<this->sweeps; s++)
		{	
			// compute res <- b - A*x
			cusp::multiply(A, x, res);
			cusp::blas::axpby(b, res, res, ValueType(1), ValueType(-1));			
			// Forward Solve: L y = r
			cusparseDcsrsv_solve(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, nrow, 1.0, descrM_lo, 
										thrust::raw_pointer_cast(&A_lo.values[0]),
										thrust::raw_pointer_cast(&A_lo.row_offsets[0]),
										thrust::raw_pointer_cast(&A_lo.column_indices[0]),	
										Analy_lo, res_, temp_y_);    
			// Back Substitution: U x = y
			cusparseDcsrsv_solve(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, nrow, 1.0, descrM_up, 
										thrust::raw_pointer_cast(&A_up.values[0]),
										thrust::raw_pointer_cast(&A_up.row_offsets[0]),
										thrust::raw_pointer_cast(&A_up.column_indices[0]),	
										Analy_up, temp_y_, e_); 
			cusp::blas::axpby(x, e, x, ValueType(1), ValueType(1));				
		}
        }
    }

template <typename IndexType, typename ValueType, typename MemorySpace>
template<typename MatrixType, typename VectorType1, typename VectorType2>
    void petsc_factor<IndexType,ValueType,MemorySpace>
    ::presmooth_reorder(const MatrixType&A, const VectorType1& b, VectorType2& x)
    {
        CUSP_PROFILE_SCOPED();
			
		if (use_mkl_sv) {
		// only CPU memory is needed, copy back from device
		temp_mkl_r = b;
		// permute due to preprocess		
		cusp::permute_array(ipr_h,temp_mkl_r,pb_h);

		// get the raw host pointer to be used in MKL
		ValueType *temp_r = thrust::raw_pointer_cast(&*pb_h.begin());		
		ValueType *temp_y = thrust::raw_pointer_cast(&*temp_mkl_y.begin());		
		ValueType *temp_x = thrust::raw_pointer_cast(&*px_h.begin());				
		IndexType *h_lo_row  = thrust::raw_pointer_cast(&*A_lo_mkl.row_offsets.begin());
		IndexType *h_lo_col  = thrust::raw_pointer_cast(&*A_lo_mkl.column_indices.begin());	
		ValueType *h_lo_val  = thrust::raw_pointer_cast(&*A_lo_mkl.values.begin());	
		IndexType *h_up_row  = thrust::raw_pointer_cast(&*A_up_mkl.row_offsets.begin());
		IndexType *h_up_col  = thrust::raw_pointer_cast(&*A_up_mkl.column_indices.begin());	
		ValueType *h_up_val  = thrust::raw_pointer_cast(&*A_up_mkl.values.begin());		
		cvar1='L';
		cvar='N';
		cvar2='U';
		mkl_dcsrtrsv(&cvar1,&cvar,&cvar2,&nrow,h_lo_val,h_lo_row,h_lo_col,temp_r,temp_y);
		cvar1='U';
		cvar='N';
		cvar2='N';
		mkl_dcsrtrsv(&cvar1,&cvar,&cvar2,&nrow,h_up_val,h_up_row,h_up_col,temp_y,temp_x);		
		cusp::array1d<ValueType,cusp::host_memory> res(A.num_rows);
		cusp::array1d<ValueType,cusp::host_memory> e(A.num_rows);
		ValueType *res_ = thrust::raw_pointer_cast(&*res.begin());	
		ValueType *e_ = thrust::raw_pointer_cast(&*e.begin());	
		IndexType *h_row  = thrust::raw_pointer_cast(&*A_mkl.row_offsets.begin());
		IndexType *h_col  = thrust::raw_pointer_cast(&*A_mkl.column_indices.begin());	
		ValueType *h_val = thrust::raw_pointer_cast(&*A_mkl.values.begin());			
		for (IndexType s = 1; s<this->sweeps; s++)
		{	
			// compute res <- b - A*x
			cvar='N';			
			mkl_dcsrgemv(&cvar,&nrow,h_val,h_row,h_col,temp_x, res_);
			cblas_daxpy(nrow,ValueType(-1), temp_r, MKL_INT(1), res_, MKL_INT(1));
			cblas_dscal(nrow,ValueType(-1),res_,MKL_INT(1));		
			cvar1='L';
			cvar='N';
			cvar2='U';
			mkl_dcsrtrsv(&cvar1,&cvar,&cvar2,&nrow,h_lo_val,h_lo_row,h_lo_col,res_,temp_y);
			cvar1='U';
			cvar='N';
			cvar2='N';
			mkl_dcsrtrsv(&cvar1,&cvar,&cvar2,&nrow,h_up_val,h_up_row,h_up_col,temp_y,e_);	
			cblas_daxpy(nrow,ValueType(1), e_, MKL_INT(1), temp_x, MKL_INT(1));	
		}		
		// permute back
		cusp::permute_array(ipc_h,px_h,temp_mkl_x);
		// copy to device
		x = temp_mkl_x;
		} else {
		// permute due to preprocess		
		cusp::permute_array(ipr_d,b,pb_d);	
		
		ValueType *temp_y_ = thrust::raw_pointer_cast(&*temp_y.begin());		
		const ValueType *b_ = thrust::raw_pointer_cast(&*pb_d.begin());	
		ValueType *x_ = thrust::raw_pointer_cast(&*px_d.begin());			
        // Forward Solve: L y = r
        cusparseDcsrsv_solve(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, nrow, 1.0, descrM_lo, 
									thrust::raw_pointer_cast(&A_lo.values[0]),
									thrust::raw_pointer_cast(&A_lo.row_offsets[0]),
									thrust::raw_pointer_cast(&A_lo.column_indices[0]),	
									Analy_lo, b_, temp_y_);    
        // Back Substitution: U x = y
        cusparseDcsrsv_solve(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, nrow, 1.0, descrM_up, 
									thrust::raw_pointer_cast(&A_up.values[0]),
									thrust::raw_pointer_cast(&A_up.row_offsets[0]),
									thrust::raw_pointer_cast(&A_up.column_indices[0]),	
									Analy_up, temp_y_, x_); 
		VectorType1 res(A_p.num_rows);
		VectorType2 e(A_p.num_rows);
		ValueType *res_ = thrust::raw_pointer_cast(&*res.begin());	
		ValueType *e_ = thrust::raw_pointer_cast(&*e.begin());	
		for (IndexType s = 1; s<this->sweeps; s++)
		{	
			// compute res <- b - A*x
			cusp::multiply(A_p, px_d, res);
			cusp::blas::axpby(pb_d, res, res, ValueType(1), ValueType(-1));			
			// Forward Solve: L y = r
			cusparseDcsrsv_solve(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, nrow, 1.0, descrM_lo, 
										thrust::raw_pointer_cast(&A_lo.values[0]),
										thrust::raw_pointer_cast(&A_lo.row_offsets[0]),
										thrust::raw_pointer_cast(&A_lo.column_indices[0]),	
										Analy_lo, res_, temp_y_);    
			// Back Substitution: U x = y
			cusparseDcsrsv_solve(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, nrow, 1.0, descrM_up, 
										thrust::raw_pointer_cast(&A_up.values[0]),
										thrust::raw_pointer_cast(&A_up.row_offsets[0]),
										thrust::raw_pointer_cast(&A_up.column_indices[0]),	
										Analy_up, temp_y_, e_); 
			cusp::blas::axpby(px_d, e, px_d, ValueType(1), ValueType(1));				
		}
		// permute back
		cusp::permute_array(ipc_d,px_d,x);		
        }
    }

template <typename IndexType, typename ValueType, typename MemorySpace>
template<typename MatrixType, typename VectorType1, typename VectorType2>
    void petsc_factor<IndexType,ValueType,MemorySpace>
    ::presmooth(const MatrixType&A, const VectorType1& b, VectorType2& x)
    {
        CUSP_PROFILE_SCOPED();
			
		if (row_identity && col_identity) {
		presmooth_natural(A, b, x);		
		} else {
		presmooth_reorder(A, b, x);
		}
    }
	
template <typename IndexType, typename ValueType, typename MemorySpace>
template<typename MatrixType, typename VectorType1, typename VectorType2>
    void petsc_factor<IndexType,ValueType,MemorySpace>
    ::postsmooth_natural(const MatrixType& A, const VectorType1& b, VectorType2& x)
    {
        CUSP_PROFILE_SCOPED();
			
		if (use_mkl_sv) {
		// only CPU memory is needed, copy back from device
		temp_mkl_r = b;
		temp_mkl_x = x;		
		
		// get the raw host pointer to be used in MKL
		ValueType *temp_r = thrust::raw_pointer_cast(&*temp_mkl_r.begin());		
		ValueType *temp_y = thrust::raw_pointer_cast(&*temp_mkl_y.begin());		
		ValueType *temp_x = thrust::raw_pointer_cast(&*temp_mkl_x.begin());		
		IndexType *h_lo_row  = thrust::raw_pointer_cast(&*A_lo_mkl.row_offsets.begin());
		IndexType *h_lo_col  = thrust::raw_pointer_cast(&*A_lo_mkl.column_indices.begin());	
		ValueType *h_lo_val  = thrust::raw_pointer_cast(&*A_lo_mkl.values.begin());	
		IndexType *h_up_row  = thrust::raw_pointer_cast(&*A_up_mkl.row_offsets.begin());
		IndexType *h_up_col  = thrust::raw_pointer_cast(&*A_up_mkl.column_indices.begin());	
		ValueType *h_up_val  = thrust::raw_pointer_cast(&*A_up_mkl.values.begin());			
		IndexType *h_row  = thrust::raw_pointer_cast(&*A_mkl.row_offsets.begin());
		IndexType *h_col  = thrust::raw_pointer_cast(&*A_mkl.column_indices.begin());	
		ValueType *h_val  = thrust::raw_pointer_cast(&*A_mkl.values.begin());
		cusp::array1d<ValueType,cusp::host_memory> res(A.num_rows);
		cusp::array1d<ValueType,cusp::host_memory> e(A.num_rows);
		ValueType *res_ = thrust::raw_pointer_cast(&*res.begin());	
		ValueType *e_ = thrust::raw_pointer_cast(&*e.begin());			
		for (IndexType s = 0; s<this->sweeps; s++)
		{	
			// compute res <- b - A*x
			cvar='N';			
			mkl_dcsrgemv(&cvar,&nrow,h_val,h_row,h_col,temp_x, res_);
			cblas_daxpy(nrow,ValueType(-1), temp_r, MKL_INT(1), res_, MKL_INT(1));
			cblas_dscal(nrow,ValueType(-1),res_,MKL_INT(1));
			cvar1='L';
			cvar='N';
			cvar2='U';
			mkl_dcsrtrsv(&cvar1,&cvar,&cvar2,&nrow,h_lo_val,h_lo_row,h_lo_col,res_,temp_y);
			cvar1='U';
			cvar='N';
			cvar2='N';
			mkl_dcsrtrsv(&cvar1,&cvar,&cvar2,&nrow,h_up_val,h_up_row,h_up_col,temp_y,e_);	
			cblas_daxpy(nrow,ValueType(1), e_, MKL_INT(1), temp_x, MKL_INT(1));	
		}		
		// copy to device
		x = temp_mkl_x;
		} else {
		ValueType *temp_y_ = thrust::raw_pointer_cast(&*temp_y.begin());		
		VectorType1 res(A.num_rows);
		VectorType2 e(A.num_rows);
		ValueType *res_ = thrust::raw_pointer_cast(&*res.begin());	
		ValueType *e_ = thrust::raw_pointer_cast(&*e.begin());	
		for(IndexType s = 0; s<this->sweeps; s++)
		{	
			// compute res <- b - A*x
			cusp::multiply(A, x, res);
			cusp::blas::axpby(b, res, res, ValueType(1), ValueType(-1));		
			// Forward Solve: L y = r
			cusparseDcsrsv_solve(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, nrow, 1.0, descrM_lo, 
										thrust::raw_pointer_cast(&A_lo.values[0]),
										thrust::raw_pointer_cast(&A_lo.row_offsets[0]),
										thrust::raw_pointer_cast(&A_lo.column_indices[0]),	
										Analy_lo, res_, temp_y_);    
			// Back Substitution: U x = y
			cusparseDcsrsv_solve(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, nrow, 1.0, descrM_up, 
										thrust::raw_pointer_cast(&A_up.values[0]),
										thrust::raw_pointer_cast(&A_up.row_offsets[0]),
										thrust::raw_pointer_cast(&A_up.column_indices[0]),	
										Analy_up, temp_y_, e_); 
			cusp::blas::axpby(x, e, x, ValueType(1), ValueType(1));									
		}		
        }
    }

template <typename IndexType, typename ValueType, typename MemorySpace>
template<typename MatrixType, typename VectorType1, typename VectorType2>
    void petsc_factor<IndexType,ValueType,MemorySpace>
    ::postsmooth_reorder(const MatrixType& A, const VectorType1& b, VectorType2& x)
    {
        CUSP_PROFILE_SCOPED();
			
		if (use_mkl_sv) {
		// only CPU memory is needed, copy back from device
		temp_mkl_r = b;
		// permute due to preprocess		
		cusp::permute_array(ipr_h,temp_mkl_r,pb_h);	
		
		// only CPU memory is needed, copy back from device		
		temp_mkl_x = x;		
		// permute 
		cusp::permute_array(pc_h,temp_mkl_x,px_h);
		
		// get the raw host pointer to be used in MKL
		ValueType *temp_r = thrust::raw_pointer_cast(&*pb_h.begin());		
		ValueType *temp_y = thrust::raw_pointer_cast(&*temp_mkl_y.begin());		
		ValueType *temp_x = thrust::raw_pointer_cast(&*px_h.begin());		
		IndexType *h_lo_row  = thrust::raw_pointer_cast(&*A_lo_mkl.row_offsets.begin());
		IndexType *h_lo_col  = thrust::raw_pointer_cast(&*A_lo_mkl.column_indices.begin());	
		ValueType *h_lo_val  = thrust::raw_pointer_cast(&*A_lo_mkl.values.begin());	
		IndexType *h_up_row  = thrust::raw_pointer_cast(&*A_up_mkl.row_offsets.begin());
		IndexType *h_up_col  = thrust::raw_pointer_cast(&*A_up_mkl.column_indices.begin());	
		ValueType *h_up_val  = thrust::raw_pointer_cast(&*A_up_mkl.values.begin());			
		IndexType *h_row  = thrust::raw_pointer_cast(&*A_mkl.row_offsets.begin());
		IndexType *h_col  = thrust::raw_pointer_cast(&*A_mkl.column_indices.begin());	
		ValueType *h_val  = thrust::raw_pointer_cast(&*A_mkl.values.begin());
		cusp::array1d<ValueType,cusp::host_memory> res(A.num_rows);
		cusp::array1d<ValueType,cusp::host_memory> e(A.num_rows);
		ValueType *res_ = thrust::raw_pointer_cast(&*res.begin());	
		ValueType *e_ = thrust::raw_pointer_cast(&*e.begin());			
		for (IndexType s = 0; s<this->sweeps; s++)
		{	
			// compute res <- b - A*x
			cvar='N';			
			mkl_dcsrgemv(&cvar,&nrow,h_val,h_row,h_col,temp_x, res_);
			cblas_daxpy(nrow,ValueType(-1), temp_r, MKL_INT(1), res_, MKL_INT(1));
			cblas_dscal(nrow,ValueType(-1),res_,MKL_INT(1));
			cvar1='L';
			cvar='N';
			cvar2='U';
			mkl_dcsrtrsv(&cvar1,&cvar,&cvar2,&nrow,h_lo_val,h_lo_row,h_lo_col,res_,temp_y);
			cvar1='U';
			cvar='N';
			cvar2='N';
			mkl_dcsrtrsv(&cvar1,&cvar,&cvar2,&nrow,h_up_val,h_up_row,h_up_col,temp_y,e_);	
			cblas_daxpy(nrow,ValueType(1), e_, MKL_INT(1), temp_x, MKL_INT(1));	
		}		
		// permute back
		cusp::permute_array(ipc_h,px_h,temp_mkl_x);
		// copy to device
		x = temp_mkl_x;
		} else {
		// permute due to preprocess		
		cusp::permute_array(ipr_d,b,pb_d);	
		
		// permute 
		cusp::permute_array(pc_d,x,px_d);

		ValueType *temp_y_ = thrust::raw_pointer_cast(&*temp_y.begin());		
		VectorType1 res(A_p.num_rows);
		VectorType2 e(A_p.num_rows);
		ValueType *res_ = thrust::raw_pointer_cast(&*res.begin());	
		ValueType *e_ = thrust::raw_pointer_cast(&*e.begin());	
		for(IndexType s = 0; s<this->sweeps; s++)
		{	
			// compute res <- b - A*x
			cusp::multiply(A_p, px_d, res);
			cusp::blas::axpby(pb_d, res, res, ValueType(1), ValueType(-1));		
			// Forward Solve: L y = r
			cusparseDcsrsv_solve(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, nrow, 1.0, descrM_lo, 
										thrust::raw_pointer_cast(&A_lo.values[0]),
										thrust::raw_pointer_cast(&A_lo.row_offsets[0]),
										thrust::raw_pointer_cast(&A_lo.column_indices[0]),	
										Analy_lo, res_, temp_y_);    
			// Back Substitution: U x = y
			cusparseDcsrsv_solve(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, nrow, 1.0, descrM_up, 
										thrust::raw_pointer_cast(&A_up.values[0]),
										thrust::raw_pointer_cast(&A_up.row_offsets[0]),
										thrust::raw_pointer_cast(&A_up.column_indices[0]),	
										Analy_up, temp_y_, e_); 
			cusp::blas::axpby(px_d, e, px_d, ValueType(1), ValueType(1));									
		}	
		// permute back
		cusp::permute_array(ipc_d,px_d,x);		
        }
    }

template <typename IndexType, typename ValueType, typename MemorySpace>
template<typename MatrixType, typename VectorType1, typename VectorType2>
    void petsc_factor<IndexType,ValueType,MemorySpace>
    ::postsmooth(const MatrixType& A, const VectorType1& b, VectorType2& x)
    {
        CUSP_PROFILE_SCOPED();
			
		if (row_identity && col_identity) {
		postsmooth_natural(A, b, x);		
		} else {
		postsmooth_reorder(A, b, x);
		}
    }
	
} // end namespace relaxation
} // end namespace cusp

