/*
 *  Author: Li Luo 
 *  Email:  li.luo@siat.ac.cn  
*/

/*! \file polynomial_sweeps.inl
 *  \brief Inline file for polynomial_sweeps.h
 */

#include <cusp/multiply.h>
#include <cusp/detail/format_utils.h>

#include <math.h>

namespace cusp
{
namespace relaxation
{
namespace detail
{
	template <typename ValueType>
	void chebyshev_polynomial_coefficients( const ValueType rho, 
						cusp::array1d<ValueType,cusp::host_memory>& coefficients,
						const ValueType lower_bound = 1.0/30.0, 
						const ValueType upper_bound = 1.1) 
	{
		const size_t degree = 3;

		ValueType x0 = lower_bound * rho;
		ValueType x1 = upper_bound * rho;
		
		// Chebyshev roots for the interval [-1,1]
		cusp::array1d<ValueType,cusp::host_memory> std_roots(degree);

		for( size_t i=0; i<degree; i++ )
			std_roots[i] = std::cos( M_PI * (ValueType(i) + 0.5)/ degree );

		// Chebyshev roots for the interval [x0,x1]
		for( size_t i=0; i<degree; i++ )
			std_roots[i] = 0.5 * (x1-x0) * (1 + std_roots[i]) + x0;

		// Compute monic polynomial coefficients of polynomial with scaled roots
		// TODO: Implement convolution method for polynomial multiplication
		coefficients.resize(degree+1);
		ValueType a = std_roots[0];
		ValueType b = std_roots[1];
		ValueType c = std_roots[2];
		coefficients[0] = 1.0;
		coefficients[1] = -(a+b+c);
		coefficients[2] = (a*b) + (b*c) + (c*a);
		coefficients[3] = -(a*b*c);

		// Scale coefficients to enforce C(0) = 1.0
		ValueType scale_factor = 1.0/coefficients.back();
		cusp::blas::scal(coefficients, scale_factor);
	}
}

// constructor
template <typename ValueType, typename MemorySpace>
    polynomial<ValueType,MemorySpace>
    ::polynomial()
    {
		default_coefficients.resize(3);
		sweeps = 1;		
		initialized = true;			
    }

template <typename ValueType, typename MemorySpace>
template<typename VectorType>
    void polynomial<ValueType,MemorySpace>
    ::setup(const VectorType& coefficients, const int sweeps_)
    {
		if (!initialized) {
		  printf("polynomial has not yet been created!\n");
		  exit(1);		
		}
		for( size_t index = 0; index < 3; index++ ) {
			default_coefficients[index] = -coefficients[index];
		}			
		sweeps = sweeps_;
    }
	
// linear_operator
template <typename ValueType, typename MemorySpace>
template<typename MatrixType, typename VectorType1, typename VectorType2>
    void polynomial<ValueType,MemorySpace>
    ::operator()(const MatrixType& A, const VectorType1& b, VectorType2& x) const
    {
        CUSP_PROFILE_SCOPED();

        polynomial<ValueType,MemorySpace>::operator()(A,b,x,default_coefficients);
    }

template <typename ValueType, typename MemorySpace>
template<typename MatrixType, typename VectorType1, typename VectorType2>
    void polynomial<ValueType,MemorySpace>
    ::presmooth(const MatrixType& A, const VectorType1& b, VectorType2& x)
    {
        CUSP_PROFILE_SCOPED();
        
        x = b;
		ValueType scale_factor = default_coefficients[0];
		cusp::blas::scal(x, scale_factor);

		VectorType1 mult(A.num_rows);

		for( size_t i=1; i<default_coefficients.size(); i++ )
		{
			scale_factor = default_coefficients[i];

			cusp::multiply(A, x, mult);
			cusp::blas::axpby(mult, b, x, ValueType(1), scale_factor);
		}
	
		VectorType1 res(A.num_rows);
		for(size_t s = 1; s<this->sweeps; s++)
		{	
			// compute res <- b - A*x
			cusp::multiply(A, x, res);
			cusp::blas::axpby(b, res, res, ValueType(1), ValueType(-1));

			scale_factor = default_coefficients[0];
			VectorType2 h(res);
			cusp::blas::scal(h, scale_factor);

			for( size_t i=1; i<default_coefficients.size(); i++ )
			{
				cusp::multiply(A, h, mult);
				scale_factor = default_coefficients[i];
				cusp::blas::axpby(mult, res, h, ValueType(1), scale_factor);
			}
			
			cusp::blas::axpby(x, h, x, ValueType(1), ValueType(1));
		}		
    }

template <typename ValueType, typename MemorySpace>
template<typename MatrixType, typename VectorType1, typename VectorType2>
    void polynomial<ValueType,MemorySpace>
    ::postsmooth(const MatrixType& A, const VectorType1& b, VectorType2& x)
    {
        CUSP_PROFILE_SCOPED();

        polynomial<ValueType,MemorySpace>::operator()(A,b,x,default_coefficients);
    }

// override default coefficients
template <typename ValueType, typename MemorySpace>
template<typename MatrixType, typename VectorType1, typename VectorType2, typename VectorType3>
    void polynomial<ValueType,MemorySpace>
    ::operator()(const MatrixType& A, const VectorType1& b, VectorType2& x, VectorType3& coefficients) const
    {
		VectorType1 residual(A.num_rows);
		for(size_t s = 0; s<this->sweeps; s++)
		{
			// compute residual <- b - A*x
			cusp::multiply(A, x, residual);
			cusp::blas::axpby(b, residual, residual, ValueType(1), ValueType(-1));

			ValueType scale_factor = coefficients[0];
			VectorType2 h(residual);
			cusp::blas::scal(h, scale_factor);

			VectorType1 mult(A.num_rows);

			for( size_t i=1; i<coefficients.size(); i++ )
			{
				cusp::multiply(A, h, mult);
				scale_factor = coefficients[i];
				cusp::blas::axpby(mult, residual, h, ValueType(1), scale_factor);
			}

			cusp::blas::axpby(x, h, x, ValueType(1), ValueType(1));
		}
    }

} // end namespace relaxation
} // end namespace cusp

