/*
 *  Author: Li Luo 
 *  Email:  li.luo@siat.ac.cn 
 */

/*! \file ilupp.h
 *  \brief relaxed by ilu++ iluc, ilut and multilevel_ilu*.
 *  triangular system for the latter is solved on CPU.
 */

#pragma once

#include <cusp/detail/config.h>
#include <cusp/linear_operator.h>
#include <cusp/csr_matrix.h>
#include <iluplusplus_interface.h>
#include <cusparse.h>
#include <mkl_blas.h>
#include <mkl_cblas.h>
#include <mkl_spblas.h>

namespace cusp
{
namespace relaxation
{

enum ilupp_type {
    ilupp_iluc,
	ilupp_ilut,
	multilevel_ilu
};

template <typename IndexType, typename ValueType, typename MemorySpace>
class ilupp : public cusp::linear_operator<ValueType, MemorySpace, IndexType>
{
	// ilupp context
	// permutation and scale vector for preprocess
	cusp::array1d<ValueType,cusp::host_memory> dr_h,dc_h;  // scale value
	cusp::array1d<IndexType,cusp::host_memory> ipr_h,pc_h,ipc_h;  // permutation indices	
	cusp::array1d<ValueType,cusp::host_memory> pb_h, px_h;	
	cusp::array1d<ValueType,MemorySpace> dr_d,dc_d;
	cusp::array1d<IndexType,MemorySpace> ipr_d,pc_d,ipc_d;	
	cusp::array1d<ValueType,MemorySpace> scaled_d; // scaled right hand side and solution
	cusp::array1d<ValueType,MemorySpace> pb_d, px_d; // permuted right hand side and solution	
	
	iluplusplus::iluplusplus_precond_parameter ilupp_param_;
	iluplusplus::multilevel_preconditioner* multilevel_ilu_;

	void init_parameter(const IndexType prepro_type, const ValueType threshold, const IndexType fill_in);
				  
	void init_parameter_multilevel(const IndexType prepro_type, const IndexType precond_no,
                  const IndexType max_levels, const ValueType mem_factor,
                  const ValueType threshold, const ValueType min_pivot,
				  bool max_fillin_is_inf, const IndexType fill_in, bool begin_total_piv);
				  
	// MKL context
	IndexType nrow;
	char cvar,cvar1,cvar2;
	// if use_mkl_sv==true, the container for matrix on host, one-based
	// preprocessed matrix for MKL
	cusp::csr_matrix<IndexType,ValueType,cusp::host_memory> A_mkl;
	// the factorized matrix
	cusp::csr_matrix<IndexType,ValueType,cusp::host_memory> A_lo_mkl;
	cusp::csr_matrix<IndexType,ValueType,cusp::host_memory> A_up_mkl;
	cusp::array1d<ValueType,cusp::host_memory> temp_mkl_r;
	cusp::array1d<ValueType,cusp::host_memory> temp_mkl_y;
	cusp::array1d<ValueType,cusp::host_memory> temp_mkl_x;
		
	// cuSAPRSE context	
    cusparseHandle_t handle;
	cusparseMatDescr_t descrM_lo;
	cusparseMatDescr_t descrM_up;
	cusparseSolveAnalysisInfo_t Analy_lo;
	cusparseSolveAnalysisInfo_t Analy_up;	
	// preprocessed matrix
	cusp::csr_matrix<IndexType,ValueType,MemorySpace> A_p;	
	// lower and upper part of factorization matrix, used in the triangular solver of cuSPARSE 
	cusp::csr_matrix<IndexType,ValueType,MemorySpace> A_lo;
	cusp::csr_matrix<IndexType,ValueType,MemorySpace> A_up;
	cusp::array1d<ValueType,MemorySpace> temp_y;
	
public:

	bool initialized;
	bool same_structure;			
	// use MKL routine mkl_dcsrtrsv to solve the triangular systems
	bool use_mkl_sv;
	bool detail_param;

	// ilupp parameters	
	ilupp_type  type;
	IndexType sweeps;
	IndexType fill_in;
	ValueType threshold;
	
    ilupp();
	
    template <typename MatrixType>
    ilupp(bool same_structure, const MatrixType& A,
			cusparseHandle_t& handle, bool use_mkl_sv=false, ilupp_type type=ilupp_iluc, 
			const IndexType prepro_type=-99999, const IndexType precond_no=110,
            const IndexType max_levels=100, const ValueType mem_factor=5.,
            const ValueType threshold=1000., const ValueType min_pivot=1e-2, 
			bool max_fillin_is_inf=false, const IndexType fill_in=0, 
			bool begin_total_piv=false, bool detail_param = false);
									  
    ~ilupp();
	
    template <typename MatrixType>
    void setup(const MatrixType& A, const IndexType sweeps);
    
    // ignores initial x
    template<typename MatrixType, typename VectorType1, typename VectorType2>
    void presmooth(const MatrixType& A, const VectorType1& b, VectorType2& x);
   
    // smooths initial x
    template<typename MatrixType, typename VectorType1, typename VectorType2>
    void postsmooth(const MatrixType& A, const VectorType1& b, VectorType2& x);

    template <typename Array1, typename Array2>
    void operator()(const Array1& b, Array2& x);
};

} // end namespace relaxation
} // end namespace cusp

#include <cusp/relaxation/detail/ilupp.inl>

