/*
 *  Copyright 2008-2009 NVIDIA Corporation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

/*! \file kpz_polynomial.h
 *  \brief Polynomail relaxation based on Kraus et al. 2010.
 */

#pragma once

#include <cusp/detail/config.h>

#include <cusp/linear_operator.h>

namespace cusp
{
namespace relaxation
{


/*! \addtogroup relaxation Relaxation 
 *  \ingroup relaxation 
 *  \{
 */

/*! \p kpz_polynomial: polynomial smoother based on 
 * "Algebraic multilevel iteration methods and the best approximation to $1/x$ in the uniform norm"
 * by Kraus, Pillwein, and Zikatanov, 2010
 *
 * This is similar to a Chebychev smoother, but should be less sensitive to properly bounding
 * the eigenvalues, and therefore require less precomputation
 * This form also doesn't compute the coefficients directly, but uses a 3-term recurrence
 * to compute the polynomial in an iterative way.
 */

template <typename ValueType, typename MemorySpace>
class kpz_polynomial 
{
    int order;
    ValueType l_inf_colsum; // estimate for largest eigenvalue

public:
    ValueType mu; // from the paper

    template <typename MatrixType>
    kpz_polynomial(const MatrixType& A, int poly_order=2);
    
    template <typename MatrixType, typename VectorType1, typename VectorType2>
    void presmooth(const MatrixType& A, const VectorType1& b, VectorType2& x);

    template <typename MatrixType, typename VectorType1, typename VectorType2>
    void postsmooth(const MatrixType& A, const VectorType1& b, VectorType2& x);

    template <typename MatrixType, typename VectorType1, typename VectorType2>
    void operator()(const MatrixType& A, const VectorType1& b, VectorType2& x);
};

/*! \}
 */

} // end namespace relaxation
} // end namespace cusp

#include <cusp/relaxation/detail/kpz_polynomial.inl>

