/*
 *  Author: Li Luo 
 *  Email:  li.luo@siat.ac.cn 
 */

/*! \file mc_ssor.h
 *  \brief MC-SSOR relaxation.
 */

#pragma once

#include <cusp/detail/config.h>
#include <cusp/linear_operator.h>
#include <cusp/csr_matrix.h>
#include <vector> 

namespace cusp
{
namespace relaxation
{

template <typename IndexType, typename ValueType, typename MemorySpace>
class mc_ssor : public cusp::linear_operator<ValueType, MemorySpace, IndexType>
{
	IndexType n;
	IndexType num_coloring;
	IndexType max_block_size;
    ValueType default_omega;
	// permuted right hand side
	cusp::array1d<ValueType, MemorySpace> p_b;
	// permuted solution	
	cusp::array1d<ValueType, MemorySpace> p_x;	
	cusp::array1d<ValueType, MemorySpace> y;	
    cusp::array1d<ValueType, MemorySpace> A_diagonal;
    cusp::array1d<ValueType, MemorySpace> ptAp_diagonal;
    cusp::array1d<ValueType, MemorySpace> temp;
	// p[i] is the original natural number of i in sorted number	
	cusp::array1d<IndexType, MemorySpace> p;
	// pt[i] is the objective number of i in original number	
	cusp::array1d<IndexType, MemorySpace> pt;
	// bk: with size = num_coloring, number of rows in each color block
	cusp::array1d<IndexType, cusp::host_memory> bk;
	// entries_offsets: with size = num_coloring+1, pointer array points to the first entry of  ith color block
	cusp::array1d<IndexType, cusp::host_memory> entries_offsets;	
	// ptr: with size = num_coloring+1, pointer array points to the first row of  ith color block
	cusp::array1d<IndexType, cusp::host_memory> ptr;

	// the permuted matrix
	cusp::csr_matrix<IndexType, ValueType, MemorySpace> ptAp;
	
public:

	bool initialized;
	bool same_structure;	
	bool red_black;
	IndexType max_colors;
    IndexType sweeps;
	IndexType mx,my,dof;
	bool 	  SOR;
	bool	  Jacobi;
	
    mc_ssor();

    template <typename MatrixType>
    mc_ssor(bool same_structure, bool red_black, const MatrixType& A, const IndexType mx, const IndexType my, const IndexType dof);

    template <typename MatrixType>
    void setup(const MatrixType& A,
	const ValueType omega=1.0, const IndexType max_colors=10,
	const IndexType sweeps=1, bool SOR=true, bool Jacobi=false);
    
    // ignores initial x
    template<typename MatrixType, typename VectorType1, typename VectorType2>
    void presmooth(const MatrixType& A, const VectorType1& b, VectorType2& x);
   
    // smooths initial x
    template<typename MatrixType, typename VectorType1, typename VectorType2>
    void postsmooth(const MatrixType& A, const VectorType1& b, VectorType2& x);

    template <typename Array1, typename Array2>
    void operator()(const Array1& b, Array2& x);
        
	template <typename Array1, typename Array2>
    void operator()(const Array1& b, Array2& x, ValueType omega);
	
};

} // end namespace relaxation
} // end namespace cusp

#include <cusp/relaxation/detail/mc_ssor.inl>

