/*
 *  Copyright 2008-2009 NVIDIA Corporation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  Author: Li Luo 
 *  Email:  li.luo@siat.ac.cn 
 */

#pragma once

#include <cusp/detail/device/arch.h>
#include <cusp/detail/device/common.h>
#include <cusp/detail/device/utils.h>
#include <cusp/detail/device/texture.h>
#include <thrust/detail/type_traits.h>

#include <thrust/device_ptr.h>

namespace cusp
{
namespace relaxation
{

//////////////////////////////////////////////////////////////////////////////
// CSR SpMV kernels based on a vector model (one warp per row)
//////////////////////////////////////////////////////////////////////////////
//
// spmv_csr_vector_device
//   Each row of the CSR matrix is assigned to a warp.  The warp computes
//   y[i] = A[i,:] * x, i.e. the dot product of the i-th row of A with 
//   the x vector, in parallel.  This division of work implies that 
//   the CSR index and data arrays (Aj and Ax) are accessed in a contiguous
//   manner (but generally not aligned).  On GT200 these accesses are
//   coalesced, unlike kernels based on the one-row-per-thread division of 
//   work.  Since an entire 32-thread warp is assigned to each row, many 
//   threads will remain idle when their row contains a small number 
//   of elements.  This code relies on implicit synchronization among 
//   threads in a warp.
//
// spmv_csr_vector_tex_device
//   Same as spmv_csr_vector_tex_device, except that the texture cache is 
//   used for accessing the x vector.
//  
//  Note: THREADS_PER_VECTOR must be one of [2,4,8,16,32]


template <typename IndexType, typename ValueType, unsigned int VECTORS_PER_BLOCK, unsigned int THREADS_PER_VECTOR, bool UseCache>
__launch_bounds__(VECTORS_PER_BLOCK * THREADS_PER_VECTOR,1)
__global__ void
block_spmv_csr_vector_kernel(const IndexType num_rows,
					   const IndexType offsets,	
                       const IndexType * Ap, 
                       const IndexType * Aj, 
                       const ValueType * Ax, 
                       const ValueType * x, 
                             ValueType * y)
{
    __shared__ volatile ValueType sdata[VECTORS_PER_BLOCK * THREADS_PER_VECTOR + THREADS_PER_VECTOR / 2];  // padded to avoid reduction conditionals
    __shared__ volatile IndexType ptrs[VECTORS_PER_BLOCK][2];
    
    const IndexType THREADS_PER_BLOCK = VECTORS_PER_BLOCK * THREADS_PER_VECTOR;

    const IndexType thread_id   = THREADS_PER_BLOCK * blockIdx.x + threadIdx.x;    // global thread index
    const IndexType thread_lane = threadIdx.x & (THREADS_PER_VECTOR - 1);          // thread index within the vector
    const IndexType vector_id   = thread_id   /  THREADS_PER_VECTOR;               // global vector index
    const IndexType vector_lane = threadIdx.x /  THREADS_PER_VECTOR;               // vector index within the block
    const IndexType num_vectors = VECTORS_PER_BLOCK * gridDim.x;                   // total number of active vectors

    for(IndexType row = vector_id; row < num_rows; row += num_vectors)
    {
        // use two threads to fetch Ap[row] and Ap[row+1]
        // this is considerably faster than the straightforward version
        if(thread_lane < 2)
            ptrs[vector_lane][thread_lane] = Ap[row + thread_lane];

        const IndexType row_start = ptrs[vector_lane][0] - offsets;                   //same as: row_start = Ap[row];
        const IndexType row_end   = ptrs[vector_lane][1] - offsets;                   //same as: row_end   = Ap[row+1];

        // initialize local sum
        ValueType sum = 0;
     
        if (THREADS_PER_VECTOR == 32 && row_end - row_start > 32)
        {
            // ensure aligned memory access to Aj and Ax

            IndexType jj = row_start - (row_start & (THREADS_PER_VECTOR - 1)) + thread_lane;

            // accumulate local sums
            if(jj >= row_start && jj < row_end)
                sum += Ax[jj] * fetch_x<UseCache>(Aj[jj], x);

            // accumulate local sums
            for(jj += THREADS_PER_VECTOR; jj < row_end; jj += THREADS_PER_VECTOR)
                sum += Ax[jj] * fetch_x<UseCache>(Aj[jj], x);
        }
        else
        {
            // accumulate local sums
            for(IndexType jj = row_start + thread_lane; jj < row_end; jj += THREADS_PER_VECTOR)
                sum += Ax[jj] * fetch_x<UseCache>(Aj[jj], x);
        }

        // store local sum in shared memory
        sdata[threadIdx.x] = sum;
        
        // reduce local sums to row sum
        if (THREADS_PER_VECTOR > 16) sdata[threadIdx.x] = sum = sum + sdata[threadIdx.x + 16];
        if (THREADS_PER_VECTOR >  8) sdata[threadIdx.x] = sum = sum + sdata[threadIdx.x +  8];
        if (THREADS_PER_VECTOR >  4) sdata[threadIdx.x] = sum = sum + sdata[threadIdx.x +  4];
        if (THREADS_PER_VECTOR >  2) sdata[threadIdx.x] = sum = sum + sdata[threadIdx.x +  2];
        if (THREADS_PER_VECTOR >  1) sdata[threadIdx.x] = sum = sum + sdata[threadIdx.x +  1];
       
        // first thread writes the result
        if (thread_lane == 0)
            y[row] = sdata[threadIdx.x];
    }
}

template <bool UseCache, unsigned int THREADS_PER_VECTOR, typename IndexType, typename ValueType>
void __block_spmv_csr_vector(const IndexType num_rows,
					   const IndexType offsets,	
                       const IndexType* Ap, 
                       const IndexType* Aj, 
                       const ValueType* Ax, 
                       const ValueType* x, 
                             ValueType* y)
{
    const size_t THREADS_PER_BLOCK  = 128;
    const size_t VECTORS_PER_BLOCK  = THREADS_PER_BLOCK / THREADS_PER_VECTOR;

    const size_t MAX_BLOCKS = cusp::detail::device::arch::max_active_blocks(block_spmv_csr_vector_kernel<IndexType, ValueType, VECTORS_PER_BLOCK, THREADS_PER_VECTOR, UseCache>, THREADS_PER_BLOCK, (size_t) 0);
    const size_t NUM_BLOCKS = std::min<size_t>(MAX_BLOCKS, cusp::detail::device::DIVIDE_INTO(num_rows, VECTORS_PER_BLOCK));
    
    if (UseCache)
        bind_x(x);

    block_spmv_csr_vector_kernel<IndexType, ValueType, VECTORS_PER_BLOCK, THREADS_PER_VECTOR, UseCache> <<<NUM_BLOCKS, THREADS_PER_BLOCK>>> 
        (num_rows, offsets, Ap, Aj, Ax, x, y);

    if (UseCache)
        unbind_x(x);
}

template <typename IndexType, typename ValueType>
void block_spmv_csr(const IndexType num_entries,
					 const IndexType num_rows,
					 const IndexType offsets,					 
                     const IndexType* Ap, 
                     const IndexType* Aj, 
                     const ValueType* Ax, 
                     const ValueType* x, 
                           ValueType* y,
						   cusp::device_memory)
{
    const IndexType nnz_per_row = num_entries / num_rows;

#ifdef CUSP_USE_TEXTURE_MEMORY
    if (nnz_per_row <=  2) { __block_spmv_csr_vector<true, 2>(num_rows, offsets, Ap, Aj, Ax, x, y); return; }
    if (nnz_per_row <=  4) { __block_spmv_csr_vector<true, 4>(num_rows, offsets, Ap, Aj, Ax, x, y); return; }
    if (nnz_per_row <=  8) { __block_spmv_csr_vector<true, 8>(num_rows, offsets, Ap, Aj, Ax, x, y); return; }
    if (nnz_per_row <= 16) { __block_spmv_csr_vector<true,16>(num_rows, offsets, Ap, Aj, Ax, x, y); return; }
    
    __block_spmv_csr_vector<true,32>(num_rows, offsets, Ap, Aj, Ax, x, y);
#else 	
    if (nnz_per_row <=  2) { __block_spmv_csr_vector<false, 2>(num_rows, offsets, Ap, Aj, Ax, x, y); return; }
    if (nnz_per_row <=  4) { __block_spmv_csr_vector<false, 4>(num_rows, offsets, Ap, Aj, Ax, x, y); return; }
    if (nnz_per_row <=  8) { __block_spmv_csr_vector<false, 8>(num_rows, offsets, Ap, Aj, Ax, x, y); return; }
    if (nnz_per_row <= 16) { __block_spmv_csr_vector<false,16>(num_rows, offsets, Ap, Aj, Ax, x, y); return; }
    
    __block_spmv_csr_vector<false,32>(num_rows, offsets, Ap, Aj, Ax, x, y);
#endif
}

template <typename IndexType, 
          typename ValueType, 
		  typename UnaryFunction,
          typename BinaryFunction1,
          typename BinaryFunction2>
void __block_spmv_csr(const IndexType num_rows,
					 const IndexType offsets,					 
                     const IndexType* Ap, 
                     const IndexType* Aj, 
                     const ValueType* Ax, 
                     const ValueType* x, 
                           ValueType* y,
			  UnaryFunction   initialize,
              BinaryFunction1 combine,
              BinaryFunction2 reduce)
{
 
    for(size_t i = 0; i < num_rows; i++)
    {
        const IndexType& row_start = Ap[i] - offsets;
        const IndexType& row_end   = Ap[i+1] - offsets;
 
        ValueType accumulator = initialize(y[i]);
 
        for (IndexType jj = row_start; jj < row_end; jj++)
        {
            const IndexType& j   = Aj[jj];
            const ValueType& Aij = Ax[jj];
            const ValueType& xj  = x[j];
 
            accumulator = reduce(accumulator, combine(Aij, xj));
        }
 
        y[i] = accumulator;
    }
}

template <typename IndexType, typename ValueType>
void block_spmv_csr(const IndexType num_entries,
					 const IndexType num_rows,
					 const IndexType offsets,					 
                     const IndexType* Ap, 
                     const IndexType* Aj, 
                     const ValueType* Ax, 
                     const ValueType* x, 
                           ValueType* y,
						   cusp::host_memory)
{
    __block_spmv_csr(num_rows, offsets, Ap, Aj, Ax, x, y,
             cusp::detail::zero_function<ValueType>(),
             thrust::multiplies<ValueType>(),
             thrust::plus<ValueType>());
}

template <typename IndexType, typename ValueType, unsigned int VECTORS_PER_BLOCK, unsigned int THREADS_PER_VECTOR, bool UseCache>
__launch_bounds__(VECTORS_PER_BLOCK * THREADS_PER_VECTOR,1)
__global__ void
triangular_forward_spmv_csr_vector_kernel(const IndexType num_rows,
					   const IndexType entries_offsets,
					   const IndexType block_offsets,					   
                       const IndexType * Ap, 
                       const IndexType * Aj, 
                       const ValueType * Ax, 
                       const ValueType * x, 
                             ValueType * y)
{
    __shared__ volatile ValueType sdata[VECTORS_PER_BLOCK * THREADS_PER_VECTOR + THREADS_PER_VECTOR / 2];  // padded to avoid reduction conditionals
    __shared__ volatile IndexType ptrs[VECTORS_PER_BLOCK][2];
    
    const IndexType THREADS_PER_BLOCK = VECTORS_PER_BLOCK * THREADS_PER_VECTOR;

    const IndexType thread_id   = THREADS_PER_BLOCK * blockIdx.x + threadIdx.x;    // global thread index
    const IndexType thread_lane = threadIdx.x & (THREADS_PER_VECTOR - 1);          // thread index within the vector
    const IndexType vector_id   = thread_id   /  THREADS_PER_VECTOR;               // global vector index
    const IndexType vector_lane = threadIdx.x /  THREADS_PER_VECTOR;               // vector index within the block
    const IndexType num_vectors = VECTORS_PER_BLOCK * gridDim.x;                   // total number of active vectors

    for(IndexType row = vector_id; row < num_rows; row += num_vectors)
    {
        // use two threads to fetch Ap[row] and Ap[row+1]
        // this is considerably faster than the straightforward version
        if(thread_lane < 2)
            ptrs[vector_lane][thread_lane] = Ap[row + thread_lane];

        const IndexType row_start = ptrs[vector_lane][0] - entries_offsets;                   //same as: row_start = Ap[row];
        const IndexType row_end   = ptrs[vector_lane][1] - entries_offsets;                   //same as: row_end   = Ap[row+1];

        // initialize local sum
        ValueType sum = 0;
     
        if (THREADS_PER_VECTOR == 32 && row_end - row_start > 32)
        {
            // ensure aligned memory access to Aj and Ax

            IndexType jj = row_start - (row_start & (THREADS_PER_VECTOR - 1)) + thread_lane;

            // accumulate local sums
            if(jj >= row_start && jj < row_end) {
				if(Aj[jj] < block_offsets)
                sum += Ax[jj] * fetch_x<UseCache>(Aj[jj], x);
				}

            // accumulate local sums
            for(jj += THREADS_PER_VECTOR; jj < row_end; jj += THREADS_PER_VECTOR)
				if(Aj[jj] < block_offsets)
                sum += Ax[jj] * fetch_x<UseCache>(Aj[jj], x);
        }
        else
        {
            // accumulate local sums
            for(IndexType jj = row_start + thread_lane; jj < row_end; jj += THREADS_PER_VECTOR)
				if(Aj[jj] < block_offsets)
                sum += Ax[jj] * fetch_x<UseCache>(Aj[jj], x);
        }

        // store local sum in shared memory
        sdata[threadIdx.x] = sum;
        
        // reduce local sums to row sum
        if (THREADS_PER_VECTOR > 16) sdata[threadIdx.x] = sum = sum + sdata[threadIdx.x + 16];
        if (THREADS_PER_VECTOR >  8) sdata[threadIdx.x] = sum = sum + sdata[threadIdx.x +  8];
        if (THREADS_PER_VECTOR >  4) sdata[threadIdx.x] = sum = sum + sdata[threadIdx.x +  4];
        if (THREADS_PER_VECTOR >  2) sdata[threadIdx.x] = sum = sum + sdata[threadIdx.x +  2];
        if (THREADS_PER_VECTOR >  1) sdata[threadIdx.x] = sum = sum + sdata[threadIdx.x +  1];
       
        // first thread writes the result
        if (thread_lane == 0)
            y[row] = sdata[threadIdx.x];
    }
}

template <bool UseCache, unsigned int THREADS_PER_VECTOR, typename IndexType, typename ValueType>
void __triangular_forward_spmv_csr_vector(const IndexType num_rows,
					   const IndexType entries_offsets,	
					   const IndexType block_offsets,
                       const IndexType* Ap, 
                       const IndexType* Aj, 
                       const ValueType* Ax, 
                       const ValueType* x, 
                             ValueType* y)
{
    const size_t THREADS_PER_BLOCK  = 128;
    const size_t VECTORS_PER_BLOCK  = THREADS_PER_BLOCK / THREADS_PER_VECTOR;

    const size_t MAX_BLOCKS = cusp::detail::device::arch::max_active_blocks(triangular_forward_spmv_csr_vector_kernel<IndexType, ValueType, VECTORS_PER_BLOCK, THREADS_PER_VECTOR, UseCache>, THREADS_PER_BLOCK, (size_t) 0);
    const size_t NUM_BLOCKS = std::min<size_t>(MAX_BLOCKS, cusp::detail::device::DIVIDE_INTO(num_rows, VECTORS_PER_BLOCK));
    
    if (UseCache)
        bind_x(x);

    triangular_forward_spmv_csr_vector_kernel<IndexType, ValueType, VECTORS_PER_BLOCK, THREADS_PER_VECTOR, UseCache> <<<NUM_BLOCKS, THREADS_PER_BLOCK>>> 
        (num_rows, entries_offsets, block_offsets, Ap, Aj, Ax, x, y);

    if (UseCache)
        unbind_x(x);
}

template <typename IndexType, typename ValueType>
void triangular_forward_spmv_csr(const IndexType num_entries,
					 const IndexType num_rows,
					 const IndexType entries_offsets,	
					 const IndexType block_offsets,
                     const IndexType* Ap, 
                     const IndexType* Aj, 
                     const ValueType* Ax, 
                     const ValueType* x, 
                           ValueType* y,
						   cusp::device_memory)
{
    const IndexType nnz_per_row = (num_entries / num_rows) / 2;

#ifdef CUSP_USE_TEXTURE_MEMORY
    if (nnz_per_row <=  2) { __triangular_forward_spmv_csr_vector<true, 2>(num_rows, entries_offsets, block_offsets, Ap, Aj, Ax, x, y); return; }
    if (nnz_per_row <=  4) { __triangular_forward_spmv_csr_vector<true, 4>(num_rows, entries_offsets, block_offsets, Ap, Aj, Ax, x, y); return; }
    if (nnz_per_row <=  8) { __triangular_forward_spmv_csr_vector<true, 8>(num_rows, entries_offsets, block_offsets, Ap, Aj, Ax, x, y); return; }
    if (nnz_per_row <= 16) { __triangular_forward_spmv_csr_vector<true,16>(num_rows, entries_offsets, block_offsets, Ap, Aj, Ax, x, y); return; }
    
    __triangular_forward_spmv_csr_vector<true,32>(num_rows, entries_offsets, block_offsets, Ap, Aj, Ax, x, y);
#else	
    if (nnz_per_row <=  2) { __triangular_forward_spmv_csr_vector<false, 2>(num_rows, entries_offsets, block_offsets, Ap, Aj, Ax, x, y); return; }
    if (nnz_per_row <=  4) { __triangular_forward_spmv_csr_vector<false, 4>(num_rows, entries_offsets, block_offsets, Ap, Aj, Ax, x, y); return; }
    if (nnz_per_row <=  8) { __triangular_forward_spmv_csr_vector<false, 8>(num_rows, entries_offsets, block_offsets, Ap, Aj, Ax, x, y); return; }
    if (nnz_per_row <= 16) { __triangular_forward_spmv_csr_vector<false,16>(num_rows, entries_offsets, block_offsets, Ap, Aj, Ax, x, y); return; }
    
    __triangular_forward_spmv_csr_vector<false,32>(num_rows, entries_offsets, block_offsets, Ap, Aj, Ax, x, y);
#endif
}

template <typename IndexType, 
          typename ValueType, 
		  typename UnaryFunction,
          typename BinaryFunction1,
          typename BinaryFunction2>
void __triangular_forward_spmv_csr(const IndexType num_rows,
					 const IndexType entries_offsets,	
					 const IndexType block_offsets,			 
                     const IndexType* Ap, 
                     const IndexType* Aj, 
                     const ValueType* Ax, 
                     const ValueType* x, 
                           ValueType* y,
			  UnaryFunction   initialize,
              BinaryFunction1 combine,
              BinaryFunction2 reduce)
{
 
    for(size_t i = 0; i < num_rows; i++)
    {
        const IndexType& row_start = Ap[i] - entries_offsets;
        const IndexType& row_end   = Ap[i+1] - entries_offsets;
 
        ValueType accumulator = initialize(y[i]);
 
        for (IndexType jj = row_start; jj < row_end; jj++)
        {
            const IndexType& j   = Aj[jj];
            const ValueType& Aij = Ax[jj];
            const ValueType& xj  = x[j];
			if(j < block_offsets) 
            accumulator = reduce(accumulator, combine(Aij, xj));
        }
 
        y[i] = accumulator;
    }
}

template <typename IndexType, typename ValueType>
void triangular_forward_spmv_csr(const IndexType num_entries,
					 const IndexType num_rows,
					 const IndexType entries_offsets,	
					 const IndexType block_offsets,
                     const IndexType* Ap, 
                     const IndexType* Aj, 
                     const ValueType* Ax, 
                     const ValueType* x, 
                           ValueType* y,
						   cusp::host_memory)
{
    __triangular_forward_spmv_csr(num_rows, entries_offsets, block_offsets, Ap, Aj, Ax, x, y,
             cusp::detail::zero_function<ValueType>(),
             thrust::multiplies<ValueType>(),
             thrust::plus<ValueType>());
}

template <typename IndexType, typename ValueType, unsigned int VECTORS_PER_BLOCK, unsigned int THREADS_PER_VECTOR, bool UseCache>
__launch_bounds__(VECTORS_PER_BLOCK * THREADS_PER_VECTOR,1)
__global__ void
triangular_backward_spmv_csr_vector_kernel(const IndexType num_rows,
					   const IndexType entries_offsets,
					   const IndexType block_offsets,					   
                       const IndexType * Ap, 
                       const IndexType * Aj, 
                       const ValueType * Ax, 
                       const ValueType * x, 
                             ValueType * y)
{
    __shared__ volatile ValueType sdata[VECTORS_PER_BLOCK * THREADS_PER_VECTOR + THREADS_PER_VECTOR / 2];  // padded to avoid reduction conditionals
    __shared__ volatile IndexType ptrs[VECTORS_PER_BLOCK][2];
    
    const IndexType THREADS_PER_BLOCK = VECTORS_PER_BLOCK * THREADS_PER_VECTOR;

    const IndexType thread_id   = THREADS_PER_BLOCK * blockIdx.x + threadIdx.x;    // global thread index
    const IndexType thread_lane = threadIdx.x & (THREADS_PER_VECTOR - 1);          // thread index within the vector
    const IndexType vector_id   = thread_id   /  THREADS_PER_VECTOR;               // global vector index
    const IndexType vector_lane = threadIdx.x /  THREADS_PER_VECTOR;               // vector index within the block
    const IndexType num_vectors = VECTORS_PER_BLOCK * gridDim.x;                   // total number of active vectors

    for(IndexType row = vector_id; row < num_rows; row += num_vectors)
    {
        // use two threads to fetch Ap[row] and Ap[row+1]
        // this is considerably faster than the straightforward version
        if(thread_lane < 2)
            ptrs[vector_lane][thread_lane] = Ap[row + thread_lane];

        const IndexType row_start = ptrs[vector_lane][0] - entries_offsets;                   //same as: row_start = Ap[row];
        const IndexType row_end   = ptrs[vector_lane][1] - entries_offsets;                   //same as: row_end   = Ap[row+1];

        // initialize local sum
        ValueType sum = 0;
     
        if (THREADS_PER_VECTOR == 32 && row_end - row_start > 32)
        {
            // ensure aligned memory access to Aj and Ax

            IndexType jj = row_start - (row_start & (THREADS_PER_VECTOR - 1)) + thread_lane;

            // accumulate local sums
            if(jj >= row_start && jj < row_end) {
				if(Aj[jj] > block_offsets)
                sum += Ax[jj] * fetch_x<UseCache>(Aj[jj], x);
				}

            // accumulate local sums
            for(jj += THREADS_PER_VECTOR; jj < row_end; jj += THREADS_PER_VECTOR)
				if(Aj[jj] > block_offsets)
                sum += Ax[jj] * fetch_x<UseCache>(Aj[jj], x);
        }
        else
        {
            // accumulate local sums
            for(IndexType jj = row_start + thread_lane; jj < row_end; jj += THREADS_PER_VECTOR)
				if(Aj[jj] > block_offsets)
                sum += Ax[jj] * fetch_x<UseCache>(Aj[jj], x);
        }

        // store local sum in shared memory
        sdata[threadIdx.x] = sum;
        
        // reduce local sums to row sum
        if (THREADS_PER_VECTOR > 16) sdata[threadIdx.x] = sum = sum + sdata[threadIdx.x + 16];
        if (THREADS_PER_VECTOR >  8) sdata[threadIdx.x] = sum = sum + sdata[threadIdx.x +  8];
        if (THREADS_PER_VECTOR >  4) sdata[threadIdx.x] = sum = sum + sdata[threadIdx.x +  4];
        if (THREADS_PER_VECTOR >  2) sdata[threadIdx.x] = sum = sum + sdata[threadIdx.x +  2];
        if (THREADS_PER_VECTOR >  1) sdata[threadIdx.x] = sum = sum + sdata[threadIdx.x +  1];
       
        // first thread writes the result
        if (thread_lane == 0)
            y[row] = sdata[threadIdx.x];
    }
}

template <bool UseCache, unsigned int THREADS_PER_VECTOR, typename IndexType, typename ValueType>
void __triangular_backward_spmv_csr_vector(const IndexType num_rows,
					   const IndexType entries_offsets,	
					   const IndexType block_offsets,
                       const IndexType* Ap, 
                       const IndexType* Aj, 
                       const ValueType* Ax, 
                       const ValueType* x, 
                             ValueType* y)
{
    const size_t THREADS_PER_BLOCK  = 128;
    const size_t VECTORS_PER_BLOCK  = THREADS_PER_BLOCK / THREADS_PER_VECTOR;

    const size_t MAX_BLOCKS = cusp::detail::device::arch::max_active_blocks(triangular_backward_spmv_csr_vector_kernel<IndexType, ValueType, VECTORS_PER_BLOCK, THREADS_PER_VECTOR, UseCache>, THREADS_PER_BLOCK, (size_t) 0);
    const size_t NUM_BLOCKS = std::min<size_t>(MAX_BLOCKS, cusp::detail::device::DIVIDE_INTO(num_rows, VECTORS_PER_BLOCK));
    
    if (UseCache)
        bind_x(x);

    triangular_backward_spmv_csr_vector_kernel<IndexType, ValueType, VECTORS_PER_BLOCK, THREADS_PER_VECTOR, UseCache> <<<NUM_BLOCKS, THREADS_PER_BLOCK>>> 
        (num_rows, entries_offsets, block_offsets, Ap, Aj, Ax, x, y);

    if (UseCache)
        unbind_x(x);
}

template <typename IndexType, typename ValueType>
void triangular_backward_spmv_csr(const IndexType num_entries,
					 const IndexType num_rows,
					 const IndexType entries_offsets,	
					 const IndexType block_offsets,
                     const IndexType* Ap, 
                     const IndexType* Aj, 
                     const ValueType* Ax, 
                     const ValueType* x, 
                           ValueType* y,
						   cusp::device_memory)
{
    const IndexType nnz_per_row = (num_entries / num_rows) / 2;
#ifdef CUSP_USE_TEXTURE_MEMORY
    if (nnz_per_row <=  2) { __triangular_backward_spmv_csr_vector<true, 2>(num_rows, entries_offsets, block_offsets, Ap, Aj, Ax, x, y); return; }
    if (nnz_per_row <=  4) { __triangular_backward_spmv_csr_vector<true, 4>(num_rows, entries_offsets, block_offsets, Ap, Aj, Ax, x, y); return; }
    if (nnz_per_row <=  8) { __triangular_backward_spmv_csr_vector<true, 8>(num_rows, entries_offsets, block_offsets, Ap, Aj, Ax, x, y); return; }
    if (nnz_per_row <= 16) { __triangular_backward_spmv_csr_vector<true,16>(num_rows, entries_offsets, block_offsets, Ap, Aj, Ax, x, y); return; }
    
    __triangular_backward_spmv_csr_vector<true,32>(num_rows, entries_offsets, block_offsets, Ap, Aj, Ax, x, y);
#else
    if (nnz_per_row <=  2) { __triangular_backward_spmv_csr_vector<false, 2>(num_rows, entries_offsets, block_offsets, Ap, Aj, Ax, x, y); return; }
    if (nnz_per_row <=  4) { __triangular_backward_spmv_csr_vector<false, 4>(num_rows, entries_offsets, block_offsets, Ap, Aj, Ax, x, y); return; }
    if (nnz_per_row <=  8) { __triangular_backward_spmv_csr_vector<false, 8>(num_rows, entries_offsets, block_offsets, Ap, Aj, Ax, x, y); return; }
    if (nnz_per_row <= 16) { __triangular_backward_spmv_csr_vector<false,16>(num_rows, entries_offsets, block_offsets, Ap, Aj, Ax, x, y); return; }
    
    __triangular_backward_spmv_csr_vector<false,32>(num_rows, entries_offsets, block_offsets, Ap, Aj, Ax, x, y);
#endif	
}

template <typename IndexType, 
          typename ValueType, 
		  typename UnaryFunction,
          typename BinaryFunction1,
          typename BinaryFunction2>
void __triangular_backward_spmv_csr(const IndexType num_rows,
					 const IndexType entries_offsets,	
					 const IndexType block_offsets,				 
                     const IndexType* Ap, 
                     const IndexType* Aj, 
                     const ValueType* Ax, 
                     const ValueType* x, 
                           ValueType* y,
			  UnaryFunction   initialize,
              BinaryFunction1 combine,
              BinaryFunction2 reduce)
{
 
    for(size_t i = 0; i < num_rows; i++)
    {
        const IndexType& row_start = Ap[i] - entries_offsets;
        const IndexType& row_end   = Ap[i+1] - entries_offsets;
 
        ValueType accumulator = initialize(y[i]);
 
        for (IndexType jj = row_start; jj < row_end; jj++)
        {
            const IndexType& j   = Aj[jj];
            const ValueType& Aij = Ax[jj];
            const ValueType& xj  = x[j];
			if(j > block_offsets) 
            accumulator = reduce(accumulator, combine(Aij, xj));
        }
 
        y[i] = accumulator;
    }
}

template <typename IndexType, typename ValueType>
void triangular_backward_spmv_csr(const IndexType num_entries,
					 const IndexType num_rows,
					 const IndexType entries_offsets,	
					 const IndexType block_offsets,
                     const IndexType* Ap, 
                     const IndexType* Aj, 
                     const ValueType* Ax, 
                     const ValueType* x, 
                           ValueType* y,
						   cusp::host_memory)
{
    __triangular_backward_spmv_csr(num_rows, entries_offsets, block_offsets, Ap, Aj, Ax, x, y,
             cusp::detail::zero_function<ValueType>(),
             thrust::multiplies<ValueType>(),
             thrust::plus<ValueType>());
}

} // end namespace relaxation
} // end namespace cusp

