/*
 *  Author: Li Luo 
 *  Email:  li.luo@siat.ac.cn 
 */

/*! \file petsc_factor.h
 *  \brief relaxed by PETSc lu, ilu.
 *  PetscInitialize((int *)0,(char ***)0,(char *)0,(char *)0);	
 *  and   PetscFinalize();    
 *	must be called at the beginning and ending for main function 
 *
 */

#pragma once

#include <cusp/detail/config.h>
#include <cusp/linear_operator.h>
#include <cusp/csr_matrix.h>
#include <petscpc.h>
#include <../src/mat/impls/aij/seq/seqcusp/cuspmatimpl.h>
#include <../src/mat/impls/aij/seq/aij.h>
#include <cusparse.h>
#include <mkl_blas.h>
#include <mkl_cblas.h>
#include <mkl_spblas.h>

namespace cusp
{
namespace relaxation
{

enum petsc_factor_type {
    petsc_factor_lu,
	petsc_factor_ilu
};

enum petsc_factor_mat_ordering_type {
    petsc_factor_mat_natural,
	petsc_factor_mat_nd,
	petsc_factor_mat_lwd,
	petsc_factor_mat_rcm,
	petsc_factor_mat_qmd
};

template <typename IndexType, typename ValueType, typename MemorySpace>
class petsc_factor : public cusp::linear_operator<ValueType, MemorySpace, IndexType>
{
	// petsc_factor context
	// indicators
    PetscBool row_identity;
    PetscBool col_identity;	
	// permutation indices for preprocess	
	cusp::array1d<IndexType,cusp::host_memory> ipr_h,pc_h,ipc_h;	
	cusp::array1d<ValueType,cusp::host_memory> pb_h, px_h;		
	cusp::array1d<IndexType,MemorySpace> ipr_d,pc_d,ipc_d;	
	cusp::array1d<ValueType,MemorySpace> pb_d, px_d; // permuted right hand side and solution		
	
	template <typename MatrixType, typename VectorType>
	PetscErrorCode setup_petsc_lower(Mat, MatrixType&, VectorType&);
	
	template <typename MatrixType, typename VectorType>
	PetscErrorCode setup_petsc_upper(Mat, MatrixType&, VectorType&, VectorType&); 
				  
	// MKL context
	IndexType nrow;
	char cvar,cvar1,cvar2;
	// if use_mkl_sv==true, the container for matrix on host, one-based
	// preprocessed matrix for MKL
	cusp::csr_matrix<IndexType,ValueType,cusp::host_memory> A_mkl;
	// the factorized matrix
	cusp::csr_matrix<IndexType,ValueType,cusp::host_memory> A_lo_mkl;
	cusp::csr_matrix<IndexType,ValueType,cusp::host_memory> A_up_mkl;
	cusp::array1d<ValueType,cusp::host_memory> temp_mkl_r;
	cusp::array1d<ValueType,cusp::host_memory> temp_mkl_y;
	cusp::array1d<ValueType,cusp::host_memory> temp_mkl_x;
		
	// cuSAPRSE context	
    cusparseHandle_t handle;
	cusparseMatDescr_t descrM_lo;
	cusparseMatDescr_t descrM_up;
	cusparseSolveAnalysisInfo_t Analy_lo;
	cusparseSolveAnalysisInfo_t Analy_up;	
	// preprocessed matrix
	cusp::csr_matrix<IndexType,ValueType,MemorySpace> A_p;	
	// lower and upper part of factorization matrix, used in the triangular solver of cuSPARSE 
	cusp::csr_matrix<IndexType,ValueType,MemorySpace> A_lo;
	cusp::csr_matrix<IndexType,ValueType,MemorySpace> A_up;
	cusp::array1d<ValueType,MemorySpace> temp_y;
	
public:

	bool initialized;
	bool same_structure;			
	// use MKL routine mkl_dcsrtrsv to solve the triangular systems
	bool use_mkl_sv;
	bool allow_diagonal_fill;

	// petsc_factor parameters	
	petsc_factor_type type;
	petsc_factor_mat_ordering_type mat_ordering_type;	
	IndexType sweeps;
	IndexType fill_in;
	ValueType fill;
	ValueType rtol;
	
    petsc_factor();
	
    template <typename MatrixType>
    petsc_factor(bool same_structure, const MatrixType& A,
			cusparseHandle_t& handle, bool use_mkl_sv=false, petsc_factor_type type=petsc_factor_ilu, 
			const IndexType fill_in=0, const ValueType fill=5.0, bool allow_diagonal_fill=false, 
			petsc_factor_mat_ordering_type mat_ordering_type=petsc_factor_mat_natural,
			const ValueType rtol=1.e-10);
									  
    ~petsc_factor();
	
    template <typename MatrixType>
    PetscErrorCode setup(const MatrixType& A, const IndexType sweeps);
    
	///////////////////////
    // ignores initial x
    template<typename MatrixType, typename VectorType1, typename VectorType2>
    void presmooth(const MatrixType& A, const VectorType1& b, VectorType2& x);
	
    template<typename MatrixType, typename VectorType1, typename VectorType2>
    void presmooth_natural(const MatrixType& A, const VectorType1& b, VectorType2& x);	
	
    template<typename MatrixType, typename VectorType1, typename VectorType2>
    void presmooth_reorder(const MatrixType& A, const VectorType1& b, VectorType2& x);		
   
	///////////////////////
    // smooths initial x
    template<typename MatrixType, typename VectorType1, typename VectorType2>
    void postsmooth(const MatrixType& A, const VectorType1& b, VectorType2& x);
	
    template<typename MatrixType, typename VectorType1, typename VectorType2>
    void postsmooth_natural(const MatrixType& A, const VectorType1& b, VectorType2& x);

    template<typename MatrixType, typename VectorType1, typename VectorType2>
    void postsmooth_reorder(const MatrixType& A, const VectorType1& b, VectorType2& x);	

	///////////////////////	
    template <typename Array1, typename Array2>
    void operator()(const Array1& b, Array2& x);
	
    template <typename Array1, typename Array2>
    void apply_natural(const Array1& b, Array2& x);

    template <typename Array1, typename Array2>
    void apply_reorder(const Array1& b, Array2& x);		
};

} // end namespace relaxation
} // end namespace cusp

#include <cusp/relaxation/detail/petsc_factor.inl>

