/*
 *  Author: Li Luo 
 *  Email:  li.luo@siat.ac.cn 
 */

/*! \file sparskit2_ilu.h
 *  \brief relaxed by sparskit2 ilu*.
 */

#pragma once

#include <cusp/detail/config.h>
#include <cusp/linear_operator.h>
#include <cusp/csr_matrix.h>
#include <cusp/relaxation/cusparse_overload.h>
#include <cusp/relaxation/mkl_overload.h>

namespace cusp
{
namespace relaxation
{

enum sparskit_ilu_type {
    sparskit_ilut,
	sparskit_ilutp,
	sparskit_ilud,
	sparskit_iludp,
	sparskit_iluk,	
	sparskit_ilu0,
	sparskit_milu0
};

template <typename IndexType, typename ValueType, typename MemorySpace>
class sparskit2_ilu : public cusp::linear_operator<ValueType, MemorySpace, IndexType>
{
	// sparskit2 context
	IndexType ierr, iwk, nvv, niw;
	
	// MKL context
	IndexType nrow;
	char cvar,cvar1,cvar2;
	// if use_mkl_sv==true, the container for matrix on host, one-based
	// the original matrix
	cusp::csr_matrix<IndexType,ValueType,cusp::host_memory> A_mkl;
	// the factorized matrix
	cusp::csr_matrix<IndexType,ValueType,cusp::host_memory> A_fact_mkl;
	cusp::array1d<ValueType,cusp::host_memory> temp_mkl_r;
	cusp::array1d<ValueType,cusp::host_memory> temp_mkl_y;
	cusp::array1d<ValueType,cusp::host_memory> temp_mkl_x;
		
	// cuSAPRSE context	
    cusparseHandle_t handle;
	cusparseMatDescr_t descrM_lo;
	cusparseMatDescr_t descrM_up;
	cusparseSolveAnalysisInfo_t Analy_lo;
	cusparseSolveAnalysisInfo_t Analy_up;	
	// lower and upper part of factorization matrix, used in the triangular solver of cuSPARSE 
	cusp::csr_matrix<IndexType,ValueType,MemorySpace> A_lo;
	cusp::csr_matrix<IndexType,ValueType,MemorySpace> A_up;
	cusp::array1d<ValueType,MemorySpace> temp_y;
	
public:

	bool initialized;
	bool same_structure;			
	// use MKL routine mkl_dcsrtrsv to solve the triangular systems
	bool use_mkl_sv;

	// sparskit2 parameters	
	sparskit_ilu_type  type;
	IndexType sweeps, mbloc;
	IndexType maxfil, lfil;  // maxfil: half of the preconditioner bandwidth; lfil: level of fill-in in iluk
	ValueType droptol, permtol, alph;
	
    sparskit2_ilu();
	
    template <typename MatrixType>
    sparskit2_ilu(bool same_structure, const MatrixType& A, cusparseHandle_t& handle, bool use_mkl_sv=false, sparskit_ilu_type type=sparskit_ilu0);

    ~sparskit2_ilu();
	
    template <typename MatrixType>
    void setup(const MatrixType& A, const IndexType maxfil=10, const IndexType lfil=0,
	const ValueType droptol=0.0001, const IndexType mbloc=0, const ValueType permtol=0.1, const ValueType alph=1.0, const IndexType sweeps=1);
    
    // ignores initial x
    template<typename MatrixType, typename VectorType1, typename VectorType2>
    void presmooth(const MatrixType& A, const VectorType1& b, VectorType2& x);
   
    // smooths initial x
    template<typename MatrixType, typename VectorType1, typename VectorType2>
    void postsmooth(const MatrixType& A, const VectorType1& b, VectorType2& x);

    template <typename Array1, typename Array2>
    void operator()(const Array1& b, Array2& x);
};

} // end namespace relaxation
} // end namespace cusp

#include <cusp/relaxation/detail/sparskit2_ilu.inl>

