/*
 *  Author: Li Luo 
 *  Email:  li.luo@siat.ac.cn 
 */

/*! \file sparskit2_ilu.inl
 *  \brief Inline file for sparskit2_ilu.h
 */
#include <cusp/fact.h>
#include <cusp/detail/utils_.h>
#include <thrust/transform.h>
#include <cusp/permutation.h>
#include <cusp/multiply.h>
#include <performance/timer.h>

namespace cusp
{
namespace relaxation
{

#ifndef PETSC_USE_REAL_DOUBLE
#define PETSC_USE_REAL_DOUBLE
#endif

#if defined(PETSC_USE_REAL_SINGLE) 
extern "C" void ilu0_(int *n, float *a, int *ja, int *ia, float *au, int *jau, int *ju, int *iw, int *fierr);
extern "C" void milu0_(int *n, float *a, int *ja, int *ia, float *au, int *jau, int *ju, int *iw, int *fierr);
extern "C" void ilut_(int *n, float *a, int *ja, int *ia, int *lfil, float *tol, float *au, int *jau, int *ju, int *nwk, float *vv, int *iw, int *fierr);
extern "C" void ilutp_(int *n, float *a, int *ja, int *ia, int *lfil, float *tol, float *permtol, int *mbloc, float *au, int *jau, int *ju, int *num_entries, float *vv, int *iw, int *iperm, int *fierr);
extern "C" void ilud_(int *n, float *a, int *ja, int *ia, float *alph, float *tol, float *au, int *jau, int *ju, int *num_entries, float *vv, int *iw, int *fierr);	
extern "C" void iludp_(int *n, float *a, int *ja, int *ia, float *alph, float *tol, float *permtol, int *mbloc, float *au, int *jau, int *ju, int *num_entries, float *vv, int *iw, int *iperm, int *fierr);
extern "C" void iluk_(int *n, float *a, int *ja, int *ia, int *lfil, float *au, int *jau, int *ju, int *levs, int *num_entries, float *vv, int *iw, int *fierr); 
#elif defined(PETSC_USE_REAL_DOUBLE)
extern "C" void ilu0_(int *n, double *a, int *ja, int *ia, double *au, int *jau, int *ju, int *iw, int *fierr);
extern "C" void milu0_(int *n, double *a, int *ja, int *ia, double *au, int *jau, int *ju, int *iw, int *fierr);
extern "C" void ilut_(int *n, double *a, int *ja, int *ia, int *lfil, double *tol, double *au, int *jau, int *ju, int *nwk, double *vv, int *iw, int *fierr);
extern "C" void ilutp_(int *n, double *a, int *ja, int *ia, int *lfil, double *tol, double *permtol, int *mbloc, double *au, int *jau, int *ju, int *num_entries, double *vv, int *iw, int *iperm, int *fierr);
extern "C" void ilud_(int *n, double *a, int *ja, int *ia, double *alph, double *tol, double *au, int *jau, int *ju, int *num_entries, double *vv, int *iw, int *fierr);	
extern "C" void iludp_(int *n, double *a, int *ja, int *ia, double *alph, double *tol, double *permtol, int *mbloc, double *au, int *jau, int *ju, int *num_entries, double *vv, int *iw, int *iperm, int *fierr);
extern "C" void iluk_(int *n, double *a, int *ja, int *ia, int *lfil, double *au, int *jau, int *ju, int *levs, int *num_entries, double *vv, int *iw, int *fierr);
#endif


// constructor
template <typename IndexType, typename ValueType, typename MemorySpace>
    sparskit2_ilu<IndexType,ValueType,MemorySpace>
    ::sparskit2_ilu()
    {
    }

template <typename IndexType, typename ValueType, typename MemorySpace>
template<typename MatrixType>
    sparskit2_ilu<IndexType,ValueType,MemorySpace>
    ::sparskit2_ilu(bool same_structure, const MatrixType& A, cusparseHandle_t& handle, bool use_mkl_sv_, sparskit_ilu_type type_)
        : handle(handle), use_mkl_sv(use_mkl_sv_), type(type_)
    {
        CUSP_PROFILE_SCOPED();

		sweeps  = 1;
		maxfil  = 10;
		lfil    = 1;		
		droptol = 1e-4;
		mbloc   = A.num_rows;
		permtol = 1e-3;
		alph    = 1.0;

		this->same_structure = same_structure;			
		if (this->same_structure) {
		nrow = A.num_rows;
		/*-------------------SPARSKIT2 routines begin----------------------*/		
		switch (type)
		{ 
			case sparskit_ilut: 
			nvv = nrow + 1;
			niw = 2 * nrow;
			break;
			case sparskit_ilutp: 
			nvv = nrow + 1;
			niw = 2 * nrow;
			break;		
			case sparskit_ilud: 
			nvv = 2 * nrow;
			niw = 2 * nrow;
			break;	
			case sparskit_iludp: 
			nvv = 2 * nrow;
			niw = 2 * nrow;
			break;			
			case sparskit_iluk: 
			nvv = nrow;
			niw = 3 * nrow;
			break;
			case sparskit_ilu0: 
			nvv = 1;
			niw = nrow;
			break;		
			case sparskit_milu0: 
			nvv = 1;
			niw = nrow;
			break;
			default:	
			nvv = 2 * nrow;
			niw = 3 * nrow;		
			break;
		}	
		/*-------------------SPARSKIT2 routines end----------------------*/		
		
		/*-------------------MKL routines begin----------------------*/		
		if (use_mkl_sv) {
		temp_mkl_r.resize(A.num_rows);
		temp_mkl_y.resize(A.num_rows);
		temp_mkl_x.resize(A.num_rows);
		}
		/*-------------------MKL routines end----------------------*/
		}
		
		/*-------------------cuSPARSE routines begin----------------------*/			
		if (!use_mkl_sv) {		
		// create a matrix description for the lower
		cusparseCreateMatDescr(&descrM_lo); 

		cusparseSetMatType (descrM_lo, CUSPARSE_MATRIX_TYPE_TRIANGULAR);
		cusparseSetMatIndexBase (descrM_lo, CUSPARSE_INDEX_BASE_ZERO);
		// NOTE! CUSPARSE_DIAG_TYPE_UNIT seems not right
		//cusparseSetMatDiagType (descrM_lo, CUSPARSE_DIAG_TYPE_UNIT);
		cusparseSetMatFillMode (descrM_lo, CUSPARSE_FILL_MODE_LOWER);	

		// create the analysis info object for the lower	
		cusparseCreateSolveAnalysisInfo(&Analy_lo);
		
		// create a matrix description for the upper
		cusparseCreateMatDescr(&descrM_up); 
		cusparseSetMatType (descrM_up, CUSPARSE_MATRIX_TYPE_TRIANGULAR);
		cusparseSetMatIndexBase (descrM_up, CUSPARSE_INDEX_BASE_ZERO);
		cusparseSetMatDiagType (descrM_up, CUSPARSE_DIAG_TYPE_NON_UNIT);
		cusparseSetMatFillMode (descrM_up, CUSPARSE_FILL_MODE_UPPER);	
			
		// create the analysis info object for the upper	
		cusparseCreateSolveAnalysisInfo(&Analy_up);
		
		if (this->same_structure) {		
		temp_y.resize(A.num_rows);	
		if (type == sparskit_ilutp || type == sparskit_iludp) {				
		px.resize(A.num_rows);	
		}			
		}		
		}
		/*-------------------cuSPARSE routines end----------------------*/		
		
		initialized = true;
	}		
	
template <typename IndexType, typename ValueType, typename MemorySpace>
template<typename MatrixType>
    void sparskit2_ilu<IndexType,ValueType,MemorySpace>
    ::setup(const MatrixType& A_, const IndexType maxfil_, const IndexType lfil_,
	const ValueType droptol_, const IndexType mbloc_, const ValueType permtol_, const ValueType alph_, const IndexType sweeps_)
    {
        CUSP_PROFILE_SCOPED();
		
		if (!initialized) {
		  printf("sparskit2_ilu has not yet been created!\n");
		  exit(1);		
		}
		
		cusp::csr_matrix<IndexType, ValueType, MemorySpace> A = A_;
		cusp::detail::sort_by_row_and_column(A);  // sort by row and column  
  
		sweeps  = sweeps_;
		maxfil  = maxfil_;
		lfil    = lfil_;		
		droptol = droptol_;
		mbloc   = mbloc_;
		permtol = permtol_;
		alph    = alph_;
	
		if (!same_structure) {
		nrow = A.num_rows;
		/*-------------------SPARSKIT2 routines begin----------------------*/		
		switch (type)
		{ 
			case sparskit_ilut: 
			nvv = nrow + 1;
			niw = 2 * nrow;
			break;
			case sparskit_ilutp: 
			nvv = nrow + 1;
			niw = 2 * nrow;
			break;		
			case sparskit_ilud: 
			nvv = 2 * nrow;
			niw = 2 * nrow;
			break;	
			case sparskit_iludp: 
			nvv = 2 * nrow;
			niw = 2 * nrow;
			break;			
			case sparskit_iluk: 
			nvv = nrow;
			niw = 3 * nrow;
			break;
			case sparskit_ilu0: 
			nvv = 1;
			niw = nrow;
			break;		
			case sparskit_milu0: 
			nvv = 1;
			niw = nrow;
			break;
			default:	
			nvv = 2 * nrow;
			niw = 3 * nrow;		
			break;
		}	
		/*-------------------SPARSKIT2 routines end----------------------*/		
		
		if (use_mkl_sv) {
			temp_mkl_r.resize(A.num_rows);
			temp_mkl_y.resize(A.num_rows);
			temp_mkl_x.resize(A.num_rows);
		} else {
			temp_y.resize(A.num_rows);	
			if (type == sparskit_ilutp || type == sparskit_iludp) {				
			px.resize(A.num_rows);	
			}	
		}
		}
		
		/*-------------------SPARSKIT2 routines begin----------------------*/
		// preallocate memory for nonzero entries in fortran
		iwk = (2*maxfil+1)*nrow-maxfil*(maxfil+1)+1;		
		
		cusp::array1d<ValueType, cusp::host_memory> au_(iwk,0);
		cusp::array1d<IndexType, cusp::host_memory> jau_(iwk,0);		
		cusp::array1d<IndexType, cusp::host_memory> ju_(nrow,0);	
		cusp::array1d<ValueType, cusp::host_memory> vv_(nvv,0);		
		cusp::array1d<IndexType, cusp::host_memory> iw_(niw,0);		
		
		ValueType *au = thrust::raw_pointer_cast(&*au_.begin());		
		IndexType *jau = thrust::raw_pointer_cast(&*jau_.begin());	
		IndexType *ju = thrust::raw_pointer_cast(&*ju_.begin());			
		ValueType *vv = thrust::raw_pointer_cast(&*vv_.begin());			
		IndexType *iw = thrust::raw_pointer_cast(&*iw_.begin());							

		cusp::csr_matrix<IndexType,ValueType,cusp::host_memory> A_host;		
		// SPARSKIT2 and MKL is one-based, transform row_offsets and column_indices to be plusing one
		cusp::detail::setup_mkl_matrix(A_host,A);

		// get the raw host pointer to be used in MKL
		IndexType *h_row  = thrust::raw_pointer_cast(&*A_host.row_offsets.begin());
		IndexType *h_col  = thrust::raw_pointer_cast(&*A_host.column_indices.begin());	
		ValueType *h_val  = thrust::raw_pointer_cast(&*A_host.values.begin());			
		
		// temporary 
		cusp::array1d<IndexType, cusp::host_memory> iperm_;
		IndexType *iperm;
		cusp::array1d<IndexType, cusp::host_memory> levs_;
		IndexType *levs;
		
		timer ilu_fact;			
		switch (type)
		{ 
			case sparskit_ilut: 
			ilut_(&nrow, h_val, h_col, h_row, &maxfil, &droptol, au, jau, ju, &iwk, vv, iw, &ierr);	
			break;
			case sparskit_ilutp:
			iperm_.resize(2*nrow);	
			iperm = thrust::raw_pointer_cast(&*iperm_.begin());						
			ilutp_(&nrow, h_val, h_col, h_row, &maxfil, &droptol, &permtol, &mbloc, au, jau, ju, &iwk, vv, iw, iperm, &ierr);	
			if (use_mkl_sv) {
			pc_h.assign(iperm_.begin(),iperm_.begin()+nrow);
			ipc_h.assign(iperm_.begin()+nrow,iperm_.end());			
			} else {
			pc_d.assign(iperm_.begin(),iperm_.begin()+nrow);
			ipc_d.assign(iperm_.begin()+nrow,iperm_.end());					
			}
			break;		
			case sparskit_ilud: 
			ilud_(&nrow, h_val, h_col, h_row, &alph, &droptol, au, jau, ju, &iwk, vv, iw, &ierr);	
			break;	
			case sparskit_iludp: 
			iperm_.resize(2*nrow);	
			iperm = thrust::raw_pointer_cast(&*iperm_.begin());						
			iludp_(&nrow, h_val, h_col, h_row, &alph, &droptol, &permtol, &mbloc, au, jau, ju, &iwk, vv, iw, iperm, &ierr);
			if (use_mkl_sv) {
			pc_h.assign(iperm_.begin(),iperm_.begin()+nrow);
			ipc_h.assign(iperm_.begin()+nrow,iperm_.end());			
			} else {
			pc_d.assign(iperm_.begin(),iperm_.begin()+nrow);
			ipc_d.assign(iperm_.begin()+nrow,iperm_.end());					
			}			
			break;			
			case sparskit_iluk: 
			levs_.resize(iwk,0);	
			levs = thrust::raw_pointer_cast(&*levs_.begin());				
			iluk_(&nrow, h_val, h_col, h_row, &lfil, au, jau, ju, levs, &iwk, vv, iw, &ierr);
			break;
			case sparskit_ilu0: 
			ilu0_(&nrow, h_val, h_col, h_row, au, jau, ju, iw, &ierr);
			break;		
			case sparskit_milu0: 
			milu0_(&nrow, h_val, h_col, h_row, au, jau, ju, iw, &ierr);
			break;
			default:	
			ilut_(&nrow, h_val, h_col, h_row, &maxfil, &droptol, au, jau, ju, &iwk, vv, iw, &ierr);		
			break;
		}		
		//std::cout << "ilu factorization in " << ilu_fact.seconds_elapsed() << " s " << std::endl;
		
		//IndexType num_entries_msr = jau[nrow] - jau[0] + nrow;		
		//std::cout << "num_entries of msr matrix: " << num_entries_msr << std::endl;		
		/*-------------------SPARSKIT2 routines end----------------------*/	
		
		cusp::csr_matrix<IndexType, ValueType, cusp::device_memory> A_fact;		
		cusp::csr_matrix<IndexType, ValueType, cusp::host_memory> A_fact_host;
		if (use_mkl_sv) {
			// used in solving triangular system
			A_mkl = A_host;
			msr_to_csr(IndexType(A.num_rows), IndexType(A.num_cols), iwk, ju, jau, au, true, A_fact_mkl);					
		} else {
			if (type == sparskit_ilutp || type == sparskit_iludp) {	
			A_p.resize(A_host.num_rows,A_host.num_cols,A_host.num_entries);
			// MKL is one-based, transform row_offsets and column_indices to be minusing one
			thrust::copy(thrust::make_transform_iterator(A_host.row_offsets.begin(), cusp::detail::ADD<IndexType>(-1)),
			thrust::make_transform_iterator(A_host.row_offsets.end(), cusp::detail::ADD<IndexType>(-1)),
			A_p.row_offsets.begin());
			thrust::copy(thrust::make_transform_iterator(A_host.column_indices.begin(), cusp::detail::ADD<IndexType>(-1)),
			thrust::make_transform_iterator(A_host.column_indices.end(), cusp::detail::ADD<IndexType>(-1)),
			A_p.column_indices.begin());
			thrust::copy(A_host.values.begin(), A_host.values.end(), A_p.values.begin());			
			}
			// zero-based factorized matrix		
			// the device version doesn't work
			msr_to_csr(IndexType(A.num_rows), IndexType(A.num_cols), iwk, ju, jau, au, false, A_fact_host);	
			A_fact = A_fact_host;
			
			cusp::detail::extract_triangle(A_fact, A_lo, A_up);					
			// Perform the analysis for the lower
			cusparseXcsrsv_analysis(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, A.num_rows, descrM_lo,
			thrust::raw_pointer_cast(&A_lo.values[0]),
			thrust::raw_pointer_cast(&A_lo.row_offsets[0]),
			thrust::raw_pointer_cast(&A_lo.column_indices[0]),	
			Analy_lo);
			// Perform the analysis for the upper
			cusparseXcsrsv_analysis(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, A.num_rows, descrM_up,
			thrust::raw_pointer_cast(&A_up.values[0]),
			thrust::raw_pointer_cast(&A_up.row_offsets[0]),
			thrust::raw_pointer_cast(&A_up.column_indices[0]),	
			Analy_up);
		}				
    }

// destructor	
template <typename IndexType, typename ValueType, typename MemorySpace>
    sparskit2_ilu<IndexType,ValueType,MemorySpace>
    ::~sparskit2_ilu()
    {
		if (!use_mkl_sv) {		
		// destroy cuSPARSE context
		cusparseDestroyMatDescr(descrM_lo);
		cusparseDestroySolveAnalysisInfo(Analy_lo);  
		cusparseDestroyMatDescr(descrM_up);
		cusparseDestroySolveAnalysisInfo(Analy_up);
		}	
    }

// linear_operator
template <typename IndexType, typename ValueType, typename MemorySpace>
template <typename Array1, typename Array2>
    void sparskit2_ilu<IndexType,ValueType,MemorySpace>
    ::operator()(const Array1& b, Array2& x)
    {
        CUSP_PROFILE_SCOPED();
		
		if (use_mkl_sv) {
		// only CPU memory is needed, copy back from device
		temp_mkl_r = b;
		// get the raw host pointer to be used in MKL
		ValueType *temp_r = thrust::raw_pointer_cast(&*temp_mkl_r.begin());		
		ValueType *temp_y = thrust::raw_pointer_cast(&*temp_mkl_y.begin());		
		ValueType *temp_x = thrust::raw_pointer_cast(&*temp_mkl_x.begin());		
		IndexType *h_fact_row  = thrust::raw_pointer_cast(&*A_fact_mkl.row_offsets.begin());
		IndexType *h_fact_col  = thrust::raw_pointer_cast(&*A_fact_mkl.column_indices.begin());	
		ValueType *h_fact_val = thrust::raw_pointer_cast(&*A_fact_mkl.values.begin());	
		cvar1='L';
		cvar='N';
		cvar2='U';
		mkl_xcsrtrsv(&cvar1,&cvar,&cvar2,&nrow,h_fact_val,h_fact_row,h_fact_col,temp_r,temp_y);
		cvar1='U';
		cvar='N';
		cvar2='N';
		mkl_xcsrtrsv(&cvar1,&cvar,&cvar2,&nrow,h_fact_val,h_fact_row,h_fact_col,temp_y,temp_x);
		if (type == sparskit_ilutp || type == sparskit_iludp) { // only Upper triangular is permuted
		cusp::permute_array(pc_h, temp_mkl_x, temp_mkl_y);		
		x = temp_mkl_y;
		} else {
		// copy the result to device
		x = temp_mkl_x;
		}
		} else {
		ValueType *temp_y_ = thrust::raw_pointer_cast(&*temp_y.begin());		
		const ValueType *b_ = thrust::raw_pointer_cast(&*b.begin());
		ValueType *x_;
		if (type == sparskit_ilutp || type == sparskit_iludp) {
		x_ = thrust::raw_pointer_cast(&*px.begin());		
		} else {
		x_ = thrust::raw_pointer_cast(&*x.begin());
		}		
        // Forward Solve: L y = r
        cusparseXcsrsv_solve(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, nrow, 1.0, descrM_lo, 
									thrust::raw_pointer_cast(&A_lo.values[0]),
									thrust::raw_pointer_cast(&A_lo.row_offsets[0]),
									thrust::raw_pointer_cast(&A_lo.column_indices[0]),	
									Analy_lo, b_, temp_y_);    
        // Back Substitution: U x = y
        cusparseXcsrsv_solve(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, nrow, 1.0, descrM_up, 
									thrust::raw_pointer_cast(&A_up.values[0]),
									thrust::raw_pointer_cast(&A_up.row_offsets[0]),
									thrust::raw_pointer_cast(&A_up.column_indices[0]),	
									Analy_up, temp_y_, x_); 	
		if (type == sparskit_ilutp || type == sparskit_iludp) { // only Upper triangular is permuted
		cusp::permute_array(pc_d, px, x);		
		} 								
        }
    }

template <typename IndexType, typename ValueType, typename MemorySpace>
template<typename MatrixType, typename VectorType1, typename VectorType2>
    void sparskit2_ilu<IndexType,ValueType,MemorySpace>
    ::presmooth(const MatrixType&A, const VectorType1& b, VectorType2& x)
    {
        CUSP_PROFILE_SCOPED();
			
		if (use_mkl_sv) {
		// only CPU memory is needed, copy back from device
		temp_mkl_r = b;

		// get the raw host pointer to be used in MKL
		ValueType *temp_r = thrust::raw_pointer_cast(&*temp_mkl_r.begin());		
		ValueType *temp_y = thrust::raw_pointer_cast(&*temp_mkl_y.begin());		
		ValueType *temp_x = thrust::raw_pointer_cast(&*temp_mkl_x.begin());				
		IndexType *h_fact_row  = thrust::raw_pointer_cast(&*A_fact_mkl.row_offsets.begin());
		IndexType *h_fact_col  = thrust::raw_pointer_cast(&*A_fact_mkl.column_indices.begin());	
		ValueType *h_fact_val = thrust::raw_pointer_cast(&*A_fact_mkl.values.begin());	
		cvar1='L';
		cvar='N';
		cvar2='U';
		mkl_xcsrtrsv(&cvar1,&cvar,&cvar2,&nrow,h_fact_val,h_fact_row,h_fact_col,temp_r,temp_y);
		cvar1='U';
		cvar='N';
		cvar2='N';
		mkl_xcsrtrsv(&cvar1,&cvar,&cvar2,&nrow,h_fact_val,h_fact_row,h_fact_col,temp_y,temp_x);		
		cusp::array1d<ValueType,cusp::host_memory> res(A_mkl.num_rows);
		cusp::array1d<ValueType,cusp::host_memory> e(A_mkl.num_rows);
		ValueType *res_ = thrust::raw_pointer_cast(&*res.begin());	
		ValueType *e_ = thrust::raw_pointer_cast(&*e.begin());	
		IndexType *h_row  = thrust::raw_pointer_cast(&*A_mkl.row_offsets.begin());
		IndexType *h_col  = thrust::raw_pointer_cast(&*A_mkl.column_indices.begin());	
		ValueType *h_val = thrust::raw_pointer_cast(&*A_mkl.values.begin());			
		for(IndexType s = 1; s<this->sweeps; s++)
		{	
			// compute res <- b - A*x
			cvar='N';			
			mkl_xcsrgemv(&cvar,&nrow,h_val,h_row,h_col,temp_x, res_);
			cblas_xaxpy(nrow,ValueType(-1), temp_r, MKL_INT(1), res_, MKL_INT(1));
			cblas_xscal(nrow,ValueType(-1),res_,MKL_INT(1));		
			cvar1='L';
			cvar='N';
			cvar2='U';
			mkl_xcsrtrsv(&cvar1,&cvar,&cvar2,&nrow,h_fact_val,h_fact_row,h_fact_col,res_,temp_y);
			cvar1='U';
			cvar='N';
			cvar2='N';
			mkl_xcsrtrsv(&cvar1,&cvar,&cvar2,&nrow,h_fact_val,h_fact_row,h_fact_col,temp_y,e_);	
			cblas_xaxpy(nrow,ValueType(1), e_, MKL_INT(1), temp_x, MKL_INT(1));	
		}		
		if (type == sparskit_ilutp || type == sparskit_iludp) { // only Upper triangular is permuted
		cusp::permute_array(pc_h, temp_mkl_x, temp_mkl_y);		
		x = temp_mkl_y;
		} else {
		// copy the result to device
		x = temp_mkl_x;
		}		
		} else {
		ValueType *temp_y_ = thrust::raw_pointer_cast(&*temp_y.begin());		
		const ValueType *b_ = thrust::raw_pointer_cast(&*b.begin());
		ValueType *x_;
		if (type == sparskit_ilutp || type == sparskit_iludp) {	
		x_ = thrust::raw_pointer_cast(&*px.begin());			
		} else {
		x_ = thrust::raw_pointer_cast(&*x.begin());		
		}	
        // Forward Solve: L y = r
        cusparseXcsrsv_solve(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, nrow, 1.0, descrM_lo, 
									thrust::raw_pointer_cast(&A_lo.values[0]),
									thrust::raw_pointer_cast(&A_lo.row_offsets[0]),
									thrust::raw_pointer_cast(&A_lo.column_indices[0]),	
									Analy_lo, b_, temp_y_);    
        // Back Substitution: U x = y
        cusparseXcsrsv_solve(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, nrow, 1.0, descrM_up, 
									thrust::raw_pointer_cast(&A_up.values[0]),
									thrust::raw_pointer_cast(&A_up.row_offsets[0]),
									thrust::raw_pointer_cast(&A_up.column_indices[0]),	
									Analy_up, temp_y_, x_); 
		VectorType1 res(A.num_rows);
		VectorType2 e(A.num_rows);
		ValueType *res_ = thrust::raw_pointer_cast(&*res.begin());	
		ValueType *e_ = thrust::raw_pointer_cast(&*e.begin());	
		for (IndexType s = 1; s<this->sweeps; s++)
		{	
			// compute res <- b - A*x
			if (type == sparskit_ilutp || type == sparskit_iludp) {			
			cusp::multiply(A_p, px, res);	
			} else {
			cusp::multiply(A, x, res);			
			}
			cusp::blas::axpby(b, res, res, ValueType(1), ValueType(-1));			
			// Forward Solve: L y = r
			cusparseXcsrsv_solve(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, nrow, 1.0, descrM_lo, 
										thrust::raw_pointer_cast(&A_lo.values[0]),
										thrust::raw_pointer_cast(&A_lo.row_offsets[0]),
										thrust::raw_pointer_cast(&A_lo.column_indices[0]),	
										Analy_lo, res_, temp_y_);    
			// Back Substitution: U x = y
			cusparseXcsrsv_solve(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, nrow, 1.0, descrM_up, 
										thrust::raw_pointer_cast(&A_up.values[0]),
										thrust::raw_pointer_cast(&A_up.row_offsets[0]),
										thrust::raw_pointer_cast(&A_up.column_indices[0]),	
										Analy_up, temp_y_, e_); 
			if (type == sparskit_ilutp || type == sparskit_iludp) {			
			cusp::blas::axpby(px, e, px, ValueType(1), ValueType(1));		
			} else {										
			cusp::blas::axpby(x, e, x, ValueType(1), ValueType(1));	
			}
		}
		if (type == sparskit_ilutp || type == sparskit_iludp) { // only Upper triangular is permuted
		cusp::permute_array(pc_d, px, x);		
		} 			
        }		
    }

template <typename IndexType, typename ValueType, typename MemorySpace>
template<typename MatrixType, typename VectorType1, typename VectorType2>
    void sparskit2_ilu<IndexType,ValueType,MemorySpace>
    ::postsmooth(const MatrixType& A, const VectorType1& b, VectorType2& x)
    {
        CUSP_PROFILE_SCOPED();
			
		if (use_mkl_sv) {
		// only CPU memory is needed, copy back from device
		temp_mkl_r = b;
		if (type == sparskit_ilutp || type == sparskit_iludp) {
		temp_mkl_y = x;		
		cusp::permute_array(ipc_h, temp_mkl_y, temp_mkl_x);			
		} else {
		temp_mkl_x = x;
		}
		
		// get the raw host pointer to be used in MKL
		ValueType *temp_r = thrust::raw_pointer_cast(&*temp_mkl_r.begin());		
		ValueType *temp_y = thrust::raw_pointer_cast(&*temp_mkl_y.begin());		
		ValueType *temp_x = thrust::raw_pointer_cast(&*temp_mkl_x.begin());		
		IndexType *h_fact_row  = thrust::raw_pointer_cast(&*A_fact_mkl.row_offsets.begin());
		IndexType *h_fact_col  = thrust::raw_pointer_cast(&*A_fact_mkl.column_indices.begin());	
		ValueType *h_fact_val = thrust::raw_pointer_cast(&*A_fact_mkl.values.begin());		
		IndexType *h_row  = thrust::raw_pointer_cast(&*A_mkl.row_offsets.begin());
		IndexType *h_col  = thrust::raw_pointer_cast(&*A_mkl.column_indices.begin());	
		ValueType *h_val = thrust::raw_pointer_cast(&*A_mkl.values.begin());
		cusp::array1d<ValueType,cusp::host_memory> res(A_mkl.num_rows);
		cusp::array1d<ValueType,cusp::host_memory> e(A_mkl.num_rows);
		ValueType *res_ = thrust::raw_pointer_cast(&*res.begin());	
		ValueType *e_ = thrust::raw_pointer_cast(&*e.begin());			
		for(IndexType s = 0; s<this->sweeps; s++)
		{	
			// compute res <- b - A*x
			cvar='N';			
			mkl_xcsrgemv(&cvar,&nrow,h_val,h_row,h_col,temp_x, res_);
			cblas_xaxpy(nrow,ValueType(-1), temp_r, MKL_INT(1), res_, MKL_INT(1));
			cblas_xscal(nrow,ValueType(-1),res_,MKL_INT(1));
			cvar1='L';
			cvar='N';
			cvar2='U';
			mkl_xcsrtrsv(&cvar1,&cvar,&cvar2,&nrow,h_fact_val,h_fact_row,h_fact_col,res_,temp_y);
			cvar1='U';
			cvar='N';
			cvar2='N';
			mkl_xcsrtrsv(&cvar1,&cvar,&cvar2,&nrow,h_fact_val,h_fact_row,h_fact_col,temp_y,e_);	
			cblas_xaxpy(nrow,ValueType(1), e_, MKL_INT(1), temp_x, MKL_INT(1));	
		}		
		if (type == sparskit_ilutp || type == sparskit_iludp) { // only Upper triangular is permuted
		cusp::permute_array(pc_h, temp_mkl_x, temp_mkl_y);		
		x = temp_mkl_y;
		} else {
		// copy the result to device
		x = temp_mkl_x;
		}			
		} else {
		if (type == sparskit_ilutp || type == sparskit_iludp) {		
		cusp::permute_array(ipc_d, x, px);				
		}
		ValueType *temp_y_ = thrust::raw_pointer_cast(&*temp_y.begin());		
		VectorType1 res(A.num_rows);
		VectorType2 e(A.num_rows);
		ValueType *res_ = thrust::raw_pointer_cast(&*res.begin());	
		ValueType *e_ = thrust::raw_pointer_cast(&*e.begin());	
		for(IndexType s = 0; s<this->sweeps; s++)
		{	
			// compute res <- b - A*x
			if (type == sparskit_ilutp || type == sparskit_iludp) {	
			cusp::multiply(A_p, px, res);			
			} else {		
			cusp::multiply(A, x, res);
			}
			cusp::blas::axpby(b, res, res, ValueType(1), ValueType(-1));		
			// Forward Solve: L y = r
			cusparseXcsrsv_solve(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, nrow, 1.0, descrM_lo, 
										thrust::raw_pointer_cast(&A_lo.values[0]),
										thrust::raw_pointer_cast(&A_lo.row_offsets[0]),
										thrust::raw_pointer_cast(&A_lo.column_indices[0]),	
										Analy_lo, res_, temp_y_);    
			// Back Substitution: U x = y
			cusparseXcsrsv_solve(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, nrow, 1.0, descrM_up, 
										thrust::raw_pointer_cast(&A_up.values[0]),
										thrust::raw_pointer_cast(&A_up.row_offsets[0]),
										thrust::raw_pointer_cast(&A_up.column_indices[0]),	
										Analy_up, temp_y_, e_); 
			if (type == sparskit_ilutp || type == sparskit_iludp) {	
			cusp::blas::axpby(px, e, px, ValueType(1), ValueType(1));				
			} else {			
			cusp::blas::axpby(x, e, x, ValueType(1), ValueType(1));	
			}
		}
		if (type == sparskit_ilutp || type == sparskit_iludp) { // only Upper triangular is permuted
		cusp::permute_array(pc_d, px, x);		
		} 		
        }		
    }

} // end namespace relaxation
} // end namespace cusp

