#include <cusp/precond/asm.h>
#include <cusp/blas.h>
#include <cusp/krylov/cg.h>
#include <cusp/gallery/poisson.h>
#include <cusp/csr_matrix.h>
#include <cusp/precond/diagonal.h>
#include <iostream>
#include <performance/timer.h>
#include <cmath>
#define PI 3.1415926

template <typename Monitor>
void report_status(Monitor& monitor)
{
    if (monitor.converged())
    {
        std::cout << "  Solver converged to " << monitor.tolerance1() << " tolerance";
        std::cout << " after " << monitor.iteration_count() << " iterations";
        std::cout << " (" << monitor.relative_residual_norm() << " final residual)" << std::endl;
    }
    else
    {
        std::cout << "  Solver reached iteration limit " << monitor.iteration_limit() << " before converging";
        std::cout << " to " << monitor.tolerance1() << " tolerance ";
        std::cout << " (" << monitor.relative_residual_norm() << " final residual)" << std::endl;
    }
}

void pq_gen(const int np, int *p, int *q)
{
    int a = 1, b = np;

    for(int i = 2; i < floor(sqrt(double(np))) + 1; i++)
        if(np % i == 0)  
        {   
            a = i;
            b = np / i;
        }
    *p = a;    
    *q = b;
}

int main(void)
{
    typedef int                 IndexType;
    typedef double               ValueType;
    typedef cusp::device_memory MemorySpace;

    // create an empty sparse matrix structure
    cusp::csr_matrix<IndexType, ValueType, MemorySpace> A;

    // create 2D Poisson problem
	int m = 0, n = 0;
	while(m < 2 || n < 2)
    {
    	printf("Please input m, n: ");
    	scanf("%d %d", &m, &n);
    }
    cusp::gallery::poisson5pt(A, m, n);
	printf("row num of A: %d\n", A.num_rows);

	// allocate storage for solution (x) and right hand side (b)	
	cusp::array1d<ValueType, MemorySpace> x_solution(A.num_rows, 1);
    // initialize starting vector to random values in [0,1)
    //cusp::copy(cusp::detail::random_reals<ValueType>(A.num_rows), x_solution);
	ValueType *x_so_cpu;	
	cudaMallocHost((void **) &x_so_cpu, (A.num_rows)*sizeof(ValueType));
		
	for(int i=0;i<A.num_rows;i++)	
	{
		x_so_cpu[i]=std::sin(0.5*PI*i/A.num_rows);
	}
	ValueType *x_so_gpu = thrust::raw_pointer_cast(&x_solution[0]);
	cudaMemcpy(x_so_gpu, x_so_cpu, (A.num_rows)*sizeof(ValueType), cudaMemcpyHostToDevice);
	cudaFreeHost(x_so_cpu);
	
    cusp::array1d<ValueType, MemorySpace> b(A.num_rows);
	cusp::multiply(A, x_solution, b);
		
#if 0
    // solve with diagonal preconditioner
    {
        std::cout << "\nSolving with diagonal preconditioner (M = D^-1)" << std::endl;
        
        cusp::array1d<ValueType, MemorySpace> x(A.num_rows, 0);

        // set stopping criteria (iteration_limit = 10000, relative_tolerance = 1e-6)
        cusp::verbose_monitor<ValueType> monitor(b, 10000, 1e-6);

        // setup preconditioner
        cusp::precond::diagonal<ValueType, MemorySpace> M(A);

        // solve			
		timer t1;	
		//cusp::krylov::gmres(A, x, b, restart, monitor, M);
		cusp::krylov::cg(A, x, b, monitor, M);
		std::cout << "\nsolve by jacobi-cg in " << t1.milliseconds_elapsed() << " ms " << std::endl;	
		
		// report status
        report_status(monitor);
		
		// check solution
        cusp::array1d<ValueType, MemorySpace> e(A.num_rows);
		blas::axpby(x_solution, x, e, ValueType(1), ValueType(-1));
		ValueType e_norm = cusp::blas::nrm2(e);
		printf("\n check solution: ||x*-x||: %f \n", e_norm );	
		
		cusp::array1d<ValueType, MemorySpace> ax(x.size());
		cusp::array1d<ValueType, MemorySpace> r(A.num_rows);
		cusp::multiply(A, x, ax);
		cusp::blas::axpby(ax, b, r, ValueType(1), ValueType(-1));
		ValueType r_norm = cusp::blas::nrm2(r);
		ValueType b_norm = cusp::blas::nrm2(b);		
		printf("\n ||b-Ax||/||b||: %f \n", r_norm / b_norm );
    }
#endif	
		
    // solve with asm preconditioner 
	int asm_type = -1;
	while(asm_type < 0 || asm_type > 1)
	{
		printf("\nPlease input asm_type: 0: basic, 1: restrict ");
		scanf("%d", &asm_type);
	}
	int overlap = -1;	
	while(overlap < 0)
	{
		printf("\nPlease input overlap(>=0): ");
		scanf("%d", &overlap);
	}
	int iter = 0;	
	while(iter < 1)
	{
		printf("\nPlease input iter(>0): ");
		scanf("%d", &iter);
	}

	{
        std::cout << "\n ******Solving with asm preconditioner, PETSc sort****** " << std::endl;
		
		int nd = 0;	
		while(nd < 1)
		{
			printf("\nPlease input nd: ");
			scanf("%d", &nd);
		}
		int p, q, m_p, n_q;// p*q=nd, 2D domain partition   
		pq_gen(nd, &p, &q);
		printf("\nPartition np into %d, %d\n", p, q);
		
		m_p = m / p;
		n_q = n / q;
	
		cusp::array1d<IndexType,MemorySpace> domains(A.num_rows, 0);
		
		timer t0;
		dim3 grid(p, q, 1);
		dim3 block(16, 16, 1);
		cusp::precond::detail::set_domains<IndexType><<<grid,block>>>(thrust::raw_pointer_cast(&*domains.begin()), m, n, m_p, n_q, 1);
        std::cout << "\nset_domains in " << t0.milliseconds_elapsed() << " ms " << std::endl;
        
        // allocate storage for solution (x) and right hand side (b)
        cusp::array1d<ValueType, MemorySpace> x(A.num_rows, 0);

        // set stopping criteria (iteration_limit = 1000, relative_tolerance = 1e-6)
        cusp::verbose_monitor<ValueType> monitor(b, 10000, 1e-6);
	
        std::cout << "\nSolving with asm preconditioner, Jacobi subsolver " << std::endl;
        // setup preconditioner
		timer t1;
        cusp::precond::additive_schwarz<IndexType, ValueType, MemorySpace> M(true,A,domains,overlap);
		M.setup(true,A,domains,(cusp::precond::asm_type)asm_type,overlap,iter,1e-7);
		std::cout << "\nconstruct preconditioner in " << t1.milliseconds_elapsed() << " ms " << std::endl;
        
        // solve
		timer t2;
        cusp::krylov::cg(A, x, b, monitor, M);
        std::cout << "\nsolve by as in " << t2.milliseconds_elapsed() << " ms " << std::endl;
		
        // report status
        report_status(monitor);
		
		// check solution
        cusp::array1d<ValueType, MemorySpace> e(A.num_rows);
		blas::axpby(x_solution, x, e, ValueType(1), ValueType(-1));
		ValueType e_norm = cusp::blas::nrm2(e);
		printf("\n check solution: ||x*-x||: %f \n", e_norm );	
		
		cusp::array1d<ValueType, MemorySpace> ax(x.size());
		cusp::array1d<ValueType, MemorySpace> r(A.num_rows);
		cusp::multiply(A, x, ax);
		cusp::blas::axpby(ax, b, r, ValueType(1), ValueType(-1));
		ValueType r_norm = cusp::blas::nrm2(r);
		ValueType b_norm = cusp::blas::nrm2(b);		
		printf("\n ||b-Ax||/||b||: %f \n", r_norm / b_norm );
    }
	
    return 0;
}

