/*
 * Copyright 1993-2011 NVIDIA Corporation.  All rights reserved.
 *
 * Please refer to the NVIDIA end user license agreement (EULA) associated
 * with this source code for terms and conditions that govern your use of
 * this software. Any use, reproduction, disclosure, or distribution of
 * this software and related documentation outside the terms of the EULA
 * is strictly prohibited.
 *
 */

/*
 * This sample implements a preconditioned conjugate gradient solver on 
 * the GPU using CUBLAS and CUSPARSE.  Relative to the conjugateGradient
 * SDK example, this demonstrates the use of cusparseScsrsv_solve()
 * for solving triangular systems.  Specifically, the preconditioned 
 * conjugate gradient method with an Incomplete Cholesky preconditioner
 * is used to solve a randomly generated symmetric, strictly diagonally 
 * dominant, tridiagonal real matrix.
 *
 * Note that the code in this example and the specific matrices used here
 * were chosen to demonstrate the use of the CUSPARSE library as simply
 * and as clearly as possible.  This is not optimized code and the input 
 * matrices have been artificially generated, hence this code should not
 * be used either as a performance guide or for benchmarking purposes. 
 * In particular, the efficient solution of tridiagonal systems is a well
 * researched topic and interested users are encouraged to consult the
 * following references:
 *
 * Author: Li Luo   
 */


// Utilities and system includes
#include <cusparse.h>
#include <cublas.h>
#include <assert.h>
#include <stdio.h>

#include <cusp/csr_matrix.h>
#include <cusp/multiply.h>
#include <cusp/io/matrix_market.h>
#include <cusp/gallery/poisson.h>
#include <performance/timer.h>
#include <cusp/print.h>
#include <performance/timer.h>

#include "math.h"
#include "mkl_rci.h"
#include "mkl_spblas.h"

#define PI 3.1415926

#include <cusp/fact.h>
#include <cusp/detail/utils_.h>

#include <thrust/transform.h>
		
const char * sSDKname	  = "conjugateGradientPrecond";


/* checkStatus: concise method for verifying CUDA return status */
int checkStatus ( cusparseStatus_t status, char *msg )
{
    if ( status != CUSPARSE_STATUS_SUCCESS ) {
        fprintf (stderr, "!!!! CUSPARSE %s ERROR \n", msg);
        return 1;
    }
    return 0;
}


/* Solve Ax=b using the conjugate gradient method a) without any preconditioning and b) using an Incomplete LU preconditioner. */
int main(int argc, char **argv)
{
	typedef cusp::device_memory MemorySpace;
	typedef double ValueType;	
	typedef cusp::csr_matrix<int, ValueType, MemorySpace> Matrix;
	
	// preclaims for CUBLAS, CUSPARSE
    cusparseHandle_t handle = 0;
    cusparseStatus_t status;	
    const int max_iter = 2000;
    int k;
    int *d_col, *d_row;
    const ValueType tol = 1e-7f;                          
    ValueType r1, r_ini, alpha, beta;
    ValueType *d_val, *d_x;
    ValueType *d_r, *d_p, *d_omega;

	// preclaims for CUSP	
	Matrix A;			
	// load a matrix stored in MatrixMarket format
	#if 1
	std::string mtx="thermal2.mtx";
	printf("Please input matrix : ");
	std::cin >> mtx;
    cusp::io::read_matrix_market_file(A, mtx);
	#else
    // create 2D Poisson problem
	int m = 0, n = 0;
	while(m < 2 || n < 2)
    {
    	printf("Please input m, n: ");
    	scanf("%d %d", &m, &n);
    }
    cusp::gallery::poisson5pt(A, m, n);
	#endif	
	
	// allocate storage for solution (x) and right hand side (b)
	cusp::array1d<ValueType, MemorySpace> x_solution(A.num_rows, 1);
    cusp::array1d<ValueType, MemorySpace> b(A.num_rows);
	
	// initialize x
	/*
	for(int i=0;i<A.num_rows;i++)	
	{
		x_solution[i]=std::sin(0.5*PI*i/A.num_rows);
	}
	*/
	// b=Ax^*
	cusp::multiply(A, x_solution, b);


    /* Get handle to the CUSPARSE context */
    status = cusparseCreate(&handle);
    if (checkStatus(status, (char*)"initialization")) return EXIT_FAILURE;
    
    cudaMalloc((void**)&d_p, A.num_rows*sizeof(ValueType));
    cudaMalloc((void**)&d_omega, A.num_rows*sizeof(ValueType));
	
	// get the raw pointer on device, need not to be free
	d_row = thrust::raw_pointer_cast(&A.row_offsets[0]);
	d_col = thrust::raw_pointer_cast(&A.column_indices[0]);
	d_val = thrust::raw_pointer_cast(&A.values[0]);

	#if 0
	{
    printf("Convergence of conjugate gradient without preconditioning: \n");
	ValueType r0;
	
	cusp::array1d<ValueType, MemorySpace> x(A.num_rows, 0);	
	cusp::array1d<ValueType, MemorySpace> b_(b);
	d_x   = thrust::raw_pointer_cast(&x[0]);
	d_r	  = thrust::raw_pointer_cast(&b_[0]);
	
    /* Description of the A matrix*/
    cusparseMatDescr_t descr = 0;
    status = cusparseCreateMatDescr(&descr); 
    if ( checkStatus ( status, (char*)"cusparseCreateMatDescr" ) ) return EXIT_FAILURE;
    cusparseSetMatType(descr,CUSPARSE_MATRIX_TYPE_GENERAL);
    cusparseSetMatIndexBase(descr,CUSPARSE_INDEX_BASE_ZERO);
	
	timer t0;	
    k = 0;
    r0 = 0;
    r1 = cublasDdot(A.num_rows, d_r, 1, d_r, 1);
	r_ini = r1;	
    while (r1 > tol*tol*r_ini && k <= max_iter) {
        k++;
        if ( k == 1 ) {
            cublasDcopy (A.num_rows, d_r, 1, d_p, 1);
        }
        else {
            beta = r1/r0;
            cublasDscal(A.num_rows, beta, d_p, 1);         
            cublasDaxpy (A.num_rows, 1.0, d_r, 1, d_p, 1) ;
        }
        cusparseDcsrmv(handle,CUSPARSE_OPERATION_NON_TRANSPOSE, A.num_rows, A.num_rows, 1.0, 
                        descr, d_val, d_row, d_col, d_p, 0.0, d_omega);
        alpha = r1/cublasDdot (A.num_rows, d_p, 1, d_omega, 1);
        cublasDaxpy (A.num_rows, alpha, d_p, 1, d_x, 1);
        cublasDaxpy (A.num_rows, -alpha, d_omega, 1, d_r, 1);
        r0 = r1;
        r1 = cublasDdot(A.num_rows, d_r, 1, d_r, 1);
        printf("  iteration = %3d, residual = %e \n", k, sqrt(r1)/sqrt(r_ini));
    }
	std::cout << "\nsolve without preconditioner in " << t0.milliseconds_elapsed()/1000. << " s " << std::endl;	
		// check solution
        cusp::array1d<ValueType, MemorySpace> e(A.num_rows);
		cusp::blas::axpby(x_solution, x, e, ValueType(1), ValueType(-1));
		ValueType e_norm = cusp::blas::nrm2(e);
		printf("\n check solution: ||x*-x||: %f \n", e_norm );	
		
		cusp::array1d<ValueType, MemorySpace> ax(A.num_rows);
		cusp::array1d<ValueType, MemorySpace> r(A.num_rows);
		cusp::multiply(A, x, ax);
		cusp::blas::axpby(ax, b, r, ValueType(1), ValueType(-1));
		ValueType r_norm = cusp::blas::nrm2(r);
		ValueType b_norm = cusp::blas::nrm2(b);		
		printf("\n ||b-Ax||/||b||: %f \n", r_norm/b_norm);
		
      status = cusparseDestroyMatDescr(descr);
	  if (checkStatus(status, (char*)"cusparseDestroyMatDescr")) return EXIT_FAILURE;
	}
	#endif
	{	
    printf("\nConvergence of conjugate gradient using Incomplete LUT preconditioning: \n");

	cusp::array1d<ValueType, MemorySpace> x(A.num_rows, 0);	
	cusp::array1d<ValueType, MemorySpace> b_(b);
	d_x   = thrust::raw_pointer_cast(&x[0]);
	d_r	  = thrust::raw_pointer_cast(&b_[0]);
	
	/*-------------------MKL routines begin----------------------*/
	int k_ = 0;
	while(k_ < 1)
    {
    	printf("Please input fill in number k: ");
    	scanf("%d", &k_);
    }

	ValueType tol_ilut_ = 2.;
	while(tol_ilut_ > 1.)
    {
    	printf("Please input filling tolerance of ilut: ");
    	scanf("%lf", &tol_ilut_);
    }
	
	ValueType dpar30 = 2.;
	while(dpar30 > 1.)
    {
    	printf("Please input dpar(30) of ilut: ");
    	scanf("%lf", &dpar30);
    }	

        bool use_mkl_sv=false;
		int mkl_sv_=-1;
		while(mkl_sv_<0)
		{
			printf("Please input mkl_sv_: 0: use cuSPARSE to solve the triangular system,  1: use MKL\n");
			scanf("%d", &mkl_sv_);
		}
		if(mkl_sv_) {
		use_mkl_sv=true;
		} else {
		use_mkl_sv=false;		
		}
		
	timer t0;		
	// preclaims for MKL		
	MKL_INT ivar, ipar[128], maxfil, *ibilut, *jbilut, ierr=0;
	ValueType tol_ilut, dpar[128], *bilut;
	char cvar,cvar1,cvar2;
		
	// initialize the parameters for ilut
	ivar = A.num_rows;	
	ipar[1]  = 6;
	ipar[5]  = 1;
	ipar[6]  = 1;	
	ipar[30] = 1;
	dpar[30] = dpar30;
	dpar[31] = 1.E-16;
	tol_ilut = tol_ilut_;
	maxfil   = (A.num_entries+A.num_rows*k_-A.num_rows)/(2*A.num_rows)+1;
	
	// allocate memory for creating preconditioner in MKL
	/*
	ValueType *h_val = (ValueType *) malloc (A.num_entries*sizeof(ValueType));
	int   *h_row = (int *) malloc ((A.num_rows+1)*sizeof(int));
	int   *h_col = (int *) malloc (A.num_entries*sizeof(int));	
	*/
	ValueType *h_val;
	int *h_col, *h_row;
	cudaMallocHost((void **) &h_val, A.num_entries*sizeof(ValueType));
	cudaMallocHost((void **) &h_col, A.num_entries*sizeof(int));	
	cudaMallocHost((void **) &h_row, (A.num_rows+1)*sizeof(int));
	
	// number of entries in factorized matrix
	int num_entries = (2*maxfil+1)*A.num_rows-maxfil*(maxfil+1)+1;
	/*
	bilut = (ValueType *) malloc (num_entries*sizeof(ValueType));	
	ibilut = (int *) malloc ((A.num_rows+1)*sizeof(int));	
	jbilut = (int *) malloc (num_entries*sizeof(int));	
	*/
	cudaMallocHost((void **) &bilut, num_entries*sizeof(ValueType));
	cudaMallocHost((void **) &ibilut, (A.num_rows+1)*sizeof(int));
	cudaMallocHost((void **) &jbilut, num_entries*sizeof(int));	
	
	// MKL is 1-based, transform row_offsets and column_indices to be plusing one
	thrust::copy(thrust::make_transform_iterator(A.row_offsets.begin(), cusp::detail::ADD<int>(1)),
	thrust::make_transform_iterator(A.row_offsets.end(), cusp::detail::ADD<int>(1)),
	h_row);
	thrust::copy(thrust::make_transform_iterator(A.column_indices.begin(), cusp::detail::ADD<int>(1)),
	thrust::make_transform_iterator(A.column_indices.end(), cusp::detail::ADD<int>(1)),
	h_col);
	thrust::copy(A.values.begin(), A.values.end(), h_val);

	// call MKL routine dcsrilut
	dcsrilut(&ivar, h_val, h_row, h_col, bilut, ibilut, jbilut, &tol_ilut, &maxfil, ipar, dpar, &ierr);
	if (ierr!=0)
	{
	  printf("Preconditioner dcsrilut has returned the ERROR code %d\n", ierr);
	  return EXIT_FAILURE;
	}

	Matrix A_fact;	
	create_fact_from_mkl((int)A.num_rows, (int)A.num_cols, num_entries, ibilut, jbilut, bilut, A_fact);
	
	Matrix A_lo;
	Matrix A_up;
	cusp::detail::extract_triangle(A_fact, A_lo, A_up);
	
	std::cout << "\ncreate preconditioner in " << t0.milliseconds_elapsed()/1000. << " s " << std::endl;	
	/*-------------------MKL routines end----------------------*/

	/*-------------------cuSPARSE routines begin----------------------*/	
    /* Description of the A matrix*/
	timer t1;	
    cusparseMatDescr_t descr = 0;
    status = cusparseCreateMatDescr(&descr); 
    if ( checkStatus ( status, (char*)"cusparseCreateMatDescr" ) ) return EXIT_FAILURE;
    cusparseSetMatType(descr,CUSPARSE_MATRIX_TYPE_GENERAL);
    cusparseSetMatIndexBase(descr,CUSPARSE_INDEX_BASE_ZERO);	
	
    // create a matrix description for the lower
    cusparseMatDescr_t descrM_lo = 0;
    status = cusparseCreateMatDescr(&descrM_lo); 
    if (checkStatus(status, (char*)"cusparseCreateMatDescr") ) return EXIT_FAILURE;
    cusparseSetMatType (descrM_lo, CUSPARSE_MATRIX_TYPE_TRIANGULAR);
    cusparseSetMatIndexBase (descrM_lo, CUSPARSE_INDEX_BASE_ZERO);
	// NOTE! CUSPARSE_DIAG_TYPE_UNIT seems not right
    //cusparseSetMatDiagType (descrM_lo, CUSPARSE_DIAG_TYPE_UNIT);
    cusparseSetMatFillMode (descrM_lo, CUSPARSE_FILL_MODE_LOWER);	

    // create the analysis info object for the lower	
	cusparseSolveAnalysisInfo_t Analy_lo = 0;
	status = cusparseCreateSolveAnalysisInfo(&Analy_lo);
    if (checkStatus (status, (char*)"cusparseCreateSolveAnalysisInfo")) return EXIT_FAILURE;	
		 
    // Perform the analysis for the lower
    status = cusparseDcsrsv_analysis(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, A.num_rows, descrM_lo,
	thrust::raw_pointer_cast(&A_lo.values[0]),
	thrust::raw_pointer_cast(&A_lo.row_offsets[0]),
	thrust::raw_pointer_cast(&A_lo.column_indices[0]),	
	Analy_lo);
    if (checkStatus(status, (char*)"susparseScsrv_analysis")) return EXIT_FAILURE;
	
	
    // create a matrix description for the upper
    cusparseMatDescr_t descrM_up = 0;
    status = cusparseCreateMatDescr(&descrM_up); 
    if (checkStatus(status, (char*)"cusparseCreateMatDescr")) return EXIT_FAILURE;
    cusparseSetMatType (descrM_up, CUSPARSE_MATRIX_TYPE_TRIANGULAR);
    cusparseSetMatIndexBase (descrM_up, CUSPARSE_INDEX_BASE_ZERO);
    cusparseSetMatDiagType (descrM_up, CUSPARSE_DIAG_TYPE_NON_UNIT);
    cusparseSetMatFillMode (descrM_up, CUSPARSE_FILL_MODE_UPPER);	
		
    // create the analysis info object for the upper	
	cusparseSolveAnalysisInfo_t Analy_up = 0;
	status = cusparseCreateSolveAnalysisInfo(&Analy_up);
    if (checkStatus (status, (char*)"cusparseCreateSolveAnalysisInfo")) return EXIT_FAILURE;	

    // Perform the analysis for the upper
    status = cusparseDcsrsv_analysis(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, A.num_rows, descrM_up,
	thrust::raw_pointer_cast(&A_up.values[0]),
	thrust::raw_pointer_cast(&A_up.row_offsets[0]),
	thrust::raw_pointer_cast(&A_up.column_indices[0]),	
	Analy_up);
    if (checkStatus(status, (char*)"cusparseDcsrv_analysis")) return EXIT_FAILURE;

	std::cout << "\n triangle analysis in " << t1.milliseconds_elapsed()/1000. << " s " << std::endl;		
	/*-------------------cuSPARSE routines end----------------------*/
	
	ValueType *d_zm1, *d_zm2, *d_rm2, *d_y;		
	cudaMalloc((void**)&d_zm1, (A.num_rows)*sizeof(ValueType));
    cudaMalloc((void**)&d_zm2, (A.num_rows)*sizeof(ValueType));
    cudaMalloc((void**)&d_rm2, (A.num_rows)*sizeof(ValueType));	
	if (!use_mkl_sv) {
    cudaMalloc((void**)&d_y, A.num_rows*sizeof(ValueType));  
	}
	ValueType *h_r,*h_y,*h_zm1;
	if (use_mkl_sv) {
	/*
	h_r = (ValueType *) malloc ((A.num_rows)*sizeof(ValueType));
	h_y = (ValueType *) malloc ((A.num_rows)*sizeof(ValueType));
	h_zm1 = (ValueType *) malloc ((A.num_rows)*sizeof(ValueType));	
	*/
	cudaMallocHost((void **) &h_r, (A.num_rows)*sizeof(ValueType));
	cudaMallocHost((void **) &h_y, (A.num_rows)*sizeof(ValueType));
	cudaMallocHost((void **) &h_zm1, (A.num_rows)*sizeof(ValueType));
	}

	timer t2;	
	k = 0;
    r1 = cublasDdot(A.num_rows, d_r, 1, d_r, 1);
	r_ini = r1;
    while (r1 > tol*tol*r_ini && k <= max_iter) {
		if (use_mkl_sv) {
		// only CPU memory is needed, copy back from device
		cudaMemcpy(h_r, d_r, (A.num_rows)*sizeof(ValueType), cudaMemcpyDeviceToHost);
		cvar1='L';
		cvar='N';
		cvar2='U';
		mkl_dcsrtrsv(&cvar1,&cvar,&cvar2,&ivar,bilut,ibilut,jbilut,h_r,h_y);
		cvar1='U';
		cvar='N';
		cvar2='N';
		mkl_dcsrtrsv(&cvar1,&cvar,&cvar2,&ivar,bilut,ibilut,jbilut,h_y,h_zm1);
		// copy the result to device
		cudaMemcpy(d_zm1, h_zm1, (A.num_rows)*sizeof(ValueType), cudaMemcpyHostToDevice);
		} else {
        // Forward Solve: L y = r
        status = cusparseDcsrsv_solve(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, A.num_rows, 1.0, descrM_lo, 
									thrust::raw_pointer_cast(&A_lo.values[0]),
									thrust::raw_pointer_cast(&A_lo.row_offsets[0]),
									thrust::raw_pointer_cast(&A_lo.column_indices[0]),	
									Analy_lo, d_r, d_y);    
        if (checkStatus(status, (char*)"cusparseDcsrv_solve")) return EXIT_FAILURE;
        // Back Substitution: U z = y
        status = cusparseDcsrsv_solve(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, A.num_rows, 1.0, descrM_up, 
									thrust::raw_pointer_cast(&A_up.values[0]),
									thrust::raw_pointer_cast(&A_up.row_offsets[0]),
									thrust::raw_pointer_cast(&A_up.column_indices[0]),	
									Analy_up, d_y, d_zm1);  
        if (checkStatus(status, (char*)"cusparseDcsrv_solve")) return EXIT_FAILURE;
        } 
        k++;
        if ( k == 1 ) {
            cublasDcopy (A.num_rows, d_zm1, 1, d_p, 1);
        }
        else {
            beta = cublasDdot(A.num_rows, d_r, 1, d_zm1, 1)/cublasDdot(A.num_rows, d_rm2, 1, d_zm2, 1);
            cublasDscal(A.num_rows, beta, d_p, 1);    
            cublasDaxpy (A.num_rows, 1.0, d_zm1, 1, d_p, 1) ;
        }
        cusparseDcsrmv(handle,CUSPARSE_OPERATION_NON_TRANSPOSE, A.num_rows, A.num_rows, 1.0, descr,
		d_val, d_row, d_col, d_p, 0.0, d_omega);
        alpha = cublasDdot(A.num_rows, d_r, 1, d_zm1, 1)/cublasDdot (A.num_rows, d_p, 1, d_omega, 1);
        cublasDaxpy (A.num_rows, alpha, d_p, 1, d_x, 1);
        cublasDcopy (A.num_rows, d_r, 1, d_rm2, 1);
        cublasDcopy (A.num_rows, d_zm1, 1, d_zm2, 1);
        cublasDaxpy (A.num_rows, -alpha, d_omega, 1, d_r, 1);
        r1 = cublasDdot(A.num_rows, d_r, 1, d_r, 1);
        printf("  iteration = %3d, residual = %e \n", k, sqrt(r1)/sqrt(r_ini));
    }
	std::cout << "\n solve with preconditioner ilut in " << t2.milliseconds_elapsed()/1000. << " s " << std::endl;
	
		// check solution
        cusp::array1d<ValueType, MemorySpace> e(A.num_rows);
		cusp::blas::axpby(x_solution, x, e, ValueType(1), ValueType(-1));
		ValueType e_norm = cusp::blas::nrm2(e);
		printf("\n check solution: ||x*-x||: %f \n", e_norm );	
		
		cusp::array1d<ValueType, MemorySpace> ax(A.num_rows);
		cusp::array1d<ValueType, MemorySpace> r(A.num_rows);
		cusp::multiply(A, x, ax);
		cusp::blas::axpby(ax, b, r, ValueType(1), ValueType(-1));
		ValueType r_norm = cusp::blas::nrm2(r);
		ValueType b_norm = cusp::blas::nrm2(b);		
		printf("\n ||b-Ax||/||b||: %f \n", r_norm/b_norm);

	// free host vectors
	/*
	delete bilut;
	delete ibilut;
	delete jbilut;	
	delete h_val;
	delete h_row;
	delete h_col;
	*/
	cudaFreeHost(bilut);
	cudaFreeHost(ibilut);
	cudaFreeHost(jbilut);
	cudaFreeHost(h_val);
	cudaFreeHost(h_row);
	cudaFreeHost(h_col);
	
	if (use_mkl_sv) {
	/*
	delete h_r;
	delete h_y;
	delete h_zm1;
	*/
	cudaFreeHost(h_r);
	cudaFreeHost(h_y);
	cudaFreeHost(h_zm1);
	}	
	
    cudaFree(d_zm1);
    cudaFree(d_zm2);
    cudaFree(d_rm2);
	if (!use_mkl_sv) {
    cudaFree(d_y);
	}
	
      status = cusparseDestroyMatDescr(descr);
	  if (checkStatus(status, (char*)"cusparseDestroyMatDescr")) return EXIT_FAILURE;
      status = cusparseDestroyMatDescr(descrM_lo);
	  if (checkStatus(status, (char*)"cusparseDestroyMatDescr")) return EXIT_FAILURE;
      status = cusparseDestroySolveAnalysisInfo(Analy_lo);
	  if (checkStatus(status, (char*)"cusparseDestroySolveAnalysisInfo")) return EXIT_FAILURE;	  
      status = cusparseDestroyMatDescr(descrM_up);
	  if (checkStatus(status, (char*)"cusparseDestroyMatDescr")) return EXIT_FAILURE;
      status = cusparseDestroySolveAnalysisInfo(Analy_up);
	  if (checkStatus(status, (char*)"cusparseDestroySolveAnalysisInfo")) return EXIT_FAILURE;	 
	}
	
    cusparseDestroy(handle);
    cudaFree(d_p);
    cudaFree(d_omega);
}
