


#include <iostream>
#include <cusp/io/matrix_market.h>
#include <cusp/monitor.h>
#include <cusp/krylov/gmres.h>
#include <cusp/relaxation/jacobi.h>
#include <cusp/precond/diagonal.h>
#include <cusp/gallery/poisson.h>
#include <performance/timer.h>
#include <cusp/print.h>
#define PI 3.1415926

#include <thrust/transform.h>

double bicubic_coefficients[16][16]={{1.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0},
										{0.0,0.0,0.0,0.0,1.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0},
										{-3.0,3.0,0.0,0.0,-2.0,-1.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0},
										{2.0,-2.0,0.0,0.0,1.0,1.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0},
										{0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,1.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0},
										{0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,1.0,0.0,0.0,0.0},
										{0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,-3.0,3.0,0.0,0.0,-2.0,-1.0,0.0,0.0},
										{0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,2.0,-2.0,0.0,0.0,1.0,1.0,0.0,0.0},
										{-3.0,0.0,3.0,0.0,0.0,0.0,0.0,0.0,-2.0,0.0,-1.0,0.0,0.0,0.0,0.0,0.0},
										{0.0,0.0,0.0,0.0,-3.0,0.0,3.0,0.0,0.0,0.0,0.0,0.0,-2.0,0.0,-1.0,0.0},
										{9.0,-9.0,-9.0,9.0,6.0,3.0,-6.0,-3.0,6.0,-6.0,3.0,-3.0,4.0,2.0,2.0,1.0},
										{-6.0,6.0,6.0,-6.0,-3.0,-3.0,3.0,3.0,-4.0,4.0,-2.0,2.0,-2.0,-2.0,-1.0,-1.0},
										{2.0,0.0,-2.0,0.0,0.0,0.0,0.0,0.0,1.0,0.0,1.0,0.0,0.0,0.0,0.0,0.0},
										{0.0,0.0,0.0,0.0,2.0,0.0,-2.0,0.0,0.0,0.0,0.0,0.0,1.0,0.0,1.0,0.0},
										{-6.0,6.0,6.0,-6.0,-4.0,-2.0,4.0,2.0,-3.0,3.0,-3.0,3.0,-2.0,-1.0,-2.0,-1.0},
										{4.0,-4.0,-4.0,4.0,2.0,2.0,-2.0,-2.0,2.0,-2.0,2.0,-2.0,1.0,1.0,1.0,1.0}};

// where to perform the computation
typedef cusp::device_memory MemorySpace;
// which floating point type to use
typedef double ValueType;
typedef int   IndexType;

//#include <cusp/elementwise.h>

template <typename Monitor>
void report_status(Monitor& monitor)
{
    if (monitor.converged())
    {
        std::cout << "  Solver converged to " << monitor.tolerance1() << " tolerance";
        std::cout << " after " << monitor.iteration_count() << " iterations";
        std::cout << " (" << monitor.relative_residual_norm() << " final residual)" << std::endl;
    }
    else
    {
        std::cout << "  Solver reached iteration limit " << monitor.iteration_limit() << " before converging";
        std::cout << " to " << monitor.tolerance1() << " tolerance ";
        std::cout << " (" << monitor.relative_residual_norm() << " final residual)" << std::endl;
    }
}

template <typename IndexType>
struct red_black_multicoloring_functor
{

    red_black_multicoloring_functor() {}

    template <typename Tuple>
    __host__ __device__
    IndexType operator()(const Tuple& t)
    {
        const IndexType num = thrust::get<0>(t);
        const IndexType m = thrust::get<1>(t); // number of columns
        const IndexType n = thrust::get<2>(t); // number of rows
        const IndexType dof = thrust::get<3>(t); // degree of freedom

		const IndexType nd = num / dof;  // scalar mesh point
		const IndexType j = nd / m;      // row index
		const IndexType i = nd % m;		 // col index
        return (j%2)^(i%2);
    }
};

template< typename Array1d, typename IndexType >
void red_black_multicoloring(Array1d &coloring, const IndexType m, const IndexType n, const IndexType dof)
{
    CUSP_PROFILE_SCOPED();	

    thrust::transform(thrust::make_zip_iterator(thrust::make_tuple(coloring.begin(), thrust::make_constant_iterator(m), thrust::make_constant_iterator(n), thrust::make_constant_iterator(dof))),
                      thrust::make_zip_iterator(thrust::make_tuple(coloring.end(),   thrust::make_constant_iterator(m), thrust::make_constant_iterator(n), thrust::make_constant_iterator(dof))),
                      coloring.begin(),
                      red_black_multicoloring_functor<IndexType>());	
}

// compute bicubic coefficient c[4][4]
template <typename ValueType, typename IndexType>
void bicubic_coefficient(ValueType phy[], ValueType dphy1[], ValueType dphy2[], ValueType dphy12[], ValueType coe[][4])
{
	IndexType i,j,k,l;
	ValueType xx,cl[16],x[16];
	
	for (i=0; i<4; i++) {
		x[i]    = phy[i];
		x[i+4]  = dphy1[i];	
        x[i+8]  = dphy2[i];
		x[i+12] = dphy12[i];		
	}

	for (i=0; i<16; i++) {
		xx = 0.0;
		for (k=0; k<16; k++) {
			xx = xx + bicubic_coefficients[i][k]*x[k];
		}
		cl[i] = xx;
	}
	
	l = 0;
	for (i=0; i<4; i++) {
		for (j=0; j<4; j++) {
			coe[i][j] = cl[l];
			l++;
		}
	}
}

// bicubic interpolation in [0,1]x[0,1]
template <typename ValueType, typename IndexType>
ValueType bicubic(ValueType phy[], ValueType dphy1[], ValueType dphy2[], ValueType dphy12[], ValueType x, ValueType y)
{
	IndexType i;
	ValueType phy_inter, coe[4][4];
	
	bicubic_coefficient<ValueType, IndexType>(phy, dphy1, dphy2, dphy12, coe); 
	
	phy_inter = 0.0;
	for (i=3; i>=0; i--) {
		phy_inter = x*phy_inter+((coe[i][3]*y+coe[i][2])*y+coe[i][1])*y+coe[i][0];
	}
	return phy_inter;
}

// Hermite interpolation in [0,1]
template <typename ValueType>
ValueType hermite(ValueType eta[2], ValueType deta[2], ValueType x)
{
	return  (x-1.0)*(x-1.0)*(1.0+2.0*x)*eta[0]+
	        x*x*(3.0-2.0*x)*eta[1]+
			x*(x-1.0)*(x-1.0)*deta[0]+
			x*x*(x-1.0)*deta[1];
}

int main(void)
{

	
    // create an empty sparse matrix structure.
    typedef cusp::csr_matrix<int, ValueType, MemorySpace> Matrix;
	
    // create an empty sparse matrix structure.
    Matrix A;
    // load a matrix stored in MatrixMarket format
    //cusp::io::read_matrix_market_file(A, "A.mtx");

    // create 2D Poisson problem
	// load a matrix stored in MatrixMarket format
	#if 0
	std::string mtx="/home/liluo/UF_collection/chem_master1.mtx";
	//printf("Please input matrix : ");
	//std::cin >> mtx;
	timer t1;		
    cusp::io::read_matrix_market_file(A, mtx);
	std::cout << "\nread matrix in " << t1.milliseconds_elapsed() << " ms " << std::endl;
	#else
	int m = 129, n = 129;
/*	
while(m < 2 || n < 2)
    {
    	printf("Please input m, n: ");
    	scanf("%d %d", &m, &n);
    }
*/
    cusp::gallery::poisson5pt(A, m, n);
	#endif
	
	// allocate storage for solution (x) and right hand side (b)	
	cusp::array1d<ValueType, MemorySpace> x_solution(A.num_rows, 1);
    // initialize starting vector to random values in [0,1)
    //cusp::copy(cusp::detail::random_reals<ValueType>(A.num_rows), x_solution);
	ValueType *x_so_cpu;	
	cudaMallocHost((void **) &x_so_cpu, (A.num_rows)*sizeof(ValueType));
		
	for(int i=0;i<A.num_rows;i++)	
	{
		x_so_cpu[i]=std::sin(0.5*PI*i/A.num_rows);
	}
	ValueType *x_so_gpu = thrust::raw_pointer_cast(&x_solution[0]);
	cudaMemcpy(x_so_gpu, x_so_cpu, (A.num_rows)*sizeof(ValueType), cudaMemcpyHostToDevice);
	cudaFreeHost(x_so_cpu);
	
    cusp::array1d<ValueType, MemorySpace> b(A.num_rows);
	cusp::multiply(A, x_solution, b);
	

    // solve with diagonal preconditioner
	{
        std::cout << "\nSolving with diagonal preconditioner (M = D^-1)" << std::endl;
        
        cusp::array1d<ValueType, MemorySpace> x(A.num_rows, 0);

        // set stopping criteria (iteration_limit = 10000, relative_tolerance = 1e-6)
        cusp::verbose_monitor<ValueType> monitor(b, 10000, 1e-6);

        // setup preconditioner
        cusp::precond::diagonal<ValueType, MemorySpace> M(A);

        // solve			
		int restart = 40;
		//while(restart < 1)
		//{
		//	printf("\nPlease input restart: ");
		//	scanf("%d", &restart);
		//}
		timer t1;	
		cusp::krylov::gmres(A, x, b, restart, monitor, M);
		std::cout << "\nsolve by jacobi-gmres in " << t1.seconds_elapsed() << " s " << std::endl;	
		
		// report status
        report_status(monitor);
		
		// check solution
        cusp::array1d<ValueType, MemorySpace> e(A.num_rows);
		blas::axpby(x_solution, x, e, ValueType(1), ValueType(-1));
		ValueType e_norm = cusp::blas::nrm2(e);
		printf("\n check solution: ||x*-x||: %f \n", e_norm );	
		
		cusp::array1d<ValueType, MemorySpace> ax(x.size());
		cusp::array1d<ValueType, MemorySpace> r(A.num_rows);
		cusp::multiply(A, x, ax);
		cusp::blas::axpby(ax, b, r, ValueType(1), ValueType(-1));
		ValueType r_norm = cusp::blas::nrm2(r);
		ValueType b_norm = cusp::blas::nrm2(b);		
		printf("\n ||b-Ax||/||b||: %f \n", r_norm / b_norm );
    }

	/*
	IndexType dof=3;
	cusp::array1d<IndexType, MemorySpace> coloring(m*n*dof);
	thrust::sequence(coloring.begin(), coloring.end());
	cusp::print(coloring);	
	timer t1;	
	red_black_multicoloring(coloring, m, n, dof);
	std::cout << "\n time:  " << t1.seconds_elapsed() << " s " << std::endl;		
	cusp::print(coloring);
	*/
	
	ValueType y[4],y1[4],y2[4],y12[4];
	y[0]=0.;y[1]=0.;y[2]=1.;y[3]=0.;
	y1[0]=0.;y1[1]=0.;y1[2]=1.;y1[3]=1.;	
	y2[0]=0.;y2[1]=0.;y2[2]=2.;y2[3]=0.;	
	y12[0]=0.;y12[1]=0.;y12[2]=2.;y12[3]=2.;
	ValueType ans_bicubic = bicubic<ValueType, IndexType>(y, y1, y2, y12, 0.5, 0.5);
	printf("\n %f\n", ans_bicubic);
	
	ValueType xx[2], xx1[2], pi=3.1415926;
	xx[0] = 0.; xx[1] = 1.;
	xx1[0] = 0.; xx1[1] = 2.;
	ValueType ans_hermite = hermite(xx, xx1, 0.5);
	printf("\n %f\n", ans_hermite);
	
    return 0;
}

