


#include <iostream>
#include <cusp/io/matrix_market.h>
#include <cusp/gallery/poisson.h>
#include <performance/timer.h>
#include <cusp/print.h>
#define PI 3.1415926

#include <cusp/detail/utils_.h>
#include <thrust/transform.h>

template <typename IndexType>
struct stencil_coarsening_functor
{
    template <typename Tuple>
    __host__ __device__
    void operator()(Tuple t)
    {
        const IndexType num = thrust::get<0>(t);
        const IndexType m = thrust::get<1>(t); // number of columns in fine mesh
        const IndexType n = thrust::get<2>(t); // number of rows in fine mesh
        const IndexType m_c = thrust::get<3>(t); // number of columns in fine mesh
        const IndexType n_c = thrust::get<4>(t); // number of rows in fine mesh		
        const IndexType dof = thrust::get<5>(t); // degree of freedom
        const IndexType ratioi = thrust::get<6>(t); // refine ratio of columns	
        const IndexType ratioj = thrust::get<7>(t); // refine ratio of rows
		
		const IndexType nd = num / dof;  // scalar mesh point
		const IndexType nd_k = num % dof; // the field
		const IndexType j = nd / m;      // row index
		const IndexType i = nd % m;		 // col index
		
        const IndexType j_c = (j/ratioj);    // coarse grid index	
		const IndexType i_c = (i/ratioi);    // coarse grid index

        thrust::get<8>(t) = (j_c*ratioj == j)&&(i_c*ratioi == i); // right on the coarse point, return 0 or 1
        thrust::get<9>(t) = thrust::get<8>(t) * (j_c*m_c*dof+i_c*dof+nd_k); // if on the coarse point, retrun the index of coarse mesh
	}
};

template< typename Array1d, typename IndexType >
void stencil_coarsening(Array1d &stencil, Array1d &map, const IndexType m, const IndexType n, const IndexType m_c, const IndexType n_c, const IndexType dof, const IndexType ratioi, const IndexType ratioj)
{
    CUSP_PROFILE_SCOPED();	
	
	Array1d sequence_temp(m*n*dof);	
	thrust::sequence(sequence_temp.begin(), sequence_temp.end());
	
    thrust::for_each(thrust::make_zip_iterator(thrust::make_tuple(sequence_temp.begin(),
												thrust::make_constant_iterator(m),
												thrust::make_constant_iterator(n),
												thrust::make_constant_iterator(m_c),
												thrust::make_constant_iterator(n_c),
												thrust::make_constant_iterator(dof),
												thrust::make_constant_iterator(ratioi),
												thrust::make_constant_iterator(ratioj),
												stencil.begin(),
												map.begin())),
                     thrust::make_zip_iterator(thrust::make_tuple(sequence_temp.end(),				  
												thrust::make_constant_iterator(m),
												thrust::make_constant_iterator(n),
												thrust::make_constant_iterator(m_c),
												thrust::make_constant_iterator(n_c),
												thrust::make_constant_iterator(dof),
												thrust::make_constant_iterator(ratioi),
												thrust::make_constant_iterator(ratioj),
												stencil.end(),
												map.end())),
                     stencil_coarsening_functor<IndexType>());	
}

template< typename Matrix, typename Array1d >
void matrix_coarsening(const Matrix &A, Array1d &stencil, Array1d &map, Matrix &RAP, cusp::coo_format)
{
    CUSP_PROFILE_SCOPED();	
	
    typedef typename Matrix::index_type  IndexType;
	IndexType num_rows = A.num_rows;
	IndexType num_entries = A.num_entries;	
	IndexType num_rows_c = RAP.num_rows;
	
	// row_indices --> nnz
	Array1d nnz(num_rows);
	cusp::detail::indices_to_nnz(A.row_indices, nnz);
	cusp::print(nnz);

	Array1d nnz_c(num_rows_c);	
	thrust::scatter_if(nnz.begin(), nnz.end(), map.begin(), stencil.begin(), nnz_c.begin());
	cusp::print(nnz_c);	
	
	// nnz --> row_offsets on coarse
	Array1d row_offsets_c(num_rows_c+1,0);
	cusp::detail::nnz_to_offsets(nnz_c, row_offsets_c);	
	
	// resize coarse matrix
	IndexType num_entries_c = row_offsets_c[num_rows_c];		
	RAP.resize(num_rows_c,num_rows_c,num_entries_c);
	
	// row_offsets --> row_indices on coarse		
	cusp::detail::offsets_to_indices(row_offsets_c, RAP.row_indices);	
	cusp::print(RAP.row_indices);
	
	// row_indices --> row_offsets on fine
	Array1d row_offsets(num_rows+1,0);
	cusp::detail::indices_to_offsets(A.row_indices, row_offsets);
	
	// stencil for scattering COO matrix column_indices
	Array1d stencil_column_indices(num_entries,0);
    thrust::scatter(stencil.begin(),stencil.end(),row_offsets.begin(),stencil_column_indices.begin());	
    // extend
	thrust::inclusive_scan_by_key(A.row_indices.begin(), A.row_indices.end(), stencil_column_indices.begin(), stencil_column_indices.begin());  
	cusp::print(stencil_column_indices);

	// map for scattering COO matrix column_indices
	Array1d map_column_indices(num_entries,0);
	// input: 11100001100011  output: 12333334555567
	thrust::inclusive_scan(stencil_column_indices.begin(), stencil_column_indices.end(), map_column_indices.begin());	
	// input: 12333334555567  output: 01222223444456
	thrust::transform(map_column_indices.begin(), map_column_indices.end(), map_column_indices.begin(), cusp::detail::ADD<IndexType>(-1));	
	// input: 01222223444456  output: 01200003400056   
	thrust::transform(map_column_indices.begin(), map_column_indices.end(), stencil_column_indices.begin(), map_column_indices.begin(), thrust::multiplies<IndexType>());
	cusp::print(map_column_indices);
	
	// scatter
	thrust::scatter_if(A.column_indices.begin(), A.column_indices.end(), map_column_indices.begin(), stencil_column_indices.begin(), RAP.column_indices.begin());
	thrust::scatter_if(A.values.begin(), A.values.end(), map_column_indices.begin(), stencil_column_indices.begin(), RAP.values.begin());	
	cusp::print(RAP);
}

/* obtain coarse matrix RAP from fine matrix A on structure mesh */
template< typename Matrix, typename Array1d >
void matrix_coarsening(const Matrix &A, Array1d &stencil, Array1d &map, Matrix &RAP)
{
    CUSP_PROFILE_SCOPED();	
	
	matrix_coarsening(A, stencil, map, RAP, typename Matrix::format( ));
}

int main(void)
{
// where to perform the computation
typedef cusp::device_memory MemorySpace;
// which floating point type to use
typedef double ValueType;
typedef int   IndexType;
	
    // create an empty sparse matrix structure.
    typedef cusp::coo_matrix<int, ValueType, MemorySpace> Matrix;
	
    // create an empty sparse matrix structure.
    Matrix A;
    // load a matrix stored in MatrixMarket format
    //cusp::io::read_matrix_market_file(A, "A.mtx");

    // create 2D Poisson problem
	// load a matrix stored in MatrixMarket format
	#if 0
	std::string mtx="/home/liluo/UF_collection/chem_master1.mtx";
	//printf("Please input matrix : ");
	//std::cin >> mtx;
	timer t1;		
    cusp::io::read_matrix_market_file(A, mtx);
	std::cout << "\nread matrix in " << t1.milliseconds_elapsed() << " ms " << std::endl;
	#else
	int mx = 0, my = 0;
	while(mx < 2 || my < 2)
    {
    	printf("Please input mx, my: ");
    	scanf("%d %d", &mx, &my);
    }
    cusp::gallery::poisson5pt(A, mx, my);
	#endif
	
	int ratioi = 0, ratioj = 0;
	while(ratioi < 2 || ratioj < 2)
    {
    	printf("Please input ratioi, ratioj: ");
    	scanf("%d %d", &ratioi, &ratioj);
    }	
	IndexType dof=0;
	while(dof<1)
	{
		printf("\nPlease input dof: ");
		scanf("%d", &dof);
	}
	/* size of coarse grid */
	IndexType Mx=(mx/ratioi)+((mx%ratioi)?1:0), My=(my/ratioj)+((my%ratioj)?1:0);
	
	/* create stencil and map */	
	cusp::array1d<IndexType, MemorySpace> stencil(mx*my*dof);
	cusp::array1d<IndexType, MemorySpace> map(mx*my*dof);	
	timer t1;	
	stencil_coarsening(stencil, map, mx, my, Mx, My, dof, ratioi, ratioj);
	std::cout << "\n time:  " << t1.seconds_elapsed() << " s " << std::endl;		
	cusp::print(stencil);
	cusp::print(map);	

	
	cusp::array1d<ValueType, MemorySpace> vector(mx*my*dof);
	thrust::sequence(vector.begin(), vector.end());	
	thrust::transform(vector.begin(), vector.end(), vector.begin(), cusp::detail::ADD<ValueType>(0.5));	
	cusp::print(vector);	
	
	cusp::array1d<ValueType, MemorySpace> vector_c(Mx*My*dof);
	/* perform restriction */
	thrust::scatter_if(vector.begin(), vector.end(), map.begin(), stencil.begin(), vector_c.begin());
	cusp::print(vector_c);	
	
	/*----------------------------------------------------*/
    //Matrix RAP(Mx*My*dof,Mx*My*dof,1);
	//matrix_coarsening(A, stencil, map, RAP);
	
    return 0;
}

