#include <iostream>
#include <cusp/io/matrix_market.h>
#include <cusp/monitor.h>
#include <cusp/precond/diagonal.h>
#include <cusp/detail/spectral_radius.h>
#include <cusp/krylov/cg.h>
#include <cusp/krylov/gmres.h>
#include <cusp/krylov/bicgstab.h>
#include <cusp/relaxation/jacobi_.h>
#include <cusp/relaxation/mc_ssor.h>
#include <cusp/gallery/poisson.h>
#include <performance/timer.h>
#include <cusp/print.h>
#include <cmath>
#define PI 3.1415926


template <typename Monitor>
void report_status(Monitor& monitor)
{
    if (monitor.converged())
    {
        std::cout << "  Solver converged to " << monitor.tolerance1() << " tolerance";
        std::cout << " after " << monitor.iteration_count() << " iterations";
        std::cout << " (" << monitor.relative_residual_norm() << " final residual)" << std::endl;
    }
    else
    {
        std::cout << "  Solver reached iteration limit " << monitor.iteration_limit() << " before converging";
        std::cout << " to " << monitor.tolerance1() << " tolerance ";
        std::cout << " (" << monitor.relative_residual_norm() << " final residual)" << std::endl;
    }
}

template <typename MatrixType>
struct Dinv_A : public cusp::linear_operator<typename MatrixType::value_type, typename MatrixType::memory_space>
{
    const MatrixType& A;
    const cusp::precond::diagonal<typename MatrixType::value_type, typename MatrixType::memory_space> Dinv;

    Dinv_A(const MatrixType& A)
        : A(A), Dinv(A),
          cusp::linear_operator<typename MatrixType::value_type, typename MatrixType::memory_space>(A.num_rows, A.num_cols, A.num_entries + A.num_rows)
          {}

    template <typename Array1, typename Array2>
    void operator()(const Array1& x, Array2& y) const
    {
        cusp::multiply(A,x,y);
        cusp::multiply(Dinv,y,y);
    }
};

template <typename MatrixType>
double estimate_rho_Dinv_A(const MatrixType& A)
{
    typedef typename MatrixType::value_type   ValueType;
    typedef typename MatrixType::memory_space MemorySpace;

    Dinv_A<MatrixType> Dinv_A(A);

    return cusp::detail::ritz_spectral_radius(Dinv_A, 8);
}

template <typename ValueType>
struct mc_ssor_postsmooth_functor
{
    ValueType omega;

    mc_ssor_postsmooth_functor(ValueType omega) : omega(omega) {}

    template <typename Tuple>
    __host__ __device__
    ValueType operator()(const Tuple& t)
    {
        const ValueType x = thrust::get<0>(t);
        const ValueType d = thrust::get<1>(t);
        const ValueType b = thrust::get<2>(t);
        const ValueType y = thrust::get<3>(t);

        return x + omega * (b - y) / d;
    }
};

int main(void)
{
	// where to perform the computation
	typedef cusp::device_memory MemorySpace;
	// which floating point type to use
	typedef double ValueType;
	
    // create an empty sparse matrix structure.
    typedef cusp::csr_matrix<int, ValueType, MemorySpace> Matrix;
	
    // create an empty sparse matrix structure.
    Matrix A;

        bool matrix_market=false;
		int mm=-1;
		while(mm<0)
		{
			printf("Please input mm: 0: poisson2d,  1: UF collection \n");
			scanf("%d", &mm);
		}
		if(mm) {
		matrix_market=true;
		} else {
		matrix_market=false;		
		}
		
	std::string mtx="5pt_10x10.mtx";
	int m = 0, n = 0;
	
	if (matrix_market)
	{
    // load a matrix stored in MatrixMarket format
	printf("Please input matrix : ");
	std::cin >> mtx;
    cusp::io::read_matrix_market_file(A, mtx);	
	} else {
	while(m < 2 || n < 2)
    {
    	printf("Please input grid size (m, n): ");
    	scanf("%d %d", &m, &n);
    }	
    // create 2D Poisson problem
    cusp::gallery::poisson5pt(A, m, n);	
	}
	
	// allocate storage for solution (x) and right hand side (b)	
	cusp::array1d<ValueType, MemorySpace> x_solution(A.num_rows, 1);
    // initialize starting vector to random values in [0,1)
    //cusp::copy(cusp::detail::random_reals<ValueType>(A.num_rows), x_solution);
	ValueType *x_so_cpu;	
	cudaMallocHost((void **) &x_so_cpu, (A.num_rows)*sizeof(ValueType));
		
	for(int i=0;i<A.num_rows;i++)	
	{
		x_so_cpu[i]=std::sin(0.5*PI*i/A.num_rows);
	}
	ValueType *x_so_gpu = thrust::raw_pointer_cast(&x_solution[0]);
	cudaMemcpy(x_so_gpu, x_so_cpu, (A.num_rows)*sizeof(ValueType), cudaMemcpyHostToDevice);
	cudaFreeHost(x_so_cpu);
	
    cusp::array1d<ValueType, MemorySpace> b(A.num_rows);
	cusp::multiply(A, x_solution, b);
	
#if 0
    // Solve without preconditioning
    {
        std::cout << "\nSolving with no preconditioner" << std::endl;
    
        // allocate storage for solution (x) and right hand side (b)
        cusp::array1d<ValueType, MemorySpace> x(A.num_rows, 0);
        cusp::array1d<ValueType, MemorySpace> b(A.num_rows, 1);

        // set stopping criteria (iteration_limit = 100, relative_tolerance = 1e-6)
        cusp::default_monitor<ValueType> monitor(b, 1000, 1e-6);
        
        // solve
        cusp::krylov::cg(A, x, b, monitor);
		
		// report status
        report_status(monitor);

    }
#endif	

#if 0
    // solve with jacobi iteration
    {
        std::cout << "\nSolving with jacobi iteration" << std::endl;
		
        // allocate storage for solution (x) and right hand side (b)
        cusp::array1d<ValueType, MemorySpace> x(A.num_rows, 0);
		int sweeps=0; 
		while(sweeps < 1)
		{
			printf("\nPlease input sweeps: ");
			scanf("%d", &sweeps);
		}		
		ValueType rho_DinvA = estimate_rho_Dinv_A(A);
		ValueType omega = ValueType(4.0/3.0) / rho_DinvA;
		timer t0;	
		cusp::relaxation::jacobi<ValueType,MemorySpace> jacobi(true,A);
		jacobi.setup(A,omega,sweeps);
		std::cout << "\ncreate jacobi relaxation in " << t0.seconds_elapsed() << " s " << std::endl;	
		
		timer t1;		
		jacobi.postsmooth(A,b,x);
		std::cout << "\nsolve by jacobi relaxation in " << t1.milliseconds_elapsed() << " ms " << std::endl;	
		
		// check solution
        cusp::array1d<ValueType, MemorySpace> r(A.num_rows);
		cusp::blas::axpby(x_solution, x, r, ValueType(1), ValueType(-1));
		ValueType r_norm = cusp::blas::nrm2(r);
		printf("\n check solution: ||x*-x||: %f \n", r_norm );	
		
		cusp::array1d<ValueType, MemorySpace> ax(x.size());
		cusp::multiply(A, x, ax);
		cusp::blas::axpby(ax, b, r, ValueType(1), ValueType(-1));
		r_norm = cusp::blas::nrm2(r);
		printf("\n post: ||b-Ax||: %f \n", r_norm );	
    }
#endif

#if 0
    // solve with diagonal preconditioner
    {
        std::cout << "\nSolving with diagonal preconditioner (M = D^-1)" << std::endl;
        
        cusp::array1d<ValueType, MemorySpace> x(A.num_rows, 0);

        // set stopping criteria (iteration_limit = 10000, relative_tolerance = 1e-6)
        cusp::verbose_monitor<ValueType> monitor(b, 10000, 1e-6);

        // setup preconditioner
        cusp::precond::diagonal<ValueType, MemorySpace> M(A);

        // solve			
        //cusp::krylov::cg(A, x, b, monitor, M);
		int restart = 0;
		while(restart < 1)
		{
			printf("\nPlease input restart: ");
			scanf("%d", &restart);
		}
		timer t1;	
		cusp::krylov::gmres(A, x, b, restart, monitor, M);
		//cusp::krylov::bicgstab(A, x, b, monitor, M);
		std::cout << "\nsolve by jacobi-gmres in " << t1.seconds_elapsed() << " s " << std::endl;	
		
		// report status
        report_status(monitor);
		
		// check solution
        cusp::array1d<ValueType, MemorySpace> e(A.num_rows);
		blas::axpby(x_solution, x, e, ValueType(1), ValueType(-1));
		ValueType e_norm = cusp::blas::nrm2(e);
		printf("\n check solution: ||x*-x||: %f \n", e_norm );	
		
		cusp::array1d<ValueType, MemorySpace> ax(x.size());
		cusp::array1d<ValueType, MemorySpace> r(A.num_rows);
		cusp::multiply(A, x, ax);
		cusp::blas::axpby(ax, b, r, ValueType(1), ValueType(-1));
		ValueType r_norm = cusp::blas::nrm2(r);
		ValueType b_norm = cusp::blas::nrm2(b);		
		printf("\n ||b-Ax||/||b||: %f \n", r_norm / b_norm );
    }
#endif

    // solve with mc-ssor iteration
    {
        std::cout << "\nSolving with mc-ssor iteration" << std::endl;
		
        cusp::array1d<ValueType, MemorySpace> x(A.num_rows, 0);
		int sweeps=0; 
		while(sweeps < 1)
		{
			printf("\nPlease input sweeps: ");
			scanf("%d", &sweeps);
		}	
		        
		bool red_black=false;
		int rb=-1;
		while(rb<0)
		{
			printf("Please input : 0: not use,  1: use red black for coloring \n");
			scanf("%d", &rb);
		}
		if(rb) {
		red_black=true;
		} else {
		red_black=false;		
		}
		
		ValueType rho_DinvA = estimate_rho_Dinv_A(A);
		ValueType omega = ValueType(4.0/3.0) / rho_DinvA;
		int dof=1;
		
		//ValueType omega = 1.0;
		timer t0;	
		cusp::relaxation::mc_ssor<int,ValueType,MemorySpace> mc_ssor(true, red_black, A, m, n, dof);
		mc_ssor.setup(A, omega, 10, sweeps, false, true);
		std::cout << "\ncreate mc-ssor relaxation in " << t0.seconds_elapsed() << " s " << std::endl;		
	
		//mc_ssor.postsmooth(A,b,x);
		
		// solve the linear system A * x = b with the GMRES
		int restart = 0;
		while(restart < 1)
		{
			printf("\nPlease input restart: ");
			scanf("%d", &restart);
		}
		// set stopping criteria (iteration_limit = 10000, relative_tolerance = 1e-6)
        cusp::verbose_monitor<ValueType> monitor(b, 10000, 1e-6);		
		timer t1;	
		//cusp::krylov::gmres(A, x, b, restart, monitor, mc_ssor);
		//cusp::krylov::bicgstab(A, x, b, monitor, mc_ssor);
		cusp::krylov::cg(A, x, b, monitor, mc_ssor);
		std::cout << "\nsolve by mc_ssor-krylov in " << t1.seconds_elapsed() << " s " << std::endl;	

		// report status
        report_status(monitor);
		
		// check solution
        cusp::array1d<ValueType, MemorySpace> e(A.num_rows);
		blas::axpby(x_solution, x, e, ValueType(1), ValueType(-1));
		ValueType e_norm = cusp::blas::nrm2(e);
		printf("\n check solution: ||x*-x||: %f \n", e_norm );	
		
		cusp::array1d<ValueType, MemorySpace> ax(x.size());
		cusp::array1d<ValueType, MemorySpace> r(A.num_rows);
		cusp::multiply(A, x, ax);
		cusp::blas::axpby(ax, b, r, ValueType(1), ValueType(-1));
		ValueType r_norm = cusp::blas::nrm2(r);
		ValueType b_norm = cusp::blas::nrm2(b);		
		printf("\n ||b-Ax||/||b||: %f \n", r_norm / b_norm );
    }
	
    return 0;
}

