#include <cusp/dia_matrix.h>
#include <cusp/precond/multigrid.h>
#include <cusp/krylov/cg.h>
#include <cusp/krylov/gmres.h>
#include <cusp/krylov/bicgstab.h>
#include <cusp/gallery/poisson.h>
#include <cusp/csr_matrix.h>
#include <cusp/io/matrix_market.h>
#include <performance/timer.h>
#include <cmath>
#define PI 3.1415926

#include <iostream>

template <typename Monitor>
void report_status(Monitor& monitor)
{
    if (monitor.converged())
    {
        std::cout << "  Solver converged to " << monitor.tolerance1() << " tolerance";
        std::cout << " after " << monitor.iteration_count() << " iterations";
        std::cout << " (" << monitor.relative_residual_norm() << " final residual)" << std::endl;
    }
    else
    {
        std::cout << "  Solver reached iteration limit " << monitor.iteration_limit() << " before converging";
        std::cout << " to " << monitor.tolerance1() << " tolerance ";
        std::cout << " (" << monitor.relative_residual_norm() << " final residual)" << std::endl;
    }
}


int main(void)
{
    typedef int                 IndexType;
    typedef double               ValueType;
    typedef cusp::device_memory   MemorySpace;

    PetscInitialize((int *)0,(char ***)0,(char *)0,(char *)0);			
    // create an empty sparse matrix structure
    cusp::csr_matrix<IndexType, ValueType, MemorySpace> A;
	
        bool matrix_market=false;
		int mm=-1;
		while(mm<0)
		{
			printf("Please input mm: 0: poisson2d,  1: UF collection \n");
			scanf("%d", &mm);
		}
		if(mm) {
		matrix_market=true;
		} else {
		matrix_market=false;		
		}
		
	std::string mtx="5pt_10x10.mtx";
	int m = 0, n = 0;
	
	if (matrix_market)
	{
    // load a matrix stored in MatrixMarket format
	printf("Please input matrix : ");
	std::cin >> mtx;
    cusp::io::read_matrix_market_file(A, mtx);	
	} else {
	while(m < 2 || n < 2)
    {
    	printf("Please input grid size (m, n): ");
    	scanf("%d %d", &m, &n);
    }	
    // create 2D Poisson problem
    cusp::gallery::poisson5pt(A, m, n);	
	}

	// allocate storage for solution (x) and right hand side (b)	
	cusp::array1d<ValueType, MemorySpace> x_solution(A.num_rows, 1);
    // initialize starting vector to random values in [0,1)
    //cusp::copy(cusp::detail::random_reals<ValueType>(A.num_rows), x_solution);

	ValueType *x_so_cpu;	
	cudaMallocHost((void **) &x_so_cpu, (A.num_rows)*sizeof(ValueType));
		
	for(int i=0;i<A.num_rows;i++)	
	{
		x_so_cpu[i]=std::sin(0.5*PI*i/A.num_rows);
	}
	ValueType *x_so_gpu = thrust::raw_pointer_cast(&x_solution[0]);
	cudaMemcpy(x_so_gpu, x_so_cpu, (A.num_rows)*sizeof(ValueType), cudaMemcpyHostToDevice);
	cudaFreeHost(x_so_cpu);

    cusp::array1d<ValueType, MemorySpace> b(A.num_rows);


		int nk=0;
		while(nk < 1)
		{
			printf("\nPlease input test times: ");
			scanf("%d", &nk);
		}
	for(int k=0;k<nk;k++)
    // solve with multigrid algebraic multigrid preconditioner
    {

		typedef cusp::precond::jacobi_policy<IndexType,ValueType,MemorySpace> smoother_jacobi;
		typedef cusp::precond::mc_ssor_policy<IndexType,ValueType,MemorySpace> smoother_mc_ssor;
		typedef cusp::precond::polynomial_policy<IndexType,ValueType,MemorySpace> smoother_polynomial;
		typedef cusp::precond::hierarchical_as_policy<IndexType,ValueType,MemorySpace> smoother_hierarchical_as;
		typedef cusp::precond::mkl_ilu0_policy<IndexType,ValueType,MemorySpace> smoother_mkl_ilu0;
		typedef cusp::precond::sparskit2_ilu_policy<IndexType,ValueType,MemorySpace> smoother_sparskit2_ilu;
		typedef cusp::precond::ilupp_policy<IndexType,ValueType,MemorySpace> smoother_ilupp;
		typedef cusp::precond::petsc_factor_policy<IndexType,ValueType,MemorySpace> smoother_petsc_factor;
		
		typedef cusp::precond::multigrid<IndexType, ValueType, MemorySpace, smoother_jacobi> multigrid_jacobi;
		typedef cusp::precond::multigrid<IndexType, ValueType, MemorySpace, smoother_mc_ssor> multigrid_mc_ssor;
		typedef cusp::precond::multigrid<IndexType, ValueType, MemorySpace, smoother_polynomial> multigrid_polynomial;
		typedef cusp::precond::multigrid<IndexType, ValueType, MemorySpace, smoother_hierarchical_as> multigrid_as;
		typedef cusp::precond::multigrid<IndexType, ValueType, MemorySpace, smoother_mkl_ilu0> multigrid_mkl_ilu0;
		typedef cusp::precond::multigrid<IndexType, ValueType, MemorySpace, smoother_sparskit2_ilu> multigrid_sparskit2_ilu;	
		typedef cusp::precond::multigrid<IndexType, ValueType, MemorySpace, smoother_ilupp> multigrid_ilupp;	
		typedef cusp::precond::multigrid<IndexType, ValueType, MemorySpace, smoother_petsc_factor> multigrid_petsc_factor;	

		
		void *M;
	
        std::cout << "\nSolving with multigrid preconditioner..." << std::endl;
        bool use_bilinear=false;
		int ub=-1;
		while(ub<0)
		{
			printf("\nPlease input iterpolation type: 0: use smoothed aggregation,  1: use bilinear\n");
			scanf("%d", &ub);
		}
		if(ub) {
		use_bilinear=true;
		} else {
		use_bilinear=false;		
		}
		bool no_postsmooth = false;
		int np=-1;
		while(np<0)
		{
			printf("\nPlease input: 0: with postsmooth,  1: without postsmooth\n");
			scanf("%d", &np);
		}
		if(np) {
		no_postsmooth=true;
		} else {
		no_postsmooth=false;		
		}
		
		int mg_type = -1;
		while(mg_type < 0)
		{
			printf("\nPlease input mg type: 0: multiplicative,  1: cascade,  2: full ");
			scanf("%d", &mg_type);
		}		
		
        // setup preconditioner
		int levels=0;
		while(levels < 1)
		{
			printf("\nPlease input levels(>0): ");
			scanf("%d", &levels);
		}
		
		int sweeps=0;
		while(sweeps < 1)
		{
			printf("\nPlease input smoothing sweeps: ");
			scanf("%d", &sweeps);
		}
		int refine=0;
		if(ub) {
		while(refine < 1)
		{
			printf("\nPlease input refine ratio: ");
			scanf("%d", &refine);
		}
		}
		int smoothe_type=-1;
		while(smoothe_type < 0)
		{
			printf("\nPlease input smoother_type\n 0: jacobi  1: mc-ssor  2: polynomial  3: additive schwarz  4: ilu0  5: sparskit2 ilu  6: ilupp  7: petsc_factor\n");
			scanf("%d", &smoothe_type);
		}		
		
		// smoothed by multi color symmetric Gauss-Seidel
		bool SOR=false;
		int sor_=-1;		
		bool Jacobi=true;
		int max_colors=0;			
		int ja_=-1;
		bool red_black=false;
		int rb=-1;
		if (smoothe_type == 1) {
		while(sor_<0)
		{
			printf("\nPlease input: 0: MC-SSOR,  1: MC-SOR ");
			scanf("%d", &sor_);
		}
		if(sor_) {
		SOR=true;
		} else {
		SOR=false;		
		}
		while(ja_<0)
		{
			printf("\nPlease input: 0: MC-SSOR,  1: Use Jacobi instead of MC-SSOR when colors > max_colors ");
			scanf("%d", &ja_);
		}
		if(ja_) {
		Jacobi=true;
		} else {
		Jacobi=false;		
		}	
		if (ja_){
		while(max_colors < 1)
		{
			printf("\nPlease input max_colors: ");
			scanf("%d", &max_colors);
		}
		}
		while(rb<0)
		{
			printf("\nPlease input: 0: not use,  1: red-black multicoloring ");
			scanf("%d", &rb);
		}
		if(rb) {
		red_black=true;
		} else {
		red_black=false;		
		}
		}
		
		int order=0; 		
		if (smoothe_type == 2) {
		while(order < 1)
		{
			printf("\nPlease input polynomial order: ");
			scanf("%d", &order);
		}
		}	
		
		// smoothed by asm
		bool as_smoother=false;
		int overlap = -1;	
		int jacobi_iter = -1;			
		int re=-1;
		int nd=0;
		bool restrict=false;
		
		if(smoothe_type == 3)
		{
		as_smoother=true;
		while(re<0)
		{
			printf("\nPlease input: 0: basic ASM smoothing,  1: restrict ASM smoothing ");
			scanf("%d", &re);
		}
		if(re) {
		restrict=true;
		} else {
		restrict=false;	
		}
		while(overlap < 0)
		{
			printf("\nPlease input overlap(>=0): ");
			scanf("%d", &overlap);
		}
		while(jacobi_iter < 0)
		{
			printf("\nPlease input jacobi_iter(>0): ");
			scanf("%d", &jacobi_iter);
		}		
		if (use_bilinear) {
		while(nd < 1)
		{
			printf("\nPlease input number of domain(>0): ");
			scanf("%d", &nd);
		}
		}		
		}
		
	    cusparseHandle_t handle = 0;
		/* Get handle to the CUSPARSE context */
		cusparseCreate(&handle);

		bool use_mkl_sv=false;
		int sv=-1;		
		if(smoothe_type == 4)
		{	
		while(sv<0)
		{
			printf("\nPlease input: 0: use cuSPARSE,  1: use MKL to solve triangular system\n");
			scanf("%d", &sv);
		}
		if(sv) {
		use_mkl_sv=true;
		} else {
		use_mkl_sv=false;		
		}		
		}
		
		// smoothed by sparskit ilu 
		int sparskit_ilu_type = -1;
		int maxfil=0;		
		ValueType droptol=-0.0001;			
		ValueType permtol=-0.01;	
		int mbloc = A.num_rows;			
		ValueType alph=-1.0;		
		int lfil=-1;	
		
		if(smoothe_type == 5)
		{				
		while(sparskit_ilu_type < 0 || sparskit_ilu_type > 6)
		{
			printf("\nPlease input sparskit_ilu_type: 0: ilut, 1: ilutp, 2: ilud, 3: iludp, 4: iluk, 5: ilu0, 6: milu0 ");
			scanf("%d", &sparskit_ilu_type);
		}		
		while(maxfil < 1)
		{
			printf("\nPlease input maxfil: ");
			scanf("%d", &maxfil);
		}
		if (sparskit_ilu_type < 4) {
		while(droptol < 0.)
		{
			printf("\nPlease input droptol: ");
			scanf("%lf", &droptol);
		}	
		}
		if (sparskit_ilu_type == 1 || sparskit_ilu_type == 3) {
		while(permtol < 0.)
		{
			printf("\nPlease input permtol: ");
			scanf("%lf", &permtol);
		}	
		}		
		if (sparskit_ilu_type == 2 || sparskit_ilu_type == 3) {
		while(alph < 0.)
		{
			printf("\nPlease input alph: ");
			scanf("%lf", &alph);
		}	
		}			
		if (sparskit_ilu_type == 4) {
		while(lfil < 0)
		{
			printf("\nPlease input lfil: ");
			scanf("%d", &lfil);
		}	
		}	
		while(sv<0)
		{
			printf("\nPlease input: 0: use cuSPARSE,  1: use MKL to solve triangular system\n");
			scanf("%d", &sv);
		}
		if(sv) {
		use_mkl_sv=true;
		} else {
		use_mkl_sv=false;		
		}			
		}		
		
		// smoothed by ilupp
		int ilupp_ilu_type = -1;		
		int fill_in=-1;		
		ValueType threshold=-1000.;		
		bool detail_param=false;
		int dp=-1;		
		int prepro_type=-1;
		int precond_no;
		int ilupp_max_levels = 0;
		ValueType mem_factor=-0.0001;	
		ValueType min_pivot=-0.0001;	
		bool max_fillin_is_inf=false;	
		int max_fi_inf=-1;		
		bool begin_total_piv=false;	
		int begin_tp=-1;			
		if(smoothe_type == 6)
		{
		while(ilupp_ilu_type < 0 || ilupp_ilu_type > 2)
		{
			printf("\nPlease input ilupp_ilu_type: 0: iluc, 1: ilut, 2: multilevel_ilu ");
			scanf("%d", &ilupp_ilu_type);
		}			
		while(fill_in < 0)
		{
			printf("\nPlease input fill_in: ");
			scanf("%d", &fill_in);
		}
		while(threshold < 0.)
		{
			printf("\nPlease input threshold: ");
			scanf("%lf", &threshold);
		}	
		while(prepro_type < 0)
		{
			printf("\nPlease input prepro_type: ");
			scanf("%d", &prepro_type);
		}		
		if (ilupp_ilu_type == 0 || ilupp_ilu_type == 1) {
		while(sv<0)
		{
			printf("\nPlease input: 0: use cuSPARSE,  1: use MKL to solve triangular system\n");
			scanf("%d", &sv);
		}
		if(sv) {
		use_mkl_sv=true;
		} else {
		use_mkl_sv=false;		
		}
		}
		if (ilupp_ilu_type == 2) {
		while(dp<0)
		{
			printf("\nPlease input detail_param: 0: disable detail parameters,  1: enable\n");
			scanf("%d", &dp);
		}
		if(dp) {
		detail_param=true;
		} else {
		detail_param=false;		
		}
		}		

		if (ilupp_ilu_type == 2) {
		printf("\nPlease input precond_no: ");
		scanf("%d", &precond_no);	
		while(mem_factor < 0.)
		{
			printf("\nPlease input mem_factor: ");
			scanf("%lf", &mem_factor);
		}	
		
		if (detail_param) {		
		while(ilupp_max_levels < 1)
		{
			printf("\nPlease input max_levels(>0): ");
			scanf("%d", &ilupp_max_levels);
		}				
		while(min_pivot < 0.)
		{
			printf("\nPlease input min_pivot: ");
			scanf("%lf", &min_pivot);
		}	
		while(max_fi_inf<0)
		{
			printf("\nPlease input max_fillin_is_inf: 0: max_fillin_is_inf is false,  1: max_fillin_is_inf is true\n");
			scanf("%d", &max_fi_inf);
		}
		if(max_fi_inf) {
		max_fillin_is_inf=true;
		} else {
		max_fillin_is_inf=false;		
		}	
		while(begin_tp<0)
		{
			printf("\nPlease input begin_total_piv: 0: begin_total_piv is false,  1: begin_total_piv is true\n");
			scanf("%d", &begin_tp);
		}
		if(begin_tp) {
		begin_total_piv=true;
		} else {
		begin_total_piv=false;		
		}		
		} else {
		ilupp_max_levels = 100;
		mem_factor = 3.0;	
		min_pivot  = 1e-2;	
		max_fillin_is_inf=true;	
		begin_total_piv=true;	
		}	
		}
		}
		
		int petsc_factor_type = -1;		
		int petsc_factor_fill_in=-1;
		ValueType rtol=-0.0001;			
		ValueType petsc_factor_fill=-0.0001;	
		int mat_ordering_type = -1;
		bool allow_diagonal_fill=false;
		int df=-1;	
		if (smoothe_type == 7) {		
		while(petsc_factor_type < 0 || petsc_factor_type > 1)
		{
			printf("\nPlease input type: 0: lu, 1: ilu ");
			scanf("%d", &petsc_factor_type);
		}
		while(sv<0)
		{
			printf("\nPlease input: 0: use cuSPARSE,  1: use MKL to solve triangular system\n");
			scanf("%d", &sv);
		}
		if(sv) {
		use_mkl_sv=true;
		} else {
		use_mkl_sv=false;		
		}		
		while(petsc_factor_fill_in < 0)
		{
			printf("\nPlease input petsc_factor_fill_in: ");
			scanf("%d", &petsc_factor_fill_in);
		}
		while(petsc_factor_fill < 0.)
		{
			printf("\nPlease input petsc_factor_fill: ");
			scanf("%lf", &petsc_factor_fill);
		}
		while(rtol < 0.)
		{
			printf("\nPlease input rtol: ");
			scanf("%lf", &rtol);
		}			
		while(df<0)
		{
			printf("\nPlease input allow_diagonal_fill: 0: not allow,  1: allow diagonal fill\n");
			scanf("%d", &df);
		}
		if(df) {
		allow_diagonal_fill=true;
		} else {
		allow_diagonal_fill=false;		
		}	
		while(mat_ordering_type < 0 || mat_ordering_type > 4)
		{
			printf("\nPlease input mat_ordering_type: 0: natural, 1: nd, 2: lwd, 3: rcm, 4: qmd\n ");
			scanf("%d", &mat_ordering_type);
		}	
		}
		
		// parameters for coarest problem
		int coarse_solver_type = -1;
		while(coarse_solver_type < 0 || coarse_solver_type > 3)
		{
			printf("\nPlease input coarse solver type: 0: cusp lu, 1: magma lu, 2: dss lu, 3: petsc lu\n ");
			scanf("%d", &coarse_solver_type);
		}			
		bool symmetric=false;
		if(coarse_solver_type==2) {
		int sy_=-1;
		while(sy_<0)
		{
			printf("\nPlease input: 0: if the matrix is unsymmetric,  1: symmetric ");
			scanf("%d", &sy_);
		}
		if(sy_) {
		symmetric=true;
		} else {
		symmetric=false;	
		}
		}
		bool positive_definite=false;
		if(coarse_solver_type==2) {
		int pd=-1;
		while(pd<0)
		{
			printf("\nPlease input: 0: if the matrix is indefinite,  1: positive definite ");
			scanf("%d", &pd);
		}
		if(pd) {
		positive_definite=true;
		} else {
		positive_definite=false;	
		}
		}

		
		timer tc;		
		if (smoothe_type == 0)
		M = new multigrid_jacobi(false, A, handle, 
		(cusp::precond::mg_type)mg_type, levels, red_black,
		use_bilinear, 1, m, n,
		refine, refine,
		nd, as_smoother, no_postsmooth,	
		symmetric, positive_definite,
		order, overlap, 
		use_mkl_sv, (cusp::relaxation::sparskit_ilu_type)sparskit_ilu_type,
		(cusp::relaxation::ilupp_type)ilupp_ilu_type, 
			prepro_type, precond_no,
			ilupp_max_levels, mem_factor,
			threshold, min_pivot, 
			max_fillin_is_inf, fill_in,
			begin_total_piv, detail_param,
			(cusp::relaxation::petsc_factor_type)petsc_factor_type, petsc_factor_fill_in,
			petsc_factor_fill, allow_diagonal_fill,
			(cusp::relaxation::petsc_factor_mat_ordering_type)mat_ordering_type, rtol, (cusp::precond::coarse_solver_type)coarse_solver_type);	
		else if(smoothe_type == 1)
		M = new multigrid_mc_ssor(false, A, handle, 
		(cusp::precond::mg_type)mg_type, levels, red_black,
		use_bilinear, 1, m, n,
		refine, refine,
		nd, as_smoother, no_postsmooth,	
		symmetric, positive_definite,
		order, overlap, 
		use_mkl_sv, (cusp::relaxation::sparskit_ilu_type)sparskit_ilu_type,
		(cusp::relaxation::ilupp_type)ilupp_ilu_type, 
			prepro_type, precond_no,
			ilupp_max_levels, mem_factor,
			threshold, min_pivot, 
			max_fillin_is_inf, fill_in,
			begin_total_piv, detail_param,
			(cusp::relaxation::petsc_factor_type)petsc_factor_type, petsc_factor_fill_in,
			petsc_factor_fill, allow_diagonal_fill,
			(cusp::relaxation::petsc_factor_mat_ordering_type)mat_ordering_type, rtol, (cusp::precond::coarse_solver_type)coarse_solver_type);	
		else if(smoothe_type == 2)
		M = new multigrid_polynomial(false, A, handle, 
		(cusp::precond::mg_type)mg_type, levels, red_black,
		use_bilinear, 1, m, n,
		refine, refine,
		nd, as_smoother, no_postsmooth,	
		symmetric, positive_definite,
		order, overlap, 
		use_mkl_sv, (cusp::relaxation::sparskit_ilu_type)sparskit_ilu_type,
		(cusp::relaxation::ilupp_type)ilupp_ilu_type, 
			prepro_type, precond_no,
			ilupp_max_levels, mem_factor,
			threshold, min_pivot, 
			max_fillin_is_inf, fill_in,
			begin_total_piv, detail_param,
			(cusp::relaxation::petsc_factor_type)petsc_factor_type, petsc_factor_fill_in,
			petsc_factor_fill, allow_diagonal_fill,
			(cusp::relaxation::petsc_factor_mat_ordering_type)mat_ordering_type, rtol, (cusp::precond::coarse_solver_type)coarse_solver_type);	
		else if(smoothe_type == 3)
		M = new multigrid_as(false, A, handle, 
		(cusp::precond::mg_type)mg_type, levels, red_black,
		use_bilinear, 1, m, n,
		refine, refine,
		nd, as_smoother, no_postsmooth,	
		symmetric, positive_definite,
		order, overlap, 
		use_mkl_sv, (cusp::relaxation::sparskit_ilu_type)sparskit_ilu_type,
		(cusp::relaxation::ilupp_type)ilupp_ilu_type, 
			prepro_type, precond_no,
			ilupp_max_levels, mem_factor,
			threshold, min_pivot, 
			max_fillin_is_inf, fill_in,
			begin_total_piv, detail_param,
			(cusp::relaxation::petsc_factor_type)petsc_factor_type, petsc_factor_fill_in,
			petsc_factor_fill, allow_diagonal_fill,
			(cusp::relaxation::petsc_factor_mat_ordering_type)mat_ordering_type, rtol, (cusp::precond::coarse_solver_type)coarse_solver_type);	
		else if(smoothe_type == 4)
		M = new multigrid_mkl_ilu0(false, A, handle, 
		(cusp::precond::mg_type)mg_type, levels, red_black,
		use_bilinear, 1, m, n,
		refine, refine,
		nd, as_smoother, no_postsmooth,	
		symmetric, positive_definite,
		order, overlap, 
		use_mkl_sv, (cusp::relaxation::sparskit_ilu_type)sparskit_ilu_type,
		(cusp::relaxation::ilupp_type)ilupp_ilu_type, 
			prepro_type, precond_no,
			ilupp_max_levels, mem_factor,
			threshold, min_pivot, 
			max_fillin_is_inf, fill_in,
			begin_total_piv, detail_param,
			(cusp::relaxation::petsc_factor_type)petsc_factor_type, petsc_factor_fill_in,
			petsc_factor_fill, allow_diagonal_fill,
			(cusp::relaxation::petsc_factor_mat_ordering_type)mat_ordering_type, rtol, (cusp::precond::coarse_solver_type)coarse_solver_type);	
		else if(smoothe_type == 5)		
		M = new multigrid_sparskit2_ilu(false, A, handle, 
		(cusp::precond::mg_type)mg_type, levels, red_black,
		use_bilinear, 1, m, n,
		refine, refine,
		nd, as_smoother, no_postsmooth,	
		symmetric, positive_definite,
		order, overlap, 
		use_mkl_sv, (cusp::relaxation::sparskit_ilu_type)sparskit_ilu_type,
		(cusp::relaxation::ilupp_type)ilupp_ilu_type, 
			prepro_type, precond_no,
			ilupp_max_levels, mem_factor,
			threshold, min_pivot, 
			max_fillin_is_inf, fill_in,
			begin_total_piv, detail_param,
			(cusp::relaxation::petsc_factor_type)petsc_factor_type, petsc_factor_fill_in,
			petsc_factor_fill, allow_diagonal_fill,
			(cusp::relaxation::petsc_factor_mat_ordering_type)mat_ordering_type, rtol, (cusp::precond::coarse_solver_type)coarse_solver_type);	
		else if(smoothe_type == 6)		
		M = new multigrid_ilupp(false, A, handle, 
		(cusp::precond::mg_type)mg_type, levels, red_black,
		use_bilinear, 1, m, n,
		refine, refine,
		nd, as_smoother, no_postsmooth,	
		symmetric, positive_definite,
		order, overlap, 
		use_mkl_sv, (cusp::relaxation::sparskit_ilu_type)sparskit_ilu_type,
		(cusp::relaxation::ilupp_type)ilupp_ilu_type, 
			prepro_type, precond_no,
			ilupp_max_levels, mem_factor,
			threshold, min_pivot, 
			max_fillin_is_inf, fill_in,
			begin_total_piv, detail_param,
			(cusp::relaxation::petsc_factor_type)petsc_factor_type, petsc_factor_fill_in,
			petsc_factor_fill, allow_diagonal_fill,
			(cusp::relaxation::petsc_factor_mat_ordering_type)mat_ordering_type, rtol, (cusp::precond::coarse_solver_type)coarse_solver_type);	
		else if(smoothe_type == 7)		
		M = new multigrid_petsc_factor(false, A, handle, 
		(cusp::precond::mg_type)mg_type, levels, red_black,
		use_bilinear, 1, m, n,
		refine, refine,
		nd, as_smoother, no_postsmooth,	
		symmetric, positive_definite,
		order, overlap, 
		use_mkl_sv, (cusp::relaxation::sparskit_ilu_type)sparskit_ilu_type,
		(cusp::relaxation::ilupp_type)ilupp_ilu_type, 
			prepro_type, precond_no,
			ilupp_max_levels, mem_factor,
			threshold, min_pivot, 
			max_fillin_is_inf, fill_in,
			begin_total_piv, detail_param,
			(cusp::relaxation::petsc_factor_type)petsc_factor_type, petsc_factor_fill_in,
			petsc_factor_fill, allow_diagonal_fill,
			(cusp::relaxation::petsc_factor_mat_ordering_type)mat_ordering_type, rtol, (cusp::precond::coarse_solver_type)coarse_solver_type);		
		
		float tc_ = tc.seconds_elapsed();
        std::cout << "\nconstruct in " << tc_ << " s " << std::endl;
		
        // solve
		int solve=-1;
		while(solve<0 || solve>3)
		{
			printf("\nPlease input solver, 0: cg  1: gmres  2: bicgstab ");
			scanf("%d", &solve);
		}
		int restart = 0;

		int nl_=0;
		while(nl_ < 1)
		{
			printf("\nPlease input nonlinear test times: ");
			scanf("%d", &nl_);
		}		
		for (int nl=0; nl<nl_; nl++) {
		printf("the %d step\n",nl);
		#if 0
		cusp::dia_matrix<int,ValueType,cusp::host_memory> diag(A.num_rows,A.num_cols,2*A.num_rows-2,2);
		diag.diagonal_offsets[0] = 0;
		diag.diagonal_offsets[1] = 2;
		for (int row=0; row<A.num_rows; row++) {
		diag.values(row,0) = (ValueType)(nl+1);
		diag.values(row,1) = (ValueType)(nl+1);		
		}
		#else
		cusp::dia_matrix<int,ValueType,cusp::host_memory> diag(A.num_rows,A.num_cols,A.num_rows,1);
		diag.diagonal_offsets[0] = 0;
		for (int row=0; row<A.num_rows; row++) {
		diag.values(row,0) = (ValueType)(nl+1);
		}		
		#endif
		cusp::dia_matrix<int,ValueType,MemorySpace> diag_d = diag;
		
		cusp::csr_matrix<IndexType, ValueType, MemorySpace> A_temp;
		cusp::multiply(A,diag_d,A_temp);
		
		cusp::multiply(A_temp, x_solution, b);		

		cusp::array1d<ValueType, MemorySpace> x(A_temp.num_rows, 0);
		
		// setup each time
				timer ts;
			if (smoothe_type == 0)
			((multigrid_jacobi*)M)->setup(A_temp, 0.,
				1, 1e-8, 0., 		
				sweeps,
				max_colors, SOR, Jacobi, 
				jacobi_iter, restrict, 
				maxfil, lfil, droptol,
				mbloc, permtol, alph);
			else if(smoothe_type == 1)	
			((multigrid_mc_ssor*)M)->setup(A_temp, 0.,
				1, 1e-8, 0., 		
				sweeps,
				max_colors, SOR, Jacobi, 
				jacobi_iter, restrict, 
				maxfil, lfil, droptol,
				mbloc, permtol, alph);
			else if(smoothe_type == 2)
			((multigrid_polynomial*)M)->setup(A_temp, 0.,
				1, 1e-8, 0., 		
				sweeps,
				max_colors, SOR, Jacobi, 
				jacobi_iter, restrict, 
				maxfil, lfil, droptol,
				mbloc, permtol, alph);		
			else if(smoothe_type == 3)
			((multigrid_as*)M)->setup(A_temp, 0.,
				1, 1e-8, 0., 		
				sweeps,
				max_colors, SOR, Jacobi, 
				jacobi_iter, restrict, 
				maxfil, lfil, droptol,
				mbloc, permtol, alph);	
			else if(smoothe_type == 4)
			((multigrid_mkl_ilu0*)M)->setup(A_temp, 0.,
				1, 1e-8, 0., 		
				sweeps,
				max_colors, SOR, Jacobi, 
				jacobi_iter, restrict, 
				maxfil, lfil, droptol,
				mbloc, permtol, alph);				
			else if(smoothe_type == 5)
			((multigrid_sparskit2_ilu*)M)->setup(A_temp, 0.,
				1, 1e-8, 0., 		
				sweeps,
				max_colors, SOR, Jacobi, 
				jacobi_iter, restrict, 
				maxfil, lfil, droptol,
				mbloc, permtol, alph);		
			else if(smoothe_type == 6)
			((multigrid_ilupp*)M)->setup(A_temp, 0.,
				1, 1e-8, 0., 		
				sweeps,
				max_colors, SOR, Jacobi, 
				jacobi_iter, restrict, 
				maxfil, lfil, droptol,
				mbloc, permtol, alph);	
			else if(smoothe_type == 7)		
			((multigrid_petsc_factor*)M)->setup(A_temp, 0.,
				1, 1e-8, 0., 		
				sweeps,
				max_colors, SOR, Jacobi, 
				jacobi_iter, restrict, 
				maxfil, lfil, droptol,
				mbloc, permtol, alph);	
				
		float ts_ = ts.seconds_elapsed();
        std::cout << "\nsetup in " << ts_ << " s " << std::endl;
		float tsetup;
		if (nl_==1) {
		tsetup = tc_+ts_;
        std::cout << "\nsetup all in " << tsetup << " s " << std::endl;		
		}
		
        // set stopping criteria (iteration_limit = 1000, relative_tolerance = 1e-7)
        cusp::verbose_monitor<ValueType> monitor(b, 10000, 1e-7);		
		
		timer t1;
		switch (solve)
		{
		case 0:
			if (smoothe_type == 0)
			cusp::krylov::cg(A_temp, x, b, monitor, *(multigrid_jacobi*)M);
			else if(smoothe_type == 1)	
			cusp::krylov::cg(A_temp, x, b, monitor, *(multigrid_mc_ssor*)M);	
			else if(smoothe_type == 2)
			cusp::krylov::cg(A_temp, x, b, monitor, *(multigrid_polynomial*)M);		
			else if(smoothe_type == 3)
			cusp::krylov::cg(A_temp, x, b, monitor, *(multigrid_as*)M);	
			else if(smoothe_type == 4)
			cusp::krylov::cg(A_temp, x, b, monitor, *(multigrid_mkl_ilu0*)M);				
			else if(smoothe_type == 5)
			cusp::krylov::cg(A_temp, x, b, monitor, *(multigrid_sparskit2_ilu*)M);			
			else if(smoothe_type == 6)
			cusp::krylov::cg(A_temp, x, b, monitor, *(multigrid_ilupp*)M);		
			else if(smoothe_type == 7)
			cusp::krylov::cg(A_temp, x, b, monitor, *(multigrid_petsc_factor*)M);				
			break;
		case 1:
			while(restart < 1)
			{
			printf("\nPlease input restart: ");
			scanf("%d", &restart);
			}
			if (smoothe_type == 0)
			cusp::krylov::gmres(A_temp, x, b, restart, monitor, *(multigrid_jacobi*)M);
			else if(smoothe_type == 1)	
			cusp::krylov::gmres(A_temp, x, b, restart, monitor, *(multigrid_mc_ssor*)M);	
			else if(smoothe_type == 2)
			cusp::krylov::gmres(A_temp, x, b, restart, monitor, *(multigrid_polynomial*)M);		
			else if(smoothe_type == 3)
			cusp::krylov::gmres(A_temp, x, b, restart, monitor, *(multigrid_as*)M);	
			else if(smoothe_type == 4)
			cusp::krylov::gmres(A_temp, x, b, restart, monitor, *(multigrid_mkl_ilu0*)M);
			else if(smoothe_type == 5)
			cusp::krylov::gmres(A_temp, x, b, restart, monitor, *(multigrid_sparskit2_ilu*)M);		
			else if(smoothe_type == 6)
			cusp::krylov::gmres(A_temp, x, b, restart, monitor, *(multigrid_ilupp*)M);	
			else if(smoothe_type == 7)
			cusp::krylov::gmres(A_temp, x, b, restart, monitor, *(multigrid_petsc_factor*)M);				
			break;
		case 2:
			if (smoothe_type == 0)
			cusp::krylov::bicgstab(A_temp, x, b, monitor, *(multigrid_jacobi*)M);
			else if(smoothe_type == 1)	
			cusp::krylov::bicgstab(A_temp, x, b, monitor, *(multigrid_mc_ssor*)M);	
			else if(smoothe_type == 2)
			cusp::krylov::bicgstab(A_temp, x, b, monitor, *(multigrid_polynomial*)M);		
			else if(smoothe_type == 3)
			cusp::krylov::bicgstab(A_temp, x, b, monitor, *(multigrid_as*)M);
			else if(smoothe_type == 4)
			cusp::krylov::bicgstab(A_temp, x, b, monitor, *(multigrid_mkl_ilu0*)M);	
			else if(smoothe_type == 5)
			cusp::krylov::bicgstab(A_temp, x, b, monitor, *(multigrid_sparskit2_ilu*)M);
			else if(smoothe_type == 6)
			cusp::krylov::bicgstab(A_temp, x, b, monitor, *(multigrid_ilupp*)M);	
			else if(smoothe_type == 7)
			cusp::krylov::bicgstab(A_temp, x, b, monitor, *(multigrid_petsc_factor*)M);				
			break;	
		default:
			if (smoothe_type == 0)
			cusp::krylov::cg(A_temp, x, b, monitor, *(multigrid_jacobi*)M);
			else if(smoothe_type == 1)	
			cusp::krylov::cg(A_temp, x, b, monitor, *(multigrid_mc_ssor*)M);	
			else if(smoothe_type == 2)
			cusp::krylov::cg(A_temp, x, b, monitor, *(multigrid_polynomial*)M);		
			else if(smoothe_type == 3)
			cusp::krylov::cg(A_temp, x, b, monitor, *(multigrid_as*)M);	
			else if(smoothe_type == 4)
			cusp::krylov::cg(A_temp, x, b, monitor, *(multigrid_mkl_ilu0*)M);
			else if(smoothe_type == 5)
			cusp::krylov::cg(A_temp, x, b, monitor, *(multigrid_sparskit2_ilu*)M);	
			else if(smoothe_type == 6)
			cusp::krylov::cg(A_temp, x, b, monitor, *(multigrid_ilupp*)M);	
			else if(smoothe_type == 7)
			cusp::krylov::cg(A_temp, x, b, monitor, *(multigrid_petsc_factor*)M);				
			break;
		}
		float t1_ = t1.seconds_elapsed();
        std::cout << "\nsolve in " << t1_ << " s " << std::endl;
 		float tall;
		if (nl_==1) {
		tall = tc_+ts_+t1_;
        std::cout << "\nsolve all in " << tall << " s " << std::endl;		
		}
		
        // report status
        report_status(monitor);

		// check solution
        cusp::array1d<ValueType, MemorySpace> e(A_temp.num_rows);
		blas::axpby(x_solution, x, e, ValueType(1), ValueType(-1));
		ValueType e_norm = cusp::blas::nrm2(e);
		printf("\n check solution: ||x*-x||: %f \n", e_norm );	
		
		cusp::array1d<ValueType, MemorySpace> ax(x.size());
		cusp::array1d<ValueType, MemorySpace> r(A_temp.num_rows);
		cusp::multiply(A_temp, x, ax);
		cusp::blas::axpby(ax, b, r, ValueType(1), ValueType(-1));
		ValueType r_norm = cusp::blas::nrm2(r);
		ValueType b_norm = cusp::blas::nrm2(b);		
		printf("\n ||b-Ax||/||b||: %f \n", r_norm / b_norm );
		}
		
		if (smoothe_type == 0)
		delete (multigrid_jacobi*)M;
		else if(smoothe_type == 1)	
		delete (multigrid_mc_ssor*)M;
		else if(smoothe_type == 2)
		delete (multigrid_polynomial*)M;
		else if(smoothe_type == 3)	
		delete (multigrid_as*)M;
		else if(smoothe_type == 4)	
		delete (multigrid_mkl_ilu0*)M;
		else if(smoothe_type == 5)	
		delete (multigrid_sparskit2_ilu*)M;		
		else if(smoothe_type == 6)	
		delete (multigrid_ilupp*)M;	
		else if(smoothe_type == 7)	
		delete (multigrid_petsc_factor*)M;	
		
		cusparseDestroy(handle);	
    }

  PetscFinalize();
}
