#include <cusp/io/matrix_market.h>
#include <cusp/precond/diagonal.h>
#include <cusp/precond/smoothed_aggregation_new.h>
//#include <cusp/precond/pamg.h>  // pamg internal called smoothed_aggregation.h, which conflicts with smoothed_aggregation_new.h
#include <cusp/precond/ainv.h>
//#include <cusp/precond/asm_new.h>   //Notes: the new smoothed_aggregation.h has already compiled the additive schwarz solver in relaxation, they must not be comiled together
#include <cusp/blas.h>
#include <cusp/krylov/cg.h>
#include <cusp/krylov/gmres.h>
#include <cusp/krylov/bicgstab.h>
#include <cusp/gallery/poisson.h>
#include <cusp/coo_matrix.h>
#include <cusp/csr_matrix.h>
#include <culapackdevice.h>
//#include "magma_overload.h" // Some of functions have already been defined at smoothed_aggregation_new.h
#include <iostream>
#include "mytimer.h"
#include "../../performance/timer.h"

template <typename Monitor>
void report_status(Monitor& monitor)
{
	#ifdef PRECON_NORM
    if (monitor.converged())
    {
        std::cout << "  Solver converged to " << monitor.tolerance2() << " tolerance";
        std::cout << " after " << monitor.iteration_count() << " iterations";
        std::cout << " (" << monitor.relative_z_norm() << " final ||z||/||b||)" << std::endl;
    }
    else
    {
        std::cout << "  Solver reached iteration limit " << monitor.iteration_limit() << " before converging";
        std::cout << " to " << monitor.tolerance2() << " tolerance ";
        std::cout << " (" << monitor.relative_z_norm() << " final ||z||/||b||)" << std::endl;
    }
	#else
	if (monitor.converged())
    {
        std::cout << "  Solver converged to " << monitor.tolerance1() << " tolerance";
        std::cout << " after " << monitor.iteration_count() << " iterations";
        std::cout << " (" << monitor.relative_residual_norm() << " final ||r||/||b||)" << std::endl;
    }
    else
    {
        std::cout << "  Solver reached iteration limit " << monitor.iteration_limit() << " before converging";
        std::cout << " to " << monitor.tolerance1() << " tolerance ";
        std::cout << " (" << monitor.relative_residual_norm() << " final ||r||/||b||)" << std::endl;
    }
	#endif
}

void pq_gen(const int np, int *p, int *q)
{
    int a = 1, b = np;

    for(int i = 2; i < floor(sqrt(double(np))) + 1; i++)
        if(np % i == 0)  
        {   
            a = i;
            b = np / i;
        }
    *p = a;    
    *q = b;
}

int main(void)
{
	#ifdef PRECISION_S
    typedef float               ValueType;
	#endif
	#ifdef PRECISION_D
    typedef double              ValueType;
	#endif

    typedef int                 IndexType;
	#ifdef DEVICE_MEMORY	
    typedef cusp::device_memory MemorySpace;
	#endif
	#ifdef HOST_MEMORY
    typedef cusp::host_memory MemorySpace;	
	#endif
    // create an empty sparse matrix structure
    cusp::coo_matrix<IndexType, ValueType, MemorySpace> A;

	#ifdef MATRIX_MARKET
    // load a matrix stored in MatrixMarket format
	std::string mtx="5pt_10x10.mtx";
	printf("Please input matrix : ");
	std::cin >> mtx;
    cusp::io::read_matrix_market_file(A, mtx);
	#else
    // create 2D Poisson problem
	int m = 0, n = 0;
	while(m < 2 || n < 2)
    {
    	printf("Please input m, n: ");
    	scanf("%d %d", &m, &n);
    }
    cusp::gallery::poisson5pt(A, m, n);
	#endif
	printf("A.num_rows: %d, A.num_cols: %d\n", A.num_rows, A.num_cols);	
	
	int solver = -1;
	while(solver < 0)
    {
    	printf("Please input solver: 0: cg,  1: gmres,  2: bicgstab\n");
    	scanf("%d", &solver);
    }
	
	int restart = -1;
	if(solver == 1) 
	{
	printf("gmres solver is chosen\n");
	while(restart < 0)
    {
    	printf("Please input restart: ");
    	scanf("%d", &restart);
    }
	}
	
	int iter_outer = 10000;
	
    mytimer time;
    begin_timer(&time); 
    double old_time, elapsed_time;    	

    // solve without preconditioner
    {
        std::cout << "\n ******Solving with no preconditioner******" << std::endl;
    
        // allocate storage for solution (x) and right hand side (b)
        cusp::array1d<ValueType, MemorySpace> x(A.num_rows, 0);
        cusp::array1d<ValueType, MemorySpace> b(A.num_rows, 1);

        // set stopping criteria (iteration_limit = 1000, relative_tolerance = 1e-6)
        cusp::default_monitor<ValueType> monitor(b, iter_outer, 1e-6, 1e-10);
        
        // solve
		timer t;
		if(solver == 0)
        cusp::krylov::cg(A, x, b, monitor);
		else if(solver == 1) 
		cusp::krylov::gmres(A, x, b,restart, monitor);	
		else if(solver == 2) 
		cusp::krylov::bicgstab(A, x, b, monitor);
		std::cout << "\nsolve without preconditioner time " << t.milliseconds_elapsed() << " ms " << std::endl;

        // report status
        report_status(monitor);
		
		// check solution
		cusp::array1d<ValueType,MemorySpace> y(A.num_rows);	
		cusp::array1d<ValueType,MemorySpace> r(A.num_rows);			
		// y <- Ax
		cusp::multiply(A, x, y);
		// r <- b - A*x
		blas::axpby(b, y, r, ValueType(1), ValueType(-1));
		ValueType r_norm = cusp::blas::nrm2(r);
		ValueType b_norm = cusp::blas::nrm2(b);	
		printf("\n check solution: ||b-Ax||/||b||: %e \n", r_norm / b_norm );
    }
	#if 0
	#ifdef PRECISION_S
	#ifndef MATRIX_MARKET
    // solve with asm preconditioner 
	{
	std::cout << "\n ******Solving with asm preconditioner****** " << std::endl;
	int asm_type = -1, solver_type = -1;
	while(asm_type < 0 || asm_type > 1)
	{
		printf("\nPlease input asm_type: ");
		scanf("%d", &asm_type);
	}
	while(solver_type < 0 || solver_type > 2)
	{
		printf("\nPlease input solver_type: ");
		scanf("%d", &solver_type);
	}
	int overlap = -1;	
	while(overlap < 0)
	{
		printf("\nPlease input overlap(>=0): ");
		scanf("%d", &overlap);
	}
	int iter = 0;	
	while(iter < 1)
	{
		printf("\nPlease input iter(>0): ");
		scanf("%d", &iter);
	}
	#if 0
	{
        std::cout << "\n *****Solving with asm preconditioner, natural sort****** " << std::endl;
		int num_domain = int(std::sqrt((double)A.num_rows));
		printf("\nPartition num_rows into %d\n", num_domain);
		cusp::array1d<IndexType,MemorySpace> domains(A.num_rows, 0);
		for(int i=0; i<num_domain; i++)
			for(int j=0; j<num_domain; j++)
			{
				domains[i*num_domain+j] = i;
			}
			
        // allocate storage for solution (x) and right hand side (b)
        cusp::array1d<ValueType, MemorySpace> x(A.num_rows, 0);
        cusp::array1d<ValueType, MemorySpace> b(A.num_rows, 1);

        // set stopping criteria (iteration_limit = 1000, relative_tolerance = 1e-6)
        cusp::default_monitor<ValueType> monitor(b, iter_outer, 1e-6, 1e-10);		
	
        std::cout << "\nSolving with asm preconditioner, Jacobi subsolver " << std::endl;		

        // setup preconditioner
        cusp::precond::restricted_additive_schwarz<IndexType, ValueType, MemorySpace> M(A,domains,(cusp::precond::asm_type_t)asm_type,overlap,(cusp::precond::asm_local_solver_t)solver_type,10,iter,1e-5,1e-10,1e-5);        
        
		// solve
		old_time = end_timer(&time); 
        if(solver == 0)
        cusp::krylov::cg(A, x, b, monitor,M);
		else if(solver == 1) 
		cusp::krylov::gmres(A, x, b,restart, monitor,M);	
		else if(solver == 2) 
		cusp::krylov::bicgstab(A, x, b, monitor,M);
		elapsed_time = elapsed_timer(&time, old_time);
		printf("\nsolve with preconditioning natural sort asm time: %f s\n", elapsed_time);
        // report status
        report_status(monitor);
      
    }
	#endif
	{
        std::cout << "\n ******Solving with asm preconditioner, PETSc sort****** " << std::endl;
		
		int nd = 0;	
		while(nd < 1)
		{
			printf("\nPlease input nd: ");
			scanf("%d", &nd);
		}
		int p, q, m_p, n_q;// p*q=nd, 2D domain partition   
		pq_gen(nd, &p, &q);
		printf("\nGrid size: %d, %d, Partition np into %d, %d\n", m, n, p, q);
		
		m_p = m / p;
		n_q = n / q;
	
		cusp::array1d<IndexType,MemorySpace> domains(A.num_rows, 0);
		
		timer t0;
		dim3 grid(p, q, 1);
		dim3 block(16, 16, 1);
		cusp::precond::detail::set_domains<IndexType><<<grid,block>>>(thrust::raw_pointer_cast(&*domains.begin()), m, n, m_p, n_q);
        std::cout << "\nset_domains in " << t0.milliseconds_elapsed() << " ms " << std::endl;
		/*
		for(int j = 0; j < q; j++)
		{	for(int i = 0; i < p; i++)
			{
				// subdomain size 
				//((m % p) > i) = 0 or 1, handle the case when not integer divide, smart adding one more node, for good load balance
				int nn = n_q + ((n % q) > j);
				int mm = m_p + ((m % p) > i);            
				// traversal at subdomain, two for
				for(int jj = 0; jj < nn; jj++)// each row
				{	for(int ii = 0; ii < mm; ii++) //each node in one row
					{
						int domain_index = j * p + i;
						
						int horizontal_offset; // original horizontal offset (native sort)
						if(m % p) // when not integer divide
						{
							if(i > (m % p) - 1) // implement the corresponding "smart adding" strategy
								horizontal_offset = (m % p) * (m_p + 1) + (i - (m % p)) * m_p + ii; // behind
							else
								horizontal_offset = i * (m_p + 1) + ii; // before
						}
						else
							horizontal_offset = i * m_p + ii;
						
						int vertical_offset;
						if(n % q)
						{
							if(j > (n % q) - 1)
								vertical_offset = (n % q) * (n_q + 1) + (j - (n % q)) * n_q + jj;
							else
								vertical_offset = j * (n_q + 1) + jj;
						}
						else
							vertical_offset = j * n_q + jj;
							
						int native_sort_index = vertical_offset * m + horizontal_offset;
						
						domains[native_sort_index] = domain_index;
					}
				}
			}
		}
		*/
		//for(int i = 0; i < A.num_rows; i++)
		//	printf("%d ", (int)domains[i]);
		//printf("\n");
		/*
		int num_domain = int(std::sqrt((double)A.num_rows));
		for(int i=0; i<num_domain; i++)
		{	for(int j=0; j<num_domain; j++)
			{
				printf("%d ", (int)domains[i*num_domain+j]);
			}
			if(i < num_domain - 1) printf("\n next row %d : \n", i+1);
		}
		*/	
        
        // allocate storage for solution (x) and right hand side (b)
        cusp::array1d<ValueType, MemorySpace> x(A.num_rows, 0);
        cusp::array1d<ValueType, MemorySpace> b(A.num_rows, 1);

        // set stopping criteria (iteration_limit = 1000, relative_tolerance = 1e-6)
        cusp::default_monitor<ValueType> monitor(b, iter_outer, 1e-6, 1e-10);
	
        std::cout << "\nSolving with asm preconditioner, Jacobi subsolver " << std::endl;
        // setup preconditioner
		timer t1;
        cusp::precond::restricted_additive_schwarz<IndexType, ValueType, MemorySpace> M(A,domains,(cusp::precond::asm_type_t)asm_type,overlap,(cusp::precond::asm_local_solver_t)solver_type,10,iter,1e-5,1e-10,1e-5);
        std::cout << "\nconstruct preconditioner in " << t1.milliseconds_elapsed() << " ms " << std::endl;
        
        // solve
		timer t2;
		old_time = end_timer(&time); 
        if(solver == 0)
        cusp::krylov::cg(A, x, b, monitor,M);
		else if(solver == 1) 
		cusp::krylov::gmres(A, x, b,restart, monitor,M);	
		else if(solver == 2) 
		cusp::krylov::bicgstab(A, x, b, monitor,M);
		elapsed_time = elapsed_timer(&time, old_time);
		std::cout << "\nsolve with preconditioning PETSc sort asm time " << t2.milliseconds_elapsed() << " ms " << std::endl;
		printf("\nsolve with preconditioning PETSc sort asm time: %f s\n", elapsed_time);
		
        // report status
        report_status(monitor);
		
		// check solution
		cusp::array1d<ValueType,MemorySpace> y(A.num_rows);	
		cusp::array1d<ValueType,MemorySpace> r(A.num_rows);			
		// y <- Ax
		cusp::multiply(A, x, y);
		// r <- b - A*x
		blas::axpby(b, y, r, ValueType(1), ValueType(-1));
		ValueType r_norm = cusp::blas::nrm2(r);
		ValueType b_norm = cusp::blas::nrm2(b);	
		printf("\n check solution: ||b-Ax||/||b||: %e \n", r_norm / b_norm );	
       
	    #ifdef PRINT_X
	    printf("\nGlobal x:\n");
		for(int i = 0; i < x.size(); i++)
			printf("%d, %f \n", i, (float)x[i]);
		#endif
    }
	}
	#endif
	#endif
    #endif
	// solve with diagonal preconditioner
    {
        std::cout << "\nSolving with diagonal preconditioner (M = D^-1)" << std::endl;
        
        // allocate storage for solution (x) and right hand side (b)
        cusp::array1d<ValueType, MemorySpace> x(A.num_rows, 0);
        cusp::array1d<ValueType, MemorySpace> b(A.num_rows, 1);

        // set stopping criteria (iteration_limit = 1000, relative_tolerance = 1e-6)
        cusp::default_monitor<ValueType> monitor(b, iter_outer, 1e-6, 1e-10);

        // setup preconditioner
		timer t1;
        cusp::precond::diagonal<ValueType, MemorySpace> M(A);
        std::cout << "\nconstruct preconditioner in " << t1.milliseconds_elapsed() << " ms " << std::endl;
		
        // solve
		timer t2;
        if(solver == 0)
        cusp::krylov::cg(A, x, b, monitor,M);
		else if(solver == 1) 
		cusp::krylov::gmres(A, x, b,restart, monitor,M);	
		else if(solver == 2) 
		cusp::krylov::bicgstab(A, x, b, monitor,M);
		std::cout << "\nsolve with diagonal preconditioner time " << t2.milliseconds_elapsed() << " ms " << std::endl;
		
        // report status
        report_status(monitor);
		
		// check solution
		cusp::array1d<ValueType,MemorySpace> y(A.num_rows);	
		cusp::array1d<ValueType,MemorySpace> r(A.num_rows);			
		// y <- Ax
		cusp::multiply(A, x, y);
		// r <- b - A*x
		blas::axpby(b, y, r, ValueType(1), ValueType(-1));
		ValueType r_norm = cusp::blas::nrm2(r);
		ValueType b_norm = cusp::blas::nrm2(b);	
		printf("\n check solution: ||b-Ax||/||b||: %e \n", r_norm / b_norm );	
       
	    #ifdef PRINT_X
	    printf("\nGlobal x:\n");
		for(int i = 0; i < x.size(); i++)
			printf("%d, %f \n", i, (float)x[i]);
		#endif
    }
	#ifdef PRECISION_S
	// solve with single precision preconditioner
    {
        std::cout << "\nSolving with SA preconditioner in single precision" << std::endl;
		
		ValueType theta = -1.;
		while(theta < 0)
		{
			printf("\nPlease input theta: ");
			scanf("%lf", &theta);
		}
        
        // allocate storage for solution (x) and right hand side (b)
        cusp::array1d<ValueType, MemorySpace> x(A.num_rows, 0);
        cusp::array1d<ValueType, MemorySpace> b(A.num_rows, 1);
        cusp::array1d<float, MemorySpace> x_d2s(x);		
        cusp::array1d<float, MemorySpace> b_d2s(b);	
		cusp::coo_matrix<IndexType, float, MemorySpace> A_float(A);
		cusp::array1d<ValueType, MemorySpace> x_s2d;

        // set stopping criteria (iteration_limit = 1000, relative_tolerance = 1e-6)
        cusp::default_monitor<ValueType> monitor(b, iter_outer, 1e-6, 1e-10);

        // setup preconditioner
		timer t1;
        cusp::precond::smoothed_aggregation<IndexType, float, MemorySpace> M_float(A_float, theta);
        std::cout << "\nconstruct preconditioner in " << t1.milliseconds_elapsed() << " ms " << std::endl;
		
        // solve
		timer t2;
        if(solver == 0)
        cusp::krylov::cg(A_float, x_d2s, b_d2s, monitor,M_float);
		else if(solver == 1) 
		cusp::krylov::gmres(A_float, x_d2s, b_d2s,restart, monitor,M_float);	
		else if(solver == 2) 
		cusp::krylov::bicgstab(A_float, x_d2s, b_d2s, monitor,M_float);
		std::cout << "\nsolve with SA preconditioner time " << t2.milliseconds_elapsed() << " ms " << std::endl;
		
		x_s2d = x_d2s;
		
        // report status
        report_status(monitor);
		
		// check solution
		cusp::array1d<ValueType,MemorySpace> y(A.num_rows);	
		cusp::array1d<ValueType,MemorySpace> r(A.num_rows);			
		// y <- Ax
		cusp::multiply(A, x_s2d, y);
		// r <- b - A*x
		blas::axpby(b, y, r, ValueType(1), ValueType(-1));
		ValueType r_norm = cusp::blas::nrm2(r);
		ValueType b_norm = cusp::blas::nrm2(b);	
		printf("\n check solution: ||b-Ax||/||b||: %e \n", r_norm / b_norm );	
       
	    #ifdef PRINT_X
	    printf("\nGlobal x:\n");
		for(int i = 0; i < x_s2d.size(); i++)
			printf("%d, %f \n", i, (float)x_s2d[i]);
		#endif
    }
	#endif
	
    // solve with SA preconditioner
    {
        std::cout << "\nSolving with SA preconditioner" << std::endl;
		
		ValueType theta = -1.;
		while(theta < 0)
		{
			printf("\nPlease input theta: ");
			scanf("%lf", &theta);
		}
        
        // allocate storage for solution (x) and right hand side (b)
        cusp::array1d<ValueType, MemorySpace> x(A.num_rows, 0);
        cusp::array1d<ValueType, MemorySpace> b(A.num_rows, 1);

        // set stopping criteria (iteration_limit = 1000, relative_tolerance = 1e-6)
        cusp::default_monitor<ValueType> monitor(b, iter_outer, 1e-6, 1e-10);

        // setup preconditioner
		timer t1;
        cusp::precond::smoothed_aggregation<IndexType, ValueType, MemorySpace> M(A, theta);
        std::cout << "\nconstruct preconditioner in " << t1.milliseconds_elapsed() << " ms " << std::endl;
		
        // solve
		timer t2;
        if(solver == 0)
        cusp::krylov::cg(A, x, b, monitor,M);
		else if(solver == 1) 
		cusp::krylov::gmres(A, x, b,restart, monitor,M);	
		else if(solver == 2) 
		cusp::krylov::bicgstab(A, x, b, monitor,M);
		std::cout << "\nsolve with SA preconditioner time " << t2.milliseconds_elapsed() << " ms " << std::endl;
		
        // report status
        report_status(monitor);
		
		// check solution
		cusp::array1d<ValueType,MemorySpace> y(A.num_rows);	
		cusp::array1d<ValueType,MemorySpace> r(A.num_rows);			
		// y <- Ax
		cusp::multiply(A, x, y);
		// r <- b - A*x
		blas::axpby(b, y, r, ValueType(1), ValueType(-1));
		ValueType r_norm = cusp::blas::nrm2(r);
		ValueType b_norm = cusp::blas::nrm2(b);	
		printf("\n check solution: ||b-Ax||/||b||: %e \n", r_norm / b_norm );	
       
	    #ifdef PRINT_X
	    printf("\nGlobal x:\n");
		for(int i = 0; i < x.size(); i++)
			printf("%d, %f \n", i, (float)x[i]);
		#endif
		
		IndexType flops;
		M.get_flops(&flops);
		printf("The flops of PCApply is: %d \n",flops);
		
		double time;
		M.get_time(&time);
		printf("Time of PCApply is: %f \n",time);	
    }
	#if 0
    // solve with PAMG preconditioner
    {
        std::cout << "\nSolving with PAMG preconditioner" << std::endl;
        
        // allocate storage for solution (x) and right hand side (b)
        cusp::array1d<ValueType, MemorySpace> x(A.num_rows, 0);
        cusp::array1d<ValueType, MemorySpace> b(A.num_rows, 1);

        // set stopping criteria (iteration_limit = 1000, relative_tolerance = 1e-6)
        cusp::default_monitor<ValueType> monitor(b, iter_outer, 1e-6, 1e-10);

        // setup preconditioner
		timer t1;
        cusp::precond::pamg_parameters<IndexType,ValueType> parms;
      
        parms.coarsening_unknowns_threshold = 100;
        parms.smoother_sweep_count = 3;
        parms.level_count = 15;
        parms.max_path_length = 32;
        ValueType alpha = 0.0;
        cusp::precond::pamg<IndexType, ValueType, MemorySpace> M(A, parms, alpha, 0);
        std::cout << "\nconstruct preconditioner in " << t1.milliseconds_elapsed() << " ms " << std::endl;
		
        // solve
		timer t2;
        if(solver == 0)
        cusp::krylov::cg(A, x, b, monitor,M);
		else if(solver == 1) 
		cusp::krylov::gmres(A, x, b,restart, monitor,M);	
		else if(solver == 2) 
		cusp::krylov::bicgstab(A, x, b, monitor,M);
		std::cout << "\nsolve with PAMG preconditioner time " << t2.milliseconds_elapsed() << " ms " << std::endl;
		
        // report status
        report_status(monitor);
		
		// check solution
		cusp::array1d<ValueType,MemorySpace> y(A.num_rows);	
		cusp::array1d<ValueType,MemorySpace> r(A.num_rows);			
		// y <- Ax
		cusp::multiply(A, x, y);
		// r <- b - A*x
		blas::axpby(b, y, r, ValueType(1), ValueType(-1));
		ValueType r_norm = cusp::blas::nrm2(r);
		ValueType b_norm = cusp::blas::nrm2(b);	
		printf("\n check solution: ||b-Ax||/||b||: %e \n", r_norm / b_norm );	
       
	    #ifdef PRINT_X
	    printf("\nGlobal x:\n");
		for(int i = 0; i < x.size(); i++)
			printf("%d, %f \n", i, (float)x[i]);
		#endif
    }
	#endif
    // solve with AINV preconditioner
    {
        std::cout << "\nSolving with AINV preconditioner" << std::endl;
		int scaled = -1;
		while(scaled < 0 || scaled > 1)
		{
			printf("\nPlease input scaled bool: ");
			scanf("%d", &scaled);
		}		
        
        // allocate storage for solution (x) and right hand side (b)
        cusp::array1d<ValueType, MemorySpace> x(A.num_rows, 0);
        cusp::array1d<ValueType, MemorySpace> b(A.num_rows, 1);

        // set stopping criteria (iteration_limit = 1000, relative_tolerance = 1e-6)
        cusp::default_monitor<ValueType> monitor(b, iter_outer, 1e-6, 1e-10);
		
		ValueType drop_tolerance=0.05;
		int 	  nonzero_per_row=-1;
		bool 	  lin_dropping=true;
		int 	  lin_param=5;
		void 	  *M;
		
        // setup preconditioner
		timer t1;
		if(scaled == 0){
        M = new cusp::precond::bridson_ainv<ValueType, MemorySpace>(A, drop_tolerance, nonzero_per_row, lin_dropping, lin_param);		
		}else if(scaled == 1){
        M = new cusp::precond::scaled_bridson_ainv<ValueType, MemorySpace>(A, drop_tolerance, nonzero_per_row, lin_dropping, lin_param);		
		}

        std::cout << "\nconstruct preconditioner in " << t1.milliseconds_elapsed() << " ms " << std::endl;
		
        // solve
		timer t2;
		if(scaled == 0){
		if(solver == 0)
        cusp::krylov::cg(A, x, b, monitor,*(cusp::precond::bridson_ainv<ValueType, MemorySpace>*)M);
		else if(solver == 1) 
		cusp::krylov::gmres(A, x, b,restart, monitor,*(cusp::precond::bridson_ainv<ValueType, MemorySpace>*)M);	
		else if(solver == 2) 
		cusp::krylov::bicgstab(A, x, b, monitor,*(cusp::precond::bridson_ainv<ValueType, MemorySpace>*)M);
		}else if(scaled == 1){
		if(solver == 0)
        cusp::krylov::cg(A, x, b, monitor,*(cusp::precond::scaled_bridson_ainv<ValueType, MemorySpace>*)M);
		else if(solver == 1) 
		cusp::krylov::gmres(A, x, b,restart, monitor,*(cusp::precond::scaled_bridson_ainv<ValueType, MemorySpace>*)M);	
		else if(solver == 2) 
		cusp::krylov::bicgstab(A, x, b, monitor,*(cusp::precond::scaled_bridson_ainv<ValueType, MemorySpace>*)M);
		}	
		std::cout << "\nsolve with AINV preconditioner time " << t2.milliseconds_elapsed() << " ms " << std::endl;
		
        // report status
        report_status(monitor);
		
		// check solution
		cusp::array1d<ValueType,MemorySpace> y(A.num_rows);	
		cusp::array1d<ValueType,MemorySpace> r(A.num_rows);			
		// y <- Ax
		cusp::multiply(A, x, y);
		// r <- b - A*x
		blas::axpby(b, y, r, ValueType(1), ValueType(-1));
		ValueType r_norm = cusp::blas::nrm2(r);
		ValueType b_norm = cusp::blas::nrm2(b);	
		printf("\n check solution: ||b-Ax||/||b||: %e \n", r_norm / b_norm );	
       
	    #ifdef PRINT_X
	    printf("\nGlobal x:\n");
		for(int i = 0; i < x.size(); i++)
			printf("%d, %f \n", i, (float)x[i]);
		#endif
		
		if(scaled == 0){
        delete (cusp::precond::bridson_ainv<ValueType, MemorySpace>*)M;
		}else if(scaled == 1){
        delete (cusp::precond::scaled_bridson_ainv<ValueType, MemorySpace>*)M;
		}		
    }	
	#if 0
	{
      std::cout << "\n ******Solving with MAGMA_LAPACK(MKL LAPACK77) LU solver****** " << std::endl;	
		
		if( CUBLAS_STATUS_SUCCESS != cublasInit( ) ) {
		fprintf(stderr, "CUBLAS: Not initialized\n"); exit(-1);
		}		
		
	  // convert to dense matrix
	  	printf("\nConverting to dense matrix");
		timer t7;	
	    cusp::array2d<ValueType, MemorySpace, cusp::column_major> A_dense(A);
		std::cout << "\nMatrix conversion time " << t7.milliseconds_elapsed() << " ms " << std::endl;
		printf("\n Memory requirested : %4.2f Mb", float(A_dense.values.size() * sizeof(ValueType))/(1024*1024));
    
        // allocate storage for solution (x) and right hand side (b)
		printf("\n n=%d, lda=%d", A_dense.num_rows, A_dense.pitch);
        cusp::array1d<ValueType, MemorySpace> b(A_dense.num_rows, 1);
        cusp::array1d<ValueType, MemorySpace> x(b);
		cusp::array1d<int,		 cusp::host_memory> pivot(A_dense.num_rows, 0);
        
		IndexType n    = A_dense.num_rows;
		IndexType m	   = n;
		IndexType lda  = A_dense.pitch;
		IndexType ldb  = n;
		IndexType info;
		const char  *up = MagmaUpperStr;
		const char  *low = MagmaLowerStr;
		const char  *trans = MagmaNoTransStr;
		const char  *unit = MagmaUnitStr;
		const char  *nounit = MagmaNonUnitStr;
		
		ValueType *dense_array   = thrust::raw_pointer_cast(&*A_dense.values.begin());	
		ValueType *h_dense_array = (ValueType*)malloc(n*n * sizeof(ValueType));                   
		if (h_dense_array == 0) {							
			fprintf (stderr, "!!!! Malloc failed for: %s\n", h_dense_array );		
			exit(-1);								
		}
		ValueType *x_array 	 = thrust::raw_pointer_cast(&*x.begin());
		ValueType *h_x_array = (ValueType*)malloc(A_dense.num_rows * sizeof(ValueType));                   
		if (h_x_array == 0) {							
			fprintf (stderr, "!!!! Malloc failed for: %s\n", h_x_array );		
			exit(-1);								
		}
		cublasGetMatrix(n, n, sizeof(ValueType),dense_array, A_dense.values.size(), h_dense_array, n*n);
		cublasGetVector(n, sizeof(ValueType), x_array, 1, h_x_array, 1);
		IndexType *pivot_array = thrust::raw_pointer_cast(&*pivot.begin());

        // solve
		printf("\n hehe");
		timer t8;
		printf("\nCalling lapackf77_sgetrf");
        //lapackf77_xgetrf( &m, &n, h_dense_array,&n,pivot_array, &info);
		lapackf77_xpotrf(low, &n, h_dense_array, &n, &info);
		printf("\nCalling blasf77_strsv");		
        blasf77_xtrsv( low, trans, unit, &n, h_dense_array,&n, h_x_array,&ldb);
        blasf77_xtrsv( up, trans, nounit, &n, h_dense_array,&n,h_x_array,&ldb);
		std::cout << "\nsolve with MAGMA_LAPACK LU solver time " << t8.milliseconds_elapsed() << " ms " << std::endl;	
		
		// check solution
		cublasSetVector(n, sizeof(ValueType), h_x_array, 1, x_array, 1);
		cusp::array1d<ValueType,MemorySpace> y(A_dense.num_rows);	
		cusp::array1d<ValueType,MemorySpace> r(A_dense.num_rows);			
		// y <- Ax
		cusp::multiply(A, x, y);
		// r <- b - A*x
		blas::axpby(b, y, r, ValueType(1), ValueType(-1));
		ValueType r_norm = cusp::blas::nrm2(r);
		ValueType b_norm = cusp::blas::nrm2(b);	
		printf("\n check solution: ||b-Ax||/||b||: %e \n", r_norm / b_norm );	
		
		cublasShutdown();
		free(h_dense_array);
		free(h_x_array);
		
		#ifdef PRINT_X
	    printf("\nGlobal x:\n");
		for(int i = 0; i < x.size(); i++)
			printf("%d, %f \n", i, (float)x[i]);
		#endif
	}
	#endif
	#if 0
	#ifdef PRECISION_S
	void checkStatus(culaStatus status)
	{
		char buf[80];

		if(!status)
			return;

		culaGetErrorInfoString(status, culaGetErrorInfo(), buf, sizeof(buf));
		printf("%s\n", buf);

		culaShutdown();
		exit(EXIT_FAILURE);
	}	
	{
      std::cout << "\n ******Solving with CULA LU solver****** " << std::endl;
	  
		culaStatus status;
		
		printf("\nInitializing CULA");
		status = culaInitialize();
		checkStatus(status);
		
	  // convert to dense matrix
	  	printf("\nConverting to dense matrix");
		timer t3;	
	    cusp::array2d<ValueType, MemorySpace, cusp::column_major> A_dense(A);
		std::cout << "\nMatrix conversion time " << t3.milliseconds_elapsed() << " ms " << std::endl;
		printf("\n Memory requirested : %4.2f Mb", float(A_dense.values.size() * sizeof(ValueType))/(1024*1024));
    
        // allocate storage for solution (x) and right hand side (b)
		printf("\n n=%d, lda=%d", A_dense.num_rows, A_dense.pitch);
        cusp::array1d<ValueType, MemorySpace> b(A_dense.num_rows, 1);
        cusp::array1d<ValueType, MemorySpace> x(b);
		cusp::array1d<int,		 MemorySpace> pivot(A_dense.num_rows, 0);
        
		IndexType n    = A_dense.num_rows;
		IndexType nrhs = 1;
		IndexType lda  = A_dense.pitch;
		IndexType ldb  = n;
		ValueType *dense_array = thrust::raw_pointer_cast(&*A_dense.values.begin());
		IndexType *pivot_array = thrust::raw_pointer_cast(&*pivot.begin());
		ValueType *x_array 	   = thrust::raw_pointer_cast(&*x.begin());
		
		#ifdef PRINT_DENSE_A
	    printf("\nArray of dense A:\n");
		for(int i = 0; i < A_dense.values.size(); i++)
			printf("%d, %f \n", i, (float)(A_dense.values)[i]);
		#endif
		
        // solve
		timer t4;
		printf("\nCalling culaDeviceSgesv");
		status = culaDeviceSgesv(n,nrhs,dense_array,lda,pivot_array,x_array,ldb);
		checkStatus(status);
		std::cout << "\nsolve with CULA LU solver time " << t4.milliseconds_elapsed() << " ms " << std::endl;

		// check solution
		cusp::array1d<ValueType,MemorySpace> y(A_dense.num_rows);	
		cusp::array1d<ValueType,MemorySpace> r(A_dense.num_rows);			
		// y <- Ax
		cusp::multiply(A, x, y);
		// r <- b - A*x
		blas::axpby(b, y, r, ValueType(1), ValueType(-1));
		ValueType r_norm = cusp::blas::nrm2(r);
		ValueType b_norm = cusp::blas::nrm2(b);	
		printf("\n check solution: ||b-Ax||/||b||: %e \n", r_norm / b_norm );	
		
		printf("\nShutting down CULA\n");
		culaShutdown();
		
		#ifdef PRINT_X
	    printf("\nGlobal x:\n");
		for(int i = 0; i < x.size(); i++)
			printf("%d, %f \n", i, (float)x[i]);
		#endif
	} 
	#endif
	#if 0
	{
      std::cout << "\n ******Solving with CULA Cholesky solver****** " << std::endl;
	  
		culaStatus status;
		
		printf("\nInitializing CULA");
		status = culaInitialize();
		checkStatus(status);
		
	  // convert to dense matrix
	  	printf("\nConverting to dense matrix");
		timer t5;	
	    cusp::array2d<ValueType, MemorySpace, cusp::column_major> A_dense(A);
		std::cout << "\nMatrix conversion time " << t5.milliseconds_elapsed() << " ms " << std::endl;
		printf("\n Memory requirested : %4.2f Mb", float(A_dense.values.size() * sizeof(ValueType))/(1024*1024));
    
        // allocate storage for solution (x) and right hand side (b)
		printf("\n n=%d, lda=%d", A_dense.num_rows, A_dense.pitch);
        cusp::array1d<ValueType, MemorySpace> b(A_dense.num_rows, 1);
        cusp::array1d<ValueType, MemorySpace> x(b);
        
		IndexType n    = A_dense.num_rows;
		IndexType nrhs = 1;
		IndexType lda  = A_dense.pitch;
		IndexType ldb  = n;
		ValueType *dense_array = thrust::raw_pointer_cast(&*A_dense.values.begin());
		ValueType *x_array 	   = thrust::raw_pointer_cast(&*x.begin());
		
		#ifdef PRINT_DENSE_A
	    printf("\nArray of dense A:\n");
		for(int i = 0; i < A_dense.values.size(); i++)
			printf("%d, %f \n", i, (float)(A_dense.values)[i]);
		#endif
		
        // solve
		timer t6;
		printf("\nCalling culaDeviceSposv");
		// culaDeviceSposv not provided by CULA basic version
		status = culaDeviceSposv('L',n,nrhs,dense_array,lda,x_array,ldb);
		// or the same function as the two below:
		//status = culaDeviceSpotrf('L',n,dense_array,lda);
		//status = culaDeviceSpotrs('L',n,nrhs,dense_array,lda,x_array,ldb);
		checkStatus(status);
		std::cout << "\nsolve with CULA Cholesky solver time " << t6.milliseconds_elapsed() << " ms " << std::endl;

		// check solution
		cusp::array1d<ValueType,MemorySpace> y(A_dense.num_rows);	
		cusp::array1d<ValueType,MemorySpace> r(A_dense.num_rows);			
		// y <- Ax
		cusp::multiply(A, x, y);
		// r <- b - A*x
		blas::axpby(b, y, r, ValueType(1), ValueType(-1));
		ValueType r_norm = cusp::blas::nrm2(r);
		ValueType b_norm = cusp::blas::nrm2(b);	
		printf("\n check solution: ||b-Ax||/||b||: %e \n", r_norm / b_norm );	
		
		printf("\nShutting down CULA\n");
		culaShutdown();
		
		#ifdef PRINT_X
	    printf("\nGlobal x:\n");
		for(int i = 0; i < x.size(); i++)
			printf("%d, %f \n", i, (float)x[i]);
		#endif
	}
	#endif
	#endif
	#if 0	
	{
      std::cout << "\n ******Solving with MAGMA LU solver****** " << std::endl;
		
		if( CUBLAS_STATUS_SUCCESS != cublasInit( ) ) {
		fprintf(stderr, "CUBLAS: Not initialized\n"); exit(-1);
		}		
		
	  // convert to dense matrix
	  	printf("\nConverting to dense matrix");
		timer t7;	
	    cusp::array2d<ValueType, MemorySpace, cusp::column_major> A_dense(A);
		std::cout << "\nMatrix conversion time " << t7.milliseconds_elapsed() << " ms " << std::endl;
		printf("\n Memory requirested : %4.2f Mb", float(A_dense.values.size() * sizeof(ValueType))/(1024*1024));
    
        // allocate storage for solution (x) and right hand side (b)
		printf("\n n=%d, lda=%d", A_dense.num_rows, A_dense.pitch);
        cusp::array1d<ValueType, MemorySpace> b(A_dense.num_rows, 1);
        cusp::array1d<ValueType, MemorySpace> x(b);
		cusp::array1d<int,		 cusp::host_memory> pivot(A_dense.num_rows, 0);
        
		IndexType n    = A_dense.num_rows;
		IndexType nrhs = 1;
		IndexType lda  = A_dense.pitch;
		IndexType ldb  = n;
		IndexType info;
		char trans = MagmaNoTrans;
		
		ValueType *dense_array = thrust::raw_pointer_cast(&*A_dense.values.begin());
		IndexType *pivot_array = thrust::raw_pointer_cast(&*pivot.begin());
		ValueType *x_array 	   = thrust::raw_pointer_cast(&*x.begin());
		
		#ifdef PRINT_DENSE_A
	    printf("\nArray of dense A:\n");
		for(int i = 0; i < A_dense.values.size(); i++)
			printf("%d, %f \n", i, (float)(A_dense.values)[i]);
		#endif
		
        // solve
		timer t8;
		printf("\nCalling magma_xgetrf_gpu and magma_xgetrs_gpu");
        magma_xgetrf_gpu( n, n, dense_array,lda,pivot_array, &info);
        magma_xgetrs_gpu( trans, n, nrhs, dense_array,lda,pivot_array,x_array,ldb,&info);
		std::cout << "\nsolve with MAGMA LU solver time " << t8.milliseconds_elapsed() << " ms " << std::endl;	
		
		// check solution
		cusp::array1d<ValueType,MemorySpace> y(A_dense.num_rows);	
		cusp::array1d<ValueType,MemorySpace> r(A_dense.num_rows);			
		// y <- Ax
		cusp::multiply(A, x, y);
		// r <- b - A*x
		blas::axpby(b, y, r, ValueType(1), ValueType(-1));
		ValueType r_norm = cusp::blas::nrm2(r);
		ValueType b_norm = cusp::blas::nrm2(b);	
		printf("\n check solution: ||b-Ax||/||b||: %e \n", r_norm / b_norm );	
		
		cublasShutdown();
		
		#ifdef PRINT_X
	    printf("\nGlobal x:\n");
		for(int i = 0; i < x.size(); i++)
			printf("%d, %f \n", i, (float)x[i]);
		#endif
	}	
	{
      std::cout << "\n ******Solving with MAGMA Cholesky solver****** " << std::endl;
		
		if( CUBLAS_STATUS_SUCCESS != cublasInit( ) ) {
		fprintf(stderr, "CUBLAS: Not initialized\n"); exit(-1);
		}		
		
	  // convert to dense matrix
	  	printf("\nConverting to dense matrix");
		timer t9;	
	    cusp::array2d<ValueType, MemorySpace, cusp::column_major> A_dense(A);
		std::cout << "\nMatrix conversion time " << t9.milliseconds_elapsed() << " ms " << std::endl;
		printf("\n Memory requirested : %4.2f Mb", float(A_dense.values.size() * sizeof(ValueType))/(1024*1024));
    
        // allocate storage for solution (x) and right hand side (b)
		printf("\n n=%d, lda=%d", A_dense.num_rows, A_dense.pitch);
        cusp::array1d<ValueType, MemorySpace> b(A_dense.num_rows, 1);
        cusp::array1d<ValueType, MemorySpace> x(b);
        
		IndexType n    = A_dense.num_rows;
		IndexType nrhs = 1;
		IndexType lda  = A_dense.pitch;
		IndexType ldb  = n;
		IndexType info;
		const char *uplo = MagmaLowerStr;
		
		ValueType *dense_array = thrust::raw_pointer_cast(&*A_dense.values.begin());
		ValueType *x_array 	   = thrust::raw_pointer_cast(&*x.begin());
		
		#ifdef PRINT_DENSE_A
	    printf("\nArray of dense A:\n");
		for(int i = 0; i < A_dense.values.size(); i++)
			printf("%d, %f \n", i, (float)(A_dense.values)[i]);
		#endif
		
        // solve
		timer t10;
		printf("\nCalling magma_xpotrf_gpu and magma_xpotrs_gpu");
		magma_xpotrf_gpu(uplo[0], n, dense_array, lda, &info);
		magma_xpotrs_gpu(uplo[0], n, nrhs, dense_array, lda, x_array, ldb, &info);
		std::cout << "\nsolve with MAGMA Cholesky solver time " << t10.milliseconds_elapsed() << " ms " << std::endl;
		
		// check solution
		cusp::array1d<ValueType,MemorySpace> y(A_dense.num_rows);	
		cusp::array1d<ValueType,MemorySpace> r(A_dense.num_rows);			
		// y <- Ax
		cusp::multiply(A, x, y);
		// r <- b - A*x
		blas::axpby(b, y, r, ValueType(1), ValueType(-1));
		ValueType r_norm = cusp::blas::nrm2(r);
		ValueType b_norm = cusp::blas::nrm2(b);	
		printf("\n check solution: ||b-Ax||/||b||: %e \n", r_norm / b_norm );	
		
		cublasShutdown();
		
		#ifdef PRINT_X
	    printf("\nGlobal x:\n");
		for(int i = 0; i < x.size(); i++)
			printf("%d, %f \n", i, (float)x[i]);
		#endif
	}		
	#ifdef PRECISION_D
	{
      std::cout << "\n ******Solving with MAGMA LU solver by iterative refinement****** " << std::endl;
	  		
		if( CUBLAS_STATUS_SUCCESS != cublasInit( ) ) {
		fprintf(stderr, "CUBLAS: Not initialized\n"); exit(-1);
		}		
		
	    // convert to dense matrix
	  	printf("\nConverting to dense matrix");
		timer t11;	
	    cusp::array2d<ValueType, MemorySpace, cusp::column_major> A_dense(A);
		std::cout << "\nMatrix conversion time " << t11.milliseconds_elapsed() << " ms " << std::endl;
		printf("\n Memory requirested : %4.2f Mb", float(A_dense.values.size() * sizeof(ValueType))/(1024*1024));
    
        // allocate storage for solution (x) and right hand side (b)
		printf("\n n=%d, lda=%d", A_dense.num_rows, A_dense.pitch);
        cusp::array1d<ValueType, MemorySpace> b(A_dense.num_rows, 1);
        cusp::array1d<ValueType, MemorySpace> x(A_dense.num_rows, 0);
		cusp::array1d<int,		 cusp::host_memory> h_pivot(A_dense.num_rows, 0);
		cusp::array1d<int,		 cusp::device_memory> d_pivot(A_dense.num_rows, 0);
        
		IndexType n    = A_dense.num_rows;
		IndexType nrhs = 1;
		IndexType lda  = A_dense.pitch;
		IndexType ldb  = n;
		IndexType ldx  = ldb;
		IndexType iter,info;
		char trans = MagmaNoTrans;
		
		// allocate storage for work space
		cusp::array1d<float, 	 cusp::device_memory> d_WORKS(lda*(n+nrhs));
		cusp::array1d<double, 	 cusp::device_memory> d_WORKD(n+nrhs);
		
		ValueType *dense_array   = thrust::raw_pointer_cast(&*A_dense.values.begin());
		ValueType *b_array 	     = thrust::raw_pointer_cast(&*b.begin());		
		ValueType *x_array 	     = thrust::raw_pointer_cast(&*x.begin());
		IndexType *h_pivot_array = thrust::raw_pointer_cast(&*h_pivot.begin());
		IndexType *d_pivot_array = thrust::raw_pointer_cast(&*d_pivot.begin());
		float     *d_WORKS_array = thrust::raw_pointer_cast(&*d_WORKS.begin());
		double	  *d_WORKD_array = thrust::raw_pointer_cast(&*d_WORKD.begin());
		
		#ifdef PRINT_DENSE_A
	    printf("\nArray of dense A:\n");
		for(int i = 0; i < A_dense.values.size(); i++)
			printf("%d, %f \n", i, (float)(A_dense.values)[i]);
		#endif
		
        // solve
		timer t12;
		printf("\nCalling magma_dsgesv_gpu");
		magma_dsgesv_gpu( trans, n, nrhs, 
			  dense_array, lda, h_pivot_array, d_pivot_array, 
			  b_array, ldb, x_array, ldx, 
			  d_WORKD_array, d_WORKS_array, &iter, &info);
		if (info < 0)
            printf("Argument %d of magma_dsgesv_gpu had an illegal value.\n", -info);	
		std::cout << "\nsolve with MAGMA LU solver by iterative refinement time " << t12.milliseconds_elapsed() << " ms " << std::endl;	
		
		// check solution
		cusp::array1d<ValueType,MemorySpace> y(A_dense.num_rows);	
		cusp::array1d<ValueType,MemorySpace> r(A_dense.num_rows);			
		// y <- Ax
		cusp::multiply(A, x, y);
		// r <- b - A*x
		blas::axpby(b, y, r, ValueType(1), ValueType(-1));
		ValueType r_norm = cusp::blas::nrm2(r);
		ValueType b_norm = cusp::blas::nrm2(b);	
		printf("\n check solution: ||b-Ax||/||b||: %e \n", r_norm / b_norm );	
		
		cublasShutdown();
		
		#ifdef PRINT_X
	    printf("\nGlobal x:\n");
		for(int i = 0; i < x.size(); i++)
			printf("%d, %f \n", i, (float)x[i]);
		#endif
	}	
	{
      std::cout << "\n ******Solving with MAGMA Cholesky solver by iterative refinement****** " << std::endl;
	  		
		if( CUBLAS_STATUS_SUCCESS != cublasInit( ) ) {
		fprintf(stderr, "CUBLAS: Not initialized\n"); exit(-1);
		}		
		
	    // convert to dense matrix
	  	printf("\nConverting to dense matrix");
		timer t13;	
	    cusp::array2d<ValueType, MemorySpace, cusp::column_major> A_dense(A);
		std::cout << "\nMatrix conversion time " << t13.milliseconds_elapsed() << " ms " << std::endl;
		printf("\n Memory requirested : %4.2f Mb", float(A_dense.values.size() * sizeof(ValueType))/(1024*1024));
    
        // allocate storage for solution (x) and right hand side (b)
		printf("\n n=%d, lda=%d", A_dense.num_rows, A_dense.pitch);
        cusp::array1d<ValueType, MemorySpace> b(A_dense.num_rows, 1);
        cusp::array1d<ValueType, MemorySpace> x(A_dense.num_rows, 0);
		cusp::array1d<int,		 cusp::host_memory> h_pivot(A_dense.num_rows, 0);
		cusp::array1d<int,		 cusp::device_memory> d_pivot(A_dense.num_rows, 0);
        
		IndexType n    = A_dense.num_rows;
		IndexType nrhs = 1;
		IndexType lda  = A_dense.pitch;
		IndexType ldb  = n;
		IndexType ldx  = ldb;
		IndexType iter,info;
		const char *uplo = MagmaLowerStr;
		
		// allocate storage for work space
		cusp::array1d<float, 	 cusp::device_memory> d_WORKS(lda*(n+nrhs));
		cusp::array1d<double, 	 cusp::device_memory> d_WORKD(n+nrhs);
		
		ValueType *dense_array   = thrust::raw_pointer_cast(&*A_dense.values.begin());
		ValueType *b_array 	     = thrust::raw_pointer_cast(&*b.begin());		
		ValueType *x_array 	     = thrust::raw_pointer_cast(&*x.begin());
		IndexType *h_pivot_array = thrust::raw_pointer_cast(&*h_pivot.begin());
		IndexType *d_pivot_array = thrust::raw_pointer_cast(&*d_pivot.begin());
		float     *d_WORKS_array = thrust::raw_pointer_cast(&*d_WORKS.begin());
		double	  *d_WORKD_array = thrust::raw_pointer_cast(&*d_WORKD.begin());
		
		#ifdef PRINT_DENSE_A
	    printf("\nArray of dense A:\n");
		for(int i = 0; i < A_dense.values.size(); i++)
			printf("%d, %f \n", i, (float)(A_dense.values)[i]);
		#endif
		
        // solve
		timer t14;
		printf("\nCalling magma_dsposv_gpu");
		magma_dsposv_gpu(uplo[0], n, nrhs, dense_array, lda, b_array, ldb, x_array, ldx, 
			 d_WORKD_array, d_WORKS_array, &iter, &info);
		if (info < 0)
            printf("Argument %d of magma_dsposv_gpu had an illegal value.\n", -info);	
		std::cout << "\nsolve with MAGMA Cholesky solver by iterative refinement time " << t14.milliseconds_elapsed() << " ms " << std::endl;	
		
		// check solution
		cusp::array1d<ValueType,MemorySpace> y(A_dense.num_rows);	
		cusp::array1d<ValueType,MemorySpace> r(A_dense.num_rows);			
		// y <- Ax
		cusp::multiply(A, x, y);
		// r <- b - A*x
		blas::axpby(b, y, r, ValueType(1), ValueType(-1));
		ValueType r_norm = cusp::blas::nrm2(r);
		ValueType b_norm = cusp::blas::nrm2(b);	
		printf("\n check solution: ||b-Ax||/||b||: %e \n", r_norm / b_norm );	
		
		cublasShutdown();
		
		#ifdef PRINT_X
	    printf("\nGlobal x:\n");
		for(int i = 0; i < x.size(); i++)
			printf("%d, %f \n", i, (float)x[i]);
		#endif
	}	
	#endif
	#endif
	
    return 0;
}

