#include <cusparse.h>
#include <cusp/print.h>
#include <cusp/relaxation/ilupp.h>
#include <cusp/io/matrix_market.h>
#include <cusp/blas.h>
#include <cusp/krylov/cg.h>
#include <cusp/krylov/gmres.h>
#include <cusp/gallery/poisson.h>
#include <cusp/csr_matrix.h>
#include <cusp/precond/diagonal.h>
#include <cusp/permutation.h>
#include <iostream>
#include <performance/timer.h>
#include <cmath>

#include <petscpc.h>
#include <../src/mat/impls/aij/seq/seqcusp/cuspmatimpl.h>
#include <../src/mat/impls/aij/seq/aij.h>

#define PI 3.1415926

template <typename MatrixType, typename VectorType>
PetscErrorCode setup_petsc_lower(Mat, MatrixType&, VectorType&,
 cusparseHandle_t&, cusparseStatus_t&, cusparseMatDescr_t&, cusparseSolveAnalysisInfo_t&, bool);
 
template <typename MatrixType, typename VectorType>
PetscErrorCode setup_petsc_upper(Mat, MatrixType&, VectorType&, VectorType&,
 cusparseHandle_t&, cusparseStatus_t&, cusparseMatDescr_t&, cusparseSolveAnalysisInfo_t&, bool); 
 
template <typename Monitor>
void report_status(Monitor& monitor)
{
    if (monitor.converged())
    {
        std::cout << "  Solver converged to " << monitor.tolerance1() << " tolerance";
        std::cout << " after " << monitor.iteration_count() << " iterations";
        std::cout << " (" << monitor.relative_residual_norm() << " final residual)" << std::endl;
    }
    else
    {
        std::cout << "  Solver reached iteration limit " << monitor.iteration_limit() << " before converging";
        std::cout << " to " << monitor.tolerance1() << " tolerance ";
        std::cout << " (" << monitor.relative_residual_norm() << " final residual)" << std::endl;
    }
}

template <typename MatrixType, typename VectorType>
PetscErrorCode setup_petsc_lower(Mat Factor, MatrixType& lower, VectorType& ipr,
 cusparseHandle_t& handle, cusparseStatus_t& status, cusparseMatDescr_t& descrM_lo, cusparseSolveAnalysisInfo_t& Analy_lo, bool use_mkl_sv)
{
	PetscErrorCode ierr;  	
	PetscInt i, nz, nzLower, offset, rowOffset, n = Factor->rmap->n;
	PetscInt *AiLo, *AjLo;
	PetscScalar *AALo;
	Mat_SeqAIJ        *a = (Mat_SeqAIJ*)Factor->data;
	const PetscInt    *ai = a->i,*aj = a->j,*vi;
	const MatScalar   *aa = a->a,*v;
	IS isrow = a->row;
	const PetscInt   *r;
	VectorType pr;

    PetscFunctionBegin;	
      // first figure out the number of nonzeros in the lower triangular matrix including 1's on the diagonal
      nzLower=n+ai[n]-ai[1];
      
      // Allocate Space for the lower triangular matrix 	  
      ierr = PetscMalloc((n+1)*sizeof(PetscInt),&AiLo);CHKERRQ(ierr);
      ierr = PetscMalloc(nzLower*sizeof(PetscInt),&AjLo);CHKERRQ(ierr);	  
      ierr = PetscMalloc(nzLower*sizeof(PetscScalar),&AALo);CHKERRQ(ierr);	  	
      
      // fill the lower triangular matrix
      AiLo[0]=(PetscInt) 0;
      AiLo[n]=nzLower;
      AjLo[0]=(PetscInt) 0;
      AALo[0]=(MatScalar) 1.0;
      v    = aa;
      vi   = aj;
      offset=1;
      rowOffset=1;
      for (i=1; i<n; i++) {
		nz  = ai[i+1] - ai[i];
		// additional 1 for the term on the diagonal
		AiLo[i]=rowOffset;
		rowOffset+=nz+1;

		ierr = PetscMemcpy(&(AjLo[offset]), vi, nz*sizeof(PetscInt));CHKERRQ(ierr);
		ierr = PetscMemcpy(&(AALo[offset]), v, nz*sizeof(MatScalar));CHKERRQ(ierr);
		
		offset+=nz;
		AjLo[offset]=(PetscInt) i;
		AALo[offset]=(MatScalar) 1.0;
		offset+=1;
		
		v  += nz;
		vi += nz;
      }    

    lower.resize(n,n,nzLower);	  
	if (use_mkl_sv) {  
	// MKL is one-based, transform row_offsets and column_indices to be plusing one
	thrust::copy(thrust::make_transform_iterator(AiLo, cusp::detail::ADD<PetscInt>(1)),
	thrust::make_transform_iterator(AiLo+n+1, cusp::detail::ADD<PetscInt>(1)),
	lower.row_offsets.begin());
	thrust::copy(thrust::make_transform_iterator(AjLo, cusp::detail::ADD<PetscInt>(1)),
	thrust::make_transform_iterator(AjLo+nzLower, cusp::detail::ADD<PetscInt>(1)),
	lower.column_indices.begin());
	thrust::copy(AALo, AALo+nzLower, lower.values.begin());	
	} else {
    // The Lower triangular matrix
    lower.row_offsets.assign(AiLo,AiLo+n+1);
    lower.column_indices.assign(AjLo,AjLo+nzLower);
    lower.values.assign(AALo,AALo+nzLower);
      
    // cusparse analysis 
	status = cusparseDcsrsv_analysis(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, n, descrM_lo,
	thrust::raw_pointer_cast(&lower.values[0]),
	thrust::raw_pointer_cast(&lower.row_offsets[0]),
	thrust::raw_pointer_cast(&lower.column_indices[0]),	
	Analy_lo);
	}

	// free the temporary memory
    ierr = PetscFree(AiLo);CHKERRQ(ierr); 
    ierr = PetscFree(AjLo);CHKERRQ(ierr); 
    ierr = PetscFree(AALo);CHKERRQ(ierr);   	
	
	// get the inverse permutation from PETSc ordering 
	ierr = ISGetIndices(isrow,&r);CHKERRQ(ierr);
	pr.assign(&r[0], &r[0]+n);
    ierr = ISRestoreIndices(isrow,&r);CHKERRQ(ierr); 
	ipr.resize(pr.size());
	cusp::get_inverse_permutation(pr, ipr);	
	
    PetscFunctionReturn(0);	  	
}

template <typename MatrixType, typename VectorType>
PetscErrorCode setup_petsc_upper(Mat Factor, MatrixType& upper, VectorType& pc, VectorType& ipc,
 cusparseHandle_t& handle, cusparseStatus_t& status, cusparseMatDescr_t& descrM_up, cusparseSolveAnalysisInfo_t& Analy_up, bool use_mkl_sv)
{
	PetscErrorCode ierr;  	
	PetscInt i, nz, nzUpper, offset, n = Factor->rmap->n;
    PetscInt *AiUp, *AjUp;
    PetscScalar *AAUp;
	Mat_SeqAIJ        *a = (Mat_SeqAIJ*)Factor->data;
    const PetscInt    *aj = a->j,*adiag = a->diag,*vi;
    const MatScalar   *aa = a->a,*v;
	IS iscol = a->icol;
	const PetscInt   *c;

    PetscFunctionBegin;	
      // next, figure out the number of nonzeros in the upper triangular matrix
      nzUpper = adiag[0]-adiag[n];
      
      // Allocate Space for the upper triangular matrix
      ierr = PetscMalloc((n+1)*sizeof(PetscInt),&AiUp);CHKERRQ(ierr);
      ierr = PetscMalloc(nzUpper*sizeof(PetscInt),&AjUp);CHKERRQ(ierr);	  
      ierr = PetscMalloc(nzUpper*sizeof(PetscScalar),&AAUp);CHKERRQ(ierr);	
      
      // Fill the upper triangular matrix
      AiUp[0]=(PetscInt) 0;
      AiUp[n]=nzUpper;
      offset = nzUpper;
      for (i=n-1; i>=0; i--) {
		v   = aa + adiag[i+1] + 1;
		vi  = aj + adiag[i+1] + 1;
		
		// number of elements NOT on the diagonal
		nz = adiag[i] - adiag[i+1]-1;
		
		// decrement the offset
		offset -= (nz+1);
		
		// first, set the diagonal elements
		AjUp[offset] = (PetscInt) i;
		AAUp[offset] = 1./v[nz];
		AiUp[i] = AiUp[i+1] - (nz+1);
		
		ierr = PetscMemcpy(&(AjUp[offset+1]), vi, nz*sizeof(PetscInt));CHKERRQ(ierr);
		ierr = PetscMemcpy(&(AAUp[offset+1]), v, nz*sizeof(MatScalar));CHKERRQ(ierr);
      }  
	  
      upper.resize(n,n,nzUpper);
	  if (use_mkl_sv) {  
	  // MKL is one-based, transform row_offsets and column_indices to be plusing one
	  thrust::copy(thrust::make_transform_iterator(AiUp, cusp::detail::ADD<PetscInt>(1)),
	  thrust::make_transform_iterator(AiUp+n+1, cusp::detail::ADD<PetscInt>(1)),
	  upper.row_offsets.begin());
	  thrust::copy(thrust::make_transform_iterator(AjUp, cusp::detail::ADD<PetscInt>(1)),
	  thrust::make_transform_iterator(AjUp+nzUpper, cusp::detail::ADD<PetscInt>(1)),
	  upper.column_indices.begin());
	  thrust::copy(AAUp, AAUp+nzUpper, upper.values.begin());	
	  } else {	  
      // The Upper triangular matrix
      upper.row_offsets.assign(AiUp,AiUp+n+1);
      upper.column_indices.assign(AjUp,AjUp+nzUpper);
      upper.values.assign(AAUp,AAUp+nzUpper);
    
      // cusparse analysis 
	  status = cusparseDcsrsv_analysis(handle, CUSPARSE_OPERATION_NON_TRANSPOSE, n, descrM_up,
	  thrust::raw_pointer_cast(&upper.values[0]),
	  thrust::raw_pointer_cast(&upper.row_offsets[0]),
	  thrust::raw_pointer_cast(&upper.column_indices[0]),	
	  Analy_up);
	  }
	  // free the temporary memory	  
      ierr = PetscFree(AiUp);CHKERRQ(ierr); 
      ierr = PetscFree(AjUp);CHKERRQ(ierr); 
      ierr = PetscFree(AAUp);CHKERRQ(ierr);  	
	
	  // get the inverse permutation from PETSc ordering 
	  ierr = ISGetIndices(iscol,&c);CHKERRQ(ierr);
	  pc.assign(&c[0], &c[0]+n);
      ierr = ISRestoreIndices(iscol,&c);CHKERRQ(ierr); 
	  ipc.resize(pc.size());
	  cusp::get_inverse_permutation(pc, ipc);	
	
    PetscFunctionReturn(0);	  	
}

int main(int argc,char **argv)
{
    typedef int                  IndexType;
    typedef double               ValueType;
    typedef cusp::device_memory MemorySpace;

    PetscInitialize((int *)0,(char ***)0,(char *)0,(char *)0);	
    // create an empty sparse matrix structure
    cusp::csr_matrix<IndexType, ValueType, MemorySpace> A;

        bool matrix_market=false;
		int mm=-1;
		while(mm<0)
		{
			printf("Please input mm: 0: poisson2d,  1: UF collection \n");
			scanf("%d", &mm);
		}
		if(mm) {
		matrix_market=true;
		} else {
		matrix_market=false;		
		}
		
	std::string mtx="5pt_10x10.mtx";
	int m = 0, n = 0;
	
	if (matrix_market)
	{
    // load a matrix stored in MatrixMarket format
	printf("Please input matrix : ");
	std::cin >> mtx;
    cusp::io::read_matrix_market_file(A, mtx);	
	} else {
	while(m < 2 || n < 2)
    {
    	printf("Please input grid size (m, n): ");
    	scanf("%d %d", &m, &n);
    }	
    // create 2D Poisson problem
    cusp::gallery::poisson5pt(A, m, n);	
	}

	// allocate storage for solution (x) and right hand side (b)	
	cusp::array1d<ValueType, MemorySpace> x_solution(A.num_rows, 1);
    // initialize starting vector to random values in [0,1)
    //cusp::copy(cusp::detail::random_reals<ValueType>(A.num_rows), x_solution);
	ValueType *x_so_cpu;	
	cudaMallocHost((void **) &x_so_cpu, (A.num_rows)*sizeof(ValueType));
		
	for(int i=0;i<A.num_rows;i++)	
	{
		x_so_cpu[i]=std::sin(0.5*PI*i/A.num_rows);
	}
	ValueType *x_so_gpu = thrust::raw_pointer_cast(&x_solution[0]);
	cudaMemcpy(x_so_gpu, x_so_cpu, (A.num_rows)*sizeof(ValueType), cudaMemcpyHostToDevice);
	cudaFreeHost(x_so_cpu);
	
    cusp::array1d<ValueType, MemorySpace> b(A.num_rows);
	cusp::multiply(A, x_solution, b);

	
	{
    cusparseHandle_t handle = 0;
    cusparseStatus_t status;
	cusparseMatDescr_t descrM_lo;
	cusparseMatDescr_t descrM_up;
	cusparseSolveAnalysisInfo_t Analy_lo;
	cusparseSolveAnalysisInfo_t Analy_up;	

	cusp::csr_matrix<IndexType,ValueType,cusp::host_memory> A_mkl;
	// the factorized matrix
	cusp::csr_matrix<IndexType,ValueType,cusp::host_memory> A_lo_mkl;
	cusp::csr_matrix<IndexType,ValueType,cusp::host_memory> A_up_mkl;
	
	cusp::csr_matrix<IndexType,ValueType,cusp::device_memory> A_lo;
	cusp::csr_matrix<IndexType,ValueType,cusp::device_memory> A_up;
	cusp::array1d<ValueType,cusp::device_memory> temp_y;
	cusp::array1d<IndexType,cusp::host_memory> ipr_h,pc_h,ipc_h;	
	cusp::array1d<IndexType,cusp::device_memory> ipr_d,pc_d,ipc_d;
	
    PetscErrorCode ierr;  	
	Mat            A_CPU,Factor; 
	PC			   pc;	
	CUSPMATRIX     *A_cusp = new CUSPMATRIX(A);
	printf("rows: %d, cols: %d, nnzeros: %d\n",A_cusp->num_rows,A_cusp->num_cols,A_cusp->num_entries); 	
    bool use_mkl_sv = true;
	
		cusparseCreate(&handle);
		temp_y.resize(A_cusp->num_rows);	
		
		// create a matrix description for the lower
		status = cusparseCreateMatDescr(&descrM_lo); 

		cusparseSetMatType (descrM_lo, CUSPARSE_MATRIX_TYPE_TRIANGULAR);
		cusparseSetMatIndexBase (descrM_lo, CUSPARSE_INDEX_BASE_ZERO);
		// NOTE! CUSPARSE_DIAG_TYPE_UNIT seems not right
		//cusparseSetMatDiagType (descrM_lo, CUSPARSE_DIAG_TYPE_UNIT);
		cusparseSetMatFillMode (descrM_lo, CUSPARSE_FILL_MODE_LOWER);	

		// create the analysis info object for the lower	
		status = cusparseCreateSolveAnalysisInfo(&Analy_lo);
		
		// create a matrix description for the upper
		status = cusparseCreateMatDescr(&descrM_up); 
		cusparseSetMatType (descrM_up, CUSPARSE_MATRIX_TYPE_TRIANGULAR);
		cusparseSetMatIndexBase (descrM_up, CUSPARSE_INDEX_BASE_ZERO);
		cusparseSetMatDiagType (descrM_up, CUSPARSE_DIAG_TYPE_NON_UNIT);
		cusparseSetMatFillMode (descrM_up, CUSPARSE_FILL_MODE_UPPER);	
			
		// create the analysis info object for the upper	
		status = cusparseCreateSolveAnalysisInfo(&Analy_up);		
		
		
		
	  ierr = MatCUSPCreateFromGPU(&A_CPU, A_cusp);CHKERRQ(ierr);  
	  printf("Copy to CPU AIJ format Succeed.\n"); 
  
      ierr = PCCreate(PETSC_COMM_SELF,&pc);CHKERRQ(ierr);  
	  ierr = PCSetType(pc,PCILU);CHKERRQ(ierr);  
	  ierr = PCFactorSetFill(pc,3);CHKERRQ(ierr); 	 	  
	  ierr = PCSetOperators(pc,A_CPU,A_CPU,DIFFERENT_NONZERO_PATTERN);CHKERRQ(ierr); 	
	  ierr = PCSetUp(pc);CHKERRQ(ierr); 
	  ierr = PCFactorGetMatrix(pc,&Factor);CHKERRQ(ierr); // Need not to be destroyed 
	  
	  if (use_mkl_sv) {
	  ierr = setup_petsc_lower(Factor, A_lo_mkl, ipr_h,
		handle, status, descrM_lo, Analy_lo, use_mkl_sv);CHKERRQ(ierr);

	  ierr = setup_petsc_upper(Factor, A_up_mkl, pc_h, ipc_h,
        handle, status, descrM_up, Analy_up, use_mkl_sv);CHKERRQ(ierr);
	  } else {	
	  ierr = setup_petsc_lower(Factor, A_lo, ipr_d,
		handle, status, descrM_lo, Analy_lo, use_mkl_sv);CHKERRQ(ierr);

	  ierr = setup_petsc_upper(Factor, A_up, pc_d, ipc_d,
        handle, status, descrM_up, Analy_up, use_mkl_sv);CHKERRQ(ierr);
		}
	
    if (use_mkl_sv)	{
	cusp::print(A_lo_mkl.values);  
	cusp::print(A_lo_mkl.row_offsets);  	  
	cusp::print(A_lo_mkl.column_indices); 

	printf("ipr:\n"); 
	cusp:: print(ipr_h);
	
	cusp::print(A_up_mkl.values);  
	cusp::print(A_up_mkl.row_offsets);  	  
	cusp::print(A_up_mkl.column_indices); 

	printf("pc:\n"); 
	cusp:: print(pc_h);	
	printf("ipc:\n"); 
	cusp:: print(ipc_h);		
	} else {
	cusp::print(A_lo.values);  
	cusp::print(A_lo.row_offsets);  	  
	cusp::print(A_lo.column_indices); 

	printf("ipr:\n"); 
	cusp:: print(ipr_d);
	
	cusp::print(A_up.values);  
	cusp::print(A_up.row_offsets);  	  
	cusp::print(A_up.column_indices); 

	printf("pc:\n"); 
	cusp:: print(pc_d);	
	printf("ipc:\n"); 
	cusp:: print(ipc_d);		
	}
	
	if (use_mkl_sv)	{ 
	cusp::coo_matrix<IndexType, ValueType, cusp::host_memory> ptAqt_temp, A_coo = A;	
	cusp::compute_ptAqt(A_coo, ipr_h, pc_h, ptAqt_temp);
	cusp::csr_matrix<IndexType, ValueType, cusp::host_memory> ptAqt;		
	// setup matrix on host for MKL, redundant sort is unnecessary
	cusp::detail::setup_mkl_matrix(ptAqt,ptAqt_temp);	

	cusp::print(ptAqt.values);  
	cusp::print(ptAqt.row_offsets);  	  
	cusp::print(ptAqt.column_indices); 		
	} else {
	cusp::coo_matrix<IndexType, ValueType, MemorySpace> ptAqt_temp, A_coo = A;	
	cusp::compute_ptAqt(A_coo, ipr_d, pc_d, ptAqt_temp);
	cusp::csr_matrix<IndexType, ValueType, MemorySpace> ptAqt = ptAqt_temp;		

	cusp::print(ptAqt.values);  
	cusp::print(ptAqt.row_offsets);  	  
	cusp::print(ptAqt.column_indices); 	
	}	
	
	
	
	
	//////////////////////////

	n = Factor->rmap->n;
	CUSPINTARRAYGPU pr,prt;	
	PetscInt   *r;	
    ierr = PetscMalloc(n*sizeof(PetscInt),&r);CHKERRQ(ierr);
	r[0] = 3; r[1] = 5; r[2] = 12; r[3] = 7; r[4] = 9; r[5] = 6; r[6] = 1; r[7] = 8;
	r[8] = 15; r[9] = 2; r[10] = 11; r[11] = 10; r[12] = 0; r[13] = 14; r[14] = 13; r[15] = 4;
	pr.assign(&r[0], &r[0]+n);
	ierr = PetscFree(r);CHKERRQ(ierr); 
	
	cusp::print(pr);

	CUSPARRAY xx(n),yy(n);
	thrust::sequence(xx.begin(),xx.end());
    thrust::copy(thrust::make_permutation_iterator(xx.begin(), pr.begin()),
	       thrust::make_permutation_iterator(xx.end(),   pr.end()),
	       yy.begin());
	cusp::print(yy);

	prt.resize(pr.size());
	cusp::get_inverse_permutation(pr, prt);
	cusp::print(prt); 	
    cusp::permute_array(prt,xx,yy);		
	cusp::print(yy); 
	
	/* destroy */  
	  ierr = PCDestroy(&pc);CHKERRQ(ierr);	
	  ierr = MatDestroy(&A_CPU);CHKERRQ(ierr);
  
	  delete A_cusp;	
		// destroy cuSPARSE context
		status = cusparseDestroyMatDescr(descrM_lo);
		status = cusparseDestroySolveAnalysisInfo(Analy_lo);  
		status = cusparseDestroyMatDescr(descrM_up);
		status = cusparseDestroySolveAnalysisInfo(Analy_up);	  
	  cusparseDestroy(handle);			  
	}  

  PetscFinalize();
}

