#include <cusp/precond/smoothed_aggregation_new_bilinear_shell.h>
#include <cusp/krylov/cg.h>
#include <cusp/gallery/poisson.h>
#include <cusp/io/matrix_market.h>
#include <cusp/csr_matrix.h>
#include "../../performance/timer.h"
#include <cmath>
#define PI 3.1415926

#include <iostream>

template <typename Monitor>
void report_status(Monitor& monitor)
{
    if (monitor.converged())
    {
        std::cout << "  Solver converged to " << monitor.tolerance1() << " tolerance";
        std::cout << " after " << monitor.iteration_count() << " iterations";
        std::cout << " (" << monitor.relative_residual_norm() << " final residual)" << std::endl;
    }
    else
    {
        std::cout << "  Solver reached iteration limit " << monitor.iteration_limit() << " before converging";
        std::cout << " to " << monitor.tolerance1() << " tolerance ";
        std::cout << " (" << monitor.relative_residual_norm() << " final residual)" << std::endl;
    }
}

int main(void)
{
    typedef int                 IndexType;
    typedef double               ValueType;
    typedef cusp::device_memory MemorySpace;
	typedef cusp::precond::sa_jacobi_policy<ValueType,MemorySpace> smoother_jacobi;
	
    // create an empty sparse matrix structure
    cusp::coo_matrix<IndexType, ValueType, MemorySpace> A;
	
	int m = 2, n = 2;
	#if 1
	std::string mtx="5pt_10x10.mtx";
	printf("Please input matrix : ");
	std::cin >> mtx;
    cusp::io::read_matrix_market_file(A, mtx);
	#else
	while(m < 2 || n < 2)
    {
    	printf("Please input m, n: ");
    	scanf("%d %d", &m, &n);
    }
    cusp::gallery::poisson5pt(A, m, n);
	#endif

	// allocate storage for solution (x) and right hand side (b)
	cusp::array1d<ValueType, MemorySpace> x_solution(A.num_rows, 1);
	/*
	for(int i=0;i<A.num_rows;i++)	
	{
		x_solution[i]=std::sin(0.5*PI*i/A.num_rows);
	}
	*/
        cusp::array1d<ValueType, MemorySpace> b(A.num_rows);
		cusp::multiply(A, x_solution, b);
		
	/*
    // solve without preconditioning
    {
        std::cout << "\nSolving with no preconditioner..." << std::endl;
    
        // allocate storage for solution (x) and right hand side (b)
        cusp::array1d<ValueType, MemorySpace> x(A.num_rows, 0);

        // set stopping criteria (iteration_limit = 1000, relative_tolerance = 1e-6)
        cusp::default_monitor<ValueType> monitor(b, 1000, 1e-6);
        
        // solve
        cusp::krylov::cg(A, x, b, monitor);

        // report status
        report_status(monitor);
    }
	*/

    // solve with smoothed aggregation algebraic multigrid preconditioner
    {
        std::cout << "\nSolving with smoothed aggregation/bilinear preconditioner..." << std::endl;
        
	int ratioi = 0, ratioj = 0;
	while(ratioi < 2 || ratioj < 2)
    {
    	printf("Please input ratioi, ratioj: ");
    	scanf("%d %d", &ratioi, &ratioj);
    }
	
        // allocate storage for solution (x) and right hand side (b)
        cusp::array1d<ValueType, MemorySpace> x(A.num_rows, 0);

        // set stopping criteria (iteration_limit = 10000, relative_tolerance = 1e-7)
        cusp::verbose_monitor<ValueType> monitor(b, 10000, 1e-7);

        // setup preconditioner
        cusp::precond::smoothed_aggregation<IndexType, ValueType, MemorySpace, smoother_jacobi> M(A, 0, 100, false, 3, 3, 60,1,1e-5,0,1,m,n,ratioi,ratioj,false);
        
        // solve
		timer t0;
        cusp::krylov::cg(A, x, b, monitor, M);
        std::cout << "\nsolve by as in " << t0.seconds_elapsed() << " s " << std::endl;
        // report status
        report_status(monitor);
        
		// check solution
        cusp::array1d<ValueType, MemorySpace> e(A.num_rows);
		blas::axpby(x_solution, x, e, ValueType(1), ValueType(-1));
		ValueType e_norm = cusp::blas::nrm2(e);
		printf("\n check solution: ||x*-x||: %f \n", e_norm );	
		
		cusp::array1d<ValueType, MemorySpace> ax(x.size());
		cusp::array1d<ValueType, MemorySpace> r(A.num_rows);
		cusp::multiply(A, x, ax);
		cusp::blas::axpby(ax, b, r, ValueType(1), ValueType(-1));
		ValueType r_norm = cusp::blas::nrm2(r);
		ValueType b_norm = cusp::blas::nrm2(b);		
		printf("\n ||b-Ax||/||b||: %f \n", r_norm / b_norm );
    }

    return 0;
}
