/*
nvcc -m64 --compiler-options -fno-strict-aliasing  --compiler-options -funroll-loops  -use_fast_math -DADD_ --compiler-options '-fPIC' -DUNIX -O3 -arch=sm_13 -I/home/liluo/lib/cuda/include -I/home/liluo/lib/cusp-library -I/home/liluo/lib/thrust -o s_a_text s_a_text.cu -L/home/liluo/lib/cuda/lib64 -lcudart -lcuda -lcublas -lcusparse -DCUSP_USE_TEXTURE_MEMORY
*/

#include <cusp/precond/smoothed_aggregation.h>
#include <cusp/krylov/cg.h>
#include <cusp/gallery/poisson.h>
#include <cusp/csr_matrix.h>
#include <cusp/io/matrix_market.h>
#include <performance/timer.h>
#include <iostream>
#include <cmath>
#define PI 3.1415926

template <typename Monitor>
void report_status(Monitor& monitor)
{
    if (monitor.converged())
    {
        std::cout << "  Solver converged to " << monitor.tolerance() << " tolerance";
        std::cout << " after " << monitor.iteration_count() << " iterations";
        std::cout << " (" << monitor.residual_norm() << " final residual)" << std::endl;
    }
    else
    {
        std::cout << "  Solver reached iteration limit " << monitor.iteration_limit() << " before converging";
        std::cout << " to " << monitor.tolerance() << " tolerance ";
        std::cout << " (" << monitor.residual_norm() << " final residual)" << std::endl;
    }
}

int main(void)
{
    typedef int                 IndexType;
    typedef double               ValueType;
    typedef cusp::device_memory MemorySpace;
	
    // create an empty sparse matrix structure
    cusp::csr_matrix<IndexType, ValueType, MemorySpace> A;

	#if 1
	std::string mtx="5pt_10x10.mtx";
	printf("Please input matrix : ");
	std::cin >> mtx;
    cusp::io::read_matrix_market_file(A, mtx);
	#else
	int m = 0, n = 0;
	while(m < 2 || n < 2)
    {
    	printf("Please input m, n: ");
    	scanf("%d %d", &m, &n);
    }
    cusp::gallery::poisson5pt(A, m, n);
	#endif

	// allocate storage for solution (x) and right hand side (b)	
	cusp::array1d<ValueType, MemorySpace> x_solution(A.num_rows, 1);
    // initialize starting vector to random values in [0,1)
    //cusp::copy(cusp::detail::random_reals<ValueType>(A.num_rows), x_solution);
	ValueType *x_so_cpu;	
	cudaMallocHost((void **) &x_so_cpu, (A.num_rows)*sizeof(ValueType));
		
	for(int i=0;i<A.num_rows;i++)	
	{
		x_so_cpu[i]=std::sin(0.5*PI*i/A.num_rows);
	}
	ValueType *x_so_gpu = thrust::raw_pointer_cast(&x_solution[0]);
	cudaMemcpy(x_so_gpu, x_so_cpu, (A.num_rows)*sizeof(ValueType), cudaMemcpyHostToDevice);
	cudaFreeHost(x_so_cpu);
	
    cusp::array1d<ValueType, MemorySpace> b(A.num_rows);
	cusp::multiply(A, x_solution, b);
		
    // solve without preconditioning
    {
        std::cout << "\nSolving with no preconditioner..." << std::endl;
    
        // allocate storage for solution (x) and right hand side (b)
        cusp::array1d<ValueType, MemorySpace> x(A.num_rows, 0);

        // set stopping criteria (iteration_limit = 10000, relative_tolerance = 1e-7)
        cusp::verbose_monitor<ValueType> monitor(b, 10000, 1e-7);
        
        // solve
		timer t1;	
        cusp::krylov::cg(A, x, b, monitor);
		std::cout << "\nsolve by cg without preconditioner in " << t1.seconds_elapsed() << " s " << std::endl;	
		
        // report status
        report_status(monitor);
		
		// check solution
        cusp::array1d<ValueType, MemorySpace> e(A.num_rows);
		blas::axpby(x_solution, x, e, ValueType(1), ValueType(-1));
		ValueType e_norm = cusp::blas::nrm2(e);
		printf("\n check solution: ||x*-x||: %f \n", e_norm );	
		
		cusp::array1d<ValueType, MemorySpace> ax(x.size());
		cusp::array1d<ValueType, MemorySpace> r(A.num_rows);
		cusp::multiply(A, x, ax);
		cusp::blas::axpby(ax, b, r, ValueType(1), ValueType(-1));
		ValueType r_norm = cusp::blas::nrm2(r);
		ValueType b_norm = cusp::blas::nrm2(b);		
		printf("\n ||b-Ax||/||b||: %f \n", r_norm / b_norm );
    }

    // solve with smoothed aggregation algebraic multigrid preconditioner
    {
        std::cout << "\nSolving with smoothed aggregation preconditioner..." << std::endl;
        
        // allocate storage for solution (x) and right hand side (b)
        cusp::array1d<ValueType, MemorySpace> x(A.num_rows, 0);

        // set stopping criteria (iteration_limit = 10000, relative_tolerance = 1e-7)
        cusp::verbose_monitor<ValueType> monitor(b, 10000, 1e-7);

        // setup preconditioner
		timer t0;	
        cusp::precond::smoothed_aggregation<IndexType, ValueType, MemorySpace> M(A);
		std::cout << "\ncreate preconditioner in " << t0.seconds_elapsed() << " s " << std::endl;	
        
        // solve
		timer t1;	
        cusp::krylov::cg(A, x, b, monitor, M);
		std::cout << "\nsolve by cg with SA preconditioner in " << t1.seconds_elapsed() << " s " << std::endl;	
        
        // report status
        report_status(monitor);

		// check solution
        cusp::array1d<ValueType, MemorySpace> e(A.num_rows);
		blas::axpby(x_solution, x, e, ValueType(1), ValueType(-1));
		ValueType e_norm = cusp::blas::nrm2(e);
		printf("\n check solution: ||x*-x||: %f \n", e_norm );	
		
		cusp::array1d<ValueType, MemorySpace> ax(x.size());
		cusp::array1d<ValueType, MemorySpace> r(A.num_rows);
		cusp::multiply(A, x, ax);
		cusp::blas::axpby(ax, b, r, ValueType(1), ValueType(-1));
		ValueType r_norm = cusp::blas::nrm2(r);
		ValueType b_norm = cusp::blas::nrm2(b);		
		printf("\n ||b-Ax||/||b||: %f \n", r_norm / b_norm );
		
        // print hierarchy information
        std::cout << "\nPreconditioner statistics" << std::endl;
        //M.print();
    }

    return 0;
}

